!
! mo_art_init_radiation
! This module provides initialization of radiation
!
! ICON
!
! ---------------------------------------------------------------
! Copyright (C) 2004-2024, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
! Contact information: icon-model.org
!
! See AUTHORS.TXT for a list of authors
! See LICENSES/ for license information
! SPDX-License-Identifier: BSD-3-Clause
! ---------------------------------------------------------------

MODULE mo_art_init_radiation
  USE mo_kind,                          ONLY: wp
  USE mo_art_read_xml,                  ONLY: art_read_visparams_xml, &
                                            & art_read_dia_factors_xml
  USE mo_art_config,                    ONLY: art_config
  USE mo_exception,                     ONLY: finish  ! TEMP !!!
  IMPLICIT NONE
    
  PRIVATE

  CHARACTER(len=*), PARAMETER :: routine = 'mo_art_init_radiation'
  
  PUBLIC :: art_init_radiation
  
  CONTAINS
  
  SUBROUTINE art_init_radiation(jg,mode_name,ext_coeff,ssa_coeff,asy_coeff,   &
    &                           ext_param, ssa_param, asy_param, ext_default, &
    &                           ssa_default, asy_default, dia_min_factor,     &
    &                           dia_max_factor)
  !<
  ! SUBROUTINE art_init_radiation
  ! This subroutine initializes the radiation
  ! Part of Module: mo_art_init_radiation
  ! Author: Daniel Rieger, Philipp Gasch, Carolin Walter, KIT
  ! Initial Release: 2015-12-15
  !>
    INTEGER, INTENT(IN)            :: &
      &  jg
    CHARACTER(LEN=*), INTENT(in)   :: &
      &  mode_name
    REAL(wp),POINTER,INTENT(inout) :: &
      &  ext_coeff(:),                &
      &  ssa_coeff(:),                &
      &  asy_coeff(:),                &
      &  ext_param(:,:),              &
      &  ssa_param(:,:),              &
      &  asy_param(:,:),              &
      &  ext_default(:,:),            &
      &  ssa_default(:,:),            &
      &  asy_default(:,:),            &
      &  dia_min_factor(:),           &
      &  dia_max_factor(:)          
  ! Local Variables
    INTEGER ::                        &
      &  jspec,                       & !< Loop indices
      &  npoly                          !< degree of polynom+1 used for diame-
                                        !  ter parameterization of coefficients
    LOGICAL ::                        &
      &  is_aerodyn                     !< determines if aerodyn is used
      
    is_aerodyn = .TRUE.

    jspec=30
    npoly=4
    ALLOCATE(ext_coeff(jspec))
    ALLOCATE(ssa_coeff(jspec))
    ALLOCATE(asy_coeff(jspec))

    ALLOCATE(ext_param(jspec,npoly))
    ALLOCATE(ssa_param(jspec,npoly))
    ALLOCATE(asy_param(jspec,npoly))

    ALLOCATE(ext_default(jspec,2))
    ALLOCATE(ssa_default(jspec,2))
    ALLOCATE(asy_default(jspec,2))

    ALLOCATE(dia_min_factor(1))
    ALLOCATE(dia_max_factor(1))

    ! NOTE: If these tables are updated here, the same has to be done for the
    !       singular values in mo_art_aero_optical_props.f90 
    !(This routine calculated the AOD for singular wavelengths)
    SELECT CASE(TRIM(mode_name))
    !! This table currently contains optics of inorganic acids and coated ash/dust 
    !! sulfur aerosols 
      !! aitken mode
!        CASE('so4_sol_ait')
      CASE('sol_ait')
        ext_coeff=(/0.02871_wp,0.01809_wp,0.06639_wp,0.06385_wp,0.04190_wp,       &
          &         0.10976_wp,0.20250_wp,0.39568_wp,0.40459_wp,0.13125_wp,       &
          &         0.13642_wp,0.19499_wp,0.19894_wp,0.21817_wp,0.23121_wp,       &
          &         0.30845_wp,0.30845_wp,0.10585_wp,0.00833_wp,0.00529_wp,       &
          &         0.00373_wp,0.00538_wp,0.01560_wp,0.02586_wp,0.09897_wp,       &
          &         0.24575_wp,0.60659_wp,1.43449_wp,4.08078_wp,0.22117_wp /)
        ssa_coeff=(/0.00000_wp,0.00002_wp,0.00002_wp,0.00001_wp,0.00004_wp,       &
          &         0.00005_wp,0.00006_wp,0.00004_wp,0.00002_wp,0.00004_wp,       &
          &         0.00013_wp,0.00014_wp,0.00023_wp,0.00029_wp,0.00036_wp,       &
          &         0.00040_wp,0.00040_wp,0.00741_wp,0.06974_wp,0.20538_wp,       &
          &         0.45469_wp,0.92180_wp,0.96070_wp,0.99968_wp,1.00000_wp,       &
          &         1.00000_wp,1.00000_wp,1.00000_wp,1.00000_wp,0.00014_wp /)
        asy_coeff=(/1.00000_wp,1.00000_wp,1.00000_wp,1.00000_wp,0.00014_wp,       &
          &         0.00089_wp,0.00117_wp,0.00115_wp,0.00125_wp,0.00170_wp,       &
          &         0.00228_wp,0.00317_wp,0.00400_wp,0.00463_wp,0.00529_wp,       &
          &         0.00702_wp,0.00702_wp,0.01005_wp,0.01492_wp,0.02098_wp,       &
          &         0.02577_wp,0.04151_wp,0.06290_wp,0.08424_wp,0.14913_wp,       &
          &         0.21072_wp,0.28762_wp,0.36437_wp,0.46741_wp,0.00272_wp /)

!! accumulation mode
!        CASE('so4_sol_acc')
      CASE('sol_acc')
        ext_coeff=(/0.02872_wp,0.01815_wp,0.06673_wp,0.06418_wp,0.04228_wp,       &
          &         0.11113_wp,0.20611_wp,0.40169_wp,0.40688_wp,0.13265_wp,       &
          &         0.13971_wp,0.20027_wp,0.20658_wp,0.22810_wp,0.24350_wp,       &
          &         0.32553_wp,0.32553_wp,0.11875_wp,0.03493_wp,0.05899_wp,       &
          &         0.08270_wp,0.20152_wp,0.45240_wp,0.70334_wp,1.77808_wp,       &
          &         3.06857_wp,5.00644_wp,7.36070_wp,10.25832_wp,0.22676_wp /)
        ssa_coeff=(/0.00015_wp,0.00148_wp,0.00144_wp,0.00095_wp,0.00279_wp,       &
          &         0.00313_wp,0.00393_wp,0.00245_wp,0.00137_wp,0.00261_wp,       &
          &         0.00775_wp,0.00864_wp,0.01348_wp,0.01665_wp,0.02028_wp,       &
          &         0.02121_wp,0.02121_wp,0.22991_wp,0.74581_wp,0.92057_wp,       &
          &         0.97199_wp,0.99749_wp,0.99870_wp,0.99999_wp,1.00000_wp,       &
          &         1.00000_wp,1.00000_wp,1.00000_wp,1.00000_wp,0.00849_wp /)
        asy_coeff=(/0.00056_wp,0.00313_wp,0.00564_wp,0.00706_wp,0.00970_wp,       &
          &         0.01440_wp,0.01905_wp,0.01783_wp,0.01880_wp,0.02563_wp,       &
          &         0.03447_wp,0.04591_wp,0.05672_wp,0.06442_wp,0.07217_wp,       &
          &         0.08948_wp,0.08948_wp,0.12143_wp,0.16762_wp,0.21427_wp,       &
          &         0.24542_wp,0.32461_wp,0.39060_wp,0.45659_wp,0.56569_wp,       &
          &         0.62403_wp,0.66664_wp,0.68228_wp,0.68491_wp,0.03920_wp /)
!! !!! sol_coa mode is missing
!
      CASE('sol_coa') ! This is copy-paste of seasb 
        ext_coeff=(/0.06207_wp,0.16129_wp,0.19155_wp,0.16685_wp,0.15007_wp,       &  
          &         0.15534_wp,0.20853_wp,0.24018_wp,0.23625_wp,0.26487_wp,       &  
          &         0.28126_wp,0.30344_wp,0.31832_wp,0.32455_wp,0.32558_wp,       &  
          &         0.32477_wp,0.32477_wp,0.30436_wp,0.31335_wp,0.30331_wp,       &  
          &         0.29818_wp,0.28915_wp,0.00000_wp,0.27465_wp,0.26579_wp,       &  
          &         0.26162_wp,0.26030_wp,0.25550_wp,0.25153_wp,0.27339_wp /)
        ssa_coeff=(/0.11635_wp,0.41183_wp,0.41645_wp,0.36938_wp,0.37299_wp,       &  
          &         0.63652_wp,0.79532_wp,0.80199_wp,0.83141_wp,0.84405_wp,       &  
          &         0.72976_wp,0.92350_wp,0.91731_wp,0.92854_wp,0.95219_wp,       &  
          &         0.84336_wp,0.84336_wp,0.67158_wp,0.96248_wp,0.97269_wp,       &  
          &         0.98742_wp,0.99005_wp,0.00000_wp,0.99760_wp,1.00000_wp,       &  
          &         1.00000_wp,1.00000_wp,0.99990_wp,0.99526_wp,0.85179_wp /)
        asy_coeff=(/0.23821_wp,0.58885_wp,0.65669_wp,0.73266_wp,0.77698_wp,       &  
          &         0.80782_wp,0.78361_wp,0.77607_wp,0.80724_wp,0.79911_wp,       &  
          &         0.81181_wp,0.79066_wp,0.77804_wp,0.77065_wp,0.76180_wp,       &  
          &         0.77478_wp,0.77478_wp,0.87834_wp,0.79962_wp,0.78291_wp,       &  
          &         0.78224_wp,0.77931_wp,0.00000_wp,0.79435_wp,0.80723_wp,       &  
          &         0.81546_wp,0.82392_wp,0.82318_wp,0.82978_wp,0.78740_wp /)        
              
!! ash insol accumulation mode
!        CASE('ash_insol_acc')
      CASE('insol_acc')
        ext_coeff=(/0.05259_wp,0.10150_wp,0.19431_wp,0.10277_wp,0.11910_wp,       &
          &         0.51961_wp,0.54460_wp,0.32293_wp,0.09055_wp,0.09215_wp,       &
          &         0.16407_wp,0.23015_wp,0.29160_wp,0.32684_wp,0.41156_wp,       &
          &         0.65373_wp,0.65373_wp,0.76660_wp,0.97555_wp,1.18041_wp,       &
          &         1.40476_wp,1.61062_wp,1.67312_wp,1.79298_wp,1.67589_wp,       &
          &         1.60644_wp,1.49926_wp,1.44206_wp,1.43603_wp,0.41156_wp /)
        ssa_coeff=(/0.04006_wp,0.06169_wp,0.03602_wp,0.08923_wp,0.27222_wp,       &
          &         0.14415_wp,0.10740_wp,0.05327_wp,0.27489_wp,0.39957_wp,       &
          &         0.61227_wp,0.74961_wp,0.77353_wp,0.79408_wp,0.81133_wp,       &
          &         0.70444_wp,0.70444_wp,0.98066_wp,0.98649_wp,0.98804_wp,       &
          &         0.99049_wp,0.99121_wp,0.99125_wp,0.99064_wp,0.98988_wp,       &
          &         0.98669_wp,0.98156_wp,0.96690_wp,0.96146_wp,0.81133_wp /)
        asy_coeff=(/0.07560_wp,0.11793_wp,0.13537_wp,0.17575_wp,0.23553_wp,       &
          &         0.17171_wp,0.19280_wp,0.28594_wp,0.39678_wp,0.41799_wp,       &
          &         0.48750_wp,0.53270_wp,0.55838_wp,0.57102_wp,0.59803_wp,       &
          &         0.65298_wp,0.65298_wp,0.66194_wp,0.67549_wp,0.69161_wp,       &
          &         0.70039_wp,0.70707_wp,0.71120_wp,0.69139_wp,0.69478_wp,       &
          &         0.68957_wp,0.71290_wp,0.73915_wp,0.75347_wp,0.59803_wp /)
!! ash insol coarse mode
!        CASE('ash_insol_coa')
      CASE('insol_coa')
        ext_coeff=(/0.13660_wp,0.15261_wp,0.13696_wp,0.14750_wp,0.17679_wp,       &
          &         0.16921_wp,0.16103_wp,0.13106_wp,0.15017_wp,0.16522_wp,       &
          &         0.18030_wp,0.18447_wp,0.18296_wp,0.18324_wp,0.18130_wp,       &
          &         0.17188_wp,0.17188_wp,0.17242_wp,0.16678_wp,0.16826_wp,       &
          &         0.16364_wp,0.16259_wp,0.16075_wp,0.15459_wp,0.15298_wp,       &
          &         0.15129_wp,0.14943_wp,0.14866_wp,0.14726_wp,0.18130_wp /)
        ssa_coeff=(/0.47743_wp,0.45472_wp,0.36681_wp,0.52128_wp,0.56743_wp,       &
          &         0.41511_wp,0.40189_wp,0.36723_wp,0.66563_wp,0.71916_wp,       &
          &         0.73765_wp,0.76397_wp,0.75205_wp,0.75175_wp,0.73536_wp,       &
          &         0.60566_wp,0.60566_wp,0.91288_wp,0.92251_wp,0.92039_wp,       &
          &         0.92658_wp,0.92277_wp,0.92170_wp,0.91454_wp,0.91610_wp,       &
          &         0.90148_wp,0.87338_wp,0.81503_wp,0.79107_wp,0.73536_wp /)
        asy_coeff=(/0.53160_wp,0.64872_wp,0.74838_wp,0.75548_wp,0.69469_wp,       &
          &         0.78330_wp,0.81618_wp,0.86939_wp,0.88900_wp,0.87300_wp,       &
          &         0.84893_wp,0.82494_wp,0.82388_wp,0.82031_wp,0.82378_wp,       &
          &         0.88148_wp,0.88148_wp,0.73400_wp,0.73960_wp,0.74079_wp,       &
          &         0.75837_wp,0.76940_wp,0.77663_wp,0.79756_wp,0.81412_wp,       &
          &         0.81912_wp,0.83439_wp,0.85476_wp,0.86322_wp,0.82378_wp /)
!! ash mixed accumulation mode
!        CASE('ash_mixed_acc')
      CASE('mixed_acc')
        ext_coeff=(/0.04731_wp,0.08720_wp,0.18097_wp,0.08664_wp,0.07276_wp,       &
          &         0.41004_wp,0.50196_wp,0.34081_wp,0.06058_wp,0.05038_wp,       &
          &         0.05903_wp,0.05723_wp,0.06806_wp,0.07189_wp,0.08792_wp,       &
          &         0.21365_wp,0.21365_wp,0.11732_wp,0.17783_wp,0.26975_wp,       &
          &         0.42464_wp,0.70725_wp,0.85086_wp,1.93800_wp,3.09623_wp,       &
          &         4.70122_wp,6.01615_wp,7.48345_wp,7.89588_wp,0.08792_wp /)
        ssa_coeff=(/0.00121_wp,0.00219_wp,0.00147_wp,0.00370_wp,0.01553_wp,       &
          &         0.01049_wp,0.00741_wp,0.00334_wp,0.02217_wp,0.03974_wp,       &
          &         0.10533_wp,0.20916_wp,0.25186_wp,0.28820_wp,0.34065_wp,       &
          &         0.26309_wp,0.26309_wp,0.92185_wp,0.95695_wp,0.96946_wp,       &
          &         0.98170_wp,0.98774_wp,0.98920_wp,0.99394_wp,0.99572_wp,       &
          &         0.99619_wp,0.99572_wp,0.99380_wp,0.99278_wp,0.34065_wp /)
        asy_coeff=(/0.02033_wp,0.03435_wp,0.03976_wp,0.05253_wp,0.07783_wp,       &
          &         0.05806_wp,0.06094_wp,0.08656_wp,0.13482_wp,0.14586_wp,       &
          &         0.18695_wp,0.22016_wp,0.24072_wp,0.25214_wp,0.27669_wp,       &
          &         0.32117_wp,0.32117_wp,0.37086_wp,0.40449_wp,0.44083_wp,       &
          &         0.48075_wp,0.52533_wp,0.54137_wp,0.60919_wp,0.64623_wp,       &
          &         0.66742_wp,0.68337_wp,0.69213_wp,0.68718_wp,0.27669_wp /)
!! ash mixed coarse mode
!        CASE('ash_mixed_coa')
      CASE('mixed_coa')
        ext_coeff=(/0.13980_wp,0.18195_wp,0.17791_wp,0.17467_wp,0.24752_wp,       &
          &         0.27436_wp,0.25671_wp,0.18340_wp,0.17857_wp,0.20659_wp,       &
          &         0.27097_wp,0.30053_wp,0.31043_wp,0.31451_wp,0.32047_wp,       &
          &         0.31410_wp,0.31410_wp,0.30629_wp,0.31355_wp,0.30923_wp,       &
          &         0.29940_wp,0.28861_wp,0.29394_wp,0.28390_wp,0.27662_wp,       &
          &         0.26733_wp,0.26428_wp,0.25868_wp,0.25618_wp,0.32047_wp /)
        ssa_coeff=(/0.44070_wp,0.41573_wp,0.29348_wp,0.48220_wp,0.59125_wp,       &
          &         0.37717_wp,0.35010_wp,0.28690_wp,0.65709_wp,0.73068_wp,       &
          &         0.78002_wp,0.81763_wp,0.80947_wp,0.81319_wp,0.80185_wp,       &
          &         0.67223_wp,0.67223_wp,0.94769_wp,0.95384_wp,0.95300_wp,       &
          &         0.95510_wp,0.95282_wp,0.95254_wp,0.94829_wp,0.94652_wp,       &
          &         0.93797_wp,0.91810_wp,0.87297_wp,0.85401_wp,0.80185_wp /)
        asy_coeff=(/0.43815_wp,0.54352_wp,0.63807_wp,0.67138_wp,0.61619_wp,       &
          &         0.64937_wp,0.70029_wp,0.81249_wp,0.84838_wp,0.83660_wp,       &
          &         0.81791_wp,0.79809_wp,0.79584_wp,0.79269_wp,0.79380_wp,       &
          &         0.84011_wp,0.84011_wp,0.71778_wp,0.70060_wp,0.71839_wp,       &
          &         0.71416_wp,0.73644_wp,0.73492_wp,0.74936_wp,0.77653_wp,       &
          &         0.80033_wp,0.80926_wp,0.82814_wp,0.83712_wp,0.79380_wp /)

!! ash giant
!        CASE('ash_giant')
      CASE('giant')
        ext_coeff=(/0.09196_wp,0.08886_wp,0.08055_wp,0.08660_wp,0.08265_wp,       &
          &         0.08085_wp,0.07844_wp,0.07462_wp,0.08577_wp,0.08552_wp,       &
          &         0.07874_wp,0.07773_wp,0.07711_wp,0.07705_wp,0.07636_wp,       &
          &         0.07445_wp,0.07445_wp,0.07504_wp,0.07422_wp,0.07310_wp,       &
          &         0.07338_wp,0.07252_wp,0.07114_wp,0.07081_wp,0.06996_wp,       &
          &         0.06923_wp,0.06934_wp,0.06860_wp,0.06856_wp,0.07636_wp /)
        ssa_coeff=(/0.47706_wp,0.47017_wp,0.45239_wp,0.51921_wp,0.49079_wp,       &
          &         0.47412_wp,0.46897_wp,0.46946_wp,0.60722_wp,0.62725_wp,       &
          &         0.61125_wp,0.63904_wp,0.62906_wp,0.63021_wp,0.61772_wp,       &
          &         0.53426_wp,0.53426_wp,0.83957_wp,0.85418_wp,0.84836_wp,       &
          &         0.85832_wp,0.85347_wp,0.85058_wp,0.84459_wp,0.84246_wp,       &
          &         0.81888_wp,0.77834_wp,0.69958_wp,0.67458_wp,0.61772_wp/)
        asy_coeff=(/0.69978_wp,0.80666_wp,0.86242_wp,0.85759_wp,0.85526_wp,       &
          &         0.89361_wp,0.90978_wp,0.90971_wp,0.92811_wp,0.91188_wp,       &
          &         0.90146_wp,0.88755_wp,0.89420_wp,0.89407_wp,0.90232_wp,       &
          &         0.94428_wp,0.94428_wp,0.81511_wp,0.81220_wp,0.81600_wp,       &
          &         0.82369_wp,0.82932_wp,0.83346_wp,0.84748_wp,0.85078_wp,       &
          &         0.85824_wp,0.87292_wp,0.89689_wp,0.90543_wp,0.90232_wp /)
       CASE DEFAULT
         is_aerodyn = .FALSE.
         
    END SELECT ! mode
    IF(.NOT.is_aerodyn) THEN
      CALL art_read_dia_factors_xml(jg, dia_min_factor, dia_max_factor)

      IF(art_config(jg)%iart_nonsph > 0) THEN  ! Ellipsoid mixture 		
        SELECT CASE(TRIM(mode_name))
          CASE ('dusta') 
            ext_coeff=(/0.01466_wp,0.03788_wp,0.07123_wp,0.10277_wp,0.07884_wp, &
              &         0.07469_wp,0.12498_wp,0.17458_wp,0.06007_wp,0.15342_wp, &
              &         0.08172_wp,0.10230_wp,0.13210_wp,0.15031_wp,0.18210_wp, &
              &         0.24438_wp,0.24438_wp,0.29232_wp,0.40495_wp,0.50912_wp, &
              &         0.63469_wp,0.83443_wp,0.97270_wp,1.19868_wp,1.46633_wp, &
              &         1.55188_wp,1.54680_wp,1.48922_wp,1.40726_wp,0.05869_wp /)
            ssa_coeff=(/0.00063_wp,0.01298_wp,0.01023_wp,0.00735_wp,0.01102_wp, &
              &         0.11811_wp,0.10446_wp,0.02539_wp,0.04252_wp,0.04561_wp, &
              &         0.11211_wp,0.14784_wp,0.41848_wp,0.44308_wp,0.55940_wp, &
              &         0.60555_wp,0.60555_wp,0.94352_wp,0.98000_wp,0.98156_wp, &
              &         0.98270_wp,0.98353_wp,0.99579_wp,0.99557_wp,0.97905_wp, &
              &         0.97362_wp,0.91873_wp,0.82928_wp,0.79290_wp,0.02137_wp /)
            asy_coeff=(/0.00581_wp,0.03922_wp,0.05281_wp,0.06070_wp,0.07925_wp, &
              &         0.12858_wp,0.14375_wp,0.12952_wp,0.17432_wp,0.18858_wp, &
              &         0.23973_wp,0.28863_wp,0.34455_wp,0.36489_wp,0.39578_wp, &
              &         0.44114_wp,0.44114_wp,0.50244_wp,0.54755_wp,0.57673_wp, &
              &         0.60478_wp,0.63908_wp,0.65685_wp,0.68028_wp,0.70389_wp, &
              &         0.70921_wp,0.72883_wp,0.76517_wp,0.78500_wp,0.16414_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                         ssa_param, asy_param, ext_default,      &
              &                         ssa_default, asy_default)
          CASE('dustb')
            ext_coeff=(/0.01628_wp,0.06851_wp,0.08874_wp,0.09475_wp,0.08027_wp, &
              &         0.17303_wp,0.20881_wp,0.13748_wp,0.07699_wp,0.13349_wp, &
              &         0.12366_wp,0.14617_wp,0.25122_wp,0.26201_wp,0.29877_wp, &
              &         0.31716_wp,0.31716_wp,0.34266_wp,0.34620_wp,0.34386_wp, &
              &         0.33775_wp,0.32477_wp,0.31148_wp,0.29343_wp,0.27450_wp, &
              &         0.26111_wp,0.24915_wp,0.24072_wp,0.23220_wp,0.06216_wp /)
            ssa_coeff=(/0.03663_wp,0.26308_wp,0.17976_wp,0.12437_wp,0.16075_wp, &
              &         0.50095_wp,0.43063_wp,0.21750_wp,0.30858_wp,0.28286_wp, &
              &         0.45829_wp,0.48632_wp,0.69160_wp,0.69111_wp,0.74741_wp, &
              &         0.73870_wp,0.73870_wp,0.94545_wp,0.97249_wp,0.96851_wp, &
              &         0.96349_wp,0.95490_wp,0.98482_wp,0.98052_wp,0.91325_wp, &
              &         0.89490_wp,0.75626_wp,0.64315_wp,0.62112_wp,0.19757_wp /)
            asy_coeff=(/0.06316_wp,0.23584_wp,0.30452_wp,0.34333_wp,0.41168_wp, &
              &         0.45793_wp,0.49030_wp,0.53259_wp,0.63203_wp,0.64174_wp, &
              &         0.70330_wp,0.74824_wp,0.73538_wp,0.74682_wp,0.74972_wp, &
              &         0.76895_wp,0.76895_wp,0.72036_wp,0.71041_wp,0.71009_wp, &
              &         0.70997_wp,0.71190_wp,0.70052_wp,0.70780_wp,0.75303_wp, &
              &         0.77597_wp,0.84854_wp,0.90375_wp,0.91856_wp,0.62150_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                         ssa_param, asy_param, ext_default,      &
              &                         ssa_default, asy_default)
          CASE('dustc')
            ext_coeff=(/0.02196_wp,0.09305_wp,0.08501_wp,0.07370_wp,0.06822_wp, &
              &         0.13738_wp,0.13545_wp,0.09148_wp,0.07363_wp,0.09336_wp, &
              &         0.10474_wp,0.11196_wp,0.14569_wp,0.14486_wp,0.15356_wp, &
              &         0.14957_wp,0.14957_wp,0.14354_wp,0.13651_wp,0.13239_wp, &
              &         0.12850_wp,0.12368_wp,0.11823_wp,0.11393_wp,0.11002_wp, &
              &         0.10610_wp,0.10269_wp,0.10037_wp,0.09810_wp,0.05638_wp /)
            ssa_coeff=(/0.16239_wp,0.43691_wp,0.33867_wp,0.26375_wp,0.30797_wp, &
              &         0.54466_wp,0.48946_wp,0.35710_wp,0.45151_wp,0.41031_wp, &
              &         0.54869_wp,0.55388_wp,0.63535_wp,0.62608_wp,0.66079_wp, &
              &         0.63680_wp,0.63680_wp,0.88237_wp,0.93674_wp,0.92923_wp, &
              &         0.92093_wp,0.90845_wp,0.96686_wp,0.95993_wp,0.85359_wp, &
              &         0.82835_wp,0.66563_wp,0.58668_wp,0.57587_wp,0.33563_wp /)
            asy_coeff=(/0.15092_wp,0.43030_wp,0.55380_wp,0.60826_wp,0.67704_wp, &
              &         0.66333_wp,0.70460_wp,0.76634_wp,0.83536_wp,0.83465_wp, &
              &         0.86068_wp,0.88055_wp,0.83999_wp,0.84648_wp,0.83563_wp, &
              &         0.84899_wp,0.84899_wp,0.74020_wp,0.72235_wp,0.73137_wp, &
              &         0.74270_wp,0.76041_wp,0.74462_wp,0.75900_wp,0.81744_wp, &
              &         0.83483_wp,0.90864_wp,0.93828_wp,0.94324_wp,0.83624_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                         ssa_param, asy_param, ext_default,      &
              &                         ssa_default, asy_default)
          ! Seasalt optical properties are currently based on 70% RH
          CASE ('seasa')
            ext_coeff=(/0.04822_wp,0.08393_wp,0.11915_wp,0.13776_wp,0.12483_wp, &
              &         0.05083_wp,0.02969_wp,0.03225_wp,0.02828_wp,0.02893_wp, &
              &         0.07075_wp,0.01846_wp,0.02343_wp,0.02396_wp,0.02274_wp, &
              &         0.07956_wp,0.07956_wp,0.28816_wp,0.07255_wp,0.12841_wp, &
              &         0.17467_wp,0.37511_wp,0.00000_wp,1.05993_wp,2.38199_wp, &
              &         3.77815_wp,5.58322_wp,7.44464_wp,10.0916_wp,0.03059_wp /)
            ssa_coeff=(/0.00003_wp,0.00046_wp,0.00085_wp,0.00078_wp,0.00089_wp, & 
              &         0.00421_wp,0.01514_wp,0.02350_wp,0.03534_wp,0.05654_wp, & 
              &         0.04674_wp,0.26871_wp,0.35651_wp,0.46646_wp,0.63783_wp, & 
              &         0.53863_wp,0.53863_wp,0.23202_wp,0.88896_wp,0.96005_wp, & 
              &         0.98711_wp,0.99458_wp,0.00000_wp,0.99939_wp,1.00000_wp, & 
              &         1.00000_wp,1.00000_wp,1.00000_wp,0.99985_wp,0.21407_wp /)
            asy_coeff=(/0.00130_wp,0.01104_wp,0.01818_wp,0.02279_wp,0.02758_wp, &  
              &         0.03904_wp,0.05055_wp,0.06068_wp,0.07248_wp,0.08676_wp, &  
              &         0.10219_wp,0.13204_wp,0.15465_wp,0.16988_wp,0.18459_wp, &  
              &         0.22292_wp,0.22292_wp,0.25503_wp,0.31480_wp,0.36593_wp, &  
              &         0.39708_wp,0.46656_wp,0.00000_wp,0.56909_wp,0.64595_wp, &  
              &         0.68314_wp,0.71287_wp,0.72566_wp,0.73141_wp,0.11137_wp /)
          CASE('seasb')
            ext_coeff=(/0.06207_wp,0.16129_wp,0.19155_wp,0.16685_wp,0.15007_wp, &  
              &         0.15534_wp,0.20853_wp,0.24018_wp,0.23625_wp,0.26487_wp, &  
              &         0.28126_wp,0.30344_wp,0.31832_wp,0.32455_wp,0.32558_wp, &  
              &         0.32477_wp,0.32477_wp,0.30436_wp,0.31335_wp,0.30331_wp, &  
              &         0.29818_wp,0.28915_wp,0.00000_wp,0.27465_wp,0.26579_wp, &  
              &         0.26162_wp,0.26030_wp,0.25550_wp,0.25153_wp,0.27339_wp /)
            ssa_coeff=(/0.11635_wp,0.41183_wp,0.41645_wp,0.36938_wp,0.37299_wp, &  
              &         0.63652_wp,0.79532_wp,0.80199_wp,0.83141_wp,0.84405_wp, &  
              &         0.72976_wp,0.92350_wp,0.91731_wp,0.92854_wp,0.95219_wp, &  
              &         0.84336_wp,0.84336_wp,0.67158_wp,0.96248_wp,0.97269_wp, &  
              &         0.98742_wp,0.99005_wp,0.00000_wp,0.99760_wp,1.00000_wp, &  
              &         1.00000_wp,1.00000_wp,0.99990_wp,0.99526_wp,0.85179_wp /)
            asy_coeff=(/0.23821_wp,0.58885_wp,0.65669_wp,0.73266_wp,0.77698_wp, &  
              &         0.80782_wp,0.78361_wp,0.77607_wp,0.80724_wp,0.79911_wp, &  
              &         0.81181_wp,0.79066_wp,0.77804_wp,0.77065_wp,0.76180_wp, &  
              &         0.77478_wp,0.77478_wp,0.87834_wp,0.79962_wp,0.78291_wp, &  
              &         0.78224_wp,0.77931_wp,0.00000_wp,0.79435_wp,0.80723_wp, &  
              &         0.81546_wp,0.82392_wp,0.82318_wp,0.82978_wp,0.78740_wp /)
          CASE('seasc') 
            ext_coeff=(/0.03895_wp,0.05011_wp,0.04807_wp,0.04617_wp,0.04545_wp, &  
              &         0.04738_wp,0.04650_wp,0.04578_wp,0.04537_wp,0.04479_wp, &  
              &         0.04428_wp,0.04386_wp,0.04350_wp,0.04333_wp,0.04315_wp, &  
              &         0.04276_wp,0.04276_wp,0.04231_wp,0.04200_wp,0.04219_wp, &  
              &         0.04190_wp,0.04125_wp,0.00000_wp,0.04102_wp,0.04070_wp, &  
              &         0.04044_wp,0.04023_wp,0.04018_wp,0.04010_wp,0.04461_wp /)
            ssa_coeff=(/0.34174_wp,0.49177_wp,0.49411_wp,0.48954_wp,0.49226_wp, &  
              &         0.54901_wp,0.58570_wp,0.57397_wp,0.59929_wp,0.59788_wp, &  
              &         0.53901_wp,0.70592_wp,0.68855_wp,0.71493_wp,0.78247_wp, &  
              &         0.66143_wp,0.66143_wp,0.54728_wp,0.81993_wp,0.86360_wp, &  
              &         0.93183_wp,0.94547_wp,0.00000_wp,0.98643_wp,0.99998_wp, &  
              &         0.99999_wp,0.99999_wp,0.99943_wp,0.97363_wp,0.64594_wp /)
            asy_coeff=(/0.56051_wp,0.87081_wp,0.89970_wp,0.93208_wp,0.94656_wp, &  
              &         0.93011_wp,0.89775_wp,0.90283_wp,0.90573_wp,0.90791_wp, &  
              &         0.93784_wp,0.88582_wp,0.89119_wp,0.88486_wp,0.86713_wp, &  
              &         0.90493_wp,0.90493_wp,0.95357_wp,0.88791_wp,0.87962_wp, &  
              &         0.86228_wp,0.85880_wp,0.00000_wp,0.85444_wp,0.85332_wp, &  
              &         0.85235_wp,0.85423_wp,0.85121_wp,0.85324_wp,0.89822_wp /)
          ! basalt  
          CASE ('asha')
            ext_coeff=(/0.01029_wp,0.06843_wp,0.10774_wp,0.05670_wp,0.03488_wp, &
              &         0.13370_wp,0.34300_wp,0.34064_wp,0.08029_wp,0.03469_wp, &
              &         0.04363_wp,0.07360_wp,0.11626_wp,0.15157_wp,0.19614_wp, &
              &         0.29664_wp,0.29664_wp,0.47732_wp,0.67157_wp,0.85705_wp, &
              &         0.99811_wp,1.17145_wp,1.23222_wp,1.18546_wp,0.97256_wp, &
              &         0.89711_wp,0.86410_wp,0.84482_wp,0.81639_wp,0.15076_wp /)
            ssa_coeff=(/0.04522_wp,0.01909_wp,0.02827_wp,0.04213_wp,0.18570_wp, &
              &         0.24148_wp,0.40871_wp,0.06948_wp,0.04807_wp,0.30705_wp, &
              &         0.71432_wp,0.89840_wp,0.94915_wp,0.96178_wp,0.96876_wp, &
              &         0.97180_wp,0.97180_wp,0.97868_wp,0.98481_wp,0.98661_wp, &
              &         0.98770_wp,0.98857_wp,0.98873_wp,0.98732_wp,0.98382_wp, &
              &         0.98090_wp,0.97536_wp,0.96398_wp,0.95173_wp,0.53271_wp /)
            asy_coeff=(/-0.08363_wp,0.03022_wp,0.04553_wp,0.06031_wp,0.09018_wp,&
              &         0.14041_wp,-0.01847_wp,0.11733_wp,0.16480_wp,0.22282_wp,&
              &         0.29370_wp,0.37114_wp,0.43783_wp,0.47564_wp,0.51169_wp, &
              &         0.57117_wp,0.57117_wp,0.63030_wp,0.66441_wp,0.68690_wp, &
              &         0.70188_wp,0.71261_wp,0.71195_wp,0.69389_wp,0.65498_wp, &
              &         0.66850_wp,0.71053_wp,0.74421_wp,0.77015_wp,0.26051_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                         ssa_param, asy_param, ext_default,      &
              &                         ssa_default, asy_default)
          CASE('ashb')
            ext_coeff=(/0.01951_wp,0.10213_wp,0.14399_wp,0.08370_wp,0.11445_wp, &
              &         0.26786_wp,0.31389_wp,0.22326_wp,0.07816_wp,0.08470_wp, &
              &         0.15609_wp,0.23764_wp,0.30306_wp,0.33958_wp,0.37135_wp, &
              &         0.40988_wp,0.40988_wp,0.43934_wp,0.44453_wp,0.43079_wp, &
              &         0.41236_wp,0.38347_wp,0.37315_wp,0.35638_wp,0.33205_wp, &
              &         0.32610_wp,0.31880_wp,0.31341_wp,0.30834_wp,0.23517_wp /)
            ssa_coeff=(/0.49167_wp,0.21336_wp,0.22553_wp,0.28751_wp,0.61474_wp, &
              &         0.53775_wp,0.55668_wp,0.25995_wp,0.20388_wp,0.67277_wp, &
              &         0.89484_wp,0.95720_wp,0.97298_wp,0.97636_wp,0.97749_wp, &
              &         0.97416_wp,0.97416_wp,0.97288_wp,0.97341_wp,0.97032_wp, &
              &         0.96831_wp,0.96485_wp,0.96279_wp,0.96052_wp,0.96031_wp, &
              &         0.95726_wp,0.94883_wp,0.92722_wp,0.89869_wp,0.68637_wp /)
            asy_coeff=(/0.02220_wp,0.23995_wp,0.30094_wp,0.40195_wp,0.46663_wp, &
              &         0.45066_wp,0.27680_wp,0.49241_wp,0.63824_wp,0.70305_wp, &
              &         0.71643_wp,0.72142_wp,0.72350_wp,0.72321_wp,0.72084_wp, &
              &         0.72410_wp,0.72410_wp,0.71521_wp,0.69830_wp,0.68482_wp, &
              &         0.68338_wp,0.68394_wp,0.68355_wp,0.70443_wp,0.74962_wp, &
              &         0.76357_wp,0.78195_wp,0.79732_wp,0.81653_wp,0.60757_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                         ssa_param, asy_param, ext_default,      &
              &                         ssa_default, asy_default)
          CASE('ashc')
            ext_coeff=(/0.04922_wp,0.11885_wp,0.13144_wp,0.09776_wp,0.14542_wp, &
              &         0.18597_wp,0.16511_wp,0.12868_wp,0.06916_wp,0.11047_wp, &
              &         0.15947_wp,0.18502_wp,0.19256_wp,0.19096_wp,0.18989_wp, &
              &         0.18575_wp,0.18575_wp,0.17730_wp,0.16892_wp,0.16353_wp, &
              &         0.16246_wp,0.15572_wp,0.15354_wp,0.14984_wp,0.14631_wp, &
              &         0.14404_wp,0.14129_wp,0.14030_wp,0.13874_wp,0.15756_wp /)
            ssa_coeff=(/0.73377_wp,0.38306_wp,0.37281_wp,0.45419_wp,0.68020_wp, &
              &         0.55185_wp,0.62946_wp,0.40361_wp,0.34481_wp,0.76108_wp, &
              &         0.89933_wp,0.94257_wp,0.95514_wp,0.95706_wp,0.95657_wp, &
              &         0.94488_wp,0.94488_wp,0.93778_wp,0.93875_wp,0.93437_wp, &
              &         0.93086_wp,0.92928_wp,0.92852_wp,0.92627_wp,0.92624_wp, &
              &         0.92086_wp,0.90516_wp,0.87042_wp,0.82430_wp,0.72664_wp /)
            asy_coeff=(/0.19415_wp,0.47105_wp,0.55620_wp,0.67797_wp,0.64138_wp, &
              &         0.60749_wp,0.51854_wp,0.72256_wp,0.85545_wp,0.84999_wp, &
              &         0.80227_wp,0.76362_wp,0.73888_wp,0.72626_wp,0.72222_wp, &
              &         0.71733_wp,0.71733_wp,0.71284_wp,0.71215_wp,0.72522_wp, &
              &         0.72654_wp,0.75040_wp,0.76042_wp,0.77519_wp,0.79337_wp, &
              &         0.80289_wp,0.81662_wp,0.83307_wp,0.85041_wp,0.72677_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                         ssa_param, asy_param, ext_default,      &
              &                         ssa_default, asy_default)                
          CASE('soot')  ! for coated soot with cmd 150 nm                                                    
            ext_coeff=(/0.02798_wp,0.04357_wp,0.07070_wp,0.06855_wp,0.06691_wp, & 
              &         0.07682_wp,0.10673_wp,0.14505_wp,0.10137_wp,0.10492_wp, & 
              &         0.15276_wp,0.16537_wp,0.16575_wp,0.18104_wp,0.22133_wp, &  
              &         0.41903_wp,0.41903_wp,0.44823_wp,0.62627_wp,0.78567_wp, & 
              &         1.07149_wp,1.60497_wp,1.74174_wp,3.20452_wp,3.96404_wp, &
              &         4.91859_wp,5.47225_wp,5.75379_wp,5.74157_wp,0.22133_wp /)
            ssa_coeff=(/0.00575_wp,0.01301_wp,0.01887_wp,0.02423_wp,0.04016_wp, &  
              &         0.08756_wp,0.07662_wp,0.07049_wp,0.15767_wp,0.19228_wp, & 
              &         0.25590_wp,0.37061_wp,0.49131_wp,0.51698_wp,0.55695_wp, & 
              &         0.50590_wp,0.50590_wp,0.78077_wp,0.81316_wp,0.82936_wp, & 
              &         0.84899_wp,0.86657_wp,0.86892_wp,0.87902_wp,0.87573_wp, & 
              &         0.86023_wp,0.84043_wp,0.81975_wp,0.79398_wp,0.55695_wp /)
            asy_coeff=(/0.05005_wp,0.08136_wp,0.10740_wp,0.11710_wp,0.13954_wp, &  
              &         0.18880_wp,0.19647_wp,0.20504_wp,0.24318_wp,0.25883_wp, &  
              &         0.30814_wp,0.34589_wp,0.37174_wp,0.38357_wp,0.40762_wp, &  
              &         0.46008_wp,0.46008_wp,0.50090_wp,0.53429_wp,0.55652_wp, & 
              &         0.58663_wp,0.62427_wp,0.63127_wp,0.68381_wp,0.70145_wp, & 
              &         0.72133_wp,0.73165_wp,0.74246_wp,0.75681_wp,0.40762_wp /)                    
        END SELECT                                                             
      ELSE ! Use the Mie calculation values for sphere                         
        SELECT CASE(TRIM(mode_name))                                         
          CASE ('dusta')                                                   
            ext_coeff=(/0.03914_wp,0.07147_wp,0.24330_wp,0.12663_wp,0.06118_wp, &
              &         0.08651_wp,0.36024_wp,0.21755_wp,0.08985_wp,0.12419_wp, &
              &         0.12099_wp,0.12680_wp,0.18674_wp,0.23555_wp,0.26436_wp, &
              &         0.34447_wp,0.34447_wp,0.48004_wp,0.64865_wp,0.75089_wp, &
              &         0.91529_wp,1.08391_wp,1.19900_wp,1.28417_wp,1.29794_wp, &
              &         1.21452_wp,1.11214_wp,1.05320_wp,1.01873_wp,0.17882_wp /)
            ssa_coeff=(/0.00657_wp,0.04161_wp,0.05717_wp,0.01334_wp,0.03427_wp, &
              &         0.38078_wp,0.17910_wp,0.08544_wp,0.05902_wp,0.09332_wp, &
              &         0.17613_wp,0.32187_wp,0.54596_wp,0.65046_wp,0.68157_wp, &
              &         0.78941_wp,0.78941_wp,0.96390_wp,0.98789_wp,0.99310_wp, &
              &         0.99480_wp,0.99563_wp,0.99629_wp,0.99433_wp,0.98715_wp, &
              &         0.96089_wp,0.89239_wp,0.81109_wp,0.72469_wp,0.32988_wp /)
            asy_coeff=(/0.02407_wp,0.06958_wp,0.05530_wp,0.09538_wp,0.13865_wp, &
              &         0.25604_wp,0.14630_wp,0.17188_wp,0.26890_wp,0.30976_wp, &
              &         0.35832_wp,0.43575_wp,0.48018_wp,0.50210_wp,0.52094_wp, &
              &         0.56232_wp,0.56232_wp,0.59984_wp,0.62604_wp,0.63771_wp, &
              &         0.65218_wp,0.66258_wp,0.66548_wp,0.66666_wp,0.66267_wp, &
              &         0.66872_wp,0.70569_wp,0.75166_wp,0.80737_wp,0.36098_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                         ssa_param, asy_param, ext_default,      &
              &                         ssa_default, asy_default)
          CASE('dustb')
            ext_coeff=(/0.06081_wp,0.13716_wp,0.22164_wp,0.09852_wp,0.07535_wp, &
              &         0.25305_wp,0.26751_wp,0.12866_wp,0.09643_wp,0.13481_wp, &
              &         0.16005_wp,0.19943_wp,0.26038_wp,0.27920_wp,0.28019_wp, &
              &         0.27941_wp,0.27941_wp,0.27049_wp,0.25445_wp,0.24655_wp, &
              &         0.23820_wp,0.22922_wp,0.22237_wp,0.21919_wp,0.21296_wp, &
              &         0.20839_wp,0.20466_wp,0.20184_wp,0.19986_wp,0.20667_wp /)
            ssa_coeff=(/0.14661_wp,0.38117_wp,0.38236_wp,0.16556_wp,0.27042_wp, &
              &         0.62925_wp,0.46296_wp,0.43010_wp,0.29684_wp,0.38326_wp, &
              &         0.49089_wp,0.61809_wp,0.68991_wp,0.70763_wp,0.71090_wp, &
              &         0.75494_wp,0.75494_wp,0.92513_wp,0.96389_wp,0.97568_wp, &
              &         0.97819_wp,0.97846_wp,0.97941_wp,0.97106_wp,0.94078_wp, &
              &         0.85709_wp,0.70975_wp,0.61249_wp,0.56440_wp,0.55522_wp /)
            asy_coeff=(/0.14609_wp,0.31221_wp,0.30373_wp,0.48861_wp,0.59303_wp, &
              &         0.53084_wp,0.50828_wp,0.61711_wp,0.77585_wp,0.79588_wp, &
              &         0.81065_wp,0.82759_wp,0.78554_wp,0.75922_wp,0.75937_wp, &
              &         0.74161_wp,0.74161_wp,0.67396_wp,0.66212_wp,0.66645_wp, &
              &         0.67522_wp,0.69524_wp,0.71865_wp,0.73385_wp,0.76588_wp, &
              &         0.80691_wp,0.86620_wp,0.90856_wp,0.93324_wp,0.73176_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                          ssa_param, asy_param, ext_default,     &
              &                          ssa_default, asy_default)
          CASE('dustc') 
            ext_coeff=(/0.07502_wp,0.12133_wp,0.12503_wp,0.06892_wp,0.07001_wp, &
              &         0.12644_wp,0.11870_wp,0.08043_wp,0.07531_wp,0.09520_wp, &
              &         0.10614_wp,0.11498_wp,0.11181_wp,0.10850_wp,0.10724_wp, &
              &         0.10466_wp,0.10466_wp,0.10231_wp,0.10095_wp,0.09986_wp, &
              &         0.09871_wp,0.09680_wp,0.09604_wp,0.09444_wp,0.09300_wp, &
              &         0.09182_wp,0.09069_wp,0.09011_wp,0.08950_wp,0.10506_wp /)
            ssa_coeff=(/0.30442_wp,0.49143_wp,0.51583_wp,0.31094_wp,0.40433_wp, &
              &         0.55449_wp,0.53170_wp,0.55951_wp,0.41027_wp,0.47723_wp, &
              &         0.53155_wp,0.57790_wp,0.56163_wp,0.55698_wp,0.55827_wp, &
              &         0.60553_wp,0.60553_wp,0.84528_wp,0.92381_wp,0.94859_wp, &
              &         0.95652_wp,0.95838_wp,0.96404_wp,0.94446_wp,0.88959_wp, &
              &         0.76147_wp,0.60998_wp,0.55905_wp,0.54875_wp,0.53805_wp /)
            asy_coeff=(/0.27896_wp,0.51733_wp,0.56182_wp,0.75763_wp,0.81537_wp, &
              &         0.69421_wp,0.71473_wp,0.78146_wp,0.90715_wp,0.90860_wp, &
              &         0.90365_wp,0.89605_wp,0.85583_wp,0.84299_wp,0.84791_wp, &
              &         0.83873_wp,0.83873_wp,0.76140_wp,0.74974_wp,0.74866_wp, &
              &         0.75812_wp,0.76595_wp,0.77397_wp,0.78639_wp,0.81417_wp, &
              &         0.86044_wp,0.91737_wp,0.94002_wp,0.94556_wp,0.84304_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
                     &                   ssa_param, asy_param, ext_default,     &
                     &                   ssa_default, asy_default)
      
          ! Seasalt optical properties 
          CASE ('seasa')
            ext_coeff=(/0.04822_wp,0.08393_wp,0.11915_wp,0.13776_wp,0.12483_wp, &
              &         0.05083_wp,0.02969_wp,0.03225_wp,0.02828_wp,0.02893_wp, &
              &         0.07075_wp,0.01846_wp,0.02343_wp,0.02396_wp,0.02274_wp, &
              &         0.07956_wp,0.07956_wp,0.28816_wp,0.07255_wp,0.12841_wp, &
              &         0.17467_wp,0.37511_wp,0.00000_wp,1.05993_wp,2.38199_wp, &
              &         3.77815_wp,5.58322_wp,7.44464_wp,10.0916_wp,0.03059_wp /)
            ssa_coeff=(/0.00003_wp,0.00046_wp,0.00085_wp,0.00078_wp,0.00089_wp, & 
              &         0.00421_wp,0.01514_wp,0.02350_wp,0.03534_wp,0.05654_wp, & 
              &         0.04674_wp,0.26871_wp,0.35651_wp,0.46646_wp,0.63783_wp, & 
              &         0.53863_wp,0.53863_wp,0.23202_wp,0.88896_wp,0.96005_wp, & 
              &         0.98711_wp,0.99458_wp,0.00000_wp,0.99939_wp,1.00000_wp, & 
              &         1.00000_wp,1.00000_wp,1.00000_wp,0.99985_wp,0.21407_wp /)
            asy_coeff=(/0.00130_wp,0.01104_wp,0.01818_wp,0.02279_wp,0.02758_wp, &  
              &         0.03904_wp,0.05055_wp,0.06068_wp,0.07248_wp,0.08676_wp, &  
              &         0.10219_wp,0.13204_wp,0.15465_wp,0.16988_wp,0.18459_wp, &  
              &         0.22292_wp,0.22292_wp,0.25503_wp,0.31480_wp,0.36593_wp, &  
              &         0.39708_wp,0.46656_wp,0.00000_wp,0.56909_wp,0.64595_wp, &  
              &         0.68314_wp,0.71287_wp,0.72566_wp,0.73141_wp,0.11137_wp /)
          CASE('seasb')
            ext_coeff=(/0.06207_wp,0.16129_wp,0.19155_wp,0.16685_wp,0.15007_wp, &  
              &         0.15534_wp,0.20853_wp,0.24018_wp,0.23625_wp,0.26487_wp, &  
              &         0.28126_wp,0.30344_wp,0.31832_wp,0.32455_wp,0.32558_wp, &  
              &         0.32477_wp,0.32477_wp,0.30436_wp,0.31335_wp,0.30331_wp, &  
              &         0.29818_wp,0.28915_wp,0.00000_wp,0.27465_wp,0.26579_wp, &  
              &         0.26162_wp,0.26030_wp,0.25550_wp,0.25153_wp,0.27339_wp /)
            ssa_coeff=(/0.11635_wp,0.41183_wp,0.41645_wp,0.36938_wp,0.37299_wp, &  
              &         0.63652_wp,0.79532_wp,0.80199_wp,0.83141_wp,0.84405_wp, &  
              &         0.72976_wp,0.92350_wp,0.91731_wp,0.92854_wp,0.95219_wp, &  
              &         0.84336_wp,0.84336_wp,0.67158_wp,0.96248_wp,0.97269_wp, &  
              &         0.98742_wp,0.99005_wp,0.00000_wp,0.99760_wp,1.00000_wp, &  
              &         1.00000_wp,1.00000_wp,0.99990_wp,0.99526_wp,0.85179_wp /)
            asy_coeff=(/0.23821_wp,0.58885_wp,0.65669_wp,0.73266_wp,0.77698_wp, &  
              &         0.80782_wp,0.78361_wp,0.77607_wp,0.80724_wp,0.79911_wp, &  
              &         0.81181_wp,0.79066_wp,0.77804_wp,0.77065_wp,0.76180_wp, &  
              &         0.77478_wp,0.77478_wp,0.87834_wp,0.79962_wp,0.78291_wp, &  
              &         0.78224_wp,0.77931_wp,0.00000_wp,0.79435_wp,0.80723_wp, &  
              &         0.81546_wp,0.82392_wp,0.82318_wp,0.82978_wp,0.78740_wp /)
          CASE('seasc') 
            ext_coeff=(/0.03895_wp,0.05011_wp,0.04807_wp,0.04617_wp,0.04545_wp, &  
              &         0.04738_wp,0.04650_wp,0.04578_wp,0.04537_wp,0.04479_wp, &  
              &         0.04428_wp,0.04386_wp,0.04350_wp,0.04333_wp,0.04315_wp, &  
              &         0.04276_wp,0.04276_wp,0.04231_wp,0.04200_wp,0.04219_wp, &  
              &         0.04190_wp,0.04125_wp,0.00000_wp,0.04102_wp,0.04070_wp, &  
              &         0.04044_wp,0.04023_wp,0.04018_wp,0.04010_wp,0.04461_wp /)
            ssa_coeff=(/0.34174_wp,0.49177_wp,0.49411_wp,0.48954_wp,0.49226_wp, &  
              &         0.54901_wp,0.58570_wp,0.57397_wp,0.59929_wp,0.59788_wp, &  
              &         0.53901_wp,0.70592_wp,0.68855_wp,0.71493_wp,0.78247_wp, &  
              &         0.66143_wp,0.66143_wp,0.54728_wp,0.81993_wp,0.86360_wp, &  
              &         0.93183_wp,0.94547_wp,0.00000_wp,0.98643_wp,0.99998_wp, &  
              &         0.99999_wp,0.99999_wp,0.99943_wp,0.97363_wp,0.64594_wp /)
            asy_coeff=(/0.56051_wp,0.87081_wp,0.89970_wp,0.93208_wp,0.94656_wp, &  
              &         0.93011_wp,0.89775_wp,0.90283_wp,0.90573_wp,0.90791_wp, &  
              &         0.93784_wp,0.88582_wp,0.89119_wp,0.88486_wp,0.86713_wp, &  
              &         0.90493_wp,0.90493_wp,0.95357_wp,0.88791_wp,0.87962_wp, &  
              &         0.86228_wp,0.85880_wp,0.00000_wp,0.85444_wp,0.85332_wp, &  
              &         0.85235_wp,0.85423_wp,0.85121_wp,0.85324_wp,0.89822_wp /)
          ! basaltic volcanic ash  
          CASE ('asha')
            ext_coeff=(/0.01029_wp,0.06843_wp,0.10774_wp,0.05670_wp,0.03488_wp, &
              &         0.13370_wp,0.34300_wp,0.34064_wp,0.08029_wp,0.03469_wp, &
              &         0.04363_wp,0.07360_wp,0.11626_wp,0.15157_wp,0.19614_wp, &
              &         0.29664_wp,0.29664_wp,0.47732_wp,0.67157_wp,0.85705_wp, &
              &         0.99811_wp,1.17145_wp,1.23222_wp,1.18546_wp,0.97256_wp, &
              &         0.89711_wp,0.86410_wp,0.84482_wp,0.81639_wp,0.15076_wp /)
            ssa_coeff=(/0.04522_wp,0.01909_wp,0.02827_wp,0.04213_wp,0.18570_wp, &
              &         0.24148_wp,0.40871_wp,0.06948_wp,0.04807_wp,0.30705_wp, &
              &         0.71432_wp,0.89840_wp,0.94915_wp,0.96178_wp,0.96876_wp, &
              &         0.97180_wp,0.97180_wp,0.97868_wp,0.98481_wp,0.98661_wp, &
              &         0.98770_wp,0.98857_wp,0.98873_wp,0.98732_wp,0.98382_wp, &
              &         0.98090_wp,0.97536_wp,0.96398_wp,0.95173_wp,0.53271_wp /)
            asy_coeff=(/-0.08363_wp,0.03022_wp,0.04553_wp,0.06031_wp,0.09018_wp,&
              &         0.14041_wp,-0.01847_wp,0.11733_wp,0.16480_wp,0.22282_wp,&
              &         0.29370_wp,0.37114_wp,0.43783_wp,0.47564_wp,0.51169_wp, &
              &         0.57117_wp,0.57117_wp,0.63030_wp,0.66441_wp,0.68690_wp, &
              &         0.70188_wp,0.71261_wp,0.71195_wp,0.69389_wp,0.65498_wp, &
              &         0.66850_wp,0.71053_wp,0.74421_wp,0.77015_wp,0.26051_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                         ssa_param, asy_param, ext_default,      &
              &                         ssa_default, asy_default)               
          CASE('ashb')                                                           
            ext_coeff=(/0.01951_wp,0.10213_wp,0.14399_wp,0.08370_wp,0.11445_wp, &
              &         0.26786_wp,0.31389_wp,0.22326_wp,0.07816_wp,0.08470_wp, &
              &         0.15609_wp,0.23764_wp,0.30306_wp,0.33958_wp,0.37135_wp, &
              &         0.40988_wp,0.40988_wp,0.43934_wp,0.44453_wp,0.43079_wp, &
              &         0.41236_wp,0.38347_wp,0.37315_wp,0.35638_wp,0.33205_wp, &
              &         0.32610_wp,0.31880_wp,0.31341_wp,0.30834_wp,0.23517_wp /)
            ssa_coeff=(/0.49167_wp,0.21336_wp,0.22553_wp,0.28751_wp,0.61474_wp, &
              &         0.53775_wp,0.55668_wp,0.25995_wp,0.20388_wp,0.67277_wp, &
              &         0.89484_wp,0.95720_wp,0.97298_wp,0.97636_wp,0.97749_wp, &
              &         0.97416_wp,0.97416_wp,0.97288_wp,0.97341_wp,0.97032_wp, &
              &         0.96831_wp,0.96485_wp,0.96279_wp,0.96052_wp,0.96031_wp, &
              &         0.95726_wp,0.94883_wp,0.92722_wp,0.89869_wp,0.68637_wp /)
            asy_coeff=(/0.02220_wp,0.23995_wp,0.30094_wp,0.40195_wp,0.46663_wp, &
              &         0.45066_wp,0.27680_wp,0.49241_wp,0.63824_wp,0.70305_wp, &
              &         0.71643_wp,0.72142_wp,0.72350_wp,0.72321_wp,0.72084_wp, &
              &         0.72410_wp,0.72410_wp,0.71521_wp,0.69830_wp,0.68482_wp, &
              &         0.68338_wp,0.68394_wp,0.68355_wp,0.70443_wp,0.74962_wp, &
              &         0.76357_wp,0.78195_wp,0.79732_wp,0.81653_wp,0.60757_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,         &
              &                         ssa_param, asy_param, ext_default,      &
              &                         ssa_default, asy_default)               
          CASE('ashc')                                                           
            ext_coeff=(/0.04922_wp,0.11885_wp,0.13144_wp,0.09776_wp,0.14542_wp, &
              &         0.18597_wp,0.16511_wp,0.12868_wp,0.06916_wp,0.11047_wp, &
              &         0.15947_wp,0.18502_wp,0.19256_wp,0.19096_wp,0.18989_wp, &
              &         0.18575_wp,0.18575_wp,0.17730_wp,0.16892_wp,0.16353_wp, &
              &         0.16246_wp,0.15572_wp,0.15354_wp,0.14984_wp,0.14631_wp, &
              &         0.14404_wp,0.14129_wp,0.14030_wp,0.13874_wp,0.15756_wp /)
            ssa_coeff=(/0.73377_wp,0.38306_wp,0.37281_wp,0.45419_wp,0.68020_wp, &
              &         0.55185_wp,0.62946_wp,0.40361_wp,0.34481_wp,0.76108_wp, &
              &         0.89933_wp,0.94257_wp,0.95514_wp,0.95706_wp,0.95657_wp, &
              &         0.94488_wp,0.94488_wp,0.93778_wp,0.93875_wp,0.93437_wp, &
              &         0.93086_wp,0.92928_wp,0.92852_wp,0.92627_wp,0.92624_wp, &
              &         0.92086_wp,0.90516_wp,0.87042_wp,0.82430_wp,0.72664_wp /)
            asy_coeff=(/0.19415_wp,0.47105_wp,0.55620_wp,0.67797_wp,0.64138_wp, &
              &         0.60749_wp,0.51854_wp,0.72256_wp,0.85545_wp,0.84999_wp, &
              &         0.80227_wp,0.76362_wp,0.73888_wp,0.72626_wp,0.72222_wp, &
              &         0.71733_wp,0.71733_wp,0.71284_wp,0.71215_wp,0.72522_wp, &
              &         0.72654_wp,0.75040_wp,0.76042_wp,0.77519_wp,0.79337_wp, &
              &         0.80289_wp,0.81662_wp,0.83307_wp,0.85041_wp,0.72677_wp /)
            CALL art_read_visparams_xml(jg, TRIM(mode_name), ext_param,        &
              &                         ssa_param, asy_param, ext_default,     &
              &                         ssa_default, asy_default)               
          CASE('soot') ! for coated soot with cmd 150 nm                                                     
            ext_coeff=(/0.02798_wp,0.04357_wp,0.07070_wp,0.06855_wp,0.06691_wp, & 
              &         0.07682_wp,0.10673_wp,0.14505_wp,0.10137_wp,0.10492_wp, & 
              &         0.15276_wp,0.16537_wp,0.16575_wp,0.18104_wp,0.22133_wp, &  
              &         0.41903_wp,0.41903_wp,0.44823_wp,0.62627_wp,0.78567_wp, & 
              &         1.07149_wp,1.60497_wp,1.74174_wp,3.20452_wp,3.96404_wp, &
              &         4.91859_wp,5.47225_wp,5.75379_wp,5.74157_wp,0.22133_wp /)
            ssa_coeff=(/0.00575_wp,0.01301_wp,0.01887_wp,0.02423_wp,0.04016_wp, &  
              &         0.08756_wp,0.07662_wp,0.07049_wp,0.15767_wp,0.19228_wp, & 
              &         0.25590_wp,0.37061_wp,0.49131_wp,0.51698_wp,0.55695_wp, & 
              &         0.50590_wp,0.50590_wp,0.78077_wp,0.81316_wp,0.82936_wp, & 
              &         0.84899_wp,0.86657_wp,0.86892_wp,0.87902_wp,0.87573_wp, & 
              &         0.86023_wp,0.84043_wp,0.81975_wp,0.79398_wp,0.55695_wp /)
            asy_coeff=(/0.05005_wp,0.08136_wp,0.10740_wp,0.11710_wp,0.13954_wp, &  
              &         0.18880_wp,0.19647_wp,0.20504_wp,0.24318_wp,0.25883_wp, &  
              &         0.30814_wp,0.34589_wp,0.37174_wp,0.38357_wp,0.40762_wp, &  
              &         0.46008_wp,0.46008_wp,0.50090_wp,0.53429_wp,0.55652_wp, & 
              &         0.58663_wp,0.62427_wp,0.63127_wp,0.68381_wp,0.70145_wp, & 
              &         0.72133_wp,0.73165_wp,0.74246_wp,0.75681_wp,0.40762_wp /)
        END SELECT                                                             
      ENDIF                                                                      
    ENDIF !.NOT. is_aerodyn
  END SUBROUTINE art_init_radiation                                            
END MODULE mo_art_init_radiation                                               
