! ======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE ISOROPIA
! *** THIS SUBROUTINE IS THE MASTER ROUTINE FOR THE ISORROPIA
!     THERMODYNAMIC EQUILIBRIUM AEROSOL MODEL (VERSION 1.1 and above)
!
! ======================== ARGUMENTS / USAGE ===========================
!
!  INPUT:
!  1. [WI]
!     DOUBLE PRECISION array of length [8].
!     Concentrations, expressed in moles/m3. Depending on the type of
!     problem solved (specified in CNTRL(1)), WI contains either
!     GAS+AEROSOL or AEROSOL only concentratios.
!     WI(1) - sodium
!     WI(2) - sulfate
!     WI(3) - ammonium
!     WI(4) - nitrate
!     WI(5) - chloride
!     WI(6) - calcium
!     WI(7) - potassium
!     WI(8) - magnesium
!
!  2. [RHI]
!     DOUBLE PRECISION variable.
!     Ambient relative humidity expressed on a (0,1) scale.
!
!  3. [TEMPI]
!     DOUBLE PRECISION variable.
!     Ambient temperature expressed in Kelvins.
!
!  4. [CNTRL]
!     DOUBLE PRECISION array of length [2].
!     Parameters that control the type of problem solved.
!
!     CNTRL(1): Defines the type of problem solved.
!     0 - Forward problem is solved. In this case, array WI contains
!         GAS and AEROSOL concentrations together.
!     1 - Reverse problem is solved. In this case, array WI contains
!         AEROSOL concentrations only.
!
!     CNTRL(2): Defines the state of the aerosol
!     0 - The aerosol can have both solid+liquid phases (deliquescent)
!     1 - The aerosol is in only liquid state (metastable aerosol)
!
!  OUTPUT:
!  1. [WT]
!     DOUBLE PRECISION array of length [8].
!     Total concentrations (GAS+AEROSOL) of species, expressed in moles/m3.
!     If the foreward probelm is solved (CNTRL(1)=0), array WT is
!     identical to array WI.
!     WT(1) - total sodium
!     WT(2) - total sulfate
!     WT(3) - total ammonium
!     WT(4) - total nitrate
!     WT(5) - total chloride
!     WT(6) - total calcium
!     WT(7) - total potassium
!     WT(8) - total magnesium
!
!  2. [GAS]
!     DOUBLE PRECISION array of length [03].
!     Gaseous species concentrations, expressed in moles/m3.
!     GAS(1) - NH3
!     GAS(2) - HNO3
!     GAS(3) - HCl
!
!  3. [AERLIQ]
!     DOUBLE PRECISION array of length [15].
!     Liquid aerosol species concentrations, expressed in moles/m3.
!     AERLIQ(01) - H+(aq)
!     AERLIQ(02) - Na+(aq)
!     AERLIQ(03) - NH4+(aq)
!     AERLIQ(04) - Cl-(aq)
!     AERLIQ(05) - SO4--(aq)
!     AERLIQ(06) - HSO4-(aq)
!     AERLIQ(07) - NO3-(aq)
!     AERLIQ(08) - H2O
!     AERLIQ(09) - NH3(aq) (undissociated)
!     AERLIQ(10) - HNCl(aq) (undissociated)
!     AERLIQ(11) - HNO3(aq) (undissociated)
!     AERLIQ(12) - OH-(aq)
!     AERLIQ(13) - Ca2+(aq)
!     AERLIQ(14) - K+(aq)
!     AERLIQ(15) - Mg2+(aq)
!
!  4. [AERSLD]
!     DOUBLE PRECISION array of length [19].
!     Solid aerosol species concentrations, expressed in moles/m3.
!     AERSLD(01) - NaNO3(s)
!     AERSLD(02) - NH4NO3(s)
!     AERSLD(03) - NaCl(s)
!     AERSLD(04) - NH4Cl(s)
!     AERSLD(05) - Na2SO4(s)
!     AERSLD(06) - (NH4)2SO4(s)
!     AERSLD(07) - NaHSO4(s)
!     AERSLD(08) - NH4HSO4(s)
!     AERSLD(09) - (NH4)4H(SO4)2(s)
!     AERSLD(10) - CaSO4(s)
!     AERSLD(11) - Ca(NO3)2(s)
!     AERSLD(12) - CaCl2(s)
!     AERSLD(13) - K2SO4(s)
!     AERSLD(14) - KHSO4(s)
!     AERSLD(15) - KNO3(s)
!     AERSLD(16) - KCl(s)
!     AERSLD(17) - MgSO4(s)
!     AERSLD(18) - Mg(NO3)2(s)
!     AERSLD(19) - MgCl2(s)
!
!  5. [SCASI]
!     CHARACTER*15 variable.
!     Returns the subcase which the input corresponds to.
!
!  6. [OTHER]
!     DOUBLE PRECISION array of length [9].
!     Returns solution information.
!
!     OTHER(1): Shows if aerosol water exists.
!     0 - Aerosol is WET
!     1 - Aerosol is DRY
!
!     OTHER(2): Aerosol Sulfate ratio, defined as (in moles/m3) :
!               (total ammonia + total Na) / (total sulfate)
!
!     OTHER(3): Sulfate ratio based on aerosol properties that defines
!               a sulfate poor system:
!               (aerosol ammonia + aerosol Na) / (aerosol sulfate)
!
!     OTHER(4): Aerosol sodium ratio, defined as (in moles/m3) :
!               (total Na) / (total sulfate)
!
!     OTHER(5): Ionic strength of the aqueous aerosol (if it exists).
!
!     OTHER(6): Total number of calls to the activity coefficient
!               calculation subroutine.
!
!     OTHER(7): Sulfate ratio with crustal species, defined as (in moles/m3) :
!               (total ammonia + total crustal species + total Na) / (total sulfate)
!
!     OTHER(8): Crustal species + sodium ratio, defined as (in moles/m3) :
!               (total crustal species + total Na) / (total sulfate)
!
!     OTHER(9): Crustal species ratio, defined as (in moles/m3) :
!               (total crustal species) / (total sulfate)
!
!=======================================================================
!
!
! ICON
!
! ---------------------------------------------------------------
! Copyright (C) 2004-2024, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
! Contact information: icon-model.org
!
! See AUTHORS.TXT for a list of authors
! See LICENSES/ for license information
! SPDX-License-Identifier: BSD-3-Clause
! ---------------------------------------------------------------

MODULE art_isorropia

PRIVATE

PUBLIC :: isoropia

CONTAINS

SUBROUTINE isoropia ( wi, rhi, tempi, cntrl, wt, gas, aerliq, aersld, scasi, other)
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER, PARAMETER :: nctrl = 2, nother = 9
  DOUBLE PRECISION,  INTENT(inout) :: wi(8)
  DOUBLE PRECISION,  INTENT(inout) :: rhi
  DOUBLE PRECISION,  INTENT(inout) :: tempi
  DOUBLE PRECISION,  INTENT(inout) :: cntrl(2)
  DOUBLE PRECISION,  INTENT(inout) :: wt(8)
  DOUBLE PRECISION,  INTENT(inout) :: gas(3)
  DOUBLE PRECISION,  INTENT(inout) :: aerliq(15)
  DOUBLE PRECISION,  INTENT(inout) :: aersld(19)
  character(len=15), INTENT(inout) :: scasi
  DOUBLE PRECISION,  INTENT(inout) :: other(nother)
  INTEGER :: i

  !
  ! *** PROBLEM TYPE (0=FOREWARD, 1=REVERSE) ******************************
  !
  iprob = nint(cntrl(1))
  !
  ! *** AEROSOL STATE (0=SOLID+LIQUID, 1=METASTABLE) **********************
  !
  metstbl = nint(cntrl(2))
  !
  ! *** SOLVE FOREWARD PROBLEM ********************************************
  !
  IF (iprob==0) THEN
    IF (wi(1)+wi(2)+wi(3)+wi(4)+wi(5)+wi(6)+wi(7)+wi(8)<=tiny) THEN !Everything=0
      CALL init1(wi, rhi, tempi)
    ELSE IF (wi(1)+wi(4)+wi(5)+wi(6)+wi(7)+wi(8)<=tiny) THEN !Ca,K,Mg,Na,Cl,NO3=0
      CALL isrp1f(wi, rhi, tempi)
    ELSE IF (wi(1)+wi(5)+wi(6)+wi(7)+wi(8)<=tiny) THEN !Ca,K,Mg,Na,Cl=0
      CALL isrp2f(wi, rhi, tempi)
    ELSE IF (wi(6)+wi(7)+wi(8)<=tiny) THEN !Ca,K,Mg=0
      CALL isrp3f(wi, rhi, tempi)
    ELSE
      CALL isrp4f(wi, rhi, tempi)
    END IF
    !
    ! *** SOLVE REVERSE PROBLEM *********************************************
    !
  ELSE
    IF (wi(1)+wi(2)+wi(3)+wi(4)+wi(5)+wi(6)+wi(7)+wi(8)<=tiny) THEN !Everything=0
      CALL init1(wi, rhi, tempi)
    ELSE IF (wi(1)+wi(4)+wi(5)+wi(6)+wi(7)+wi(8)<=tiny) THEN !Ca,K,Mg,Na,Cl,NO3=0
      CALL isrp1r(wi, rhi, tempi)
    ELSE IF (wi(1)+wi(5)+wi(6)+wi(7)+wi(8)<=tiny) THEN !Ca,K,Mg,Na,Cl=0
      CALL isrp2r(wi, rhi, tempi)
    ELSE IF (wi(6)+wi(7)+wi(8)<=tiny) THEN !Ca,K,Mg=0
      CALL isrp3r(wi, rhi, tempi)
    ELSE
      CALL isrp4r(wi, rhi, tempi)
    END IF
  END IF
  !
  ! *** ADJUST MASS BALANCE ***********************************************
  !
  IF (nadj==1) CALL adjust(wi)
  !cC
  !cC *** IF METASTABLE AND NO WATER - RESOLVE AS NORMAL ********************
  !cC
  !c      IF (WATER.LE.TINY .AND. METSTBL.EQ.1) THEN
  !c         METSTBL = 0
  !c         GOTO 50
  !c      ENDIF

  !
  ! *** SAVE RESULTS TO ARRAYS (units = mole/m3) ****************************
  !
  gas(1) = gnh3 ! Gaseous aerosol species
  gas(2) = ghno3
  gas(3) = ghcl
  !
  DO i = 1, 7 ! Liquid aerosol species
    aerliq(i) = molal(i)
  END DO
  DO i = 1, ngasaq
    aerliq(7+1+i) = gasaq(i)
  END DO
  aerliq(7+1) = water*1.0D3/18.0D0
  aerliq(7+ngasaq+2) = coh
  !
  DO i = 8, 10 ! Liquid aerosol species
    aerliq(i+5) = molal(i)
  END DO
  !
  aersld(1) = cnano3 ! Solid aerosol species
  aersld(2) = cnh4no3
  aersld(3) = cnacl
  aersld(4) = cnh4cl
  aersld(5) = cna2so4
  aersld(6) = cnh42s4
  aersld(7) = cnahso4
  aersld(8) = cnh4hs4
  aersld(9) = clc
  aersld(10) = ccaso4
  aersld(11) = ccano32
  aersld(12) = ccacl2
  aersld(13) = ck2so4
  aersld(14) = ckhso4
  aersld(15) = ckno3
  aersld(16) = ckcl
  aersld(17) = cmgso4
  aersld(18) = cmgno32
  aersld(19) = cmgcl2
  !
  IF (water<=tiny) THEN ! Dry flag
    other(1) = 1.D0
  ELSE
    other(1) = 0.D0
  END IF
  !
  other(2) = sulrat ! Other stuff
  other(3) = sulratw
  other(4) = sodrat
  other(5) = ionic
  other(6) = iclact
  other(7) = so4rat
  other(8) = crnarat
  other(9) = crrat
  !
  scasi = scase
  !
  wt(1) = wi(1) ! Total gas+aerosol phase
  wt(2) = wi(2)
  wt(3) = wi(3)
  wt(4) = wi(4)
  wt(5) = wi(5)
  wt(6) = wi(6)
  wt(7) = wi(7)
  wt(8) = wi(8)


  IF (iprob>0 .AND. water>tiny) THEN
    wt(3) = wt(3) + gnh3
    wt(4) = wt(4) + ghno3
    wt(5) = wt(5) + ghcl
  END IF
  !
  RETURN
  !
END SUBROUTINE isoropia
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE SETPARM
! *** THIS SUBROUTINE REDEFINES THE SOLUTION PARAMETERS OF ISORROPIA
!
! ======================== ARGUMENTS / USAGE ===========================
!
! *** NOTE: IF NEGATIVE VALUES ARE GIVEN FOR A PARAMETER, IT IS
!     IGNORED AND THE CURRENT VALUE IS USED INSTEAD.
!
!  INPUT:
!  1. [WFTYPI]
!     INTEGER variable.
!     Defines the type of weighting algorithm for the solution in Mutual
!     Deliquescence Regions (MDR's):
!     0 - MDR's are assumed dry. This is equivalent to the approach
!         used by SEQUILIB.
!     1 - The solution is assumed "half" dry and "half" wet throughout
!         the MDR.
!     2 - The solution is a relative-humidity weighted mean of the
!         dry and wet solutions (as defined in Nenes et al., 1998)
!
!  2. [IACALCI]
!     INTEGER variable.
!     Method of activity coefficient calculation:
!     0 - Calculate coefficients during runtime
!     1 - Use precalculated tables
!
!  3. [EPSI]
!     DOUBLE PRECITION variable.
!     Defines the convergence criterion for all iterative processes
!     in ISORROPIA, except those for activity coefficient calculations
!     (EPSACTI controls that).
!
!  4. [MAXITI]
!     INTEGER variable.
!     Defines the maximum number of iterations for all iterative
!     processes in ISORROPIA, except for activity coefficient calculations
!     (NSWEEPI controls that).
!
!  5. [NSWEEPI]
!     INTEGER variable.
!     Defines the maximum number of iterations for activity coefficient
!     calculations.
!
!  6. [EPSACTI]
!     DOUBLE PRECISION variable.
!     Defines the convergence criterion for activity coefficient
!     calculations.
!
!  7. [NDIV]
!     INTEGER variable.
!     Defines the number of subdivisions needed for the initial root
!     tracking for the bisection method. Usually this parameter should
!     not be altered, but is included for completeness.
!
!  8. [NADJ]
!     INTEGER variable.
!     Forces the solution obtained to satisfy total mass balance
!     to machine precision
!     0 - No adjustment done (default)
!     1 - Do adjustment
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS   
!
!=======================================================================
!
SUBROUTINE setparm ( wftypi, iacalci, epsi, maxiti, nsweepi, epsacti, ndivi, nadji )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER, INTENT(inout) :: wftypi
  INTEGER, INTENT(inout) :: iacalci
  DOUBLE PRECISION, INTENT(inout) :: epsi
  INTEGER, INTENT(inout) :: maxiti
  INTEGER, INTENT(inout) :: nsweepi
  DOUBLE PRECISION, INTENT(inout) :: epsacti
  INTEGER, INTENT(inout) :: ndivi
  INTEGER, INTENT(inout) :: nadji
  !
  ! *** SETUP SOLUTION PARAMETERS *****************************************
  !
  IF (wftypi>=0) wftyp = wftypi
  IF (iacalci>=0) iacalc = iacalci
  IF (epsi>=zero) eps = epsi
  IF (maxiti>0) maxit = maxiti
  IF (nsweepi>0) nsweep = nsweepi
  IF (epsacti>=zero) epsact = epsacti
  IF (ndivi>0) ndiv = ndivi
  IF (nadji>=0) nadj = nadji
  !
  RETURN
END SUBROUTINE setparm

!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE GETPARM
! *** THIS SUBROUTINE OBTAINS THE CURRENT VAULES OF THE SOLUTION
!     PARAMETERS OF ISORROPIA
!
! ======================== ARGUMENTS / USAGE ===========================
!
! *** THE PARAMETERS ARE THOSE OF SUBROUTINE SETPARM
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATED BY SHANNON CAPPS         
!
!=======================================================================
!
SUBROUTINE getparm ( wftypi, iacalci, epsi, maxiti, nsweepi, epsacti, ndivi, nadji )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER, INTENT(inout) :: wftypi
  INTEGER, INTENT(inout) :: iacalci
  DOUBLE PRECISION, INTENT(inout) :: epsi
  INTEGER, INTENT(inout) :: maxiti
  INTEGER, INTENT(inout) :: nsweepi
  DOUBLE PRECISION, INTENT(inout) :: epsacti
  INTEGER, INTENT(inout) :: ndivi
  INTEGER, INTENT(inout) :: nadji

  !
  ! *** GET SOLUTION PARAMETERS *******************************************
  !
  wftypi = wftyp
  iacalci = iacalc
  epsi = eps
  maxiti = maxit
  nsweepi = nsweep
  epsacti = epsact
  ndivi = ndiv
  nadji = nadj
  !
  RETURN
END SUBROUTINE getparm
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE INIT1
! *** THIS SUBROUTINE INITIALIZES ALL GLOBAL VARIABLES FOR AMMONIUM
!     SULFATE AEROSOL SYSTEMS (SUBROUTINE ISRP1)
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATED BY SHANNON CAPPS
!
!=======================================================================
!

SUBROUTINE init1 ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  DOUBLE PRECISION :: coef
  INTEGER :: i
  INTEGER :: irh
  DOUBLE PRECISION :: t0
  DOUBLE PRECISION :: t0t
  DOUBLE PRECISION :: tcf

  REAL, PARAMETER :: ln10 = 2.3025851
  !
  ! *** SAVE INPUT VARIABLES IN COMMON BLOCK ******************************
  !
  IF (iprob==0) THEN ! FORWARD CALCULATION
    DO i = 1, ncomp
      w(i) = max(wi(i), tiny)
    END DO
  ELSE
    DO i = 1, ncomp ! REVERSE CALCULATION
      waer(i) = max(wi(i), tiny)
      w(i) = zero
    END DO
  END IF
  rh = rhi
  temp = tempi
  !
  ! *** CALCULATE EQUILIBRIUM CONSTANTS ***********************************
  !
  xk1 = 1.015E-2 ! HSO4(aq)         <==> H(aq)     + SO4(aq)
  xk21 = 57.639 ! NH3(g)           <==> NH3(aq)
  xk22 = 1.805E-5 ! NH3(aq)          <==> NH4(aq)   + OH(aq)
  xk7 = 1.817 ! (NH4)2SO4(s)     <==> 2*NH4(aq) + SO4(aq)
  xk12 = 1.382E2 ! NH4HSO4(s)       <==> NH4(aq)   + HSO4(aq)
  xk13 = 29.268 ! (NH4)3H(SO4)2(s) <==> 3*NH4(aq) + HSO4(aq) + SO4(aq)
  xkw = 1.010E-14 ! H2O              <==> H(aq)     + OH(aq)
  !
  IF (int(temp)/=298) THEN ! FOR T != 298K or 298.15K
    t0 = 298.15
    t0t = t0/temp
    coef = 1.0 + log(t0t) - t0t
    xk1 = xk1*exp(8.85*(t0t-1.0)+25.140*coef)
    xk21 = xk21*exp(13.79*(t0t-1.0)-5.393*coef)
    xk22 = xk22*exp(-1.50*(t0t-1.0)+26.920*coef)
    xk7 = xk7*exp(-2.65*(t0t-1.0)+38.570*coef)
    xk12 = xk12*exp(-2.87*(t0t-1.0)+15.830*coef)
    xk13 = xk13*exp(-5.19*(t0t-1.0)+54.400*coef)
    xkw = xkw*exp(-22.52*(t0t-1.0)+26.920*coef)
  END IF
  xk2 = xk21*xk22
  !
  ! *** CALCULATE DELIQUESCENCE RELATIVE HUMIDITIES (UNICOMPONENT) ********
  !
  drh2so4 = 0.0000D0
  drnh42s4 = 0.7997D0
  drnh4hs4 = 0.4000D0
  drlc = 0.6900D0
  IF (int(temp)/=298) THEN
    t0 = 298.15D0
    tcf = 1.0/temp - 1.0/t0
    drnh42s4 = drnh42s4*exp(80.*tcf)
    drnh4hs4 = drnh4hs4*exp(384.*tcf)
    drlc = drlc*exp(186.*tcf)
  END IF
  !
  ! *** CALCULATE MUTUAL DELIQUESCENCE RELATIVE HUMIDITIES ****************
  !
  drmlcab = 0.3780D0 ! (NH4)3H(SO4)2 & NH4HSO4
  drmlcas = 0.6900D0 ! (NH4)3H(SO4)2 & (NH4)2SO4
  !CC      IF (INT(TEMP) .NE. 298) THEN      ! For the time being.
  !CC         T0       = 298.15d0
  !CC         TCF      = 1.0/TEMP - 1.0/T0
  !CC         DRMLCAB  = DRMLCAB*EXP(507.506*TCF)
  !CC         DRMLCAS  = DRMLCAS*EXP(133.865*TCF)
  !CC      ENDIF
  !
  ! *** LIQUID PHASE ******************************************************
  !
  chno3 = zero
  chcl = zero
  ch2so4 = zero
  coh = zero
  water = tiny
  !
  DO i = 1, npair
    molalr(i) = zero
    gama(i) = 0.1
    gamin(i) = great
    gamou(i) = great
    m0(i) = 1D5
  END DO
  !
  DO i = 1, npair
    gama(i) = 0.1D0
  END DO
  !
  DO i = 1, nions
    molal(i) = zero
  END DO
  coh = zero
  !
  DO i = 1, ngasaq
    gasaq(i) = zero
  END DO
  !
  ! *** SOLID PHASE *******************************************************
  !
  cnh42s4 = zero
  cnh4hs4 = zero
  cnacl = zero
  cna2so4 = zero
  cnano3 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnahso4 = zero
  clc = zero
  ccaso4 = zero
  ccano32 = zero
  ccacl2 = zero
  ck2so4 = zero
  ckhso4 = zero
  ckno3 = zero
  ckcl = zero
  cmgso4 = zero
  cmgno32 = zero
  cmgcl2 = zero
  !
  ! *** GAS PHASE *********************************************************
  !
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  ! *** CALCULATE ZSR PARAMETERS ******************************************
  !
  irh = min(int(rh*nzsr+0.5), nzsr) ! Position in ZSR arrays
  irh = max(irh, 1)
  !
  !      M0(01) = AWSC(IRH)      ! NACl
  !      IF (M0(01) .LT. 100.0) THEN
  !         IC = M0(01)
  !         CALL KMTAB(IC,298.0,     GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                            XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(01) = M0(01)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !C
  !      M0(02) = AWSS(IRH)      ! (NA)2SO4
  !      IF (M0(02) .LT. 100.0) THEN
  !         IC = 3.0*M0(02)
  !         CALL KMTAB(IC,298.0,     XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(02) = M0(02)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !C
  !      M0(03) = AWSN(IRH)      ! NANO3
  !      IF (M0(03) .LT. 100.0) THEN
  !         IC = M0(03)
  !         CALL KMTAB(IC,298.0,     XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(03) = M0(03)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !C
  m0(04) = awas(irh) ! (NH4)2SO4
  !C      IF (M0(04) .LT. 100.0) THEN
  !C         IC = 3.0*M0(04)
  ! C        CALL KMTAB(IC,298.0,     XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(04) = M0(04)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  !      M0(05) = AWAN(IRH)      ! NH4NO3
  !      IF (M0(05) .LT. 100.0) THEN
  !         IC     = M0(05)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(05) = M0(05)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !C
  !      M0(06) = AWAC(IRH)      ! NH4CL
  !      IF (M0(06) .LT. 100.0) THEN
  !         IC = M0(06)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(06) = M0(06)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(07) = awsa(irh) ! 2H-SO4
  !C      IF (M0(07) .LT. 100.0) THEN
  !C         IC = 3.0*M0(07)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(07) = M0(07)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(08) = awsa(irh) ! H-HSO4
  !CC      IF (M0(08) .LT. 100.0) THEN     ! These are redundant, because M0(8) is not used
  !CC         IC = M0(08)
  !CC         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX)
  !CC         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX)
!CCCCC         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,XX)
  !CC         M0(08) = M0(08)*EXP(LN10*(GI0-GII))
  !CC      ENDIF
  !
  m0(09) = awab(irh) ! NH4HSO4
  !C      IF (M0(09) .LT. 100.0) THEN
  !C         IC = M0(09)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(09) = M0(09)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  !      M0(12) = AWSB(IRH)      ! NAHSO4
  !      IF (M0(12) .LT. 100.0) THEN
  !         IC = M0(12)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,GI0,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,GII,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(12) = M0(12)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(13) = awlc(irh) ! (NH4)3H(SO4)2
  !C      IF (M0(13) .LT. 100.0) THEN
  !C         IC     = 4.0*M0(13)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,GI0,XX,XX,XX,XX,GII,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         G130   = 0.2*(3.0*GI0+2.0*GII)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,GI0,XX,XX,XX,XX,GII,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         G13I   = 0.2*(3.0*GI0+2.0*GII)
  !C         M0(13) = M0(13)*EXP(LN10*SNGL(G130-G13I))
  !C      ENDIF
  !
  ! *** OTHER INITIALIZATIONS *********************************************
  !
  iclact = 0
  calaou = .TRUE.
  calain = .TRUE.
  frst = .TRUE.
  scase = '??'
  sulratw = 2.D0
  sodrat = zero
  crnarat = zero
  crrat = zero
  nofer = 0
  stkofl = .FALSE.
  DO i = 1, nerrmx
    errstk(i) = -999
    errmsg(i) = 'MESSAGE N/A'
  END DO
  !
END SUBROUTINE init1



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE INIT2
! *** THIS SUBROUTINE INITIALIZES ALL GLOBAL VARIABLES FOR AMMONIUM,
!     NITRATE, SULFATE AEROSOL SYSTEMS (SUBROUTINE ISRP2)
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATED BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE init2 ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  DOUBLE PRECISION :: coef
  INTEGER :: i
  INTEGER :: irh
  DOUBLE PRECISION :: t0
  DOUBLE PRECISION :: t0t
  DOUBLE PRECISION :: tcf

  REAL, PARAMETER :: ln10 = 2.3025851
  !
  ! *** SAVE INPUT VARIABLES IN COMMON BLOCK ******************************
  !
  IF (iprob==0) THEN ! FORWARD CALCULATION
    DO i = 1, ncomp
      w(i) = max(wi(i), tiny)
    END DO
  ELSE
    DO i = 1, ncomp ! REVERSE CALCULATION
      waer(i) = max(wi(i), tiny)
      w(i) = zero
    END DO
  END IF
  rh = rhi
  temp = tempi
  !
  ! *** CALCULATE EQUILIBRIUM CONSTANTS ***********************************
  !
  xk1 = 1.015E-2 ! HSO4(aq)         <==> H(aq)     + SO4(aq)
  xk21 = 57.639 ! NH3(g)           <==> NH3(aq)
  xk22 = 1.805E-5 ! NH3(aq)          <==> NH4(aq)   + OH(aq)
  xk4 = 2.511E6 ! HNO3(g)          <==> H(aq)     + NO3(aq) ! ISORR
  !CC      XK4  = 3.638e6   ! HNO3(g)          <==> H(aq)     + NO3(aq) ! SEQUIL
  xk41 = 2.100E5 ! HNO3(g)          <==> HNO3(aq)
  xk7 = 1.817 ! (NH4)2SO4(s)     <==> 2*NH4(aq) + SO4(aq)
  xk10 = 5.746E-17 ! NH4NO3(s)        <==> NH3(g)    + HNO3(g) ! ISORR
  !CC      XK10 = 2.985e-17 ! NH4NO3(s)        <==> NH3(g)    + HNO3(g) ! SEQUIL
  xk12 = 1.382E2 ! NH4HSO4(s)       <==> NH4(aq)   + HSO4(aq)
  xk13 = 29.268 ! (NH4)3H(SO4)2(s) <==> 3*NH4(aq) + HSO4(aq) + SO4(aq)
  xkw = 1.010E-14 ! H2O              <==> H(aq)     + OH(aq)
  !
  IF (int(temp)/=298) THEN ! FOR T != 298K or 298.15K
    t0 = 298.15D0
    t0t = t0/temp
    coef = 1.0 + log(t0t) - t0t
    xk1 = xk1*exp(8.85*(t0t-1.0)+25.140*coef)
    xk21 = xk21*exp(13.79*(t0t-1.0)-5.393*coef)
    xk22 = xk22*exp(-1.50*(t0t-1.0)+26.920*coef)
    xk4 = xk4*exp(29.17*(t0t-1.0)+16.830*coef) !ISORR
    !CC         XK4 = XK4 *EXP( 29.47*(T0T-1.0) + 16.840*COEF) ! SEQUIL
    xk41 = xk41*exp(29.17*(t0t-1.0)+16.830*coef)
    xk7 = xk7*exp(-2.65*(t0t-1.0)+38.570*coef)
    xk10 = xk10*exp(-74.38*(t0t-1.0)+6.120*coef) ! ISORR
    !CC         XK10= XK10*EXP(-75.11*(T0T-1.0) + 13.460*COEF) ! SEQUIL
    xk12 = xk12*exp(-2.87*(t0t-1.0)+15.830*coef)
    xk13 = xk13*exp(-5.19*(t0t-1.0)+54.400*coef)
    xkw = xkw*exp(-22.52*(t0t-1.0)+26.920*coef)
  END IF
  xk2 = xk21*xk22
  xk42 = xk4/xk41
  !
  ! *** CALCULATE DELIQUESCENCE RELATIVE HUMIDITIES (UNICOMPONENT) ********
  !
  drh2so4 = zero
  drnh42s4 = 0.7997D0
  drnh4hs4 = 0.4000D0
  drnh4no3 = 0.6183D0
  drlc = 0.6900D0
  IF (int(temp)/=298) THEN
    t0 = 298.15D0
    tcf = 1.0/temp - 1.0/t0
    drnh4no3 = drnh4no3*exp(852.*tcf)
    drnh42s4 = drnh42s4*exp(80.*tcf)
    drnh4hs4 = drnh4hs4*exp(384.*tcf)
    drlc = drlc*exp(186.*tcf)
    drnh4no3 = min(drnh4no3, drnh42s4) ! ADJUST FOR DRH CROSSOVER AT T<271K
  END IF
  !
  ! *** CALCULATE MUTUAL DELIQUESCENCE RELATIVE HUMIDITIES ****************
  !
  drmlcab = 0.3780D0 ! (NH4)3H(SO4)2 & NH4HSO4
  drmlcas = 0.6900D0 ! (NH4)3H(SO4)2 & (NH4)2SO4
  drmasan = 0.6000D0 ! (NH4)2SO4     & NH4NO3
  !CC      IF (INT(TEMP) .NE. 298) THEN    ! For the time being
  !CC         T0       = 298.15d0
  !CC         TCF      = 1.0/TEMP - 1.0/T0
  !CC         DRMLCAB  = DRMLCAB*EXP( 507.506*TCF)
  !CC         DRMLCAS  = DRMLCAS*EXP( 133.865*TCF)
  !CC         DRMASAN  = DRMASAN*EXP(1269.068*TCF)
  !CC      ENDIF
  !
  ! *** LIQUID PHASE ******************************************************
  !
  chno3 = zero
  chcl = zero
  ch2so4 = zero
  coh = zero
  water = tiny
  !
  DO i = 1, npair
    molalr(i) = zero
    gama(i) = 0.1
    gamin(i) = great
    gamou(i) = great
    m0(i) = 1D5
  END DO
  !
  DO i = 1, npair
    gama(i) = 0.1D0
  END DO
  !
  DO i = 1, nions
    molal(i) = zero
  END DO
  coh = zero
  !
  DO i = 1, ngasaq
    gasaq(i) = zero
  END DO
  !
  ! *** SOLID PHASE ******************************************************
  !
  cnh42s4 = zero
  cnh4hs4 = zero
  cnacl = zero
  cna2so4 = zero
  cnano3 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnahso4 = zero
  clc = zero
  ccaso4 = zero
  ccano32 = zero
  ccacl2 = zero
  ck2so4 = zero
  ckhso4 = zero
  ckno3 = zero
  ckcl = zero
  cmgso4 = zero
  cmgno32 = zero
  cmgcl2 = zero
  !
  ! *** GAS PHASE ********************************************************
  !
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  ! *** CALCULATE ZSR PARAMETERS *****************************************
  !
  irh = min(int(rh*nzsr+0.5), nzsr) ! Position in ZSR arrays
  irh = max(irh, 1)
  !
  !      M0(01) = AWSC(IRH)      ! NACl
  !      IF (M0(01) .LT. 100.0) THEN
  !         IC = M0(01)
  !         CALL KMTAB(IC,298.0,     GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(01) = M0(01)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !C
  !      M0(02) = AWSS(IRH)      ! (NA)2SO4
  !      IF (M0(02) .LT. 100.0) THEN
  !         IC = 3.0*M0(02)
  !         CALL KMTAB(IC,298.0,     XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(02) = M0(02)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !CC
  !      M0(03) = AWSN(IRH)      ! NANO3
  !      IF (M0(03) .LT. 100.0) THEN
  !         IC = M0(03)
  !         CALL KMTAB(IC,298.0,     XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(03) = M0(03)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(04) = awas(irh) ! (NH4)2SO4
  !C      IF (M0(04) .LT. 100.0) THEN
  !C         IC = 3.0*M0(04)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(04) = M0(04)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(05) = awan(irh) ! NH4NO3
  !C      IF (M0(05) .LT. 100.0) THEN
  !C         IC     = M0(05)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(05) = M0(05)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  !      M0(06) = AWAC(IRH)      ! NH4CL
  !      IF (M0(06) .LT. 100.0) THEN
  !         IC = M0(06)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(06) = M0(06)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !C
  m0(07) = awsa(irh) ! 2H-SO4
  !C      IF (M0(07) .LT. 100.0) THEN
  !C         IC = 3.0*M0(07)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(07) = M0(07)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(08) = awsa(irh) ! H-HSO4
  !CC      IF (M0(08) .LT. 100.0) THEN     ! These are redundant, because M0(8) is not used
  !CC         IC = M0(08)
  !CC         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX)
  !CC         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX)
!CCCCC         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,XX)
  !CC         M0(08) = M0(08)*EXP(LN10*(GI0-GII))
  !CC      ENDIF
  !
  m0(09) = awab(irh) ! NH4HSO4
  !C      IF (M0(09) .LT. 100.0) THEN
  !C         IC = M0(09)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(09) = M0(09)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  !      M0(12) = AWSB(IRH)      ! NAHSO4
  !      IF (M0(12) .LT. 100.0) THEN
  !         IC = M0(12)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,GI0,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,GII,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(12) = M0(12)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(13) = awlc(irh) ! (NH4)3H(SO4)2
  !      IF (M0(13) .LT. 100.0) THEN
  !         IC     = 4.0*M0(13)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,GI0,XX,XX,XX,XX,GII,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         G130   = 0.2*(3.0*GI0+2.0*GII)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,GI0,XX,XX,XX,XX,GII,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         G13I   = 0.2*(3.0*GI0+2.0*GII)
  !         M0(13) = M0(13)*EXP(LN10*SNGL(G130-G13I))
  !      ENDIF
  !
  ! *** OTHER INITIALIZATIONS *********************************************
  !
  iclact = 0
  calaou = .TRUE.
  calain = .TRUE.
  frst = .TRUE.
  scase = '??'
  sulratw = 2.D0
  sodrat = zero
  crnarat = zero
  crrat = zero
  nofer = 0
  stkofl = .FALSE.
  DO i = 1, nerrmx
    errstk(i) = -999
    errmsg(i) = 'MESSAGE N/A'
  END DO
  !
END SUBROUTINE init2
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ISOINIT3
! *** THIS SUBROUTINE INITIALIZES ALL GLOBAL VARIABLES FOR AMMONIUM,
!     SODIUM, CHLORIDE, NITRATE, SULFATE AEROSOL SYSTEMS (SUBROUTINE
!     ISRP3)
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** RUPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE isoinit3 ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  DOUBLE PRECISION :: coef
  INTEGER :: i
  INTEGER :: irh
  DOUBLE PRECISION :: t0
  DOUBLE PRECISION :: t0t
  DOUBLE PRECISION :: tcf

  REAL, PARAMETER :: ln10 = 2.3025851
  !
  ! *** SAVE INPUT VARIABLES IN COMMON BLOCK ******************************
  !
  IF (iprob==0) THEN ! FORWARD CALCULATION
    DO i = 1, ncomp
      w(i) = max(wi(i), tiny)
    END DO
  ELSE
    DO i = 1, ncomp ! REVERSE CALCULATION
      waer(i) = max(wi(i), tiny)
      w(i) = zero
    END DO
  END IF
  rh = rhi
  temp = tempi
  !
  ! *** CALCULATE EQUILIBRIUM CONSTANTS ***********************************
  !
  xk1 = 1.015D-2 ! HSO4(aq)         <==> H(aq)     + SO4(aq)
  xk21 = 57.639D0 ! NH3(g)           <==> NH3(aq)
  xk22 = 1.805D-5 ! NH3(aq)          <==> NH4(aq)   + OH(aq)
  xk3 = 1.971D6 ! HCL(g)           <==> H(aq)     + CL(aq)
  xk31 = 2.500E3 ! HCL(g)           <==> HCL(aq)
  xk4 = 2.511E6 ! HNO3(g)          <==> H(aq)     + NO3(aq) ! ISORR
  !CC      XK4  = 3.638e6   ! HNO3(g)          <==> H(aq)     + NO3(aq) ! SEQUIL
  xk41 = 2.100E5 ! HNO3(g)          <==> HNO3(aq)
  xk5 = 0.4799D0 ! NA2SO4(s)        <==> 2*NA(aq)  + SO4(aq)
  xk6 = 1.086D-16 ! NH4CL(s)         <==> NH3(g)    + HCL(g)
  xk7 = 1.817D0 ! (NH4)2SO4(s)     <==> 2*NH4(aq) + SO4(aq)
  xk8 = 37.661D0 ! NACL(s)          <==> NA(aq)    + CL(aq)
  xk10 = 5.746D-17 ! NH4NO3(s)        <==> NH3(g)    + HNO3(g) ! ISORR
  !CC      XK10 = 2.985e-17 ! NH4NO3(s)        <==> NH3(g)    + HNO3(g) ! SEQUIL
  xk11 = 2.413D4 ! NAHSO4(s)        <==> NA(aq)    + HSO4(aq)
  xk12 = 1.382D2 ! NH4HSO4(s)       <==> NH4(aq)   + HSO4(aq)
  xk13 = 29.268D0 ! (NH4)3H(SO4)2(s) <==> 3*NH4(aq) + HSO4(aq) + SO4(aq)
  xk14 = 22.05D0 ! NH4CL(s)         <==> NH4(aq)   + CL(aq)
  xkw = 1.010D-14 ! H2O              <==> H(aq)     + OH(aq)
  xk9 = 11.977D0 ! NANO3(s)         <==> NA(aq)    + NO3(aq)
  !
  IF (int(temp)/=298) THEN ! FOR T != 298K or 298.15K
    t0 = 298.15D0
    t0t = t0/temp
    coef = 1.0 + log(t0t) - t0t
    xk1 = xk1*exp(8.85*(t0t-1.0)+25.140*coef)
    xk21 = xk21*exp(13.79*(t0t-1.0)-5.393*coef)
    xk22 = xk22*exp(-1.50*(t0t-1.0)+26.920*coef)
    xk3 = xk3*exp(30.20*(t0t-1.0)+19.910*coef)
    xk31 = xk31*exp(30.20*(t0t-1.0)+19.910*coef)
    xk4 = xk4*exp(29.17*(t0t-1.0)+16.830*coef) !ISORR
    !CC         XK4 = XK4 *EXP( 29.47*(T0T-1.0) + 16.840*COEF) ! SEQUIL
    xk41 = xk41*exp(29.17*(t0t-1.0)+16.830*coef)
    xk5 = xk5*exp(0.98*(t0t-1.0)+39.500*coef)
    xk6 = xk6*exp(-71.00*(t0t-1.0)+2.400*coef)
    xk7 = xk7*exp(-2.65*(t0t-1.0)+38.570*coef)
    xk8 = xk8*exp(-1.56*(t0t-1.0)+16.900*coef)
    xk9 = xk9*exp(-8.22*(t0t-1.0)+16.010*coef)
    xk10 = xk10*exp(-74.38*(t0t-1.0)+6.120*coef) ! ISORR
    !CC         XK10= XK10*EXP(-75.11*(T0T-1.0) + 13.460*COEF) ! SEQUIL
    xk11 = xk11*exp(0.79*(t0t-1.0)+14.746*coef)
    xk12 = xk12*exp(-2.87*(t0t-1.0)+15.830*coef)
    xk13 = xk13*exp(-5.19*(t0t-1.0)+54.400*coef)
    xk14 = xk14*exp(24.55*(t0t-1.0)+16.900*coef)
    xkw = xkw*exp(-22.52*(t0t-1.0)+26.920*coef)
  END IF
  xk2 = xk21*xk22
  xk42 = xk4/xk41
  xk32 = xk3/xk31
  !
  ! *** CALCULATE DELIQUESCENCE RELATIVE HUMIDITIES (UNICOMPONENT) ********
  !
  drh2so4 = zero
  drnh42s4 = 0.7997D0
  drnh4hs4 = 0.4000D0
  drlc = 0.6900D0
  drnacl = 0.7528D0
  drnano3 = 0.7379D0
  drnh4cl = 0.7710D0
  drnh4no3 = 0.6183D0
  drna2so4 = 0.9300D0
  drnahso4 = 0.5200D0
  IF (int(temp)/=298) THEN
    t0 = 298.15D0
    tcf = 1.0/temp - 1.0/t0
    drnacl = drnacl*exp(25.*tcf)
    drnano3 = drnano3*exp(304.*tcf)
    drna2so4 = drna2so4*exp(80.*tcf)
    drnh4no3 = drnh4no3*exp(852.*tcf)
    drnh42s4 = drnh42s4*exp(80.*tcf)
    drnh4hs4 = drnh4hs4*exp(384.*tcf)
    drlc = drlc*exp(186.*tcf)
    drnh4cl = drnh4cl*exp(239.*tcf)
    drnahso4 = drnahso4*exp(-45.*tcf)
    !
    ! *** ADJUST FOR DRH "CROSSOVER" AT LOW TEMPERATURES
    !
    drnh4no3 = min(drnh4no3, drnh4cl, drnh42s4, drnano3, drnacl)
    drnano3 = min(drnano3, drnacl)
    drnh4cl = min(drnh4cl, drnh42s4)
    !
  END IF
  !
  ! *** CALCULATE MUTUAL DELIQUESCENCE RELATIVE HUMIDITIES ****************
  !
  drmlcab = 0.378D0 ! (NH4)3H(SO4)2 & NH4HSO4
  drmlcas = 0.690D0 ! (NH4)3H(SO4)2 & (NH4)2SO4
  drmasan = 0.600D0 ! (NH4)2SO4     & NH4NO3
  drmg1 = 0.460D0 ! (NH4)2SO4, NH4NO3, NA2SO4, NH4CL
  drmg2 = 0.691D0 ! (NH4)2SO4, NA2SO4, NH4CL
  drmg3 = 0.697D0 ! (NH4)2SO4, NA2SO4
  drmh1 = 0.240D0 ! NA2SO4, NANO3, NACL, NH4NO3, NH4CL
  drmh2 = 0.596D0 ! NA2SO4, NANO3, NACL, NH4CL
  drmi1 = 0.240D0 ! LC, NAHSO4, NH4HSO4, NA2SO4, (NH4)2SO4
  drmi2 = 0.363D0 ! LC, NAHSO4, NA2SO4, (NH4)2SO4  - NO DATA -
  drmi3 = 0.610D0 ! LC, NA2SO4, (NH4)2SO4
  drmq1 = 0.494D0 ! (NH4)2SO4, NH4NO3, NA2SO4
  drmr1 = 0.663D0 ! NA2SO4, NANO3, NACL
  drmr2 = 0.735D0 ! NA2SO4, NACL
  drmr3 = 0.673D0 ! NANO3, NACL
  drmr4 = 0.694D0 ! NA2SO4, NACL, NH4CL
  drmr5 = 0.731D0 ! NA2SO4, NH4CL
  drmr6 = 0.596D0 ! NA2SO4, NANO3, NH4CL
  drmr7 = 0.380D0 ! NA2SO4, NANO3, NACL, NH4NO3
  drmr8 = 0.380D0 ! NA2SO4, NACL, NH4NO3
  drmr9 = 0.494D0 ! NA2SO4, NH4NO3
  drmr10 = 0.476D0 ! NA2SO4, NANO3, NH4NO3
  drmr11 = 0.340D0 ! NA2SO4, NACL, NH4NO3, NH4CL
  drmr12 = 0.460D0 ! NA2SO4, NH4NO3, NH4CL
  drmr13 = 0.438D0 ! NA2SO4, NANO3, NH4NO3, NH4CL
  !CC      IF (INT(TEMP) .NE. 298) THEN
  !CC         T0       = 298.15d0
  !CC         TCF      = 1.0/TEMP - 1.0/T0
  !CC         DRMLCAB  = DRMLCAB*EXP( 507.506*TCF)
  !CC         DRMLCAS  = DRMLCAS*EXP( 133.865*TCF)
  !CC         DRMASAN  = DRMASAN*EXP(1269.068*TCF)
  !CC         DRMG1    = DRMG1  *EXP( 572.207*TCF)
  !CC         DRMG2    = DRMG2  *EXP(  58.166*TCF)
  !CC         DRMG3    = DRMG3  *EXP(  22.253*TCF)
  !CC         DRMH1    = DRMH1  *EXP(2116.542*TCF)
  !CC         DRMH2    = DRMH2  *EXP( 650.549*TCF)
  !CC         DRMI1    = DRMI1  *EXP( 565.743*TCF)
  !CC         DRMI2    = DRMI2  *EXP(  91.745*TCF)
  !CC         DRMI3    = DRMI3  *EXP( 161.272*TCF)
  !CC         DRMQ1    = DRMQ1  *EXP(1616.621*TCF)
  !CC         DRMR1    = DRMR1  *EXP( 292.564*TCF)
  !CC         DRMR2    = DRMR2  *EXP(  14.587*TCF)
  !CC         DRMR3    = DRMR3  *EXP( 307.907*TCF)
  !CC         DRMR4    = DRMR4  *EXP(  97.605*TCF)
  !CC         DRMR5    = DRMR5  *EXP(  98.523*TCF)
  !CC         DRMR6    = DRMR6  *EXP( 465.500*TCF)
  !CC         DRMR7    = DRMR7  *EXP( 324.425*TCF)
  !CC         DRMR8    = DRMR8  *EXP(2660.184*TCF)
  !CC         DRMR9    = DRMR9  *EXP(1617.178*TCF)
  !CC         DRMR10   = DRMR10 *EXP(1745.226*TCF)
  !CC         DRMR11   = DRMR11 *EXP(3691.328*TCF)
  !CC         DRMR12   = DRMR12 *EXP(1836.842*TCF)
  !CC         DRMR13   = DRMR13 *EXP(1967.938*TCF)
  !CC      ENDIF
  !
  ! *** LIQUID PHASE ******************************************************
  !
  chno3 = zero
  chcl = zero
  ch2so4 = zero
  coh = zero
  water = tiny
  !
  DO i = 1, npair
    molalr(i) = zero
    gama(i) = 0.1
    gamin(i) = great
    gamou(i) = great
    m0(i) = 1D5
  END DO
  !
  DO i = 1, npair
    gama(i) = 0.1D0
  END DO
  !
  DO i = 1, nions
    molal(i) = zero
  END DO
  coh = zero
  !
  DO i = 1, ngasaq
    gasaq(i) = zero
  END DO
  !
  ! *** SOLID PHASE *******************************************************
  !
  cnh42s4 = zero
  cnh4hs4 = zero
  cnacl = zero
  cna2so4 = zero
  cnano3 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnahso4 = zero
  clc = zero
  ccaso4 = zero
  ccano32 = zero
  ccacl2 = zero
  ck2so4 = zero
  ckhso4 = zero
  ckno3 = zero
  ckcl = zero
  cmgso4 = zero
  cmgno32 = zero
  cmgcl2 = zero
  !
  ! *** GAS PHASE *********************************************************
  !
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  ! *** CALCULATE ZSR PARAMETERS ******************************************
  !
  irh = min(int(rh*nzsr+0.5), nzsr) ! Position in ZSR arrays
  irh = max(irh, 1)
  !
  m0(01) = awsc(irh) ! NACl
  !C      IF (M0(01) .LT. 100.0) THEN
  !C         IC = M0(01)
  !C         CALL KMTAB(IC,298.0,     GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(01) = M0(01)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(02) = awss(irh) ! (NA)2SO4
  !C      IF (M0(02) .LT. 100.0) THEN
  !C         IC = 3.0*M0(02)
  !C         CALL KMTAB(IC,298.0,     XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(02) = M0(02)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(03) = awsn(irh) ! NANO3
  !C      IF (M0(03) .LT. 100.0) THEN
  !C         IC = M0(03)
  !C         CALL KMTAB(IC,298.0,     XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  ! C        M0(03) = M0(03)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(04) = awas(irh) ! (NH4)2SO4
  !C      IF (M0(04) .LT. 100.0) THEN
  !C         IC = 3.0*M0(04)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(04) = M0(04)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(05) = awan(irh) ! NH4NO3
  !C      IF (M0(05) .LT. 100.0) THEN
  !C         IC     = M0(05)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(05) = M0(05)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(06) = awac(irh) ! NH4CL
  !C      IF (M0(06) .LT. 100.0) THEN
  !C         IC = M0(06)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(06) = M0(06)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(07) = awsa(irh) ! 2H-SO4
  !C      IF (M0(07) .LT. 100.0) THEN
  !C         IC = 3.0*M0(07)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(07) = M0(07)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(08) = awsa(irh) ! H-HSO4
  !CC      IF (M0(08) .LT. 100.0) THEN     ! These are redundant, because M0(8) is not used
  !CC         IC = M0(08)
  !CC         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX)
  !CC         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX)
!CCCCC         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,XX)
  !CC         M0(08) = M0(08)*EXP(LN10*(GI0-GII))
  !CC      ENDIF
  !
  m0(09) = awab(irh) ! NH4HSO4
  !C      IF (M0(09) .LT. 100.0) THEN
  !C         IC = M0(09)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(09) = M0(09)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(12) = awsb(irh) ! NAHSO4
  !C      IF (M0(12) .LT. 100.0) THEN
  !C         IC = M0(12)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,GI0,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,GII,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         M0(12) = M0(12)*EXP(LN10*(GI0-GII))
  !C      ENDIF
  !
  m0(13) = awlc(irh) ! (NH4)3H(SO4)2
  !C      IF (M0(13) .LT. 100.0) THEN
  !C         IC     = 4.0*M0(13)
  !C         CALL KMTAB(IC,298.0,     XX,XX,XX,GI0,XX,XX,XX,XX,GII,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         G130   = 0.2*(3.0*GI0+2.0*GII)
  !C         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,GI0,XX,XX,XX,XX,GII,XX,XX,XX,
  !C     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !C         G13I   = 0.2*(3.0*GI0+2.0*GII)
  !C         M0(13) = M0(13)*EXP(LN10*SNGL(G130-G13I))
  !C      ENDIF
  !
  ! *** OTHER INITIALIZATIONS *********************************************
  !
  iclact = 0
  calaou = .TRUE.
  calain = .TRUE.
  frst = .TRUE.
  scase = '??'
  sulratw = 2.D0
  crnarat = zero
  crrat = zero
  nofer = 0
  stkofl = .FALSE.
  DO i = 1, nerrmx
    errstk(i) = -999
    errmsg(i) = 'MESSAGE N/A'
  END DO
  !
END SUBROUTINE isoinit3
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE INIT4
! *** THIS SUBROUTINE INITIALIZES ALL GLOBAL VARIABLES FOR AMMONIUM,
!     SODIUM, CHLORIDE, NITRATE, SULFATE, CALCIUM, POTASSIUM, MAGNESIUM
!     AEROSOL SYSTEMS (SUBROUTINE ISRP4)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!

SUBROUTINE init4 ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  DOUBLE PRECISION :: coef
  INTEGER :: i
  INTEGER :: irh
  DOUBLE PRECISION :: t0
  DOUBLE PRECISION :: t0t
  DOUBLE PRECISION :: tcf
  REAL, PARAMETER :: ln10 = 2.3025851
  !
  ! *** SAVE INPUT VARIABLES IN COMMON BLOCK ******************************
  !
  IF (iprob==0) THEN ! FORWARD CALCULATION
    DO i = 1, ncomp
      w(i) = max(wi(i), tiny)
    END DO
  ELSE
    DO i = 1, ncomp ! REVERSE CALCULATION
      waer(i) = max(wi(i), tiny)
      w(i) = zero
    END DO
  END IF
  rh = rhi
  temp = tempi
  !
  ! *** CALCULATE EQUILIBRIUM CONSTANTS ***********************************
  !
  xk1 = 1.015D-2 ! HSO4(aq)         <==> H(aq)     + SO4(aq)
  xk21 = 57.639D0 ! NH3(g)           <==> NH3(aq)
  xk22 = 1.805D-5 ! NH3(aq)          <==> NH4(aq)   + OH(aq)
  xk3 = 1.971D6 ! HCL(g)           <==> H(aq)     + CL(aq)
  xk31 = 2.500E3 ! HCL(g)           <==> HCL(aq)
  xk4 = 2.511E6 ! HNO3(g)          <==> H(aq)     + NO3(aq) ! ISORR
  !      XK4  = 3.638e6   ! HNO3(g)          <==> H(aq)     + NO3(aq) ! SEQUIL
  xk41 = 2.100E5 ! HNO3(g)          <==> HNO3(aq)
  xk5 = 0.4799D0 ! NA2SO4(s)        <==> 2*NA(aq)  + SO4(aq)
  xk6 = 1.086D-16 ! NH4CL(s)         <==> NH3(g)    + HCL(g)
  xk7 = 1.817D0 ! (NH4)2SO4(s)     <==> 2*NH4(aq) + SO4(aq)
  xk8 = 37.661D0 ! NACL(s)          <==> NA(aq)    + CL(aq)
  !      XK10 = 5.746D-17 ! NH4NO3(s)        <==> NH3(g)    + HNO3(g) ! ISORR
  xk10 = 4.199D-17 ! NH4NO3(s)        <==> NH3(g)    + HNO3(g) ! (Mozurkewich, 1993)
  !      XK10 = 2.985e-17 ! NH4NO3(s)        <==> NH3(g)    + HNO3(g) ! SEQUIL
  xk11 = 2.413D4 ! NAHSO4(s)        <==> NA(aq)    + HSO4(aq)
  xk12 = 1.382D2 ! NH4HSO4(s)       <==> NH4(aq)   + HSO4(aq)
  xk13 = 29.268D0 ! (NH4)3H(SO4)2(s) <==> 3*NH4(aq) + HSO4(aq) + SO4(aq)
  xk14 = 22.05D0 ! NH4CL(s)         <==> NH4(aq)   + CL(aq)
  xkw = 1.010D-14 ! H2O              <==> H(aq)     + OH(aq)
  xk9 = 11.977D0 ! NANO3(s)         <==> NA(aq)    + NO3(aq)
  !CC
  xk15 = 6.067D5 ! CA(NO3)2(s)      <==> CA(aq)    + 2NO3(aq)
  xk16 = 7.974D11 ! CACL2(s)         <==> CA(aq)    + 2CL(aq)
  xk17 = 1.569D-2 ! K2SO4(s)         <==> 2K(aq)    + SO4(aq)
  xk18 = 24.016 ! KHSO4(s)         <==> K(aq)     + HSO4(aq)
  xk19 = 0.872 ! KNO3(s)          <==> K(aq)     + NO3(aq)
  xk20 = 8.680 ! KCL(s)           <==> K(aq)     + CL(aq)
  xk23 = 1.079D5 ! MGS04(s)         <==> MG(aq)    + SO4(aq)
  xk24 = 2.507D15 ! MG(NO3)2(s)      <==> MG(aq)    + 2NO3(aq)
  xk25 = 9.557D21 ! MGCL2(s)         <==> MG(aq)    + 2CL(aq)
  !      XK26 = 4.299D-7  ! CO2(aq) + H2O    <==> HCO3(aq)  + H(aq)
  !      XK27 = 4.678D-11 ! HCO3(aq)         <==> CO3(aq)   + H(aq)

  !
  IF (int(temp)/=298) THEN ! FOR T != 298K or 298.15K
    t0 = 298.15D0
    t0t = t0/temp
    coef = 1.0 + log(t0t) - t0t
    xk1 = xk1*exp(8.85*(t0t-1.0)+25.140*coef)
    xk21 = xk21*exp(13.79*(t0t-1.0)-5.393*coef)
    xk22 = xk22*exp(-1.50*(t0t-1.0)+26.920*coef)
    xk3 = xk3*exp(30.20*(t0t-1.0)+19.910*coef)
    xk31 = xk31*exp(30.20*(t0t-1.0)+19.910*coef)
    xk4 = xk4*exp(29.17*(t0t-1.0)+16.830*coef) !ISORR
    !         XK4 = XK4 *EXP( 29.47*(T0T-1.0) + 16.840*COEF) ! SEQUIL
    xk41 = xk41*exp(29.17*(t0t-1.0)+16.830*coef)
    xk5 = xk5*exp(0.98*(t0t-1.0)+39.500*coef)
    xk6 = xk6*exp(-71.00*(t0t-1.0)+2.400*coef)
    xk7 = xk7*exp(-2.65*(t0t-1.0)+38.570*coef)
    xk8 = xk8*exp(-1.56*(t0t-1.0)+16.900*coef)
    xk9 = xk9*exp(-8.22*(t0t-1.0)+16.010*coef)
    !         XK10= XK10*EXP(-74.38*(T0T-1.0) +  6.120*COEF) ! ISORR
    xk10 = xk10*exp(-74.7351*(t0t-1.0)+6.025*coef) ! (Mozurkewich, 1993)
    !         XK10= XK10*EXP(-75.11*(T0T-1.0) + 13.460*COEF) ! SEQUIL
    xk11 = xk11*exp(0.79*(t0t-1.0)+14.746*coef)
    xk12 = xk12*exp(-2.87*(t0t-1.0)+15.830*coef)
    xk13 = xk13*exp(-5.19*(t0t-1.0)+54.400*coef)
    xk14 = xk14*exp(24.55*(t0t-1.0)+16.900*coef)
    xkw = xkw*exp(-22.52*(t0t-1.0)+26.920*coef)
    !CC
    !         XK15= XK15 *EXP(  .0*(T0T-1.0) + .0*COEF)
    !         XK16= XK16 *EXP(  .0*(T0T-1.0) + .0*COEF)
    xk17 = xk17*exp(-9.585*(t0t-1.0)+45.81*coef)
    xk18 = xk18*exp(-8.423*(t0t-1.0)+17.96*coef)
    xk19 = xk19*exp(-14.08*(t0t-1.0)+19.39*coef)
    xk20 = xk20*exp(-6.902*(t0t-1.0)+19.95*coef)
    !         XK23= XK23 *EXP(  .0*(T0T-1.0) + .0*COEF)
    !         XK24= XK24 *EXP(  .0*(T0T-1.0) + .0*COEF)
    !         XK25= XK25 *EXP(  .0*(T0T-1.0) + .0*COEF)
    !         XK26= XK26 *EXP(-3.0821*(T0T-1.0) + 31.8139*COEF)
    !         XK27= XK27 *EXP(-5.9908*(T0T-1.0) + 38.844*COEF)

  END IF
  xk2 = xk21*xk22
  xk42 = xk4/xk41
  xk32 = xk3/xk31
  !
  ! *** CALCULATE DELIQUESCENCE RELATIVE HUMIDITIES (UNICOMPONENT) ********
  !
  drh2so4 = zero
  drnh42s4 = 0.7997D0
  drnh4hs4 = 0.4000D0
  drlc = 0.6900D0
  drnacl = 0.7528D0
  drnano3 = 0.7379D0
  drnh4cl = 0.7710D0
  drnh4no3 = 0.6183D0
  drna2so4 = 0.9300D0
  drnahso4 = 0.5200D0
  drcano32 = 0.4906D0
  drcacl2 = 0.2830D0
  drk2so4 = 0.9750D0
  drkhso4 = 0.8600D0
  drkno3 = 0.9248D0
  drkcl = 0.8426D0
  drmgso4 = 0.8613D0
  drmgno32 = 0.5400D0
  drmgcl2 = 0.3284D0
  IF (int(temp)/=298) THEN
    t0 = 298.15D0
    tcf = 1.0/temp - 1.0/t0
    drnacl = drnacl*exp(25.*tcf)
    drnano3 = drnano3*exp(304.*tcf)
    drna2so4 = drna2so4*exp(80.*tcf)
    drnh4no3 = drnh4no3*exp(852.*tcf)
    drnh42s4 = drnh42s4*exp(80.*tcf)
    drnh4hs4 = drnh4hs4*exp(384.*tcf)
    drlc = drlc*exp(186.*tcf)
    drnh4cl = drnh4cl*exp(239.*tcf)
    drnahso4 = drnahso4*exp(-45.*tcf)
    !         DRCANO32 = DRCANO32*EXP(-430.5*TCF)
    drcano32 = drcano32*exp(509.4*tcf) ! KELLY & WEXLER (2005) FOR CANO32.4H20
    !         DRCACL2  = DRCACL2 *EXP(-1121.*TCF)
    drcacl2 = drcacl2*exp(551.1*tcf) ! KELLY & WEXLER (2005) FOR CACL2.6H20
    drk2so4 = drk2so4*exp(35.6*tcf)
    !         DRKHSO4  = DRKHSO4 *EXP( 0.*TCF)
    !         DRKNO3   = DRKNO3  *EXP( 0.*TCF)
    drkcl = drkcl*exp(159.*tcf)
    drmgso4 = drmgso4*exp(-714.45*tcf)
    drmgno32 = drmgno32*exp(230.2*tcf) ! KELLY & WEXLER (2005) FOR MGNO32.6H20
    !         DRMGCL2  = DRMGCL2 *EXP(-1860.*TCF)
    drmgcl2 = drmgcl2*exp(42.23*tcf) ! KELLY & WEXLER (2005) FOR MGCL2.6H20
    !
  END IF
  !
  ! *** CALCULATE MUTUAL DELIQUESCENCE RELATIVE HUMIDITIES ****************
  !
  drmlcab = 0.378D0 ! (NH4)3H(SO4)2 & NH4HSO4
  drmlcas = 0.690D0 ! (NH4)3H(SO4)2 & (NH4)2SO4
  drmasan = 0.600D0 ! (NH4)2SO4     & NH4NO3
  drmg1 = 0.460D0 ! (NH4)2SO4, NH4NO3, NA2SO4, NH4CL
  drmg2 = 0.691D0 ! (NH4)2SO4, NA2SO4, NH4CL
  drmg3 = 0.697D0 ! (NH4)2SO4, NA2SO4
  drmh1 = 0.240D0 ! NA2SO4, NANO3, NACL, NH4NO3, NH4CL
  drmh2 = 0.596D0 ! NA2SO4, NANO3, NACL, NH4CL
  drmi1 = 0.240D0 ! LC, NAHSO4, NH4HSO4, NA2SO4, (NH4)2SO4
  drmi2 = 0.363D0 ! LC, NAHSO4, NA2SO4, (NH4)2SO4  - NO DATA -
  drmi3 = 0.610D0 ! LC, NA2SO4, (NH4)2SO4
  drmq1 = 0.494D0 ! (NH4)2SO4, NH4NO3, NA2SO4
  drmr1 = 0.663D0 ! NA2SO4, NANO3, NACL
  drmr2 = 0.735D0 ! NA2SO4, NACL
  drmr3 = 0.673D0 ! NANO3, NACL
  drmr4 = 0.694D0 ! NA2SO4, NACL, NH4CL
  drmr5 = 0.731D0 ! NA2SO4, NH4CL
  drmr6 = 0.596D0 ! NA2SO4, NANO3, NH4CL
  drmr7 = 0.380D0 ! NA2SO4, NANO3, NACL, NH4NO3
  drmr8 = 0.380D0 ! NA2SO4, NACL, NH4NO3
  drmr9 = 0.494D0 ! NA2SO4, NH4NO3
  drmr10 = 0.476D0 ! NA2SO4, NANO3, NH4NO3
  drmr11 = 0.340D0 ! NA2SO4, NACL, NH4NO3, NH4CL
  drmr12 = 0.460D0 ! NA2SO4, NH4NO3, NH4CL
  drmr13 = 0.438D0 ! NA2SO4, NANO3, NH4NO3, NH4CL
  !
  drmo1 = 0.460D0 ! (NH4)2SO4, NH4NO3, NH4Cl, NA2SO4, K2SO4, MGSO4
  drmo2 = 0.691D0 ! (NH4)2SO4, NH4Cl, NA2SO4, K2SO4, MGSO4
  drmo3 = 0.697D0 ! (NH4)2SO4, NA2SO4, K2SO4, MGSO4
  drml1 = 0.240D0 ! K2SO4, MGSO4, KHSO4, NH4HSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
  drml2 = 0.363D0 ! K2SO4, MGSO4, KHSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
  drml3 = 0.610D0 ! K2SO4, MGSO4, KHSO4, (NH4)2SO4, NA2SO4, LC
  drmm1 = 0.240D0 ! K2SO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3, NH4NO3
  drmm2 = 0.596D0 ! K2SO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3
  drmp1 = 0.200D0 ! CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4, MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
  drmp2 = 0.240D0 ! CA(NO3)2, K2SO4, KNO3, KCL, MGSO4, MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
  drmp3 = 0.240D0 ! CA(NO3)2, K2SO4, KNO3, KCL, MGSO4, MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
  drmp4 = 0.240D0 ! K2SO4, KNO3, KCL, MGSO4, MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
  drmp5 = 0.240D0 ! K2SO4, KNO3, KCL, MGSO4, NANO3, NACL, NH4NO3, NH4CL
  !C
  drmv1 = 0.494D0 ! (NH4)2SO4, NH4NO3, NA2SO4, K2SO4, MGSO4
  !C
  !C
  !      DRMO1   = 0.1D0    ! (NH4)2SO4, NH4NO3, NH4Cl, NA2SO4, K2SO4, MGSO4
  !      DRMO2   = 0.1D0    ! (NH4)2SO4, NH4Cl, NA2SO4, K2SO4, MGSO4
  !      DRMO3   = 0.1D0    ! (NH4)2SO4, NA2SO4, K2SO4, MGSO4
  !      DRML1   = 0.1D0    ! K2SO4, MGSO4, KHSO4, NH4HSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
  !      DRML2   = 0.1D0    ! K2SO4, MGSO4, KHSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
  !      DRML3   = 0.1D0    ! K2SO4, MGSO4, KHSO4, (NH4)2SO4, NA2SO4, LC
  !      DRMM1   = 0.1D0    ! K2SO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3, NH4NO3
  !      DRMM2   = 0.1D0    ! K2SO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3
  !      DRMP1   = 0.1D0    ! CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4, MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
  !      DRMP2   = 0.1D0    ! CA(NO3)2, K2SO4, KNO3, KCL, MGSO4, MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
  !      DRMP3   = 0.1D0    ! CA(NO3)2, K2SO4, KNO3, KCL, MGSO4, MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
  !      DRMP4   = 0.1D0    ! K2SO4, KNO3, KCL, MGSO4, MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
  !      DRMP5   = 0.1D0    ! K2SO4, KNO3, KCL, MGSO4, NANO3, NACL, NH4NO3, NH4CL
  !C
  !      DRMV1   = 0.1D0    ! (NH4)2SO4, NH4NO3, NA2SO4, K2SO4, MGSO4
  !
  !CC      IF (INT(TEMP) .NE. 298) THEN
  !CC         T0       = 298.15d0
  !CC         TCF      = 1.0/TEMP - 1.0/T0
  !CC         DRMLCAB  = DRMLCAB*EXP( 507.506*TCF)
  !CC         DRMLCAS  = DRMLCAS*EXP( 133.865*TCF)
  !CC         DRMASAN  = DRMASAN*EXP(1269.068*TCF)
  !CC         DRMG1    = DRMG1  *EXP( 572.207*TCF)
  !CC         DRMG2    = DRMG2  *EXP(  58.166*TCF)
  !CC         DRMG3    = DRMG3  *EXP(  22.253*TCF)
  !CC         DRMH1    = DRMH1  *EXP(2116.542*TCF)
  !CC         DRMH2    = DRMH2  *EXP( 650.549*TCF)
  !CC         DRMI1    = DRMI1  *EXP( 565.743*TCF)
  !CC         DRMI2    = DRMI2  *EXP(  91.745*TCF)
  !CC         DRMI3    = DRMI3  *EXP( 161.272*TCF)
  !CC         DRMQ1    = DRMQ1  *EXP(1616.621*TCF)
  !CC         DRMR1    = DRMR1  *EXP( 292.564*TCF)
  !CC         DRMR2    = DRMR2  *EXP(  14.587*TCF)
  !CC         DRMR3    = DRMR3  *EXP( 307.907*TCF)
  !CC         DRMR4    = DRMR4  *EXP(  97.605*TCF)
  !CC         DRMR5    = DRMR5  *EXP(  98.523*TCF)
  !CC         DRMR6    = DRMR6  *EXP( 465.500*TCF)
  !CC         DRMR7    = DRMR7  *EXP( 324.425*TCF)
  !CC         DRMR8    = DRMR8  *EXP(2660.184*TCF)
  !CC         DRMR9    = DRMR9  *EXP(1617.178*TCF)
  !CC         DRMR10   = DRMR10 *EXP(1745.226*TCF)
  !CC         DRMR11   = DRMR11 *EXP(3691.328*TCF)
  !CC         DRMR12   = DRMR12 *EXP(1836.842*TCF)
  !CC         DRMR13   = DRMR13 *EXP(1967.938*TCF)
  !CC      ENDIF
  !
  ! *** LIQUID PHASE ******************************************************
  !
  chno3 = zero
  chcl = zero
  ch2so4 = zero
  coh = zero
  water = tiny
  !
  DO i = 1, npair
    molalr(i) = zero
    gama(i) = 0.1
    gamin(i) = great
    gamou(i) = great
    m0(i) = 1D5
  END DO
  !
  DO i = 1, npair
    gama(i) = 0.1D0
  END DO
  !
  DO i = 1, nions
    molal(i) = zero
  END DO
  coh = zero
  !
  DO i = 1, ngasaq
    gasaq(i) = zero
  END DO
  !
  ! *** SOLID PHASE *******************************************************
  !
  cnh42s4 = zero
  cnh4hs4 = zero
  cnacl = zero
  cna2so4 = zero
  cnano3 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnahso4 = zero
  clc = zero
  ccaso4 = zero
  ccano32 = zero
  ccacl2 = zero
  ck2so4 = zero
  ckhso4 = zero
  ckno3 = zero
  ckcl = zero
  cmgso4 = zero
  cmgno32 = zero
  cmgcl2 = zero
  !
  ! *** GAS PHASE *********************************************************
  !
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  ! *** CALCULATE ZSR PARAMETERS ******************************************
  !
  irh = min(int(rh*nzsr+0.5), nzsr) ! Position in ZSR arrays
  irh = max(irh, 1)
  !
  m0(01) = awsc(irh) ! NACl
  !      IF (M0(01) .LT. 100.0) THEN
  !         IC = M0(01)
  !         CALL KMTAB(IC,298.0,     GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                            XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                            XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(01) = M0(01)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(02) = awss(irh) ! (NA)2SO4
  !      IF (M0(02) .LT. 100.0) THEN
  !         IC = 3.0*M0(02)
  !         CALL KMTAB(IC,298.0,     XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(02) = M0(02)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(03) = awsn(irh) ! NANO3
  !      IF (M0(03) .LT. 100.0) THEN
  !         IC = M0(03)
  !         CALL KMTAB(IC,298.0,     XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(03) = M0(03)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(04) = awas(irh) ! (NH4)2SO4
  !      IF (M0(04) .LT. 100.0) THEN
  !         IC = 3.0*M0(04)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(04) = M0(04)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(05) = awan(irh) ! NH4NO3
  !      IF (M0(05) .LT. 100.0) THEN
  !         IC     = M0(05)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(05) = M0(05)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(06) = awac(irh) ! NH4CL
  !      IF (M0(06) .LT. 100.0) THEN
  !         IC = M0(06)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(06) = M0(06)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(07) = awsa(irh) ! 2H-SO4
  !      IF (M0(07) .LT. 100.0) THEN
  !         IC = 3.0*M0(07)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(07) = M0(07)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(08) = awsa(irh) ! H-HSO4
  !CC      IF (M0(08) .LT. 100.0) THEN     ! These are redundant, because M0(8) is not used
  !CC         IC = M0(08)
  !CC         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX)
  !CC         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,XX)
!CCCCC         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,XX)
  !CC         M0(08) = M0(08)*EXP(LN10*(GI0-GII))
  !CC      ENDIF
  !
  m0(09) = awab(irh) ! NH4HSO4
  !      IF (M0(09) .LT. 100.0) THEN
  !         IC = M0(09)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,GI0,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,GII,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(09) = M0(09)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(12) = awsb(irh) ! NAHSO4
  !      IF (M0(12) .LT. 100.0) THEN
  !         IC = M0(12)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,GI0,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,GII,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(12) = M0(12)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(13) = awlc(irh) ! (NH4)3H(SO4)2
  !      IF (M0(13) .LT. 100.0) THEN
  !         IC     = 4.0*M0(13)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,GI0,XX,XX,XX,XX,GII,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         G130   = 0.2*(3.0*GI0+2.0*GII)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,GI0,XX,XX,XX,XX,GII,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,XX)
  !         G13I   = 0.2*(3.0*GI0+2.0*GII)
  !         M0(13) = M0(13)*EXP(LN10*SNGL(G130-G13I))
  !      ENDIF
  !
  m0(15) = awcn(irh) ! CA(NO3)2
  !      IF (M0(15) .LT. 100.0) THEN
  !         IC = M0(15)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             GI0,XX,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             GII,XX,XX,XX,XX,XX,XX,XX,XX)
  !         M0(15) = M0(15)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !C
  m0(16) = awcc(irh) ! CACl2
  !      IF (M0(16) .LT. 100.0) THEN
  !         IC = M0(16)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,GI0,XX,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,GII,XX,XX,XX,XX,XX,XX,XX)
  !         M0(16) = M0(16)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(17) = awps(irh) ! K2SO4
  !      IF (M0(17) .LT. 100.0) THEN
  !         IC = M0(17)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,GI0,XX,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,GII,XX,XX,XX,XX,XX,XX)
  !         M0(17) = M0(17)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(18) = awpb(irh) ! KHSO4
  !      IF (M0(18) .LT. 100.0) THEN
  !         IC = M0(18)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,GI0,XX,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,GII,XX,XX,XX,XX,XX)
  !         M0(18) = M0(18)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(19) = awpn(irh) ! KNO3
  !      IF (M0(19) .LT. 100.0) THEN
  !         IC = M0(19)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,GI0,XX,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,GII,XX,XX,XX,XX)
  !         M0(19) = M0(19)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(20) = awpc(irh) ! KCl
  !      IF (M0(20) .LT. 100.0) THEN
  !         IC = M0(20)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,GI0,XX,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,GII,XX,XX,XX)
  !         M0(20) = M0(20)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(21) = awms(irh) ! MGSO4
  !      IF (M0(21) .LT. 100.0) THEN
  !         IC = M0(21)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,GI0,XX,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,GII,XX,XX)
  !         M0(21) = M0(21)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(22) = awmn(irh) ! MG(NO3)2
  !      IF (M0(22) .LT. 100.0) THEN
  !         IC = M0(22)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,GI0,XX)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,GII,XX)
  !         M0(22) = M0(22)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  m0(23) = awmc(irh) ! MGCL2
  !      IF (M0(23) .LT. 100.0) THEN
  !         IC = M0(23)
  !         CALL KMTAB(IC,298.0,     XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,GI0)
  !         CALL KMTAB(IC,SNGL(TEMP),XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,XX,
  !     &                             XX,XX,XX,XX,XX,XX,XX,XX,GII)
  !         M0(23) = M0(23)*EXP(LN10*(GI0-GII))
  !      ENDIF
  !
  ! *** OTHER INITIALIZATIONS *********************************************
  !
  iclact = 0
  calaou = .TRUE.
  calain = .TRUE.
  frst = .TRUE.
  scase = '??'
  sulratw = 2.D0
  so4rat = 2.D0
  crnarat = 2.D0
  crrat = 2.D0
  nofer = 0
  stkofl = .FALSE.
  DO i = 1, nerrmx
    errstk(i) = -999
    errmsg(i) = 'MESSAGE N/A'
  END DO
  !
END SUBROUTINE init4
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ADJUST
! *** ADJUSTS FOR MASS BALANCE BETWEEN VOLATILE SPECIES AND SULFATE
!     FIRST CALCULATE THE EXCESS OF EACH PRECURSOR, AND IF IT EXISTS, THEN
!     ADJUST SEQUENTIALY AEROSOL PHASE SPECIES WHICH CONTAIN THE EXCESS
!     PRECURSOR.
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE adjust ( wi )
  USE mo_isrpia 
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(:)
  DOUBLE PRECISION :: excl
  DOUBLE PRECISION :: exnh4
  DOUBLE PRECISION :: exno3
  DOUBLE PRECISION :: exs4

  !
  ! *** FOR AMMONIUM *****************************************************
  !
  IF (iprob==0) THEN ! Calculate excess (solution - input)
    exnh4 = gnh3 + molal(3) + cnh4cl + cnh4no3 + cnh4hs4 + 2D0*cnh42s4 +   &
         3D0*clc - wi(3)
  ELSE
    exnh4 = molal(3) + cnh4cl + cnh4no3 + cnh4hs4 + 2D0*cnh42s4 +          &
         3D0*clc - wi(3)

  END IF
  exnh4 = max(exnh4, zero)
  IF (exnh4<tiny) GO TO 100 ! No excess NH4, go to next precursor
  !
  IF (molal(3)>exnh4) THEN ! Adjust aqueous phase NH4
    molal(3) = molal(3) - exnh4
    GO TO 100
  ELSE
    exnh4 = exnh4 - molal(3)
    molal(3) = zero
  END IF
  !
  IF (cnh4cl>exnh4) THEN ! Adjust NH4Cl(s)
    cnh4cl = cnh4cl - exnh4 ! more solid than excess
    ghcl = ghcl + exnh4 ! evaporate Cl to gas phase
    GO TO 100
  ELSE ! less solid than excess
    ghcl = ghcl + cnh4cl ! evaporate into gas phase
    exnh4 = exnh4 - cnh4cl ! reduce excess
    cnh4cl = zero ! zero salt concentration
  END IF
  !
  IF (cnh4no3>exnh4) THEN ! Adjust NH4NO3(s)
    cnh4no3 = cnh4no3 - exnh4 ! more solid than excess
    ghno3 = ghno3 + exnh4 ! evaporate NO3 to gas phase
    GO TO 100
  ELSE ! less solid than excess
    ghno3 = ghno3 + cnh4no3 ! evaporate into gas phase
    exnh4 = exnh4 - cnh4no3 ! reduce excess
    cnh4no3 = zero ! zero salt concentration
  END IF
  !
  IF (clc>3D0*exnh4) THEN ! Adjust (NH4)3H(SO4)2(s)
    clc = clc - exnh4/3D0 ! more solid than excess
    GO TO 100
  ELSE ! less solid than excess
    exnh4 = exnh4 - 3D0*clc ! reduce excess
    clc = zero ! zero salt concentration
  END IF
  !
  IF (cnh4hs4>exnh4) THEN ! Adjust NH4HSO4(s)
    cnh4hs4 = cnh4hs4 - exnh4 ! more solid than excess
    GO TO 100
  ELSE ! less solid than excess
    exnh4 = exnh4 - cnh4hs4 ! reduce excess
    cnh4hs4 = zero ! zero salt concentration
  END IF
  !
  IF (cnh42s4>exnh4) THEN ! Adjust (NH4)2SO4(s)
    cnh42s4 = cnh42s4 - exnh4 ! more solid than excess
    GO TO 100
  ELSE ! less solid than excess
    exnh4 = exnh4 - cnh42s4 ! reduce excess
    cnh42s4 = zero ! zero salt concentration
  END IF
  !
  ! *** FOR NITRATE ******************************************************
  !
100 IF (iprob==0) THEN ! Calculate excess (solution - input)
    exno3 = ghno3 + molal(7) + cnh4no3 - wi(4)
  ELSE
    exno3 = molal(7) + cnh4no3 - wi(4)
  END IF
  exno3 = max(exno3, zero)
  IF (exno3<tiny) GO TO 110 ! No excess NO3, go to next precursor
  !
  IF (molal(7)>exno3) THEN ! Adjust aqueous phase NO3
    molal(7) = molal(7) - exno3
    GO TO 110
  ELSE
    exno3 = exno3 - molal(7)
    molal(7) = zero
  END IF
  !
  IF (cnh4no3>exno3) THEN ! Adjust NH4NO3(s)
    cnh4no3 = cnh4no3 - exno3 ! more solid than excess
    gnh3 = gnh3 + exno3 ! evaporate NO3 to gas phase
    GO TO 110
  ELSE ! less solid than excess
    gnh3 = gnh3 + cnh4no3 ! evaporate into gas phase
    exno3 = exno3 - cnh4no3 ! reduce excess
    cnh4no3 = zero ! zero salt concentration
  END IF
  !
  ! *** FOR CHLORIDE *****************************************************
  !
110 IF (iprob==0) THEN ! Calculate excess (solution - input)
    excl = ghcl + molal(4) + cnh4cl - wi(5)
  ELSE
    excl = molal(4) + cnh4cl - wi(5)
  END IF
  excl = max(excl, zero)
  IF (excl<tiny) GO TO 120 ! No excess Cl, go to next precursor
  !
  IF (molal(4)>excl) THEN ! Adjust aqueous phase Cl
    molal(4) = molal(4) - excl
    GO TO 120
  ELSE
    excl = excl - molal(4)
    molal(4) = zero
  END IF
  !
  IF (cnh4cl>excl) THEN ! Adjust NH4Cl(s)
    cnh4cl = cnh4cl - excl ! more solid than excess
    ghcl = ghcl + excl ! evaporate Cl to gas phase
    GO TO 120
  ELSE ! less solid than excess
    ghcl = ghcl + cnh4cl ! evaporate into gas phase
    excl = excl - cnh4cl ! reduce excess
    cnh4cl = zero ! zero salt concentration
  END IF
  !
  ! *** FOR SULFATE ******************************************************
  !
120 exs4 = molal(5) + molal(6) + 2.D0*clc + cnh42s4 + cnh4hs4 + cna2so4 +    &
       cnahso4 - wi(2)
  exs4 = max(exs4, zero) ! Calculate excess (solution - input)
  IF (exs4<tiny) GO TO 130 ! No excess SO4, return
  !
  IF (molal(6)>exs4) THEN ! Adjust aqueous phase HSO4
    molal(6) = molal(6) - exs4
    GO TO 130
  ELSE
    exs4 = exs4 - molal(6)
    molal(6) = zero
  END IF
  !
  IF (molal(5)>exs4) THEN ! Adjust aqueous phase SO4
    molal(5) = molal(5) - exs4
    GO TO 130
  ELSE
    exs4 = exs4 - molal(5)
    molal(5) = zero
  END IF
  !
  IF (clc>2D0*exs4) THEN ! Adjust (NH4)3H(SO4)2(s)
    clc = clc - exs4/2D0 ! more solid than excess
    gnh3 = gnh3 + 1.5D0*exs4 ! evaporate NH3 to gas phase
    GO TO 130
  ELSE ! less solid than excess
    gnh3 = gnh3 + 1.5D0*clc ! evaporate NH3 to gas phase
    exs4 = exs4 - 2D0*clc ! reduce excess
    clc = zero ! zero salt concentration
  END IF
  !
  IF (cnh4hs4>exs4) THEN ! Adjust NH4HSO4(s)
    cnh4hs4 = cnh4hs4 - exs4 ! more solid than excess
    gnh3 = gnh3 + exs4 ! evaporate NH3 to gas phase
    GO TO 130
  ELSE ! less solid than excess
    gnh3 = gnh3 + cnh4hs4 ! evaporate NH3 to gas phase
    exs4 = exs4 - cnh4hs4 ! reduce excess
    cnh4hs4 = zero ! zero salt concentration
  END IF
  !
  IF (cnh42s4>exs4) THEN ! Adjust (NH4)2SO4(s)
    cnh42s4 = cnh42s4 - exs4 ! more solid than excess
    gnh3 = gnh3 + 2.D0*exs4 ! evaporate NH3 to gas phase
    GO TO 130
  ELSE ! less solid than excess
    gnh3 = gnh3 + 2.D0*cnh42s4 ! evaporate NH3 to gas phase
    exs4 = exs4 - cnh42s4 ! reduce excess
    cnh42s4 = zero ! zero salt concentration
  END IF
  !
  ! *** RETURN **********************************************************
  !
130 RETURN
END SUBROUTINE adjust

!=======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION GETASR
! *** CALCULATES THE LIMITING NH4+/SO4 RATIO OF A SULFATE POOR SYSTEM
!     (i.e. SULFATE RATIO = 2.0) FOR GIVEN SO4 LEVEL AND RH
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION getasr ( so4i, rhi )
  USE mo_aersr
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: so4i
  DOUBLE PRECISION, INTENT(inout) :: rhi
  REAL :: a1
  INTEGER :: ia1
  INTEGER :: indr
  INTEGER :: inds
  INTEGER :: indsh
  INTEGER :: indsl
  INTEGER :: iposh
  INTEGER :: iposl
  REAL :: rat
  REAL :: wf

  !  DOUBLE PRECISION :: so4i, rhi
  !CC
  !CC *** SOLVE USING FULL COMPUTATIONS, NOT LOOK-UP TABLES **************
  !CC
  !CC         W(2) = WAER(2)
  !CC         W(3) = WAER(2)*2.0001D0
  !CC         CALL CALCA2
  !CC         SULRATW = MOLAL(3)/WAER(2)
  !CC         CALL INIT1 (WI, RHI, TEMPI)   ! Re-initialize COMMON BLOCK
  !
  ! *** CALCULATE INDICES ************************************************
  !
  rat = so4i/1.E-9
  a1 = int(alog10(rat)) ! Magnitude of RAT
  ia1 = int(rat/2.5/10.0**a1)
  !
  inds = 4.0*a1 + min(ia1, 4)
  inds = min(max(0,inds), nso4s-1) + 1 ! SO4 component of IPOS
  !
  indr = int(99.0-rhi*100.0) + 1
  indr = min(max(1,indr), nrhs) ! RH component of IPOS
  !
  ! *** GET VALUE AND RETURN *********************************************
  !
  indsl = inds
  indsh = min(indsl+1, nso4s)
  iposl = (indsl-1)*nrhs + indr ! Low position in array
  iposh = (indsh-1)*nrhs + indr ! High position in array
  !
  wf = (so4i-asso4(indsl))/(asso4(indsh)-asso4(indsl)+1E-7)
  wf = min(max(wf,0.0), 1.0)
  !
  getasr = wf*asrat(iposh) + (1.0-wf)*asrat(iposl)
  !
  ! *** END OF FUNCTION GETASR *******************************************
  !
  RETURN
END FUNCTION getasr
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCHA
! *** CALCULATES CHLORIDES SPECIATION
!
!     HYDROCHLORIC ACID IN THE LIQUID PHASE IS ASSUMED A MINOR SPECIES,
!     AND DOES NOT SIGNIFICANTLY PERTURB THE HSO4-SO4 EQUILIBRIUM. THE
!     HYDROCHLORIC ACID DISSOLVED IS CALCULATED FROM THE
!     HCL(G) <-> (H+) + (CL-)
!     EQUILIBRIUM, USING THE (H+) FROM THE SULFATES.
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcha (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: alfa
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: kapa
  DOUBLE PRECISION :: x

  !C      CHARACTER ERRINF*40
  !
  ! *** CALCULATE HCL DISSOLUTION *****************************************
  !
  x = w(5)
  delt = 0.0D0
  IF (water>tiny) THEN
    kapa = molal(1)
    alfa = xk3*r*temp*(water/gama(11))**2.0
    diak = sqrt((kapa+alfa)**2.0+4.0*alfa*x)
    delt = 0.5*(-(kapa+alfa)+diak)
    !C         IF (DELT/KAPA.GT.0.1d0) THEN
    !C            WRITE (ERRINF,'(1PE10.3)') DELT/KAPA*100.0
    !C            CALL PUSHERR (0033, ERRINF)
    !C         ENDIF
  END IF
  !
  ! *** CALCULATE HCL SPECIATION IN THE GAS PHASE *************************
  !
  ghcl = max(x-delt, 0.0D0) ! GAS HCL
  !
  ! *** CALCULATE HCL SPECIATION IN THE LIQUID PHASE **********************
  !
  molal(4) = delt ! CL-
  molal(1) = molal(1) + delt ! H+
  !
  RETURN
  !
END SUBROUTINE calcha
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCHAP
! *** CALCULATES CHLORIDES SPECIATION
!
!     HYDROCHLORIC ACID IN THE LIQUID PHASE IS ASSUMED A MINOR SPECIES,
!     THAT DOES NOT SIGNIFICANTLY PERTURB THE HSO4-SO4 EQUILIBRIUM.
!     THE HYDROCHLORIC ACID DISSOLVED IS CALCULATED FROM THE
!     HCL(G) -> HCL(AQ)   AND  HCL(AQ) ->  (H+) + (CL-)
!     EQUILIBRIA, USING (H+) FROM THE SULFATES.
!
!     THIS IS THE VERSION USED BY THE INVERSE PROBLEM SOVER
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calchap (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: alfa
  DOUBLE PRECISION :: delt
  !
  ! *** IS THERE A LIQUID PHASE? ******************************************
  !
  IF (water<=tiny) RETURN
  !
  ! *** CALCULATE HCL SPECIATION IN THE GAS PHASE *************************
  !
  CALL calcclaq(molal(4), molal(1), delt)
  alfa = xk3*r*temp*(water/gama(11))**2.0
  gasaq(3) = delt
  molal(1) = molal(1) - delt
  molal(4) = molal(4) - delt
  ghcl = molal(1)*molal(4)/alfa
  !
  RETURN
  !
END SUBROUTINE calchap


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCNA
! *** CALCULATES NITRATES SPECIATION
!
!     NITRIC ACID IN THE LIQUID PHASE IS ASSUMED A MINOR SPECIES, THAT
!     DOES NOT SIGNIFICANTLY PERTURB THE HSO4-SO4 EQUILIBRIUM. THE NITRIC
!     ACID DISSOLVED IS CALCULATED FROM THE HNO3(G) -> (H+) + (NO3-)
!     EQUILIBRIUM, USING THE (H+) FROM THE SULFATES.
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcna (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: alfa
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: kapa
  DOUBLE PRECISION :: x
  !C      CHARACTER ERRINF*40
  !
  ! *** CALCULATE HNO3 DISSOLUTION ****************************************
  !
  x = w(4)
  delt = 0.0D0
  IF (water>tiny) THEN
    kapa = molal(1)
    alfa = xk4*r*temp*(water/gama(10))**2.0
    diak = sqrt((kapa+alfa)**2.0+4.0*alfa*x)
    delt = 0.5*(-(kapa+alfa)+diak)
    !C         IF (DELT/KAPA.GT.0.1d0) THEN
    !C            WRITE (ERRINF,'(1PE10.3)') DELT/KAPA*100.0
    !C            CALL PUSHERR (0019, ERRINF)    ! WARNING ERROR: NO SOLUTION
    !C         ENDIF
  END IF
  !
  ! *** CALCULATE HNO3 SPECIATION IN THE GAS PHASE ************************
  !
  ghno3 = max(x-delt, 0.0D0) ! GAS HNO3
  !
  ! *** CALCULATE HNO3 SPECIATION IN THE LIQUID PHASE *********************
  !
  molal(7) = delt ! NO3-
  molal(1) = molal(1) + delt ! H+
  !
  RETURN
  !
END SUBROUTINE calcna



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCNAP
! *** CALCULATES NITRATES SPECIATION
!
!     NITRIC ACID IN THE LIQUID PHASE IS ASSUMED A MINOR SPECIES, THAT
!     DOES NOT SIGNIFICANTLY PERTURB THE HSO4-SO4 EQUILIBRIUM. THE NITRIC
!     ACID DISSOLVED IS CALCULATED FROM THE HNO3(G) -> HNO3(AQ) AND
!     HNO3(AQ) -> (H+) + (CL-) EQUILIBRIA, USING (H+) FROM THE SULFATES.
!
!     THIS IS THE VERSION USED BY THE INVERSE PROBLEM SOVER
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcnap (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: alfa
  DOUBLE PRECISION :: delt
  !
  ! *** IS THERE A LIQUID PHASE? ******************************************
  !
  IF (water<=tiny) RETURN
  !
  ! *** CALCULATE HNO3 SPECIATION IN THE GAS PHASE ************************
  !
  CALL calcniaq(molal(7), molal(1), delt)
  alfa = xk4*r*temp*(water/gama(10))**2.0
  gasaq(3) = delt
  molal(1) = molal(1) - delt
  molal(7) = molal(7) - delt
  ghno3 = molal(1)*molal(7)/alfa

  WRITE (*, *) alfa, molal(1), molal(7), ghno3, delt
  !
  RETURN
  !
END SUBROUTINE calcnap

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCNH3
! *** CALCULATES AMMONIA IN GAS PHASE
!
!     AMMONIA IN THE GAS PHASE IS ASSUMED A MINOR SPECIES, THAT
!     DOES NOT SIGNIFICANTLY PERTURB THE AEROSOL EQUILIBRIUM.
!     AMMONIA GAS IS CALCULATED FROM THE NH3(g) + (H+)(l) <==> (NH4+)(l)
!     EQUILIBRIUM, USING (H+), (NH4+) FROM THE AEROSOL SOLUTION.
!
!     THIS IS THE VERSION USED BY THE DIRECT PROBLEM
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcnh3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: a1
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: chi1
  DOUBLE PRECISION :: chi2
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: psi
  !
  ! *** IS THERE A LIQUID PHASE? ******************************************
  !
  IF (water<=tiny) RETURN
  !
  ! *** CALCULATE NH3 SUBLIMATION *****************************************
  !
  a1 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
  chi1 = molal(3)
  chi2 = molal(1)
  !
  bb = (chi2+one/a1) ! a=1; b!=1; c!=1
  cc = -chi1/a1
  diak = sqrt(bb*bb-4.D0*cc) ! Always > 0
  psi = 0.5*(-bb+diak) ! One positive root
  psi = max(tiny, min(psi,chi1)) ! Constrict in acceptible range
  !
  ! *** CALCULATE NH3 SPECIATION IN THE GAS PHASE *************************
  !
  gnh3 = psi ! GAS HNO3
  !
  ! *** CALCULATE NH3 AFFECT IN THE LIQUID PHASE **************************
  !
  molal(3) = chi1 - psi ! NH4+
  molal(1) = chi2 + psi ! H+
  !
  RETURN
  !
END SUBROUTINE calcnh3



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCNH3P
! *** CALCULATES AMMONIA IN GAS PHASE
!
!     AMMONIA GAS IS CALCULATED FROM THE NH3(g) + (H+)(l) <==> (NH4+)(l)
!     EQUILIBRIUM, USING (H+), (NH4+) FROM THE AEROSOL SOLUTION.
!
!     THIS IS THE VERSION USED BY THE INVERSE PROBLEM SOLVER
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcnh3p (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: a1
  !
  ! *** IS THERE A LIQUID PHASE? ******************************************
  !
  IF (water<=tiny) RETURN
  !
  ! *** CALCULATE NH3 GAS PHASE CONCENTRATION *****************************
  !
  a1 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
  gnh3 = molal(3)/molal(1)/a1
  !
  RETURN
  !
END SUBROUTINE calcnh3p


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCNHA
!
!     THIS SUBROUTINE CALCULATES THE DISSOLUTION OF HCL, HNO3 AT
!     THE PRESENCE OF (H,SO4). HCL, HNO3 ARE CONSIDERED MINOR SPECIES,
!     THAT DO NOT SIGNIFICANTLY AFFECT THE EQUILIBRIUM POINT.
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcnha (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: a3
  DOUBLE PRECISION :: a4
  DOUBLE PRECISION :: c1
  DOUBLE PRECISION :: c2
  DOUBLE PRECISION :: c3
  DOUBLE PRECISION :: chi3
  DOUBLE PRECISION :: chi4
  DOUBLE PRECISION :: delcl
  DOUBLE PRECISION :: delno
  character(len=40) :: errinf
  INTEGER :: islv
  DOUBLE PRECISION :: m1
  DOUBLE PRECISION :: m2
  DOUBLE PRECISION :: m3
  DOUBLE PRECISION :: omega
  !
  ! *** SPECIAL CASE; WATER=ZERO ******************************************
  !
  IF (water<=tiny) THEN
    GO TO 100
    !
    ! *** SPECIAL CASE; HCL=HNO3=ZERO ***************************************
    !
  ELSE IF (w(5)<=tiny .AND. w(4)<=tiny) THEN
    GO TO 110
    !
    ! *** SPECIAL CASE; HCL=ZERO ********************************************
    !
  ELSE IF (w(5)<=tiny) THEN
    CALL calcna ! CALL HNO3 DISSOLUTION ROUTINE
    GO TO 110
    !
    ! *** SPECIAL CASE; HNO3=ZERO *******************************************
    !
  ELSE IF (w(4)<=tiny) THEN
    CALL calcha ! CALL HCL DISSOLUTION ROUTINE
    GO TO 110
  END IF
  !
  ! *** CALCULATE EQUILIBRIUM CONSTANTS ***********************************
  !
  a3 = xk4*r*temp*(water/gama(10))**2.0 ! HNO3
  a4 = xk3*r*temp*(water/gama(11))**2.0 ! HCL
  !
  ! *** CALCULATE CUBIC EQUATION COEFFICIENTS *****************************
  !
  delcl = zero
  delno = zero
  !
  omega = molal(1) ! H+
  chi3 = w(4) ! HNO3
  chi4 = w(5) ! HCL
  !
  c1 = a3*chi3
  c2 = a4*chi4
  c3 = a3 - a4
  !
  m1 = (c1+c2+(omega+a4)*c3)/c3
  m2 = ((omega+a4)*c2-a4*c3*chi4)/c3
  m3 = -a4*c2*chi4/c3
  !
  ! *** CALCULATE ROOTS ***************************************************
  !
  CALL poly3(m1, m2, m3, delcl, islv) ! HCL DISSOLUTION
  IF (islv/=0) THEN
    delcl = tiny ! TINY AMOUNTS OF HCL ASSUMED WHEN NO ROOT
    WRITE (errinf, '(1PE7.1)') tiny
    CALL pusherr(0022, errinf) ! WARNING ERROR: NO SOLUTION
  END IF
  delcl = min(delcl, chi4)
  !
  delno = c1*delcl/(c2+c3*delcl)
  delno = min(delno, chi3)
  !
  IF (delcl<zero .OR. delno<zero .OR. delcl>chi4 .OR. delno>chi3) THEN
    delcl = tiny ! TINY AMOUNTS OF HCL ASSUMED WHEN NO ROOT
    delno = tiny
    WRITE (errinf, '(1PE7.1)') tiny
    CALL pusherr(0022, errinf) ! WARNING ERROR: NO SOLUTION
  END IF
  !CC
  !CC *** COMPARE DELTA TO TOTAL H+ ; ESTIMATE EFFECT TO HSO4 ***************
  !CC
  !C      IF ((DELCL+DELNO)/MOLAL(1).GT.0.1d0) THEN
  !C         WRITE (ERRINF,'(1PE10.3)') (DELCL+DELNO)/MOLAL(1)*100.0
  !C         CALL PUSHERR (0021, ERRINF)
  !C      ENDIF
  !
  ! *** EFFECT ON LIQUID PHASE ********************************************
  !
  molal(1) = molal(1) + (delno+delcl) ! H+   CHANGE
  molal(4) = molal(4) + delcl ! CL-  CHANGE
  molal(7) = molal(7) + delno ! NO3- CHANGE
  !
  ! *** EFFECT ON GAS PHASE ***********************************************
  !
100 ghcl = max(w(5)-molal(4), tiny)
  ghno3 = max(w(4)-molal(7), tiny)
  !
110 RETURN
  !
END SUBROUTINE calcnha



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCNHP
!
!     THIS SUBROUTINE CALCULATES THE GAS PHASE NITRIC AND HYDROCHLORIC
!     ACID. CONCENTRATIONS ARE CALCULATED FROM THE DISSOLUTION
!     EQUILIBRIA, USING (H+), (Cl-), (NO3-) IN THE AEROSOL PHASE.
!
!     THIS IS THE VERSION USED BY THE INVERSE PROBLEM SOLVER
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcnhp (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: a3
  DOUBLE PRECISION :: a4
  DOUBLE PRECISION :: delt
  !
  ! *** IS THERE A LIQUID PHASE? ******************************************
  !
  IF (water<=tiny) RETURN
  !
  ! *** CALCULATE EQUILIBRIUM CONSTANTS ***********************************
  !
  a3 = xk3*r*temp*(water/gama(11))**2.0
  a4 = xk4*r*temp*(water/gama(10))**2.0
  molal(1) = molal(1) + waer(4) + waer(5) ! H+ increases because NO3, Cl are added.
  !
  ! *** CALCULATE CONCENTRATIONS ******************************************
  ! *** ASSUME THAT 'DELT' FROM HNO3 >> 'DELT' FROM HCL
  !
  CALL calcniaq(waer(4), molal(1)+molal(7)+molal(4), delt)
  molal(1) = molal(1) - delt
  molal(7) = waer(4) - delt ! NO3- = Waer(4) minus any turned into (HNO3aq)
  gasaq(3) = delt
  !
  CALL calcclaq(waer(5), molal(1)+molal(7)+molal(4), delt)
  molal(1) = molal(1) - delt
  molal(4) = waer(5) - delt ! Cl- = Waer(4) minus any turned into (HNO3aq)
  gasaq(2) = delt
  !
  ghno3 = molal(1)*molal(7)/a4
  ghcl = molal(1)*molal(4)/a3
  !
  RETURN
  !
END SUBROUTINE calcnhp

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCHCO3
! *** CORRECTS FOR H+ WHEN CRUSTALS ARE IN EXCESS
!
!     CARBONATES ARE IN EXCESS, HCO3- IS ASSUMED A MINOR SPECIES,
!     THE H+ CONCENTRATION IS CALCULATED FROM THE
!     CO2(aq) + H2O <-> (HCO3-) + (H+)
!     HCO3- <-> (H+) + (CO3--) EQUILIBRIUM.
!     THE CO3-- CONCENTRATION IS ASSUMED NEGLIGIBLE WITH RESPECT TO HCO3-
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
!      SUBROUTINE CALCHCO3
!      USE mo_isrpia
!      DOUBLE PRECISION KAPA
!CC      CHARACTER ERRINF*40
!C
!C *** SPECIAL CASE; WATER=ZERO ******************************************
!C
!      IF (WATER.LE.TINY) THEN
!         GOTO 521
!      ENDIF
!C
!C *** CALCULATE CO2 DISSOLUTION *****************************************
!C
!      REST = 2.D0*W(2) + W(4) + W(5)
!C
!      DELT = 0.0d0
!C      DELT2 = 0.0d0
!      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
!      KAPA = MOLAL(1)
!C
!C *** CALCULATE EQUILIBRIUM CONSTANTS ***********************************
!C
!      ALFA = XK26*RH*(WATER/1.0)                 ! CO2(aq) + H2O
!C      ALFA2 = XK27*(WATER/1.0)                    ! HCO3-
!C
!C *** CALCULATE CUBIC EQUATION COEFFICIENTS *****************************
!C
!      X  = W(1)+W(6)+W(7)+W(8) - REST          ! EXCESS OF CRUSTALS EQUALS HCO3-
!C
!      BB =-(KAPA + X + ALFA)
!      CC = KAPA*X
!      DD = BB*BB - 4.D0*CC
!C
!      IF (DD.GE.ZERO) THEN
!         SQDD  = SQRT(DD)
!         DELT  = 0.5*(-BB - SQDD)
!      ELSE
!         DELT  = ZERO
!      ENDIF
!
!      ENDIF
!C
!C *** CALCULATE H+ *****************************************************
!C
!      MOLAL(1) = KAPA - DELT             ! H+
!C
!521   RETURN
!C
!      END
!C
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCAMAQ
! *** THIS SUBROUTINE CALCULATES THE NH3(aq) GENERATED FROM (H,NH4+).
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcamaq ( nh4i, ohi, delt )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: nh4i
  DOUBLE PRECISION, INTENT(inout) :: ohi
  DOUBLE PRECISION, INTENT(inout) :: delt
  DOUBLE PRECISION :: a22
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del1
  DOUBLE PRECISION :: del2
  DOUBLE PRECISION :: om1
  DOUBLE PRECISION :: om2
  !C      CHARACTER ERRINF*40
  !
  ! *** EQUILIBRIUM CONSTANTS
  !
  a22 = xk22/xkw/water*(gama(8)/gama(9))**2. ! GAMA(NH3) ASSUMED 1
  akw = xkw*rh*water*water
  !
  ! *** FIND ROOT
  !
  om1 = nh4i
  om2 = ohi
  bb = -(om1+om2+a22*akw)
  cc = om1*om2
  dd = sqrt(bb*bb-4.D0*cc)

  del1 = 0.5D0*(-bb-dd)
  del2 = 0.5D0*(-bb+dd)
  !
  ! *** GET APPROPRIATE ROOT.
  !
  IF (del1<zero) THEN
    IF (del2>nh4i .OR. del2>ohi) THEN
      delt = zero
    ELSE
      delt = del2
    END IF
  ELSE
    delt = del1
  END IF
  !C
  !C *** COMPARE DELTA TO TOTAL NH4+ ; ESTIMATE EFFECT *********************
  !C
  !C      IF (DELTA/HYD.GT.0.1d0) THEN
  !C         WRITE (ERRINF,'(1PE10.3)') DELTA/HYD*100.0
  !C         CALL PUSHERR (0020, ERRINF)
  !C      ENDIF
  !
  RETURN
  !
END SUBROUTINE calcamaq



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCAMAQ2
!
!     THIS SUBROUTINE CALCULATES THE NH3(aq) GENERATED FROM (H,NH4+).
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcamaq2 ( ggnh3, nh4i, ohi, nh3aq )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(in)    :: ggnh3
  DOUBLE PRECISION, INTENT(inout) :: nh4i
  DOUBLE PRECISION, INTENT(inout) :: ohi
  DOUBLE PRECISION, INTENT(inout) :: nh3aq
  DOUBLE PRECISION :: a22
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: alf1
  DOUBLE PRECISION :: alf2
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: del
  !
  ! *** EQUILIBRIUM CONSTANTS
  !
  a22 = xk22/xkw/water*(gama(8)/gama(9))**2. ! GAMA(NH3) ASSUMED 1
  akw = xkw*rh*water*water
  !
  ! *** FIND ROOT
  !
  alf1 = nh4i - ggnh3
  alf2 = ggnh3
  bb = alf1 + a22*akw
  cc = -a22*akw*alf2
  del = 0.5D0*(-bb+sqrt(bb*bb-4.D0*cc))
  !
  ! *** ADJUST CONCENTRATIONS
  !
  nh4i = alf1 + del
  ohi = del
  IF (ohi<=tiny) ohi = sqrt(akw) ! If solution is neutral.
  nh3aq = alf2 - del
  !
  RETURN
  !
END SUBROUTINE calcamaq2



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCCLAQ
!
!     THIS SUBROUTINE CALCULATES THE HCL(aq) GENERATED FROM (H+,CL-).
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcclaq ( cli, hi, delt )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: cli
  DOUBLE PRECISION, INTENT(in)    :: hi
  DOUBLE PRECISION, INTENT(inout) :: delt
  DOUBLE PRECISION :: a32
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del1
  DOUBLE PRECISION :: del2
  DOUBLE PRECISION :: om1
  DOUBLE PRECISION :: om2
  !
  ! *** EQUILIBRIUM CONSTANTS
  !
  a32 = xk32*water/(gama(11))**2. ! GAMA(HCL) ASSUMED 1
  !
  ! *** FIND ROOT
  !
  om1 = cli
  om2 = hi
  bb = -(om1+om2+a32)
  cc = om1*om2
  dd = sqrt(bb*bb-4.D0*cc)

  del1 = 0.5D0*(-bb-dd)
  del2 = 0.5D0*(-bb+dd)
  !
  ! *** GET APPROPRIATE ROOT.
  !
  IF (del1<zero) THEN
    IF (del2<zero .OR. del2>cli .OR. del2>hi) THEN
      delt = zero
    ELSE
      delt = del2
    END IF
  ELSE
    delt = del1
  END IF
  !
  RETURN
  !
END SUBROUTINE calcclaq



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCCLAQ2
!
!     THIS SUBROUTINE CALCULATES THE HCL(aq) GENERATED FROM (H+,CL-).
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcclaq2 ( ggcl, cli, hi, claq )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: ggcl
  DOUBLE PRECISION, INTENT(inout) :: cli
  DOUBLE PRECISION, INTENT(inout) :: hi
  DOUBLE PRECISION, INTENT(inout) :: claq
  DOUBLE PRECISION :: a32
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: alf1
  DOUBLE PRECISION :: alf2
  DOUBLE PRECISION :: coef
  DOUBLE PRECISION :: del1
  !
  ! *** EQUILIBRIUM CONSTANTS
  !
  a32 = xk32*water/(gama(11))**2. ! GAMA(HCL) ASSUMED 1
  akw = xkw*rh*water*water
  !
  ! *** FIND ROOT
  !
  alf1 = cli - ggcl
  alf2 = ggcl
  coef = (alf1+a32)
  del1 = 0.5*(-coef+sqrt(coef*coef+4.D0*a32*alf2))
  !
  ! *** CORRECT CONCENTRATIONS
  !
  cli = alf1 + del1
  hi = del1
  IF (hi<=tiny) hi = sqrt(akw) ! If solution is neutral.
  claq = alf2 - del1
  !
  RETURN
  !
END SUBROUTINE calcclaq2



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCNIAQ
!
!     THIS SUBROUTINE CALCULATES THE HNO3(aq) GENERATED FROM (H,NO3-).
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcniaq ( no3i, hi, delt )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: no3i
  DOUBLE PRECISION, INTENT(in)    :: hi
  DOUBLE PRECISION, INTENT(inout) :: delt
  DOUBLE PRECISION :: a42
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del1
  DOUBLE PRECISION :: del2
  DOUBLE PRECISION :: om1
  DOUBLE PRECISION :: om2
  !
  ! *** EQUILIBRIUM CONSTANTS
  !
  a42 = xk42*water/(gama(10))**2. ! GAMA(HNO3) ASSUMED 1
  !
  ! *** FIND ROOT
  !
  om1 = no3i
  om2 = hi
  bb = -(om1+om2+a42)
  cc = om1*om2
  dd = sqrt(bb*bb-4.D0*cc)

  del1 = 0.5D0*(-bb-dd)
  del2 = 0.5D0*(-bb+dd)
  !
  ! *** GET APPROPRIATE ROOT.
  !
  IF (del1<zero .OR. del1>hi .OR. del1>no3i) THEN
    PRINT *, delt
    delt = zero
  ELSE
    delt = del1
    RETURN
  END IF
  !
  IF (del2<zero .OR. del2>no3i .OR. del2>hi) THEN
    delt = zero
  ELSE
    delt = del2
  END IF
  !
  RETURN
  !
END SUBROUTINE calcniaq



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCNIAQ2
!
!     THIS SUBROUTINE CALCULATES THE UNDISSOCIATED HNO3(aq)
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcniaq2 ( ggno3, no3i, hi, no3aq )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: ggno3
  DOUBLE PRECISION, INTENT(inout) :: no3i
  DOUBLE PRECISION, INTENT(inout) :: hi
  DOUBLE PRECISION, INTENT(inout) :: no3aq
  DOUBLE PRECISION :: a42
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: alf1
  DOUBLE PRECISION :: alf2
  DOUBLE PRECISION :: alf3
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: del1
  !
  ! *** EQUILIBRIUM CONSTANTS
  !
  a42 = xk42*water/(gama(10))**2. ! GAMA(HNO3) ASSUMED 1
  akw = xkw*rh*water*water
  !
  ! *** FIND ROOT
  !
  alf1 = no3i - ggno3
  alf2 = ggno3
  alf3 = hi
  !
  bb = alf3 + alf1 + a42
  cc = alf3*alf1 - a42*alf2
  del1 = 0.5*(-bb+sqrt(bb*bb-4.D0*cc))
  !
  ! *** CORRECT CONCENTRATIONS
  !
  no3i = alf1 + del1
  hi = alf3 + del1
  IF (hi<=tiny) hi = sqrt(akw) ! If solution is neutral.
  no3aq = alf2 - del1
  !
  RETURN
  !
END SUBROUTINE calcniaq2

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCMR
! *** THIS SUBROUTINE CALCULATES:
!     1. ION PAIR CONCENTRATIONS (FROM [MOLAR] ARRAY)
!     2. WATER CONTENT OF LIQUID AEROSOL PHASE (FROM ZSR CORRELATION)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcmr (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: aml5
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frnh4
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  character(len=1) sc
  DOUBLE PRECISION :: so4i
  DOUBLE PRECISION :: tots4
  !
  ! *** CALCULATE ION PAIR CONCENTRATIONS ACCORDING TO SPECIFIC CASE ****
  !
  sc = scase(1:1) ! SULRAT & SODRAT case
  !
  ! *** NH4-SO4 SYSTEM ; SULFATE POOR CASE
  !
  IF (sc=='A') THEN
    molalr(4) = molal(5) + molal(6)        ! (NH4)2SO4 - CORRECT FOR SO4 TO HSO4
    !
    ! *** NH4-SO4 SYSTEM ; SULFATE RICH CASE ; NO FREE ACID
    !
  ELSE IF (sc=='B') THEN
    so4i = molal(5) - molal(1) ! CORRECT FOR HSO4 DISSOCIATION
    hso4i = molal(6) + molal(1)
    IF (so4i<hso4i) THEN
      molalr(13) = so4i ! [LC] = [SO4]
      molalr(9) = max(hso4i-so4i, zero)    ! NH4HSO4
    ELSE
      molalr(13) = hso4i ! [LC] = [HSO4]
      molalr(4) = max(so4i-hso4i, zero)    ! (NH4)2SO4
    END IF
    !
    ! *** NH4-SO4 SYSTEM ; SULFATE RICH CASE ; FREE ACID
    !
  ELSE IF (sc=='C') THEN
    molalr(9) = molal(3)                      ! NH4HSO4
    molalr(7) = max(w(2)-w(3), zero)          ! H2SO4
    !
    ! *** NH4-SO4-NO3 SYSTEM ; SULFATE POOR CASE
    !
  ELSE IF (sc=='D') THEN
    molalr(4) = molal(5) + molal(6)           ! (NH4)2SO4
    aml5 = molal(3) - 2.D0*molalr(4)          ! "free" NH4
    molalr(5) = max(min(aml5,molal(7)), zero) ! NH4NO3 = MIN("free", NO3)
    !
    ! *** NH4-SO4-NO3 SYSTEM ; SULFATE RICH CASE ; NO FREE ACID
    !
  ELSE IF (sc=='E') THEN
    so4i = max(molal(5)-molal(1), zero)       ! FROM HSO4 DISSOCIATION
    hso4i = molal(6) + molal(1)
    IF (so4i<hso4i) THEN
      molalr(13) = so4i                       ! [LC] = [SO4]
      molalr(9) = max(hso4i-so4i, zero)       ! NH4HSO4
    ELSE
      molalr(13) = hso4i                      ! [LC] = [HSO4]
      molalr(4) = max(so4i-hso4i, zero)       ! (NH4)2SO4
    END IF
    !
    ! *** NH4-SO4-NO3 SYSTEM ; SULFATE RICH CASE ; FREE ACID
    !
  ELSE IF (sc=='F') THEN
    molalr(9) = molal(3)                              ! NH4HSO4
    molalr(7) = max(molal(5)+molal(6)-molal(3), zero) ! H2SO4
    !
    ! *** NA-NH4-SO4-NO3-CL SYSTEM ; SULFATE POOR ; SODIUM POOR CASE
    !
  ELSE IF (sc=='G') THEN
    molalr(2) = 0.5D0*molal(2)                         ! NA2SO4
    tots4 = molal(5) + molal(6)                        ! Total SO4
    molalr(4) = max(tots4-molalr(2), zero)             ! (NH4)2SO4
    frnh4 = max(molal(3)-2.D0*molalr(4), zero)
    molalr(5) = min(molal(7), frnh4)                   ! NH4NO3
    frnh4 = max(frnh4-molalr(5), zero)
    molalr(6) = min(molal(4), frnh4)                   ! NH4CL
    !
    ! *** NA-NH4-SO4-NO3-CL SYSTEM ; SULFATE POOR ; SODIUM RICH CASE
    ! *** RETREIVE DISSOLVED SALTS DIRECTLY FROM COMMON BLOCK /SOLUT/
    !
  ELSE IF (sc=='H') THEN
    molalr(1) = psi7                                   ! NACL
    molalr(2) = psi1                                   ! NA2SO4
    molalr(3) = psi8                                   ! NANO3
    molalr(4) = zero                                   ! (NH4)2SO4
    frno3 = max(molal(7)-molalr(3), zero)              ! "FREE" NO3
    frcl = max(molal(4)-molalr(1), zero)               ! "FREE" CL
    molalr(5) = min(molal(3), frno3)                   ! NH4NO3
    frnh4 = max(molal(3)-molalr(5), zero)              ! "FREE" NH3
    molalr(6) = min(frcl, frnh4)                       ! NH4CL
    !
    ! *** NA-NH4-SO4-NO3-CL SYSTEM ; SULFATE RICH CASE ; NO FREE ACID
    ! *** RETREIVE DISSOLVED SALTS DIRECTLY FROM COMMON BLOCK /SOLUT/
    !
  ELSE IF (sc=='I') THEN
    molalr(04) = psi5 ! (NH4)2SO4
    molalr(02) = psi4 ! NA2SO4
    molalr(09) = psi1 ! NH4HSO4
    molalr(12) = psi3 ! NAHSO4
    molalr(13) = psi2 ! LC
    !
    ! *** NA-NH4-SO4-NO3-CL SYSTEM ; SULFATE RICH CASE ; FREE ACID
    !
  ELSE IF (sc=='J') THEN
    molalr(09) = molal(3) ! NH4HSO4
    molalr(12) = molal(2) ! NAHSO4
    molalr(07) = molal(5) + molal(6) - molal(3) - molal(2) ! H2SO4
    molalr(07) = max(molalr(07), zero)
    !
    ! *** NA-NH4-SO4-NO3-CL-CA-K-MG SYSTEM ; SULFATE POOR ; CR+NA POOR CASE
    !
  ELSE IF (sc=='O') THEN
    molalr(2) = 0.5D0*molal(2) ! NA2SO4
    tots4 = molal(5) + molal(6) ! Total SO4
    molalr(17) = 0.5*molal(9) ! K2SO4
    molalr(21) = molal(10) ! MGSO4
    molalr(4) = max(tots4-molalr(2)-molalr(17)-molalr(21), zero) ! (NH4)2SO4
    frnh4 = max(molal(3)-2.D0*molalr(4), zero)
    molalr(5) = min(molal(7), frnh4) ! NH4NO3
    frnh4 = max(frnh4-molalr(5), zero)
    molalr(6) = min(molal(4), frnh4) ! NH4CL
    !
    ! *** NA-NH4-SO4-NO3-CL-CA-K-MG SYSTEM ; SULFATE POOR ; CR+NA RICH; CR POOR CASE
    ! *** RETREIVE DISSOLVED SALTS DIRECTLY FROM COMMON BLOCK /SOLUT/
    !
  ELSE IF (sc=='M') THEN
    molalr(1) = psi7 ! NACL
    molalr(2) = psi1 ! NA2SO4
    molalr(3) = psi8 ! NANO3
    molalr(4) = zero ! (NH4)2SO4
    frno3 = max(molal(7)-molalr(3), zero) ! "FREE" NO3
    frcl = max(molal(4)-molalr(1), zero) ! "FREE" CL
    molalr(5) = min(molal(3), frno3) ! NH4NO3
    frnh4 = max(molal(3)-molalr(5), zero) ! "FREE" NH3
    molalr(6) = min(frcl, frnh4) ! NH4CL
    molalr(17) = psi9 ! K2SO4
    molalr(21) = psi10 ! MGSO4
    !
    ! *** NA-NH4-SO4-NO3-CL-CA-K-MG SYSTEM ; SULFATE POOR ; CR+NA RICH; CR RICH CASE
    ! *** RETREIVE DISSOLVED SALTS DIRECTLY FROM COMMON BLOCK /SOLUT/
    !
  ELSE IF (sc=='P') THEN
    molalr(1) = psi7 ! NACL
    molalr(3) = psi8 ! NANO3
    molalr(15) = psi12 ! CANO32
    molalr(16) = psi17 ! CACL2
    molalr(19) = psi13 ! KNO3
    molalr(20) = psi14 ! KCL
    molalr(22) = psi15 ! MGNO32
    molalr(23) = psi16 ! MGCL2
    frno3 = max(molal(7)-molalr(3)-2.D0*molalr(15)-molalr(19)-             &
         2.D0*molalr(22), zero) ! "FREE" NO3
    frcl = max(molal(4)-molalr(1)-2.D0*molalr(16)-molalr(20)-              &
         2.D0*molalr(23), zero) ! "FREE" CL
    molalr(5) = min(molal(3), frno3) ! NH4NO3
    frnh4 = max(molal(3)-molalr(5), zero) ! "FREE" NH3
    molalr(6) = min(frcl, frnh4) ! NH4CL
    molalr(17) = psi9 ! K2SO4
    molalr(21) = psi10 ! MGSO4
    !
    ! *** NA-NH4-SO4-NO3-CL-CA-K-MG SYSTEM ; SULFATE RICH CASE ; NO FREE ACID
    !
  ELSE IF (sc=='L') THEN
    molalr(04) = psi5 ! (NH4)2SO4
    molalr(02) = psi4 ! NA2SO4
    molalr(09) = psi1 ! NH4HSO4
    molalr(12) = psi3 ! NAHSO4
    molalr(13) = psi2 ! LC
    molalr(17) = psi6 ! K2SO4
    molalr(21) = psi7 ! MGSO4
    molalr(18) = psi8 ! KHSO4
    !
    ! *** NA-NH4-SO4-NO3-CL-CA-K-MG SYSTEM ; SULFATE SUPER RICH CASE ; FREE ACID
    !
  ELSE IF (sc=='K') THEN
    molalr(09) = molal(3) ! NH4HSO4
    molalr(12) = molal(2) ! NAHSO4
    molalr(14) = molal(8) ! CASO4
    molalr(18) = molal(9) ! KHSO4
    molalr(07) = molal(5) + molal(6) - molal(3) - molal(2) - molal(8) -    &
         molal(9) - molal(10) ! H2SO4
    molalr(07) = max(molalr(07), zero)
    !
    ! ======= REVERSE PROBLEMS ===========================================
    !
    ! *** NH4-SO4-NO3 SYSTEM ; SULFATE POOR CASE
    !
  ELSE IF (sc=='N') THEN
    molalr(4) = molal(5) + molal(6) ! (NH4)2SO4
    aml5 = waer(3) - 2.D0*molalr(4) ! "free" NH4
    molalr(5) = max(min(aml5,waer(4)), zero) ! NH4NO3 = MIN("free", NO3)
    !
    ! *** NH4-SO4-NO3-NA-CL SYSTEM ; SULFATE POOR, SODIUM POOR CASE
    !
  ELSE IF (sc=='Q') THEN
    molalr(2) = psi1 ! NA2SO4
    molalr(4) = psi6 ! (NH4)2SO4
    molalr(5) = psi5 ! NH4NO3
    molalr(6) = psi4 ! NH4CL
    !
    ! *** NH4-SO4-NO3-NA-CL SYSTEM ; SULFATE POOR, SODIUM RICH CASE
    !
  ELSE IF (sc=='R') THEN
    molalr(1) = psi3 ! NACL
    molalr(2) = psi1 ! NA2SO4
    molalr(3) = psi2 ! NANO3
    molalr(4) = zero ! (NH4)2SO4
    molalr(5) = psi5 ! NH4NO3
    molalr(6) = psi4 ! NH4CL
    !
    ! *** NH4-SO4-NO3-NA-CL-CA-K-MG SYSTEM ; SULFATE POOR, CRUSTAL&SODIUM POOR CASE
    !
  ELSE IF (sc=='V') THEN
    molalr(2) = psi1 ! NA2SO4
    molalr(4) = psi6 ! (NH4)2SO4
    molalr(5) = psi5 ! NH4NO3
    molalr(6) = psi4 ! NH4CL
    molalr(17) = psi7 ! K2SO4
    molalr(21) = psi8 ! MGSO4
    !
    ! *** NH4-SO4-NO3-NA-CL-CA-K-MG SYSTEM ; SULFATE POOR, CRUSTAL&SODIUM RICH, CRUSTAL POOR CASE
    !
  ELSE IF (sc=='U') THEN
    molalr(1) = psi3 ! NACL
    molalr(2) = psi1 ! NA2SO4
    molalr(3) = psi2 ! NANO3
    molalr(5) = psi5 ! NH4NO3
    molalr(6) = psi4 ! NH4CL
    molalr(17) = psi7 ! K2SO4
    molalr(21) = psi8 ! MGSO4
    !
    ! *** NH4-SO4-NO3-NA-CL-CA-K-MG SYSTEM ; SULFATE POOR, CRUSTAL&SODIUM RICH, CRUSTAL RICH CASE
    !
  ELSE IF (sc=='W') THEN
    molalr(1) = psi7 ! NACL
    molalr(3) = psi8 ! NANO3
    molalr(5) = psi6 ! NH4NO3
    molalr(6) = psi5 ! NH4CL
    molalr(15) = psi12 ! CANO32
    molalr(16) = psi17 ! CACL2
    molalr(17) = psi9 ! K2SO4
    molalr(19) = psi13 ! KNO3
    molalr(20) = psi14 ! KCL
    molalr(21) = psi10 ! MGSO4
    molalr(22) = psi15 ! MGNO32
    molalr(23) = psi16 ! MGCL2
    !
    ! *** UNKNOWN CASE
    !
    !      ELSE
    !         CALL PUSHERR (1001, ' ') ! FATAL ERROR: CASE NOT SUPPORTED
  END IF
  !
  ! *** CALCULATE WATER CONTENT ; ZSR CORRELATION ***********************
  !
  water = zero
  DO i = 1, npair
    water = water + molalr(i)/m0(i)
  END DO
  water = max(water, tiny)
  !
  RETURN
  !
END SUBROUTINE calcmr
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCMDRH
!
!     THIS IS THE CASE WHERE THE RELATIVE HUMIDITY IS IN THE MUTUAL
!     DRH REGION. THE SOLUTION IS ASSUMED TO BE THE SUM OF TWO WEIGHTED
!     SOLUTIONS ; THE 'DRY' SOLUTION (SUBROUTINE DRYCASE) AND THE
!     'SATURATED LIQUID' SOLUTION (SUBROUTINE LIQCASE).
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcmdrh ( rhi, rhdry, rhliq,  drycase, liqcase )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: rhdry
  DOUBLE PRECISION, INTENT(inout) :: rhliq
  DOUBLE PRECISION :: clco
  DOUBLE PRECISION :: cna2so
  DOUBLE PRECISION :: cnaclo
  DOUBLE PRECISION :: cnahso
  DOUBLE PRECISION :: cnano
  DOUBLE PRECISION :: cnh42so
  DOUBLE PRECISION :: cnh4clo
  DOUBLE PRECISION :: cnh4hso
  DOUBLE PRECISION :: cnh4n3o
  DOUBLE PRECISION :: dambis
  DOUBLE PRECISION :: damchl
  DOUBLE PRECISION :: damg
  DOUBLE PRECISION :: damnit
  DOUBLE PRECISION :: damsul
  DOUBLE PRECISION :: dhag
  DOUBLE PRECISION :: dlc
  DOUBLE PRECISION :: dnag
  DOUBLE PRECISION :: dsobis
  DOUBLE PRECISION :: dsochl
  DOUBLE PRECISION :: dsonit
  DOUBLE PRECISION :: dsosul
  DOUBLE PRECISION :: ghclo
  DOUBLE PRECISION :: ghno3o
  DOUBLE PRECISION :: gnh3o
  INTEGER :: i
  DOUBLE PRECISION :: onemwf
  DOUBLE PRECISION :: wf

  !
  ! *** FIND WEIGHT FACTOR **********************************************
  !
  IF (wftyp==0) THEN
    wf = one
  ELSE IF (wftyp==1) THEN
    wf = 0.5D0
  ELSE
    wf = (rhliq-rhi)/(rhliq-rhdry)
  END IF
  onemwf = one - wf
  !
  ! *** FIND FIRST SECTION ; DRY ONE ************************************
  !
  CALL drycase
  IF (abs(onemwf)<=1D-5) GO TO 100 ! DRY AEROSOL
  !
  cnh42so = cnh42s4 ! FIRST (DRY) SOLUTION
  cnh4hso = cnh4hs4
  clco = clc
  cnh4n3o = cnh4no3
  cnh4clo = cnh4cl
  cna2so = cna2so4
  cnahso = cnahso4
  cnano = cnano3
  cnaclo = cnacl
  gnh3o = gnh3
  ghno3o = ghno3
  ghclo = ghcl
  !
  ! *** FIND SECOND SECTION ; DRY & LIQUID ******************************
  !
  cnh42s4 = zero
  cnh4hs4 = zero
  clc = zero
  cnh4no3 = zero
  cnh4cl = zero
  cna2so4 = zero
  cnahso4 = zero
  cnano3 = zero
  cnacl = zero
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  CALL liqcase ! SECOND (LIQUID) SOLUTION
  !
  ! *** ADJUST THINGS FOR THE CASE THAT THE LIQUID SUB PREDICTS DRY AEROSOL
  !
  IF (water<=tiny) THEN
    DO i = 1, nions
      molal(i) = zero ! Aqueous phase
    END DO
    water = zero
    !
    cnh42s4 = cnh42so ! Solid phase
    cna2so4 = cna2so
    cnahso4 = cnahso
    cnh4hs4 = cnh4hso
    clc = clco
    cnh4no3 = cnh4n3o
    cnano3 = cnano
    cnacl = cnaclo
    cnh4cl = cnh4clo
    !
    gnh3 = gnh3o ! Gas phase
    ghno3 = ghno3o
    ghcl = ghclo
    !
    GO TO 100
  END IF
  !
  ! *** FIND SALT DISSOLUTIONS BETWEEN DRY & LIQUID SOLUTIONS.
  !
  damsul = cnh42so - cnh42s4
  dsosul = cna2so - cna2so4
  dambis = cnh4hso - cnh4hs4
  dsobis = cnahso - cnahso4
  dlc = clco - clc
  damnit = cnh4n3o - cnh4no3
  damchl = cnh4clo - cnh4cl
  dsonit = cnano - cnano3
  dsochl = cnaclo - cnacl
  !
  ! *** FIND GAS DISSOLUTIONS BETWEEN DRY & LIQUID SOLUTIONS.
  !
  damg = gnh3o - gnh3
  dhag = ghclo - ghcl
  dnag = ghno3o - ghno3
  !
  ! *** FIND SOLUTION AT MDRH BY WEIGHTING DRY & LIQUID SOLUTIONS.
  !
  !     LIQUID
  !
  molal(1) = onemwf*molal(1) ! H+
  molal(2) = onemwf*(2.D0*dsosul+dsobis+dsonit+dsochl) ! NA+
  molal(3) = onemwf*(2.D0*damsul+damg+dambis+damchl+3.D0*dlc+damnit) ! NH4+
  molal(4) = onemwf*(damchl+dsochl+dhag) ! CL-
  molal(5) = onemwf*(damsul+dsosul+dlc-molal(6)) ! SO4-- !VB 17 Sept 2001
  molal(6) = onemwf*(molal(6)+dsobis+dambis+dlc) ! HSO4-
  molal(7) = onemwf*(damnit+dsonit+dnag) ! NO3-
  water = onemwf*water
  !
  !     SOLID
  !
  cnh42s4 = wf*cnh42so + onemwf*cnh42s4
  cna2so4 = wf*cna2so + onemwf*cna2so4
  cnahso4 = wf*cnahso + onemwf*cnahso4
  cnh4hs4 = wf*cnh4hso + onemwf*cnh4hs4
  clc = wf*clco + onemwf*clc
  cnh4no3 = wf*cnh4n3o + onemwf*cnh4no3
  cnano3 = wf*cnano + onemwf*cnano3
  cnacl = wf*cnaclo + onemwf*cnacl
  cnh4cl = wf*cnh4clo + onemwf*cnh4cl
  !
  !     GAS
  !
  gnh3 = wf*gnh3o + onemwf*gnh3
  ghno3 = wf*ghno3o + onemwf*ghno3
  ghcl = wf*ghclo + onemwf*ghcl
  !
  ! *** RETURN POINT
  !
100 RETURN
  !
END SUBROUTINE calcmdrh

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCMDRH2
!
!     THIS IS THE CASE WHERE THE RELATIVE HUMIDITY IS IN THE MUTUAL
!     DRH REGION. THE SOLUTION IS ASSUMED TO BE THE SUM OF TWO WEIGHTED
!     SOLUTIONS ; THE 'DRY' SOLUTION (SUBROUTINE DRYCASE) AND THE
!     'SATURATED LIQUID' SOLUTION (SUBROUTINE LIQCASE).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcmdrh2 ( rhi, rhdry, rhliq, drycase, liqcase )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: rhdry
  DOUBLE PRECISION, INTENT(inout) :: rhliq
  DOUBLE PRECISION :: ccac2l
  DOUBLE PRECISION :: ccan32o
  DOUBLE PRECISION :: ccaso
  DOUBLE PRECISION :: ck2so
  DOUBLE PRECISION :: ckclo
  DOUBLE PRECISION :: ckhso
  DOUBLE PRECISION :: ckn3o
  DOUBLE PRECISION :: clco
  DOUBLE PRECISION :: cmgc2l
  DOUBLE PRECISION :: cmgn32o
  DOUBLE PRECISION :: cmgso
  DOUBLE PRECISION :: cna2so
  DOUBLE PRECISION :: cnaclo
  DOUBLE PRECISION :: cnahso
  DOUBLE PRECISION :: cnano
  DOUBLE PRECISION :: cnh42so
  DOUBLE PRECISION :: cnh4clo
  DOUBLE PRECISION :: cnh4hso
  DOUBLE PRECISION :: cnh4n3o
  DOUBLE PRECISION :: dambis
  DOUBLE PRECISION :: damchl
  DOUBLE PRECISION :: damg
  DOUBLE PRECISION :: damnit
  DOUBLE PRECISION :: damsul
  DOUBLE PRECISION :: dcachl
  DOUBLE PRECISION :: dcanit
  DOUBLE PRECISION :: dcasul
  DOUBLE PRECISION :: dhag
  DOUBLE PRECISION :: dlc
  DOUBLE PRECISION :: dmgchl
  DOUBLE PRECISION :: dmgnit
  DOUBLE PRECISION :: dmgsul
  DOUBLE PRECISION :: dnag
  DOUBLE PRECISION :: dpobis
  DOUBLE PRECISION :: dpochl
  DOUBLE PRECISION :: dponit
  DOUBLE PRECISION :: dposul
  DOUBLE PRECISION :: dsobis
  DOUBLE PRECISION :: dsochl
  DOUBLE PRECISION :: dsonit
  DOUBLE PRECISION :: dsosul
  DOUBLE PRECISION :: ghclo
  DOUBLE PRECISION :: ghno3o
  DOUBLE PRECISION :: gnh3o
  INTEGER :: i
  DOUBLE PRECISION :: onemwf
  DOUBLE PRECISION :: wf

  !
  ! *** FIND WEIGHT FACTOR **********************************************
  !
  IF (wftyp==0) THEN
    wf = one
  ELSE IF (wftyp==1) THEN
    wf = 0.5D0
  ELSE
    wf = (rhliq-rhi)/(rhliq-rhdry)
  END IF
  onemwf = one - wf
  !
  ! *** FIND FIRST SECTION ; DRY ONE ************************************
  !
  CALL drycase
  IF (abs(onemwf)<=1D-5) GO TO 100 ! DRY AEROSOL
  !
  cnh42so = cnh42s4 ! FIRST (DRY) SOLUTION
  cnh4hso = cnh4hs4
  clco = clc
  cnh4n3o = cnh4no3
  cnh4clo = cnh4cl
  cna2so = cna2so4
  cnahso = cnahso4
  cnano = cnano3
  cnaclo = cnacl
  gnh3o = gnh3
  ghno3o = ghno3
  ghclo = ghcl
  !
  ccaso = ccaso4
  ck2so = ck2so4
  cmgso = cmgso4
  ckhso = ckhso4
  ccan32o = ccano32
  ccac2l = ccacl2
  ckn3o = ckno3
  ckclo = ckcl
  cmgn32o = cmgno32
  cmgc2l = cmgcl2
  !
  ! *** FIND SECOND SECTION ; DRY & LIQUID ******************************
  !
  cnh42s4 = zero
  cnh4hs4 = zero
  clc = zero
  cnh4no3 = zero
  cnh4cl = zero
  cna2so4 = zero
  cnahso4 = zero
  cnano3 = zero
  cnacl = zero
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  ccaso4 = zero
  ck2so4 = zero
  cmgso4 = zero
  ckhso4 = zero
  ccano32 = zero
  ccacl2 = zero
  ckno3 = zero
  ckcl = zero
  cmgno32 = zero
  cmgcl2 = zero
  !
  CALL liqcase ! SECOND (LIQUID) SOLUTION
  !
  ! *** ADJUST THINGS FOR THE CASE THAT THE LIQUID SUB PREDICTS DRY AEROSOL
  !
  IF (water<=tiny) THEN
    DO i = 1, nions
      molal(i) = zero ! Aqueous phase
    END DO
    water = zero
    !
    cnh42s4 = cnh42so ! Solid phase
    cna2so4 = cna2so
    cnahso4 = cnahso
    cnh4hs4 = cnh4hso
    clc = clco
    cnh4no3 = cnh4n3o
    cnano3 = cnano
    cnacl = cnaclo
    cnh4cl = cnh4clo
    !
    gnh3 = gnh3o ! Gas phase
    ghno3 = ghno3o
    ghcl = ghclo
    !
    ccaso4 = ccaso
    ck2so4 = ck2so
    cmgso4 = cmgso
    ckhso4 = ckhso
    ccano32 = ccan32o
    ccacl2 = ccac2l
    ckno3 = ckn3o
    ckcl = ckclo
    cmgno32 = cmgn32o
    cmgcl2 = cmgc2l
    !
    GO TO 100
  END IF
  !
  ! *** FIND SALT DISSOLUTIONS BETWEEN DRY & LIQUID SOLUTIONS.
  !
  damsul = cnh42so - cnh42s4
  dsosul = cna2so - cna2so4
  dambis = cnh4hso - cnh4hs4
  dsobis = cnahso - cnahso4
  dlc = clco - clc
  damnit = cnh4n3o - cnh4no3
  damchl = cnh4clo - cnh4cl
  dsonit = cnano - cnano3
  dsochl = cnaclo - cnacl
  !
  dcasul = ccaso - ccaso4
  dposul = ck2so - ck2so4
  dmgsul = cmgso - cmgso4
  dpobis = ckhso - ckhso4
  dcanit = ccan32o - ccano32
  dcachl = ccac2l - ccacl2
  dponit = ckn3o - ckno3
  dpochl = ckclo - ckcl
  dmgnit = cmgn32o - cmgno32
  dmgchl = cmgc2l - cmgcl2
  !
  ! *** FIND GAS DISSOLUTIONS BETWEEN DRY & LIQUID SOLUTIONS.
  !
  damg = gnh3o - gnh3
  dhag = ghclo - ghcl
  dnag = ghno3o - ghno3
  !
  ! *** FIND SOLUTION AT MDRH BY WEIGHTING DRY & LIQUID SOLUTIONS.
  !
  !     LIQUID
  !
  molal(1) = onemwf*molal(1) ! H+
  molal(2) = onemwf*(2.D0*dsosul+dsobis+dsonit+dsochl) ! NA+
  molal(3) = onemwf*(2.D0*damsul+damg+dambis+damchl+3.D0*dlc+damnit) ! NH4+
  molal(4) = onemwf*(damchl+dsochl+dhag+2.D0*dcachl+2.D0*dmgchl+dpochl) ! CL-
  molal(5) = onemwf*(damsul+dsosul+dlc-molal(6)+dcasul+dposul+dmgsul) ! SO4-- !VB 17 Sept 2001
  molal(6) = onemwf*(molal(6)+dsobis+dambis+dlc+dpobis) ! HSO4-
  molal(7) = onemwf*(damnit+dsonit+dnag+2.D0*dcanit+2.D0*dmgnit+dponit) ! NO3-
  molal(8) = onemwf*(dcasul+dcanit+dcachl) ! CA2+
  molal(9) = onemwf*(2.D0*dposul+dponit+dpochl+dpobis) ! K+
  molal(10) = onemwf*(dmgsul+dmgnit+dmgchl) ! MG2+
  water = onemwf*water
  !
  !     SOLID
  !
  cnh42s4 = wf*cnh42so + onemwf*cnh42s4
  cna2so4 = wf*cna2so + onemwf*cna2so4
  cnahso4 = wf*cnahso + onemwf*cnahso4
  cnh4hs4 = wf*cnh4hso + onemwf*cnh4hs4
  clc = wf*clco + onemwf*clc
  cnh4no3 = wf*cnh4n3o + onemwf*cnh4no3
  cnano3 = wf*cnano + onemwf*cnano3
  cnacl = wf*cnaclo + onemwf*cnacl
  cnh4cl = wf*cnh4clo + onemwf*cnh4cl
  !
  ccaso4 = wf*ccaso + onemwf*ccaso4
  ck2so4 = wf*ck2so + onemwf*ck2so4
  cmgso4 = wf*cmgso + onemwf*cmgso4
  ckhso4 = wf*ckhso + onemwf*ckhso4
  ccano32 = wf*ccan32o + onemwf*ccano32
  ccacl2 = wf*ccac2l + onemwf*ccacl2
  cmgno32 = wf*cmgn32o + onemwf*cmgno32
  cmgcl2 = wf*cmgc2l + onemwf*cmgcl2
  ckcl = wf*ckclo + onemwf*ckcl
  !
  !     GAS
  !
  gnh3 = wf*gnh3o + onemwf*gnh3
  ghno3 = wf*ghno3o + onemwf*ghno3
  ghcl = wf*ghclo + onemwf*ghcl
  !
  ! *** RETURN POINT
  !
100 RETURN
  !
END SUBROUTINE calcmdrh2
!

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCMDRP
!
!     THIS IS THE CASE WHERE THE RELATIVE HUMIDITY IS IN THE MUTUAL
!     DRH REGION. THE SOLUTION IS ASSUMED TO BE THE SUM OF TWO WEIGHTED
!     SOLUTIONS ; THE 'DRY' SOLUTION (SUBROUTINE DRYCASE) AND THE
!     'SATURATED LIQUID' SOLUTION (SUBROUTINE LIQCASE).   (REVERSE PROBLEM)
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcmdrp ( rhi, rhdry, rhliq, drycase, liqcase )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: rhdry
  DOUBLE PRECISION, INTENT(inout) :: rhliq
  DOUBLE PRECISION :: a2
  DOUBLE PRECISION :: a3
  DOUBLE PRECISION :: a4
  DOUBLE PRECISION :: a8
  DOUBLE PRECISION :: clco
  DOUBLE PRECISION :: cna2so
  DOUBLE PRECISION :: cnaclo
  DOUBLE PRECISION :: cnahso
  DOUBLE PRECISION :: cnano
  DOUBLE PRECISION :: cnh42so
  DOUBLE PRECISION :: cnh4clo
  DOUBLE PRECISION :: cnh4hso
  DOUBLE PRECISION :: cnh4n3o
  DOUBLE PRECISION :: dambis
  DOUBLE PRECISION :: dlc
  DOUBLE PRECISION :: dsobis
  DOUBLE PRECISION :: hien
  DOUBLE PRECISION :: hieq
  INTEGER :: i
  DOUBLE PRECISION :: onemwf
  DOUBLE PRECISION :: wf

  !
  ! *** FIND WEIGHT FACTOR **********************************************
  !
  IF (wftyp==0) THEN
    wf = one
  ELSE IF (wftyp==1) THEN
    wf = 0.5D0
  ELSE
    wf = (rhliq-rhi)/(rhliq-rhdry)
  END IF
  onemwf = one - wf
  !
  ! *** FIND FIRST SECTION ; DRY ONE ************************************
  !
  CALL drycase
  IF (abs(onemwf)<=1D-5) GO TO 100 ! DRY AEROSOL
  !
  cnh42so = cnh42s4 ! FIRST (DRY) SOLUTION
  cnh4hso = cnh4hs4
  clco = clc
  cnh4n3o = cnh4no3
  cnh4clo = cnh4cl
  cna2so = cna2so4
  cnahso = cnahso4
  cnano = cnano3
  cnaclo = cnacl
  !
  ! *** FIND SECOND SECTION ; DRY & LIQUID ******************************
  !
  cnh42s4 = zero
  cnh4hs4 = zero
  clc = zero
  cnh4no3 = zero
  cnh4cl = zero
  cna2so4 = zero
  cnahso4 = zero
  cnano3 = zero
  cnacl = zero
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  CALL liqcase ! SECOND (LIQUID) SOLUTION
  !
  ! *** ADJUST THINGS FOR THE CASE THAT THE LIQUID SUB PREDICTS DRY AEROSOL
  !
  IF (water<=tiny) THEN
    water = zero
    DO i = 1, nions
      molal(i) = zero
    END DO
    CALL drycase
    GO TO 100
  END IF
  !
  ! *** FIND SALT DISSOLUTIONS BETWEEN DRY & LIQUID SOLUTIONS.
  !
  dambis = cnh4hso - cnh4hs4
  dsobis = cnahso - cnahso4
  dlc = clco - clc
  !
  ! *** FIND SOLUTION AT MDRH BY WEIGHTING DRY & LIQUID SOLUTIONS.
  !
  ! *** SOLID
  !
  cnh42s4 = wf*cnh42so + onemwf*cnh42s4
  cna2so4 = wf*cna2so + onemwf*cna2so4
  cnahso4 = wf*cnahso + onemwf*cnahso4
  cnh4hs4 = wf*cnh4hso + onemwf*cnh4hs4
  clc = wf*clco + onemwf*clc
  cnh4no3 = wf*cnh4n3o + onemwf*cnh4no3
  cnano3 = wf*cnano + onemwf*cnano3
  cnacl = wf*cnaclo + onemwf*cnacl
  cnh4cl = wf*cnh4clo + onemwf*cnh4cl
  !
  ! *** LIQUID
  !
  water = onemwf*water
  !
  molal(2) = waer(1) - 2.D0*cna2so4 - cnahso4 - cnano3 - cnacl ! NA+
  molal(3) = waer(3) - 2.D0*cnh42s4 - cnh4hs4 - cnh4cl - 3.D0*clc -        &
       cnh4no3 ! NH4+
  molal(4) = waer(5) - cnacl - cnh4cl ! CL-
  molal(7) = waer(4) - cnano3 - cnh4no3 ! NO3-
  molal(6) = onemwf*(molal(6)+dsobis+dambis+dlc) ! HSO4-
  molal(5) = waer(2) - molal(6) - clc - cnh42s4 - cna2so4 ! SO4--
  !
  a8 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
  IF (molal(5)<=tiny) THEN
    hieq = sqrt(xkw*rh*water*water) ! Neutral solution
  ELSE
    hieq = a8*molal(6)/molal(5)
  END IF
  hien = molal(4) + molal(7) + molal(6) + 2.D0*molal(5) - molal(2) -       &
       molal(3)
  molal(1) = max(hieq, hien) ! H+
  !
  ! *** GAS (ACTIVITY COEFS FROM LIQUID SOLUTION)
  !
  a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = molal(3)/max(molal(1), tiny)/a2
  ghno3 = molal(1)*molal(7)/a3
  ghcl = molal(1)*molal(4)/a4
  !
100 RETURN
  !
END SUBROUTINE calcmdrp
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCMDRPII
!
!     THIS IS THE CASE WHERE THE RELATIVE HUMIDITY IS IN THE MUTUAL
!     DRH REGION. THE SOLUTION IS ASSUMED TO BE THE SUM OF TWO WEIGHTED
!     SOLUTIONS ; THE 'DRY' SOLUTION (SUBROUTINE DRYCASE) AND THE
!     'SATURATED LIQUID' SOLUTION (SUBROUTINE LIQCASE).   (REVERSE PROBLEM)
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE calcmdrpii ( rhi, rhdry, rhliq, drycase, liqcase )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: rhdry
  DOUBLE PRECISION, INTENT(inout) :: rhliq
  DOUBLE PRECISION :: a2
  DOUBLE PRECISION :: a3
  DOUBLE PRECISION :: a4
  DOUBLE PRECISION :: a8
  DOUBLE PRECISION :: ccac2l
  DOUBLE PRECISION :: ccan32o
  DOUBLE PRECISION :: ccaso
  DOUBLE PRECISION :: ck2so
  DOUBLE PRECISION :: ckclo
  DOUBLE PRECISION :: ckhso
  DOUBLE PRECISION :: ckn3o
  DOUBLE PRECISION :: clco
  DOUBLE PRECISION :: cmgc2l
  DOUBLE PRECISION :: cmgn32o
  DOUBLE PRECISION :: cmgso
  DOUBLE PRECISION :: cna2so
  DOUBLE PRECISION :: cnaclo
  DOUBLE PRECISION :: cnahso
  DOUBLE PRECISION :: cnano
  DOUBLE PRECISION :: cnh42so
  DOUBLE PRECISION :: cnh4clo
  DOUBLE PRECISION :: cnh4hso
  DOUBLE PRECISION :: cnh4n3o
  DOUBLE PRECISION :: dambis
  DOUBLE PRECISION :: dlc
  DOUBLE PRECISION :: dpobis
  DOUBLE PRECISION :: dsobis
  DOUBLE PRECISION :: hien
  DOUBLE PRECISION :: hieq
  INTEGER :: i
  DOUBLE PRECISION :: onemwf
  DOUBLE PRECISION :: wf

  !
  ! *** FIND WEIGHT FACTOR **********************************************
  !
  IF (wftyp==0) THEN
    wf = one
  ELSE IF (wftyp==1) THEN
    wf = 0.5D0
  ELSE
    wf = (rhliq-rhi)/(rhliq-rhdry)
  END IF
  onemwf = one - wf
  !
  ! *** FIND FIRST SECTION ; DRY ONE ************************************
  !
  CALL drycase
  IF (abs(onemwf)<=1D-5) GO TO 100 ! DRY AEROSOL
  !
  cnh42so = cnh42s4 ! FIRST (DRY) SOLUTION
  cnh4hso = cnh4hs4
  clco = clc
  cnh4n3o = cnh4no3
  cnh4clo = cnh4cl
  cna2so = cna2so4
  cnahso = cnahso4
  cnano = cnano3
  cnaclo = cnacl
  !
  ccaso = ccaso4
  ck2so = ck2so4
  cmgso = cmgso4
  ckhso = ckhso4
  ccan32o = ccano32
  ccac2l = ccacl2
  ckn3o = ckno3
  ckclo = ckcl
  cmgn32o = cmgno32
  cmgc2l = cmgcl2
  !
  ! *** FIND SECOND SECTION ; DRY & LIQUID ******************************
  !
  cnh42s4 = zero
  cnh4hs4 = zero
  clc = zero
  cnh4no3 = zero
  cnh4cl = zero
  cna2so4 = zero
  cnahso4 = zero
  cnano3 = zero
  cnacl = zero
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  ccaso4 = zero
  ck2so4 = zero
  cmgso4 = zero
  ckhso4 = zero
  ccano32 = zero
  ccacl2 = zero
  ckno3 = zero
  ckcl = zero
  cmgno32 = zero
  cmgcl2 = zero
  !
  CALL liqcase ! SECOND (LIQUID) SOLUTION
  !
  ! *** ADJUST THINGS FOR THE CASE THAT THE LIQUID SUB PREDICTS DRY AEROSOL
  !
  IF (water<=tiny) THEN
    water = zero
    DO i = 1, nions
      molal(i) = zero
    END DO
    CALL drycase
    GO TO 100
  END IF
  !
  ! *** FIND SALT DISSOLUTIONS BETWEEN DRY & LIQUID SOLUTIONS.
  !
  dambis = cnh4hso - cnh4hs4
  dsobis = cnahso - cnahso4
  dlc = clco - clc
  dpobis = ckhso - ckhso4
  !
  ! *** FIND SOLUTION AT MDRH BY WEIGHTING DRY & LIQUID SOLUTIONS.
  !
  ! *** SOLID
  !
  cnh42s4 = wf*cnh42so + onemwf*cnh42s4
  cna2so4 = wf*cna2so + onemwf*cna2so4
  cnahso4 = wf*cnahso + onemwf*cnahso4
  cnh4hs4 = wf*cnh4hso + onemwf*cnh4hs4
  clc = wf*clco + onemwf*clc
  cnh4no3 = wf*cnh4n3o + onemwf*cnh4no3
  cnano3 = wf*cnano + onemwf*cnano3
  cnacl = wf*cnaclo + onemwf*cnacl
  cnh4cl = wf*cnh4clo + onemwf*cnh4cl
  !
  ccaso4 = wf*ccaso + onemwf*ccaso4
  ck2so4 = wf*ck2so + onemwf*ck2so4
  cmgso4 = wf*cmgso + onemwf*cmgso4
  ckhso4 = wf*ckhso + onemwf*ckhso4
  ccano32 = wf*ccan32o + onemwf*ccano32
  ccacl2 = wf*ccac2l + onemwf*ccacl2
  cmgno32 = wf*cmgn32o + onemwf*cmgno32
  cmgcl2 = wf*cmgc2l + onemwf*cmgcl2
  ckcl = wf*ckclo + onemwf*ckcl
  !
  ! *** LIQUID
  !
  water = onemwf*water
  !
  molal(2) = waer(1) - 2.D0*cna2so4 - cnahso4 - cnano3 - cnacl ! NA+
  molal(3) = waer(3) - 2.D0*cnh42s4 - cnh4hs4 - cnh4cl - 3.D0*clc -        &
       cnh4no3 ! NH4+
  molal(4) = waer(5) - cnacl - cnh4cl - 2.D0*ccacl2 - 2.D0*cmgcl2 - ckcl ! CL-
  molal(7) = waer(4) - cnano3 - cnh4no3 - ckno3 - 2.D0*ccano32 -           &
       2.D0*cmgno32 ! NO3-
  molal(6) = onemwf*(molal(6)+dsobis+dambis+dlc+dpobis) ! HSO4-
  molal(5) = waer(2) - molal(6) - clc - cnh42s4 - cna2so4 - ccaso4 -       &
       ck2so4 - cmgso4 ! SO4--
  molal(8) = waer(6) - ccaso4 - ccano32 - ccacl2 ! CA++
  molal(9) = waer(7) - 2.D0*ck2so4 - ckno3 - ckcl - ckhso4 ! K+
  molal(10) = waer(8) - cmgso4 - cmgno32 - cmgcl2 ! MG++
  !
  a8 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
  IF (molal(5)<=tiny) THEN
    hieq = sqrt(xkw*rh*water*water) ! Neutral solution
  ELSE
    hieq = a8*molal(6)/molal(5)
  END IF
  hien = molal(4) + molal(7) + molal(6) + 2.D0*molal(5) - molal(2) -       &
       molal(3)
  molal(1) = max(hieq, hien) ! H+
  !
  ! *** GAS (ACTIVITY COEFS FROM LIQUID SOLUTION)
  !
  a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = molal(3)/max(molal(1), tiny)/a2
  ghno3 = molal(1)*molal(7)/a3
  ghcl = molal(1)*molal(4)/a4
  !
100 RETURN
  !
END SUBROUTINE calcmdrpii
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCHS4
! *** THIS SUBROUTINE CALCULATES THE HSO4 GENERATED FROM (H,SO4).
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!

SUBROUTINE calchs4 ( hi, so4i, hso4i, delta )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: hi
  DOUBLE PRECISION, INTENT(inout) :: so4i
  DOUBLE PRECISION, INTENT(inout) :: hso4i
  DOUBLE PRECISION, INTENT(inout) :: delta
  DOUBLE PRECISION :: a8
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delta1
  DOUBLE PRECISION :: delta2
  DOUBLE PRECISION :: sqdd

  !C      CHARACTER ERRINF*40
  !
  ! *** IF TOO LITTLE WATER, DONT SOLVE
  !
  IF (water<=1D1*tiny) THEN
    delta = zero
    RETURN
  END IF
  !
  ! *** CALCULATE HSO4 SPECIATION *****************************************
  !
  a8 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
  !
  bb = -(hi+so4i+a8)
  cc = hi*so4i - hso4i*a8
  dd = bb*bb - 4.D0*cc
  !
  IF (dd>=zero) THEN
    sqdd = sqrt(dd)
    delta1 = 0.5*(-bb+sqdd)
    delta2 = 0.5*(-bb-sqdd)
    IF (hso4i<=tiny) THEN
      delta = delta2
    ELSE IF (hi*so4i>=a8*hso4i) THEN
      delta = delta2
    ELSE IF (hi*so4i<a8*hso4i) THEN
      delta = delta1
    ELSE
      delta = zero
    END IF
  ELSE
    delta = zero
  END IF
  !CC
  !CC *** COMPARE DELTA TO TOTAL H+ ; ESTIMATE EFFECT OF HSO4 ***************
  !CC
  !C      HYD = MAX(HI, MOLAL(1))
  !C      IF (HYD.GT.TINY) THEN
  !C         IF (DELTA/HYD.GT.0.1d0) THEN
  !C            WRITE (ERRINF,'(1PE10.3)') DELTA/HYD*100.0
  !C            CALL PUSHERR (0020, ERRINF)
  !C         ENDIF
  !C      ENDIF
  !
  RETURN
  !
END SUBROUTINE calchs4


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCPH
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!

SUBROUTINE calcph ( gg, hi, ohi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(in)    :: gg
  DOUBLE PRECISION, INTENT(inout) :: hi
  DOUBLE PRECISION, INTENT(inout) :: ohi
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: cn
  DOUBLE PRECISION :: dd
  !
  akw = xkw*rh*water*water
  cn = sqrt(akw)
  !
  ! *** GG = (negative charge) - (positive charge)
  !
  IF (gg>tiny) THEN ! H+ in excess
    bb = -gg
    cc = -akw
    dd = bb*bb - 4.D0*cc
    hi = max(0.5D0*(-bb+sqrt(dd)), cn)
    ohi = akw/hi
  ELSE ! OH- in excess
    bb = gg
    cc = -akw
    dd = bb*bb - 4.D0*cc
    ohi = max(0.5D0*(-bb+sqrt(dd)), cn)
    hi = akw/ohi
  END IF
  !
  RETURN
  !
END SUBROUTINE calcph

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCACT
! *** CALCULATES MULTI-COMPONENT ACTIVITY COEFFICIENTS FROM BROMLEYS
!     METHOD. THE BINARY ACTIVITY COEFFICIENTS ARE CALCULATED BY
!     KUSIK-MEISNER RELATION (SUBROUTINE KMTAB or SUBROUTINE KMFUL).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
!
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcact (  )
  USE mo_isrpia
  IMPLICIT NONE

  !
  !      COMMON /drvinp/wi(8), rhi, tempi, iprobi, metstbli, iacalci, nadji
  !
  IF (w(1)+w(4)+w(5)+w(6)+w(7)+w(8)<=6.d0*tiny) THEN !Ca,K,Mg,Na,Cl,NO3=0
    CALL calcact1
  ELSE IF (w(1)+w(5)+w(6)+w(7)+w(8)<=5.d0*tiny) THEN !Ca,K,Mg,Na,Cl=0
    CALL calcact2
  ELSE IF (w(6)+w(7)+w(8)<=3.d0*tiny) THEN !Ca,K,Mg=0
    CALL calcact3
  ELSE
    CALL calcact4
  END IF
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE calcact

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCACT4
! *** CALCULATES MULTI-COMPONENT ACTIVITY COEFFICIENTS FROM BROMLEYS
!     METHOD FOR AN AMMONIUM-SULFATE-NITRATE-CHLORIDE-SODIUM-CALCIUM-POTASSIUM-MAGNESIUM
!     AEROSOL SYSTEM. THE BINARY ACTIVITY COEFFICIENTS ARE CALCULATED BY
!     KUSIK-MEISNER RELATION (SUBROUTINE KMTAB or SUBROUTINE KMFUL4).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
!
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!

SUBROUTINE calcact4 (  )
  USE mo_isrpia
  IMPLICIT NONE
  REAL :: agama
  REAL :: ch
  DOUBLE PRECISION :: errin
  DOUBLE PRECISION :: errou
  REAL :: f1(6)
  REAL :: f2a(4)
  REAL :: f2b(4)
  REAL :: h
  INTEGER :: i
  INTEGER :: j
  DOUBLE PRECISION :: mpl
  REAL :: sion
  DOUBLE PRECISION :: xij
  DOUBLE PRECISION :: yji
  REAL :: zmi
  REAL :: zpl
  !
  REAL :: g0(6, 4) = 0.0

  !
  DOUBLE PRECISION :: ga, gb
  ga(i, j) = (f1(i)/z(i)+f2a(j)/z(j+3))/(z(i)+z(j+3)) - h
  gb(i, j) = (f1(i)/z(i+4)+f2b(j)/z(j+3))/(z(i+4)+z(j+3)) - h
  !
  ! *** SAVE ACTIVITIES IN OLD ARRAY *************************************
  !
  IF (frst) THEN ! Outer loop
    DO i = 1, npair
      gamou(i) = gama(i)
    END DO
  END IF
  !
  DO i = 1, npair ! Inner loop
    gamin(i) = gama(i)
  END DO
  !
  ! *** CALCULATE IONIC ACTIVITY OF SOLUTION *****************************
  !
  ionic = 0.0
  DO i = 1, nions
    ionic = ionic + molal(i)*z(i)*z(i)
  END DO
  ionic = max(min(0.5*ionic/water,100.D0), tiny)
  !
  ! *** CALCULATE BINARY ACTIVITY COEFFICIENTS ***************************
  !
  !  G0(1,1)=G11;G0(1,2)=G07;G0(1,3)=G08;G0(1,4)=G10;G0(2,1)=G01;G0(2,2)=G02
  !  G0(2,3)=G12;G0(2,4)=G03;G0(3,1)=G06;G0(3,2)=G04;G0(3,3)=G09;G0(3,4)=G05
  !
  IF (iacalc==0) THEN ! K.M.; FULL
    CALL kmful4(ionic, sngl(temp), g0(2,1), g0(2,2), g0(2,4), g0(3,2),     &
         g0(3,4), g0(3,1), g0(1,2), g0(1,3), g0(3,3), g0(1,4), g0(1,1),       &
         g0(2,3), g0(4,4), g0(4,1), g0(5,2), g0(5,3), g0(5,4), g0(5,1),       &
         g0(6,2), g0(6,4), g0(6,1))
  ELSE ! K.M.; TABULATED
    CALL kmtab(ionic, sngl(temp), g0(2,1), g0(2,2), g0(2,4), g0(3,2),      &
         g0(3,4), g0(3,1), g0(1,2), g0(1,3), g0(3,3), g0(1,4), g0(1,1),       &
         g0(2,3), g0(4,4), g0(4,1), g0(5,2), g0(5,3), g0(5,4), g0(5,1),       &
         g0(6,2), g0(6,4), g0(6,1))
  END IF
  !
  ! *** CALCULATE MULTICOMPONENT ACTIVITY COEFFICIENTS *******************
  !
  agama = 0.511*(298.0/temp)**1.5 ! Debye Huckel const. at T
  sion = sqrt(ionic)
  h = agama*sion/(1+sion)
  !
  DO i = 1, 4
    f1(i) = 0.0
    f2a(i) = 0.0
    f2b(i) = 0.0
  END DO
  f1(5) = 0.0
  f1(6) = 0.0
  !
  DO i = 1, 3
    zpl = z(i)
    mpl = molal(i)/water
    DO j = 1, 4
      zmi = z(j+3)
      ch = 0.25*(zpl+zmi)*(zpl+zmi)/ionic
      xij = ch*mpl
      yji = ch*molal(j+3)/water
      f1(i) = f1(i) + sngl(yji*(g0(i,j)+zpl*zmi*h))
      f2a(j) = f2a(j) + sngl(xij*(g0(i,j)+zpl*zmi*h))
    END DO
  END DO
  !
  DO i = 4, 6
    zpl = z(i+4)
    mpl = molal(i+4)/water
    DO j = 1, 4
      zmi = z(j+3)
      IF (j==3) THEN
        IF (i==4 .OR. i==6) THEN
          GO TO 100
        END IF
      END IF
      ch = 0.25*(zpl+zmi)*(zpl+zmi)/ionic
      xij = ch*mpl
      yji = ch*molal(j+3)/water
      f1(i) = f1(i) + sngl(yji*(g0(i,j)+zpl*zmi*h))
      f2b(j) = f2b(j) + sngl(xij*(g0(i,j)+zpl*zmi*h))
100 END DO
  END DO

  !
  ! *** LOG10 OF ACTIVITY COEFFICIENTS ***********************************
  !
  gama(01) = ga(2, 1)*zz(01) ! NACL
  gama(02) = ga(2, 2)*zz(02) ! NA2SO4
  gama(03) = ga(2, 4)*zz(03) ! NANO3
  gama(04) = ga(3, 2)*zz(04) ! (NH4)2SO4
  gama(05) = ga(3, 4)*zz(05) ! NH4NO3
  gama(06) = ga(3, 1)*zz(06) ! NH4CL
  gama(07) = ga(1, 2)*zz(07) ! 2H-SO4
  gama(08) = ga(1, 3)*zz(08) ! H-HSO4
  gama(09) = ga(3, 3)*zz(09) ! NH4HSO4
  gama(10) = ga(1, 4)*zz(10) ! HNO3
  gama(11) = ga(1, 1)*zz(11) ! HCL
  gama(12) = ga(2, 3)*zz(12) ! NAHSO4
  gama(13) = 0.20*(3.0*gama(04)+2.0*gama(09)) ! LC ; SCAPE
  !C      GAMA(13) = 0.50*(GAMA(04)+GAMA(09))          ! LC ; SEQUILIB
  !C      GAMA(13) = 0.25*(3.0*GAMA(04)+GAMA(07))      ! LC ; AIM
  gama(14) = 0.0D0 ! CASO4
  gama(15) = gb(4, 4)*zz(15) ! CA(NO3)2
  gama(16) = gb(4, 1)*zz(16) ! CACL2
  gama(17) = gb(5, 2)*zz(17) ! K2SO4
  gama(18) = gb(5, 3)*zz(18) ! KHSO4
  gama(19) = gb(5, 4)*zz(19) ! KNO3
  gama(20) = gb(5, 1)*zz(20) ! KCL
  gama(21) = gb(6, 2)*zz(21) ! MGSO4
  gama(22) = gb(6, 4)*zz(22) ! MG(NO3)2
  gama(23) = gb(6, 1)*zz(23) ! MGCL2
  !
  ! *** CONVERT LOG (GAMA) COEFFICIENTS TO GAMA **************************
  !
  DO i = 1, npair
    gama(i) = max(-5.0D0, min(gama(i),5.0D0)) ! F77 LIBRARY ROUTINE
    gama(i) = 10.0**gama(i)
    !C         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
  END DO
  !
  ! *** SETUP ACTIVITY CALCULATION FLAGS ********************************
  !
  ! OUTER CALCULATION LOOP ; ONLY IF FRST=.TRUE.
  !
  IF (frst) THEN
    errou = zero ! CONVERGENCE CRITERION
    DO i = 1, npair
      errou = max(errou, abs((gamou(i)-gama(i))/gamou(i)))
    END DO
    calaou = errou >= epsact ! SETUP FLAGS
    frst = .FALSE.
  END IF
  !
  ! INNER CALCULATION LOOP ; ALWAYS
  !
  errin = zero ! CONVERGENCE CRITERION
  DO i = 1, npair
    errin = max(errin, abs((gamin(i)-gama(i))/gamin(i)))
  END DO
  calain = errin >= epsact
  !
  iclact = iclact + 1 ! Increment ACTIVITY call counter
  !
  RETURN
END SUBROUTINE calcact4
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCACT3
! *** CALCULATES MULTI-COMPONENT ACTIVITY COEFFICIENTS FROM BROMLEYS
!     METHOD FOR AN AMMONIUM-SULFATE-NITRATE-CHLORIDE-SODIUM AEROSOL SYSTEM.
!     THE BINARY ACTIVITY COEFFICIENTS ARE CALCULATED BY
!     KUSIK-MEISNER RELATION (SUBROUTINE KMTAB or SUBROUTINE KMFUL3).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
!
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!

SUBROUTINE calcact3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  REAL :: agama
  REAL :: ch
  DOUBLE PRECISION :: errin
  DOUBLE PRECISION :: errou
  REAL :: f1(3)
  REAL :: f2(4)
  REAL :: h
  INTEGER :: i
  INTEGER :: j
  DOUBLE PRECISION :: mpl
  REAL :: sion
  DOUBLE PRECISION :: xij
  DOUBLE PRECISION :: yji
  REAL :: zmi
  REAL :: zpl
  !
  REAL :: g0(6, 4)

  REAL, PARAMETER :: urf = 0.5
  !      PARAMETER (LN10=2.30258509299404568402D0)
  !
  DOUBLE PRECISION :: g
  g(i, j) = (f1(i)/z(i)+f2(j)/z(j+3))/(z(i)+z(j+3)) - h

  g0 = 0.d0

  !
  ! *** SAVE ACTIVITIES IN OLD ARRAY *************************************
  !
  IF (frst) THEN ! Outer loop
    DO i = 1, 13
      gamou(i) = gama(i)
    END DO
  END IF
  !
  DO i = 1, 13 ! Inner loop
    gamin(i) = gama(i)
  END DO
  !
  ! *** CALCULATE IONIC ACTIVITY OF SOLUTION *****************************
  !
  ionic = 0.0
  DO i = 1, 7
    ionic = ionic + molal(i)*z(i)*z(i)
  END DO
  ionic = max(min(0.5*ionic/water,100.D0), tiny)
  !
  ! *** CALCULATE BINARY ACTIVITY COEFFICIENTS ***************************
  !
  !  G0(1,1)=G11;G0(1,2)=G07;G0(1,3)=G08;G0(1,4)=G10;G0(2,1)=G01;G0(2,2)=G02
  !  G0(2,3)=G12;G0(2,4)=G03;G0(3,1)=G06;G0(3,2)=G04;G0(3,3)=G09;G0(3,4)=G05
  !
  IF (iacalc==0) THEN ! K.M.; FULL
    CALL kmful3(ionic, sngl(temp), g0(2,1), g0(2,2), g0(2,4), g0(3,2),     &
         g0(3,4), g0(3,1), g0(1,2), g0(1,3), g0(3,3), g0(1,4), g0(1,1),       &
         g0(2,3))
  ELSE ! K.M.; TABULATED
    CALL kmtab(ionic, sngl(temp), g0(2,1), g0(2,2), g0(2,4), g0(3,2),      &
         g0(3,4), g0(3,1), g0(1,2), g0(1,3), g0(3,3), g0(1,4), g0(1,1),       &
         g0(2,3), g0(4,4), g0(4,1), g0(5,2), g0(5,3), g0(5,4), g0(5,1),       &
         g0(6,2), g0(6,4), g0(6,1))
  END IF
  !
  ! *** CALCULATE MULTICOMPONENT ACTIVITY COEFFICIENTS *******************
  !
  agama = 0.511*(298.0/temp)**1.5 ! Debye Huckel const. at T
  sion = sqrt(ionic)
  h = agama*sion/(1+sion)
  !
  DO i = 1, 3
    f1(i) = 0.0
    f2(i) = 0.0
  END DO
  f2(4) = 0.0
  !
  DO i = 1, 3
    zpl = z(i)
    mpl = molal(i)/water
    DO j = 1, 4
      zmi = z(j+3)
      ch = 0.25*(zpl+zmi)*(zpl+zmi)/ionic
      xij = ch*mpl
      yji = ch*molal(j+3)/water
      f1(i) = f1(i) + sngl(yji*(g0(i,j)+zpl*zmi*h))
      f2(j) = f2(j) + sngl(xij*(g0(i,j)+zpl*zmi*h))
    END DO
  END DO
  !
  ! *** LOG10 OF ACTIVITY COEFFICIENTS ***********************************
  !
  gama(01) = g(2, 1)*zz(01) ! NACL
  gama(02) = g(2, 2)*zz(02) ! NA2SO4
  gama(03) = g(2, 4)*zz(03) ! NANO3
  gama(04) = g(3, 2)*zz(04) ! (NH4)2SO4
  gama(05) = g(3, 4)*zz(05) ! NH4NO3
  gama(06) = g(3, 1)*zz(06) ! NH4CL
  gama(07) = g(1, 2)*zz(07) ! 2H-SO4
  gama(08) = g(1, 3)*zz(08) ! H-HSO4
  gama(09) = g(3, 3)*zz(09) ! NH4HSO4
  gama(10) = g(1, 4)*zz(10) ! HNO3
  gama(11) = g(1, 1)*zz(11) ! HCL
  gama(12) = g(2, 3)*zz(12) ! NAHSO4
  gama(13) = 0.20*(3.0*gama(04)+2.0*gama(09)) ! LC ; SCAPE
  !C      GAMA(13) = 0.50*(GAMA(04)+GAMA(09))          ! LC ; SEQUILIB
  !C      GAMA(13) = 0.25*(3.0*GAMA(04)+GAMA(07))      ! LC ; AIM
  !
  ! *** CONVERT LOG (GAMA) COEFFICIENTS TO GAMA **************************
  !
  DO i = 1, 13
    gama(i) = max(-5.0D0, min(gama(i),5.0D0)) ! F77 LIBRARY ROUTINE
    gama(i) = 10.0**gama(i)
    !         GAMA(I)=EXP(LN10*GAMA(I))
    !C         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
    !         GAMA(I) = GAMIN(I)*(1.0-URF) + URF*GAMA(I)  ! Under-relax GAMA's
  END DO
  !
  ! *** SETUP ACTIVITY CALCULATION FLAGS *********************************
  !
  ! OUTER CALCULATION LOOP ; ONLY IF FRST=.TRUE.
  !
  IF (frst) THEN
    errou = zero ! CONVERGENCE CRITERION
    DO i = 1, 13
      errou = max(errou, abs((gamou(i)-gama(i))/gamou(i)))
    END DO
    calaou = errou >= epsact ! SETUP FLAGS
    frst = .FALSE.
  END IF
  !
  ! INNER CALCULATION LOOP ; ALWAYS
  !
  errin = zero ! CONVERGENCE CRITERION
  DO i = 1, 13
    errin = max(errin, abs((gamin(i)-gama(i))/gamin(i)))
  END DO
  calain = errin >= epsact
  !
  iclact = iclact + 1 ! Increment ACTIVITY call counter
  !
  RETURN
END SUBROUTINE calcact3
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCACT2
! *** CALCULATES MULTI-COMPONENT ACTIVITY COEFFICIENTS FROM BROMLEYS
!     METHOD FOR AN AMMONIUM-SULFATE-NITRATE AEROSOL SYSTEM.
!     THE BINARY ACTIVITY COEFFICIENTS ARE CALCULATED BY
!     KUSIK-MEISNER RELATION (SUBROUTINE KMTAB or SUBROUTINE KMFUL2).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
!
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcact2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  REAL :: agama
  REAL :: ch
  DOUBLE PRECISION :: errin
  DOUBLE PRECISION :: errou
  REAL :: f1(3)
  REAL :: f2(4)
  REAL :: h
  INTEGER :: i
  INTEGER :: j
  DOUBLE PRECISION :: mpl
  REAL :: sion
  DOUBLE PRECISION :: xij
  DOUBLE PRECISION :: yji
  REAL :: zmi
  REAL :: zpl
  !
  REAL :: g0(6, 4)
  REAL, PARAMETER :: urf = 0.5
  !      PARAMETER (LN10=2.30258509299404568402D0)
  !
  DOUBLE PRECISION :: g
  g(i, j) = (f1(i)/z(i)+f2(j)/z(j+3))/(z(i)+z(j+3)) - h
  !
  g0 = 0.d0
  ! *** SAVE ACTIVITIES IN OLD ARRAY *************************************
  !
  IF (frst) THEN ! Outer loop
    DO i = 7, 10
      gamou(i) = gama(i)
    END DO
    gamou(4) = gama(4)
    gamou(5) = gama(5)
    gamou(13) = gama(13)
  END IF
  !
  DO i = 7, 10 ! Inner loop
    gamin(i) = gama(i)
  END DO
  gamin(4) = gama(4)
  gamin(5) = gama(5)
  gamin(13) = gama(13)
  !
  ! *** CALCULATE IONIC ACTIVITY OF SOLUTION *****************************
  !
  ionic = 0.0
  molal(2) = zero
  molal(4) = zero
  DO i = 1, 7
    ionic = ionic + molal(i)*z(i)*z(i)
  END DO
  ionic = max(min(0.5*ionic/water,100.D0), tiny)
  !
  ! *** CALCULATE BINARY ACTIVITY COEFFICIENTS ***************************
  !
  !  G0(1,1)=G11;G0(1,2)=G07;G0(1,3)=G08;G0(1,4)=G10;G0(2,1)=G01;G0(2,2)=G02
  !  G0(2,3)=G12;G0(2,4)=G03;G0(3,1)=G06;G0(3,2)=G04;G0(3,3)=G09;G0(3,4)=G05
  !
  IF (iacalc==0) THEN ! K.M.; FULL
    CALL kmful2(ionic, sngl(temp), g0(3,2), g0(3,4), g0(1,2), g0(1,3),     &
         g0(3,3), g0(1,4))
  ELSE ! K.M.; TABULATED
    CALL kmtab(ionic, sngl(temp), g0(2,1), g0(2,2), g0(2,4), g0(3,2),      &
         g0(3,4), g0(3,1), g0(1,2), g0(1,3), g0(3,3), g0(1,4), g0(1,1),       &
         g0(2,3), g0(4,4), g0(4,1), g0(5,2), g0(5,3), g0(5,4), g0(5,1),       &
         g0(6,2), g0(6,4), g0(6,1))
  END IF
  !
  ! *** CALCULATE MULTICOMPONENT ACTIVITY COEFFICIENTS *******************
  !
  agama = 0.511*(298.0/temp)**1.5 ! Debye Huckel const. at T
  sion = sqrt(ionic)
  h = agama*sion/(1+sion)
  !
  DO i = 1, 3
    f1(i) = 0.0
    f2(i) = 0.0
  END DO
  f2(4) = 0.0
  !
  DO i = 1, 3, 2
    zpl = z(i)
    mpl = molal(i)/water
    DO j = 2, 4
      zmi = z(j+3)
      ch = 0.25*(zpl+zmi)*(zpl+zmi)/ionic
      xij = ch*mpl
      yji = ch*molal(j+3)/water
      f1(i) = f1(i) + sngl(yji*(g0(i,j)+zpl*zmi*h))
      f2(j) = f2(j) + sngl(xij*(g0(i,j)+zpl*zmi*h))
    END DO
  END DO
  !
  ! *** LOG10 OF ACTIVITY COEFFICIENTS ***********************************
  !
  !      GAMA(01) = G(2,1)*ZZ(01)                     ! NACL
  !      GAMA(02) = G(2,2)*ZZ(02)                     ! NA2SO4
  !      GAMA(03) = G(2,4)*ZZ(03)                     ! NANO3
  gama(04) = g(3, 2)*zz(04) ! (NH4)2SO4
  gama(05) = g(3, 4)*zz(05) ! NH4NO3
  !      GAMA(06) = G(3,1)*ZZ(06)                     ! NH4CL
  gama(07) = g(1, 2)*zz(07) ! 2H-SO4
  gama(08) = g(1, 3)*zz(08) ! H-HSO4
  gama(09) = g(3, 3)*zz(09) ! NH4HSO4
  gama(10) = g(1, 4)*zz(10) ! HNO3
  !      GAMA(11) = G(1,1)*ZZ(11)                     ! HCL
  !      GAMA(12) = G(2,3)*ZZ(12)                     ! NAHSO4
  gama(13) = 0.20*(3.0*gama(04)+2.0*gama(09)) ! LC ; SCAPE
  !C      GAMA(13) = 0.50*(GAMA(04)+GAMA(09))          ! LC ; SEQUILIB
  !C      GAMA(13) = 0.25*(3.0*GAMA(04)+GAMA(07))      ! LC ; AIM
  !
  ! *** CONVERT LOG (GAMA) COEFFICIENTS TO GAMA **************************
  !
  DO i = 7, 10
    gama(i) = max(-5.0D0, min(gama(i),5.0D0)) ! F77 LIBRARY ROUTINE
    gama(i) = 10.0**gama(i)
    !         GAMA(I)=EXP(LN10*GAMA(I))
    !C         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
    !         GAMA(I) = GAMIN(I)*(1.0-URF) + URF*GAMA(I)  ! Under-relax GAMA's
  END DO
  !
  gama(4) = max(-5.0D0, min(gama(4),5.0D0)) ! F77 LIBRARY ROUTINE
  gama(4) = 10.0**gama(4)
  !         GAMA(I)=EXP(LN10*GAMA(I))
  !C         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
  !         GAMA(4) = GAMIN(4)*(1.0-URF) + URF*GAMA(4)  ! Under-relax GAMA's
  !
  gama(5) = max(-5.0D0, min(gama(5),5.0D0)) ! F77 LIBRARY ROUTINE
  gama(5) = 10.0**gama(5)
  !         GAMA(I)=EXP(LN10*GAMA(I))
  !C         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
  !         GAMA(5) = GAMIN(5)*(1.0-URF) + URF*GAMA(I)  ! Under-relax GAMA's
  !
  gama(13) = max(-5.0D0, min(gama(13),5.0D0)) ! F77 LIBRARY ROUTINE
  gama(13) = 10.0**gama(13)
  !         GAMA(I)=EXP(LN10*GAMA(I))
  !C         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
  !         GAMA(13) = GAMIN(13)*(1.0-URF) + URF*GAMA(13)  ! Under-relax GAMA's
  !
  ! *** SETUP ACTIVITY CALCULATION FLAGS *********************************
  !
  ! OUTER CALCULATION LOOP ; ONLY IF FRST=.TRUE.
  !
  IF (frst) THEN
    errou = zero ! CONVERGENCE CRITERION
    DO i = 7, 10
      errou = max(errou, abs((gamou(i)-gama(i))/gamou(i)))
    END DO
    errou = max(errou, abs((gamou(4)-gama(4))/gamou(4)))
    errou = max(errou, abs((gamou(5)-gama(5))/gamou(5)))
    errou = max(errou, abs((gamou(13)-gama(13))/gamou(13)))
    !
    calaou = errou >= epsact ! SETUP FLAGS
    frst = .FALSE.
  END IF
  !
  ! INNER CALCULATION LOOP ; ALWAYS
  !
  errin = zero ! CONVERGENCE CRITERION
  DO i = 7, 10
    errin = max(errin, abs((gamin(i)-gama(i))/gamin(i)))
  END DO
  errin = max(errin, abs((gamin(4)-gama(4))/gamin(4)))
  errin = max(errin, abs((gamin(5)-gama(5))/gamin(5)))
  errin = max(errin, abs((gamin(13)-gama(13))/gamin(13)))
  calain = errin >= epsact
  !
  iclact = iclact + 1 ! Increment ACTIVITY call counter
  !
  RETURN
END SUBROUTINE calcact2
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCACT1
! *** CALCULATES MULTI-COMPONENT ACTIVITY COEFFICIENTS FROM BROMLEYS
!     METHOD FOR AN AMMONIUM-SULFATE AEROSOL SYSTEM.
!     THE BINARY ACTIVITY COEFFICIENTS ARE CALCULATED BY
!     KUSIK-MEISNER RELATION (SUBROUTINE KMTAB or SUBROUTINE KMFUL1).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
!
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcact1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  REAL :: agama
  REAL :: ch
  DOUBLE PRECISION :: errin
  DOUBLE PRECISION :: errou
  REAL :: f1(3)
  REAL :: f2(4)
  REAL :: h
  INTEGER :: i
  INTEGER :: j
  DOUBLE PRECISION :: mpl
  REAL :: sion
  DOUBLE PRECISION :: xij
  DOUBLE PRECISION :: yji
  REAL :: zmi
  REAL :: zpl
  !
  REAL :: g0(6, 4)
  REAL, PARAMETER :: urf = 0.5
  !      PARAMETER (LN10=2.30258509299404568402D0)
  !
  DOUBLE PRECISION :: g
  g(i, j) = (f1(i)/z(i)+f2(j)/z(j+3))/(z(i)+z(j+3)) - h

  g0 = 0.d0

  !
  ! *** SAVE ACTIVITIES IN OLD ARRAY *************************************
  !
  IF (frst) THEN ! Outer loop
    DO i = 7, 9
      gamou(i) = gama(i)
    END DO
    gamou(4) = gama(4)
    !         GAMOU(5) = GAMA(5)
    gamou(13) = gama(13)
  END IF
  !
  DO i = 7, 9 ! Inner loop
    gamin(i) = gama(i)
  END DO
  gamin(4) = gama(4)
  !         GAMIN(5) = GAMA(5)
  gamin(13) = gama(13)
  !
  ! *** CALCULATE IONIC ACTIVITY OF SOLUTION *****************************
  !
  ionic = 0.0
  molal(2) = zero
  molal(4) = zero
  molal(7) = zero
  DO i = 1, 7
    ionic = ionic + molal(i)*z(i)*z(i)
  END DO
  ionic = max(min(0.5*ionic/water,100.D0), tiny)
  !
  ! *** CALCULATE BINARY ACTIVITY COEFFICIENTS ***************************
  !
  !  G0(1,1)=G11;G0(1,2)=G07;G0(1,3)=G08;G0(1,4)=G10;G0(2,1)=G01;G0(2,2)=G02
  !  G0(2,3)=G12;G0(2,4)=G03;G0(3,1)=G06;G0(3,2)=G04;G0(3,3)=G09;G0(3,4)=G05
  !
  IF (iacalc==0) THEN ! K.M.; FULL
    CALL kmful1(ionic, sngl(temp), g0(3,2), g0(1,2), g0(1,3), g0(3,3))
  ELSE ! K.M.; TABULATED
    CALL kmtab(ionic, sngl(temp), g0(2,1), g0(2,2), g0(2,4), g0(3,2),      &
         g0(3,4), g0(3,1), g0(1,2), g0(1,3), g0(3,3), g0(1,4), g0(1,1),       &
         g0(2,3), g0(4,4), g0(4,1), g0(5,2), g0(5,3), g0(5,4), g0(5,1),       &
         g0(6,2), g0(6,4), g0(6,1))
  END IF
  !
  ! *** CALCULATE MULTICOMPONENT ACTIVITY COEFFICIENTS *******************
  !
  agama = 0.511*(298.0/temp)**1.5 ! Debye Huckel const. at T
  sion = sqrt(ionic)
  h = agama*sion/(1+sion)
  !
  DO i = 1, 3
    f1(i) = 0.0
    f2(i) = 0.0
  END DO
  f2(4) = 0.0
  !
  DO i = 1, 3, 2
    zpl = z(i)
    mpl = molal(i)/water
    DO j = 2, 3
      zmi = z(j+3)
      ch = 0.25*(zpl+zmi)*(zpl+zmi)/ionic
      xij = ch*mpl
      yji = ch*molal(j+3)/water
      f1(i) = f1(i) + sngl(yji*(g0(i,j)+zpl*zmi*h))
      f2(j) = f2(j) + sngl(xij*(g0(i,j)+zpl*zmi*h))
    END DO
  END DO
  !
  ! *** LOG10 OF ACTIVITY COEFFICIENTS ***********************************
  !
  !      GAMA(01) = G(2,1)*ZZ(01)                     ! NACL
  !      GAMA(02) = G(2,2)*ZZ(02)                     ! NA2SO4
  !      GAMA(03) = G(2,4)*ZZ(03)                     ! NANO3
  gama(04) = g(3, 2)*zz(04) ! (NH4)2SO4
  !      GAMA(05) = G(3,4)*ZZ(05)                     ! NH4NO3
  !      GAMA(06) = G(3,1)*ZZ(06)                     ! NH4CL
  gama(07) = g(1, 2)*zz(07) ! 2H-SO4
  gama(08) = g(1, 3)*zz(08) ! H-HSO4
  gama(09) = g(3, 3)*zz(09)                           ! NH4HSO4 
  !      GAMA(10) = G(1,4)*ZZ(10)                     ! HNO3
  !      GAMA(11) = G(1,1)*ZZ(11)                     ! HCL
  !      GAMA(12) = G(2,3)*ZZ(12)                     ! NAHSO4
  gama(13) = 0.20*(3.0*gama(04)+2.0*gama(09)) ! LC ; SCAPE
  !C      GAMA(13) = 0.50*(GAMA(04)+GAMA(09))          ! LC ; SEQUILIB
  !C      GAMA(13) = 0.25*(3.0*GAMA(04)+GAMA(07))      ! LC ; AIM
  !
  ! *** CONVERT LOG (GAMA) COEFFICIENTS TO GAMA **************************
  !
  DO i = 7, 9
    gama(i) = max(-5.0D0, min(gama(i),5.0D0)) ! F77 LIBRARY ROUTINE
    gama(i) = 10.0**gama(i)
    !         GAMA(I)=EXP(LN10*GAMA(I))
    !C         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
    !         GAMA(I) = GAMIN(I)*(1.0-URF) + URF*GAMA(I)  ! Under-relax GAMA's
  END DO
  !
  gama(4) = max(-5.0D0, min(gama(4),5.0D0)) ! F77 LIBRARY ROUTINE
  gama(4) = 10.0**gama(4)
  !         GAMA(I)=EXP(LN10*GAMA(I))
  !C         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
  !         GAMA(4) = GAMIN(4)*(1.0-URF) + URF*GAMA(4)  ! Under-relax GAMA's
  !
  !      GAMA(5)=MAX(-5.0d0, MIN(GAMA(5),5.0d0) ) ! F77 LIBRARY ROUTINE
  !         GAMA(5)=10.0**GAMA(5)
  !C         GAMA(I)=EXP(LN10*GAMA(I))
  !CC         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
  !         GAMA(5) = GAMIN(5)*(1.0-URF) + URF*GAMA(I)  ! Under-relax GAMA's
  !
  gama(13) = max(-5.0D0, min(gama(13),5.0D0)) ! F77 LIBRARY ROUTINE
  gama(13) = 10.0**gama(13)
  !         GAMA(I)=EXP(LN10*GAMA(I))
  !C         GAMA(I)=EX10(SNGL(GAMA(I)), 5.0)    ! CUTOFF SET TO [-5,5]
  !         GAMA(13) = GAMIN(13)*(1.0-URF) + URF*GAMA(13)  ! Under-relax GAMA's
  !
  ! *** SETUP ACTIVITY CALCULATION FLAGS *********************************
  !
  ! OUTER CALCULATION LOOP ; ONLY IF FRST=.TRUE.
  !
  IF (frst) THEN
    errou = zero ! CONVERGENCE CRITERION
    DO i = 7, 9
      errou = max(errou, abs((gamou(i)-gama(i))/gamou(i)))
    END DO
    errou = max(errou, abs((gamou(4)-gama(4))/gamou(4)))
    !         ERROU=MAX(ERROU, ABS((GAMOU(5)-GAMA(5))/GAMOU(5)))
    errou = max(errou, abs((gamou(13)-gama(13))/gamou(13)))
    !
    calaou = errou >= epsact ! SETUP FLAGS
    frst = .FALSE.
  END IF
  !
  ! INNER CALCULATION LOOP ; ALWAYS
  !
  errin = zero ! CONVERGENCE CRITERION
  DO i = 7, 9
    errin = max(errin, abs((gamin(i)-gama(i))/gamin(i)))
  END DO
  errin = max(errin, abs((gamin(4)-gama(4))/gamin(4)))
  !         ERRIN = MAX (ERRIN, ABS((GAMIN(5)-GAMA(5))/GAMIN(5)))
  errin = max(errin, abs((gamin(13)-gama(13))/gamin(13)))
  calain = errin >= epsact
  !
  iclact = iclact + 1 ! Increment ACTIVITY call counter
  !
  RETURN
END SUBROUTINE calcact1
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE RSTGAM
! *** RESETS ACTIVITY COEFFICIENT ARRAYS TO DEFAULT VALUE OF 0.1
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!

SUBROUTINE rstgam (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i

  !
  DO i = 1, npair
    gama(i) = 0.1
  END DO
  !
  RETURN
END SUBROUTINE rstgam
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE RSTGAMP
! *** RESETS ACTIVITY COEFFICIENT ARRAYS TO DEFAULT VALUE OF 0.1 IF
! *** GREATER THAN THE THRESHOLD VALUE.
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE rstgamp
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION ::  gmax,gthresh
  INTEGER :: i
!
  gthresh = 100.d0
  gmax    = 0.1d0
  DO i=1, npair
    gmax = MAX(gmax,gama(i))
  ENDDO
  IF ((gmax) > (gthresh)) THEN
    DO i = 1,npair
      gama(i)  = 1.d-1
      gamin(i) = great
      gamou(i) = great
    ENDDO
    calaou   = .TRUE.
    frst     = .TRUE.
  ENDIF
!
END SUBROUTINE rstgamp

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE KMFUL4
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD
!     FOR AN AMMONIUM-SULFATE-NITRATE-CHLORIDE-SODIUM-CALCIUM-POTASSIUM-MAGNESIUM
!     AEROSOL SYSTEM.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
!
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!

SUBROUTINE kmful4 ( ionic, temp, g01, g02, g03, g04, g05, g06, g07,&
     & g08, g09, g10, g11, g12, g15, g16, g17, g18, g19, g20, g21,&
     & g22, g23 )
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(in)    :: temp
  REAL, INTENT(inout) :: g01
  REAL, INTENT(inout) :: g02
  REAL, INTENT(inout) :: g03
  REAL, INTENT(inout) :: g04
  REAL, INTENT(inout) :: g05
  REAL, INTENT(inout) :: g06
  REAL, INTENT(inout) :: g07
  REAL, INTENT(inout) :: g08
  REAL, INTENT(inout) :: g09
  REAL, INTENT(inout) :: g10
  REAL, INTENT(inout) :: g11
  REAL, INTENT(inout) :: g12
  REAL, INTENT(inout) :: g15
  REAL, INTENT(inout) :: g16
  REAL, INTENT(inout) :: g17
  REAL, INTENT(inout) :: g18
  REAL, INTENT(inout) :: g19
  REAL, INTENT(inout) :: g20
  REAL, INTENT(inout) :: g21
  REAL, INTENT(inout) :: g22
  REAL, INTENT(inout) :: g23
  REAL :: cf1
  REAL :: cf2
  REAL :: sion
  REAL :: tc
  REAL :: ti
  REAL :: z01 = 1.0e+0
  REAL :: z02 = 2.0e+0
  REAL :: z03 = 1.0e+0
  REAL :: z04 = 2.0e+0
  REAL :: z05 = 1.0e+0
  REAL :: z06 = 1.0e+0
  REAL :: z07 = 2.0e+0
  REAL :: z08 = 1.0e+0
  REAL :: z10 = 1.0e+0
  REAL :: z11 = 1.0e+0
  REAL :: z15 = 2.0e+0
  REAL :: z16 = 2.0e+0
  REAL :: z17 = 2.0e+0
  REAL :: z19 = 1.0e+0
  REAL :: z20 = 1.0e+0
  REAL :: z21 = 4.0e+0
  REAL :: z22 = 2.0e+0
  REAL :: z23 = 2.0e+0
  !
  sion = sqrt(ionic)
  !
  ! *** Coefficients at 25 oC
  !
  CALL mkbi(2.230, ionic, sion, z01, g01)
  CALL mkbi(-0.19, ionic, sion, z02, g02)
  CALL mkbi(-0.39, ionic, sion, z03, g03)
  CALL mkbi(-0.25, ionic, sion, z04, g04)
  CALL mkbi(-1.15, ionic, sion, z05, g05)
  CALL mkbi(0.820, ionic, sion, z06, g06)
  CALL mkbi(-.100, ionic, sion, z07, g07)
  CALL mkbi(8.000, ionic, sion, z08, g08)
  CALL mkbi(2.600, ionic, sion, z10, g10)
  CALL mkbi(6.000, ionic, sion, z11, g11)
  CALL mkbi(0.930, ionic, sion, z15, g15)
  CALL mkbi(2.400, ionic, sion, z16, g16)
  CALL mkbi(-0.25, ionic, sion, z17, g17)
  CALL mkbi(-2.33, ionic, sion, z19, g19)
  CALL mkbi(0.920, ionic, sion, z20, g20)
  CALL mkbi(0.150, ionic, sion, z21, g21)
  CALL mkbi(2.320, ionic, sion, z22, g22)
  CALL mkbi(2.900, ionic, sion, z23, g23)
  !
  ! *** Correct for T other than 298 K
  !
  ti = temp - 273.0
  tc = ti - 25.0
  IF (abs(tc)>1.0) THEN
    cf1 = 1.125 - 0.005*ti
    cf2 = (0.125-0.005*ti)*(0.039*ionic**0.92-0.41*sion/(1.+sion))
    g01 = cf1*g01 - cf2*z01
    g02 = cf1*g02 - cf2*z02
    g03 = cf1*g03 - cf2*z03
    g04 = cf1*g04 - cf2*z04
    g05 = cf1*g05 - cf2*z05
    g06 = cf1*g06 - cf2*z06
    g07 = cf1*g07 - cf2*z07
    g08 = cf1*g08 - cf2*z08
    g10 = cf1*g10 - cf2*z10
    g11 = cf1*g11 - cf2*z11
    g15 = cf1*g15 - cf2*z15
    g16 = cf1*g16 - cf2*z16
    g17 = cf1*g17 - cf2*z17
    g19 = cf1*g19 - cf2*z19
    g20 = cf1*g20 - cf2*z20
    g21 = cf1*g21 - cf2*z21
    g22 = cf1*g22 - cf2*z22
    g23 = cf1*g23 - cf2*z23

  END IF
  !
  g09 = g06 + g08 - g11
  g12 = g01 + g08 - g11
  g18 = g08 + g20 - g11
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE kmful4
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KMFUL3
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD
!     FOR AN AMMONIUM-SULFATE-NITRATE-CHLORIDE-SODIUM AEROSOL SYSTEM.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
!
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE kmful3 ( ionic, temp, g01, g02, g03, g04, g05, g06, g07,&
     & g08, g09, g10, g11, g12 )
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(in)    :: temp
  REAL, INTENT(inout) :: g01
  REAL, INTENT(inout) :: g02
  REAL, INTENT(inout) :: g03
  REAL, INTENT(inout) :: g04
  REAL, INTENT(inout) :: g05
  REAL, INTENT(inout) :: g06
  REAL, INTENT(inout) :: g07
  REAL, INTENT(inout) :: g08
  REAL, INTENT(inout) :: g09
  REAL, INTENT(inout) :: g10
  REAL, INTENT(inout) :: g11
  REAL, INTENT(inout) :: g12
  REAL :: cf1
  REAL :: cf2
  REAL :: sion
  REAL :: tc
  REAL :: ti
  REAL :: z01 = 1.0e+0
  REAL :: z02 = 2.0e+0
  REAL :: z03 = 1.0e+0
  REAL :: z04 = 2.0e+0
  REAL :: z05 = 1.0e+0
  REAL :: z06 = 1.0e+0
  REAL :: z07 = 2.0e+0
  REAL :: z08 = 1.0e+0
  REAL :: z10 = 1.0e+0
  REAL :: z11 = 1.0e+0
  !
  sion = sqrt(ionic)
  !
  ! *** Coefficients at 25 oC
  !
  CALL mkbi(2.230, ionic, sion, z01, g01)
  CALL mkbi(-0.19, ionic, sion, z02, g02)
  CALL mkbi(-0.39, ionic, sion, z03, g03)
  CALL mkbi(-0.25, ionic, sion, z04, g04)
  CALL mkbi(-1.15, ionic, sion, z05, g05)
  CALL mkbi(0.820, ionic, sion, z06, g06)
  CALL mkbi(-.100, ionic, sion, z07, g07)
  CALL mkbi(8.000, ionic, sion, z08, g08)
  CALL mkbi(2.600, ionic, sion, z10, g10)
  CALL mkbi(6.000, ionic, sion, z11, g11)
  !
  ! *** Correct for T other than 298 K
  !
  ti = temp - 273.0
  tc = ti - 25.0
  IF (abs(tc)>1.0) THEN
    cf1 = 1.125 - 0.005*ti
    cf2 = (0.125-0.005*ti)*(0.039*ionic**0.92-0.41*sion/(1.+sion))
    g01 = cf1*g01 - cf2*z01
    g02 = cf1*g02 - cf2*z02
    g03 = cf1*g03 - cf2*z03
    g04 = cf1*g04 - cf2*z04
    g05 = cf1*g05 - cf2*z05
    g06 = cf1*g06 - cf2*z06
    g07 = cf1*g07 - cf2*z07
    g08 = cf1*g08 - cf2*z08
    g10 = cf1*g10 - cf2*z10
    g11 = cf1*g11 - cf2*z11
  END IF
  !
  g09 = g06 + g08 - g11
  g12 = g01 + g08 - g11
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE kmful3
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KMFUL2
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD
!     FOR AN AMMONIUM-SULFATE-NITRATE AEROSOL SYSTEM.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
!
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE kmful2 ( ionic, temp, g04, g05, g07, g08, g09, g10 )
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(in)    :: temp
  REAL, INTENT(inout) :: g04
  REAL, INTENT(inout) :: g05
  REAL, INTENT(inout) :: g07
  REAL, INTENT(inout) :: g08
  REAL, INTENT(inout) :: g09
  REAL, INTENT(inout) :: g10
  REAL :: g06
  REAL :: g11
  REAL :: cf1
  REAL :: cf2
  REAL :: sion
  REAL :: tc
  REAL :: ti
  ! REAL :: z01 = 1.0e+0
  ! REAL :: z02 = 2.0e+0
  ! REAL :: z03 = 1.0e+0
  REAL :: z04 = 2.0e+0
  REAL :: z05 = 1.0e+0
  REAL :: z06 = 1.0e+0
  REAL :: z07 = 2.0e+0
  REAL :: z08 = 1.0e+0
  REAL :: z10 = 1.0e+0
  REAL :: z11 = 1.0e+0
  !
  sion = sqrt(ionic)
  !
  ! *** Coefficients at 25 oC
  !
  !      CALL MKBI(2.230, IONIC, SION, Z01, G01)
  !      CALL MKBI(-0.19, IONIC, SION, Z02, G02)
  !      CALL MKBI(-0.39, IONIC, SION, Z03, G03)
  CALL mkbi(-0.25, ionic, sion, z04, g04)
  CALL mkbi(-1.15, ionic, sion, z05, g05)
  CALL mkbi(0.820, ionic, sion, z06, g06)
  CALL mkbi(-.100, ionic, sion, z07, g07)
  CALL mkbi(8.000, ionic, sion, z08, g08)
  CALL mkbi(2.600, ionic, sion, z10, g10)
  CALL mkbi(6.000, ionic, sion, z11, g11)
  !
  ! *** Correct for T other than 298 K
  !
  ti = temp - 273.0
  tc = ti - 25.0
  IF (abs(tc)>1.0) THEN
    cf1 = 1.125 - 0.005*ti
    cf2 = (0.125-0.005*ti)*(0.039*ionic**0.92-0.41*sion/(1.+sion))
    !         G01 = CF1*G01 - CF2*Z01
    !         G02 = CF1*G02 - CF2*Z02
    !         G03 = CF1*G03 - CF2*Z03
    g04 = cf1*g04 - cf2*z04
    g05 = cf1*g05 - cf2*z05
    g06 = cf1*g06 - cf2*z06
    g07 = cf1*g07 - cf2*z07
    g08 = cf1*g08 - cf2*z08
    g10 = cf1*g10 - cf2*z10
    g11 = cf1*g11 - cf2*z11
  END IF
  !
  ! original method of calculating G09     
      g09 = g06 + g08 - g11

  ! slc.debug
  !  g09 = g05 + g08 - g10
  !      G12 = G01 + G08 - G11
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE kmful2
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KMFUL1
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD
!     FOR AN AMMONIUM-SULFATE AEROSOL SYSTEM.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
!
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE kmful1 ( ionic, temp, g04, g07, g08, g09 )
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(in)    :: temp
  REAL, INTENT(inout) :: g04
  REAL, INTENT(inout) :: g07
  REAL, INTENT(inout) :: g08
  REAL, INTENT(inout) :: g09
  REAL :: g06
  REAL :: g11
  REAL :: cf1
  REAL :: cf2
  REAL :: sion
  REAL :: tc
  REAL :: ti
  ! REAL :: z01 = 1.0e+0
  ! REAL :: z02 = 2.0e+0
  ! REAL :: z03 = 1.0e+0
  REAL :: z04 = 2.0e+0
  ! REAL :: z05 = 1.0e+0
  REAL :: z06 = 1.0e+0
  REAL :: z07 = 2.0e+0
  REAL :: z08 = 1.0e+0
  ! REAL :: z10 = 1.0e+0
  REAL :: z11 = 1.0e+0
  !
  sion = sqrt(ionic)
  !
  ! *** Coefficients at 25 oC
  !
  !      CALL MKBI(2.230, IONIC, SION, Z01, G01)
  !      CALL MKBI(-0.19, IONIC, SION, Z02, G02)
  !      CALL MKBI(-0.39, IONIC, SION, Z03, G03)
  CALL mkbi(-0.25, ionic, sion, z04, g04)
  !      CALL MKBI(-1.15, IONIC, SION, Z05, G05)
  CALL mkbi(0.820, ionic, sion, z06, g06)
  CALL mkbi(-.100, ionic, sion, z07, g07)
  CALL mkbi(8.000, ionic, sion, z08, g08)
  !      CALL MKBI(2.600, IONIC, SION, Z10, G10)
  CALL mkbi(6.000, ionic, sion, z11, g11)
  !
  ! *** Correct for T other than 298 K
  !
  ti = temp - 273.0
  tc = ti - 25.0
  IF (abs(tc)>1.0) THEN
    cf1 = 1.125 - 0.005*ti
    cf2 = (0.125-0.005*ti)*(0.039*ionic**0.92-0.41*sion/(1.+sion))
    !         G01 = CF1*G01 - CF2*Z01
    !         G02 = CF1*G02 - CF2*Z02
    !         G03 = CF1*G03 - CF2*Z03
    g04 = cf1*g04 - cf2*z04
    !         G05 = CF1*G05 - CF2*Z05
    g06 = cf1*g06 - cf2*z06
    g07 = cf1*g07 - cf2*z07
    g08 = cf1*g08 - cf2*z08
    !         G10 = CF1*G10 - CF2*Z10
    g11 = cf1*g11 - cf2*z11
  END IF

   !     ! Correction - G09 is G0(3,3), which is not calculated in CALCACT1
   !     !  Use G09 from CALCACT3 to represent G09 (slc.2.2012)
    g09 = g06 + g08 - g11

  !
  !      G09 = G05 + G08 - G10   ! CALCULATED IN CALCACT1
  !      G12 = G01 + G08 - G11
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE kmful1
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE MKBI
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD.
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE mkbi ( q, ionic, sion, zip, bi )
  IMPLICIT NONE
  REAL, INTENT(in)    :: q
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(inout) :: sion
  REAL, INTENT(inout) :: zip
  REAL, INTENT(inout) :: bi
  REAL :: b
  REAL :: c
  REAL :: xx
  !
  b = .75 - .065*q
  c = 1.0
  IF (ionic<6.0) c = 1. + .055*q*exp(-.023*ionic*ionic*ionic)
  xx = -0.5107*sion/(1.+c*sion)
  bi = (1.+b*(1.+.1*ionic)**q-b)
  bi = zip*alog10(bi) + zip*xx
  !
  RETURN
END SUBROUTINE mkbi
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KMTAB
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD.
!     THE COMPUTATIONS HAVE BEEN PERFORMED AND THE RESULTS ARE STORED IN
!     LOOKUP TABLES. THE IONIC ACTIVITY 'IONIC' IS INPUT, AND THE ARRAY
!     'BINARR' IS RETURNED WITH THE BINARY COEFFICIENTS.
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE kmtab ( in, temp, g01, g02, g03, g04, g05, g06, g07, g08,&
     & g09, g10, g11, g12, g15, g16, g17, g18, g19, g20, g21, g22,&
     & g23 )
  IMPLICIT NONE
  REAL, INTENT(inout) :: in
  REAL, INTENT(in)    :: temp
  REAL, INTENT(inout) :: g01
  REAL, INTENT(inout) :: g02
  REAL, INTENT(inout) :: g03
  REAL, INTENT(inout) :: g04
  REAL, INTENT(inout) :: g05
  REAL, INTENT(inout) :: g06
  REAL, INTENT(inout) :: g07
  REAL, INTENT(inout) :: g08
  REAL, INTENT(inout) :: g09
  REAL, INTENT(inout) :: g10
  REAL, INTENT(inout) :: g11
  REAL, INTENT(inout) :: g12
  REAL, INTENT(inout) :: g15
  REAL, INTENT(inout) :: g16
  REAL, INTENT(inout) :: g17
  REAL, INTENT(inout) :: g18
  REAL, INTENT(inout) :: g19
  REAL, INTENT(inout) :: g20
  REAL, INTENT(inout) :: g21
  REAL, INTENT(inout) :: g22
  REAL, INTENT(inout) :: g23
  REAL :: binarray(23)
  REAL :: g13
  REAL :: g14
  INTEGER :: ind
  !
  ! *** Find temperature range
  !
  ind = nint((temp-198.0)/25.0) + 1
  ind = min(max(ind,1), 6)
  !
  ! *** Call appropriate routine
  !
  IF (ind==1) THEN
    CALL km198(in, binarray)
  ELSE IF (ind==2) THEN
    CALL km223(in, binarray)
  ELSE IF (ind==3) THEN
    CALL km248(in, binarray)
  ELSE IF (ind==4) THEN
    CALL km273(in, binarray)
  ELSE IF (ind==5) THEN
    CALL km298(in, binarray)
  ELSE
    CALL km323(in, binarray)
  END IF
  !
  g01 = binarray(01)
  g02 = binarray(02)
  g03 = binarray(03)
  g04 = binarray(04)
  g05 = binarray(05)
  g06 = binarray(06)
  g07 = binarray(07)
  g08 = binarray(08)
  g09 = binarray(09)
  g10 = binarray(10)
  g11 = binarray(11)
  g12 = binarray(12)
  g13 = binarray(13)
  g14 = binarray(14)
  g15 = binarray(15)
  g16 = binarray(16)
  g17 = binarray(17)
  g18 = binarray(18)
  g19 = binarray(19)
  g20 = binarray(20)
  g21 = binarray(21)
  g22 = binarray(22)
  g23 = binarray(23)
  !
  ! *** Return point; End of subroutine
  !
  RETURN
END SUBROUTINE kmtab


!      INTEGER FUNCTION IBACPOS(IN)
!C
!C     Compute the index in the binary activity coefficient array
!C     based on the input ionic strength.
!C
!C     Chris Nolte, 6/16/05
!C
!      implicit none
!      real IN
!      IF (IN .LE. 0.300000E+02) THEN
!         ibacpos = MIN(NINT( 0.200000E+02*IN) + 1, 600)
!      ELSE
!         ibacpos =   600+NINT( 0.200000E+01*IN- 0.600000E+02)
!      ENDIF
!      ibacpos = min(ibacpos, 741)
!      return
!      end

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KM198
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD.
!     THE COMPUTATIONS HAVE BEEN PERFORMED AND THE RESULTS ARE STORED IN
!     LOOKUP TABLES. THE IONIC ACTIVITY 'IN' IS INPUT, AND THE ARRAY
!     'BINARR' IS RETURNED WITH THE BINARY COEFFICIENTS.
!
!     TEMPERATURE IS 198K
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE km198 ( ionic, binarr )
  USE mo_kmcf198  
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(inout) :: binarr(23)
  INTEGER :: ipos
  !
  ! *** Find position in arrays for bincoef
  !
  IF (ionic<=0.200000E+02) THEN
    ipos = min(nint(0.200000E+02*ionic)+1, 400)
  ELSE
    ipos = 400 + nint(0.200000E+01*ionic-0.400000E+02)
  END IF
  ipos = min(ipos, 561)
  !
  ! *** Assign values to return array
  !
  binarr(01) = bnc01m(ipos)
  binarr(02) = bnc02m(ipos)
  binarr(03) = bnc03m(ipos)
  binarr(04) = bnc04m(ipos)
  binarr(05) = bnc05m(ipos)
  binarr(06) = bnc06m(ipos)
  binarr(07) = bnc07m(ipos)
  binarr(08) = bnc08m(ipos)
  binarr(09) = bnc09m(ipos)
  binarr(10) = bnc10m(ipos)
  binarr(11) = bnc11m(ipos)
  binarr(12) = bnc12m(ipos)
  binarr(13) = bnc13m(ipos)
  binarr(14) = bnc14m(ipos)
  binarr(15) = bnc15m(ipos)
  binarr(16) = bnc16m(ipos)
  binarr(17) = bnc17m(ipos)
  binarr(18) = bnc18m(ipos)
  binarr(19) = bnc19m(ipos)
  binarr(20) = bnc20m(ipos)
  binarr(21) = bnc21m(ipos)
  binarr(22) = bnc22m(ipos)
  binarr(23) = bnc23m(ipos)
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE km198

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KM223
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD.
!     THE COMPUTATIONS HAVE BEEN PERFORMED AND THE RESULTS ARE STORED IN
!     LOOKUP TABLES. THE IONIC ACTIVITY 'IN' IS INPUT, AND THE ARRAY
!     'BINARR' IS RETURNED WITH THE BINARY COEFFICIENTS.
!
!     TEMPERATURE IS 223K
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!

SUBROUTINE km223 ( ionic, binarr )
  USE mo_kmcf223
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(inout) :: binarr(23)
  INTEGER :: ipos
  !
  ! *** Find position in arrays for bincoef
  !
  IF (ionic<=0.200000E+02) THEN
    ipos = min(nint(0.200000E+02*ionic)+1, 400)
  ELSE
    ipos = 400 + nint(0.200000E+01*ionic-0.400000E+02)
  END IF
  ipos = min(ipos, 561)
  !
  ! *** Assign values to return array
  !
  binarr(01) = bnc01m(ipos)
  binarr(02) = bnc02m(ipos)
  binarr(03) = bnc03m(ipos)
  binarr(04) = bnc04m(ipos)
  binarr(05) = bnc05m(ipos)
  binarr(06) = bnc06m(ipos)
  binarr(07) = bnc07m(ipos)
  binarr(08) = bnc08m(ipos)
  binarr(09) = bnc09m(ipos)
  binarr(10) = bnc10m(ipos)
  binarr(11) = bnc11m(ipos)
  binarr(12) = bnc12m(ipos)
  binarr(13) = bnc13m(ipos)
  binarr(14) = bnc14m(ipos)
  binarr(15) = bnc15m(ipos)
  binarr(16) = bnc16m(ipos)
  binarr(17) = bnc17m(ipos)
  binarr(18) = bnc18m(ipos)
  binarr(19) = bnc19m(ipos)
  binarr(20) = bnc20m(ipos)
  binarr(21) = bnc21m(ipos)
  binarr(22) = bnc22m(ipos)
  binarr(23) = bnc23m(ipos)
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE km223

!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KM248
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD.
!     THE COMPUTATIONS HAVE BEEN PERFORMED AND THE RESULTS ARE STORED IN
!     LOOKUP TABLES. THE IONIC ACTIVITY 'IN' IS INPUT, AND THE ARRAY
!     'BINARR' IS RETURNED WITH THE BINARY COEFFICIENTS.
!
!     TEMPERATURE IS 248K
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE km248 ( ionic, binarr )
  USE mo_kmcf248
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(inout) :: binarr(23)
  INTEGER :: ipos
  !
  ! *** Find position in arrays for bincoef
  !
  IF (ionic<=0.200000E+02) THEN
    ipos = min(nint(0.200000E+02*ionic)+1, 400)
  ELSE
    ipos = 400 + nint(0.200000E+01*ionic-0.400000E+02)
  END IF
  ipos = min(ipos, 561)
  !
  ! *** Assign values to return array
  !
  binarr(01) = bnc01m(ipos)
  binarr(02) = bnc02m(ipos)
  binarr(03) = bnc03m(ipos)
  binarr(04) = bnc04m(ipos)
  binarr(05) = bnc05m(ipos)
  binarr(06) = bnc06m(ipos)
  binarr(07) = bnc07m(ipos)
  binarr(08) = bnc08m(ipos)
  binarr(09) = bnc09m(ipos)
  binarr(10) = bnc10m(ipos)
  binarr(11) = bnc11m(ipos)
  binarr(12) = bnc12m(ipos)
  binarr(13) = bnc13m(ipos)
  binarr(14) = bnc14m(ipos)
  binarr(15) = bnc15m(ipos)
  binarr(16) = bnc16m(ipos)
  binarr(17) = bnc17m(ipos)
  binarr(18) = bnc18m(ipos)
  binarr(19) = bnc19m(ipos)
  binarr(20) = bnc20m(ipos)
  binarr(21) = bnc21m(ipos)
  binarr(22) = bnc22m(ipos)
  binarr(23) = bnc23m(ipos)
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE km248

!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KM273
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD.
!     THE COMPUTATIONS HAVE BEEN PERFORMED AND THE RESULTS ARE STORED IN
!     LOOKUP TABLES. THE IONIC ACTIVITY 'IN' IS INPUT, AND THE ARRAY
!     'BINARR' IS RETURNED WITH THE BINARY COEFFICIENTS.
!
!     TEMPERATURE IS 273K
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!

SUBROUTINE km273 ( ionic, binarr )
  USE mo_kmcf273
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(inout) :: binarr(23)
  INTEGER :: ipos
  !
  ! *** Find position in arrays for bincoef
  !
  IF (ionic<=0.200000E+02) THEN
    ipos = min(nint(0.200000E+02*ionic)+1, 400)
  ELSE
    ipos = 400 + nint(0.200000E+01*ionic-0.400000E+02)
  END IF
  ipos = min(ipos, 561)
  !
  ! *** Assign values to return array
  !
  binarr(01) = bnc01m(ipos)
  binarr(02) = bnc02m(ipos)
  binarr(03) = bnc03m(ipos)
  binarr(04) = bnc04m(ipos)
  binarr(05) = bnc05m(ipos)
  binarr(06) = bnc06m(ipos)
  binarr(07) = bnc07m(ipos)
  binarr(08) = bnc08m(ipos)
  binarr(09) = bnc09m(ipos)
  binarr(10) = bnc10m(ipos)
  binarr(11) = bnc11m(ipos)
  binarr(12) = bnc12m(ipos)
  binarr(13) = bnc13m(ipos)
  binarr(14) = bnc14m(ipos)
  binarr(15) = bnc15m(ipos)
  binarr(16) = bnc16m(ipos)
  binarr(17) = bnc17m(ipos)
  binarr(18) = bnc18m(ipos)
  binarr(19) = bnc19m(ipos)
  binarr(20) = bnc20m(ipos)
  binarr(21) = bnc21m(ipos)
  binarr(22) = bnc22m(ipos)
  binarr(23) = bnc23m(ipos)
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE km273


!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KM298
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD.
!     THE COMPUTATIONS HAVE BEEN PERFORMED AND THE RESULTS ARE STORED IN
!     LOOKUP TABLES. THE IONIC ACTIVITY 'IN' IS INPUT, AND THE ARRAY
!     'BINARR' IS RETURNED WITH THE BINARY COEFFICIENTS.
!
!     TEMPERATURE IS 298K
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE km298 ( ionic, binarr )
  USE mo_kmcf298
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(inout) :: binarr(23)
  INTEGER :: ipos
  !
  ! *** Find position in arrays for bincoef
  !
  IF (ionic<=0.200000E+02) THEN
    ipos = min(nint(0.200000E+02*ionic)+1, 400)
  ELSE
    ipos = 400 + nint(0.200000E+01*ionic-0.400000E+02)
  END IF
  ipos = min(ipos, 561)
  !
  ! *** Assign values to return array
  !
  binarr(01) = bnc01m(ipos)
  binarr(02) = bnc02m(ipos)
  binarr(03) = bnc03m(ipos)
  binarr(04) = bnc04m(ipos)
  binarr(05) = bnc05m(ipos)
  binarr(06) = bnc06m(ipos)
  binarr(07) = bnc07m(ipos)
  binarr(08) = bnc08m(ipos)
  binarr(09) = bnc09m(ipos)
  binarr(10) = bnc10m(ipos)
  binarr(11) = bnc11m(ipos)
  binarr(12) = bnc12m(ipos)
  binarr(13) = bnc13m(ipos)
  binarr(14) = bnc14m(ipos)
  binarr(15) = bnc15m(ipos)
  binarr(16) = bnc16m(ipos)
  binarr(17) = bnc17m(ipos)
  binarr(18) = bnc18m(ipos)
  binarr(19) = bnc19m(ipos)
  binarr(20) = bnc20m(ipos)
  binarr(21) = bnc21m(ipos)
  binarr(22) = bnc22m(ipos)
  binarr(23) = bnc23m(ipos)
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE km298

!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE KM323
! *** CALCULATES BINARY ACTIVITY COEFFICIENTS BY KUSIK-MEISSNER METHOD.
!     THE COMPUTATIONS HAVE BEEN PERFORMED AND THE RESULTS ARE STORED IN
!     LOOKUP TABLES. THE IONIC ACTIVITY 'IN' IS INPUT, AND THE ARRAY
!     'BINARR' IS RETURNED WITH THE BINARY COEFFICIENTS.
!
!     TEMPERATURE IS 323K
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!

SUBROUTINE km323 ( ionic, binarr )
  USE mo_kmcf223
  IMPLICIT NONE
  REAL, INTENT(inout) :: ionic
  REAL, INTENT(inout) :: binarr(23)
  INTEGER :: ipos
  !
  ! *** Find position in arrays for bincoef
  !
  IF (ionic<=0.200000E+02) THEN
    ipos = min(nint(0.200000E+02*ionic)+1, 400)
  ELSE
    ipos = 400 + nint(0.200000E+01*ionic-0.400000E+02)
  END IF
  ipos = min(ipos, 561)
  !
  ! *** Assign values to return array
  !
  binarr(01) = bnc01m(ipos)
  binarr(02) = bnc02m(ipos)
  binarr(03) = bnc03m(ipos)
  binarr(04) = bnc04m(ipos)
  binarr(05) = bnc05m(ipos)
  binarr(06) = bnc06m(ipos)
  binarr(07) = bnc07m(ipos)
  binarr(08) = bnc08m(ipos)
  binarr(09) = bnc09m(ipos)
  binarr(10) = bnc10m(ipos)
  binarr(11) = bnc11m(ipos)
  binarr(12) = bnc12m(ipos)
  binarr(13) = bnc13m(ipos)
  binarr(14) = bnc14m(ipos)
  binarr(15) = bnc15m(ipos)
  binarr(16) = bnc16m(ipos)
  binarr(17) = bnc17m(ipos)
  binarr(18) = bnc18m(ipos)
  binarr(19) = bnc19m(ipos)
  binarr(20) = bnc20m(ipos)
  binarr(21) = bnc21m(ipos)
  binarr(22) = bnc22m(ipos)
  binarr(23) = bnc23m(ipos)
  !
  ! *** Return point ; End of subroutine
  !
  RETURN
END SUBROUTINE km323

!C*************************************************************************
!C
!C  TOOLBOX LIBRARY v.1.0 (May 1995)
!C
!C  Program unit   : SUBROUTINE CHRBLN
!C  Purpose        : Position of last non-blank character in a string
!C  Author         : Athanasios Nenes
!C
!C  ======================= ARGUMENTS / USAGE =============================
!C
!C  STR        is the CHARACTER variable containing the string examined
!C  IBLK       is a INTEGER variable containing the position of last non
!C             blank character. If string is all spaces (ie '   '), then
!C             the value returned is 1.
!C
!C  EXAMPLE:
!C             STR = 'TEST1.DAT     '
!C             CALL CHRBLN (STR, IBLK)
!C
!C  after execution of this code segment, "IBLK" has the value "9", which
!C  is the position of the last non-blank character of "STR".
!C
!C***********************************************************************
!C

SUBROUTINE chrbln ( str, iblk )
  IMPLICIT NONE
  CHARACTER(len=*), INTENT(inout) :: str
  INTEGER, INTENT(inout) :: iblk
  INTEGER :: i
  INTEGER :: ilen
  !C
  !C***********************************************************************
  !
  iblk = 1 ! Substring pointer (default=1)
  ilen = len(str) ! Length of string
  DO i = ilen, 1, -1
    IF (str(i:i)/=' ' .AND. str(i:i)/=char(0)) THEN
      iblk = i
      RETURN
    END IF
  END DO
  RETURN
  !
END SUBROUTINE chrbln


!C*************************************************************************
!C
!C  TOOLBOX LIBRARY v.1.0 (May 1995)
!C
!C  Program unit   : SUBROUTINE SHFTRGHT
!C  Purpose        : RIGHT-JUSTIFICATION FUNCTION ON A STRING
!C  Author         : Athanasios Nenes
!C
!C  ======================= ARGUMENTS / USAGE =============================
!C
!C  STRING     is the CHARACTER variable with the string to be justified
!C
!C  EXAMPLE:
!C             STRING    = 'AAAA    '
!C             CALL SHFTRGHT (STRING)
!C
!C  after execution of this code segment, STRING contains the value
!C  '    AAAA'.
!C
!C*************************************************************************
!C
SUBROUTINE shftrght(chr)
  IMPLICIT NONE
  character(len=*), INTENT(inout) :: chr
  INTEGER :: i
  INTEGER :: i1
  INTEGER :: i2
  !C
  !C***********************************************************************
  !
  i1 = len(chr) ! Total length of string
  CALL chrbln(chr, i2) ! Position of last non-blank character
  IF (i2==i1) RETURN
  !
  DO i = i2, 1, -1 ! Shift characters
    chr(i1+i-i2:i1+i-i2) = chr(i:i)
    chr(i:i) = ' '
  END DO
  RETURN
  !
END SUBROUTINE shftrght




!C*************************************************************************
!C
!C  TOOLBOX LIBRARY v.1.0 (May 1995)
!C
!C  Program unit   : SUBROUTINE RPLSTR
!C  Purpose        : REPLACE CHARACTERS OCCURING IN A STRING
!C  Author         : Athanasios Nenes
!C
!C  ======================= ARGUMENTS / USAGE =============================
!C
!C  STRING     is the CHARACTER variable with the string to be edited
!C  OLD        is the old character which is to be replaced
!C  NEW        is the new character which OLD is to be replaced with
!C  IERR       is 0 if everything went well, is 1 if 'NEW' contains 'OLD'.
!C             In this case, this is invalid, and no change is done.
!C
!C  EXAMPLE:
!C             STRING    = 'AAAA'
!C             OLD       = 'A'
!C             NEW       = 'B'
!C             CALL RPLSTR (STRING, OLD, NEW)
!C
!C  after execution of this code segment, STRING contains the value
!C  'BBBB'.
!C
!C*************************************************************************
!C
SUBROUTINE rplstr(string, old, new, ierr)
  IMPLICIT NONE
  character(len=*), intent(Inout) :: string
  character(len=*), intent(in)    :: old
  character(len=*), intent(in)    :: new
  integer, intent(inout) :: ierr
  INTEGER :: ilo
  INTEGER :: ip
  !C
  !C***********************************************************************
  !
  ! *** INITIALIZE ********************************************************
  !
  ilo = len(old)
  !
  ! *** CHECK AND SEE IF 'NEW' CONTAINS 'OLD', WHICH CANNOT ***************
  !
  ip = index(new, old)
  IF (ip/=0) THEN
    ierr = 1
    RETURN
  ELSE
    ierr = 0
  END IF
  !
  ! *** PROCEED WITH REPLACING *******************************************
  !
100 ip = index(string, old) ! SEE IF 'OLD' EXISTS IN 'STRING'
  IF (ip==0) RETURN ! 'OLD' DOES NOT EXIST ; RETURN
  string(ip:ip+ilo-1) = new ! REPLACE SUBSTRING 'OLD' WITH 'NEW'
  GO TO 100 ! GO FOR NEW OCCURANCE OF 'OLD'
  !
END SUBROUTINE rplstr


!C*************************************************************************
!C
!C  TOOLBOX LIBRARY v.1.0 (May 1995)
!C
!C  Program unit   : SUBROUTINE INPTD
!C  Purpose        : Prompts user for a value (DOUBLE). A default value
!C                   is provided, so if user presses <Enter>, the default
!C                   is used.
!C  Author         : Athanasios Nenes
!C
!C  ======================= ARGUMENTS / USAGE =============================
!C
!C  VAR        is the DOUBLE PRECISION variable which value is to be saved
!C  DEF        is a DOUBLE PRECISION variable, with the default value of VAR.
!C  PROMPT     is a CHARACTER varible containing the prompt string.
!C  PRFMT      is a CHARACTER variable containing the FORMAT specifier
!C             for the default value DEF.
!C  IERR       is an INTEGER error flag, and has the values:
!C             0 - No error detected.
!C             1 - Invalid FORMAT and/or Invalid default value.
!C             2 - Bad value specified by user
!C
!C  EXAMPLE:
!C             CALL INPTD (VAR, 1.0D0, 'Give value for A ', '*', Ierr)
!C
!C  after execution of this code segment, the user is prompted for the
!C  value of variable VAR. If <Enter> is pressed (ie no value is specified)
!C  then 1.0 is assigned to VAR. The default value is displayed in free-
!C  format. The error status is specified by variable Ierr
!C
!C***********************************************************************
!C
SUBROUTINE inptd(var, def, prompt, prfmt, ierr)
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: var
  DOUBLE PRECISION, INTENT(inout) :: def
  character(len=*) :: prompt
  character(len=*) :: prfmt
  character(len=128) :: buffer
  INTEGER :: iend
  INTEGER :: ierr
  !C
  !C***********************************************************************

  !
  ierr = 0
  !
  ! *** WRITE DEFAULT VALUE TO WORK BUFFER *******************************
  !
  WRITE (buffer, FMT=prfmt, ERR=110) def
  CALL chrbln(buffer, iend)
  !
  ! *** PROMPT USER FOR INPUT AND READ IT ********************************
  !
  WRITE (*, *) prompt, ' [', buffer(1:iend), ']: '
  READ (*, '(A)', ERR=120, END=120) buffer
  CALL chrbln(buffer, iend)
  !
  ! *** READ DATA OR SET DEFAULT ? ****************************************
  !
  IF (iend==1 .AND. buffer(1:1)==' ') THEN
    var = def
  ELSE
    READ (buffer, *, ERR=120, END=120) var
  END IF
  !
  ! *** RETURN POINT ******************************************************
  !
100 RETURN
  !
  ! *** ERROR HANDLER *****************************************************
  !
110 ierr = 1 ! Bad FORMAT and/or bad default value
  GO TO 100
  !
120 ierr = 2 ! Bad number given by user
  GO TO 100
  !
END SUBROUTINE inptd


!C*************************************************************************
!C
!C  TOOLBOX LIBRARY v.1.0 (May 1995)
!C
!C  Program unit   : SUBROUTINE Pushend
!C  Purpose        : Positions the pointer of a sequential file at its end
!C                   Simulates the ACCESS='APPEND' clause of a F77L OPEN
!C                   statement with Standard Fortran commands.
!C
!C  ======================= ARGUMENTS / USAGE =============================
!C
!C  Iunit      is a INTEGER variable, the file unit which the file is
!C             connected to.
!C
!C  EXAMPLE:
!C             CALL PUSHEND (10)
!C
!C  after execution of this code segment, the pointer of unit 10 is
!C  pushed to its end.
!C
!C***********************************************************************
!C
SUBROUTINE pushend ( iunit )
  IMPLICIT NONE
  INTEGER, INTENT(inout) :: iunit
  LOGICAL :: opned
  !C
  !C***********************************************************************
  !
  ! *** INQUIRE IF Iunit CONNECTED TO FILE ********************************
  !
  INQUIRE (UNIT=iunit, OPENED=opned)
  IF (.NOT. opned) GO TO 120
  !
  ! *** Iunit CONNECTED, PUSH POINTER TO END ******************************
  !
100 READ (iunit, '()', ERR=110, END=110)
  GO TO 100
  !
  ! *** RETURN POINT ******************************************************
  !
110 BACKSPACE (iunit)
120 RETURN
END SUBROUTINE pushend



!C*************************************************************************
!C
!C  TOOLBOX LIBRARY v.1.0 (May 1995)
!C
!C  Program unit   : SUBROUTINE APPENDEXT
!C  Purpose        : Fix extension in file name string
!C
!C  ======================= ARGUMENTS / USAGE =============================
!C
!C  Filename   is the CHARACTER variable with the file name
!C  Defext     is the CHARACTER variable with extension (including '.',
!C             ex. '.DAT')
!C  Overwrite  is a LOGICAL value, .TRUE. overwrites any existing extension
!C             in "Filename" with "Defext", .FALSE. puts "Defext" only if
!C             there is no extension in "Filename".
!C
!C  EXAMPLE:
!C             FILENAME1 = 'TEST.DAT'
!C             FILENAME2 = 'TEST.DAT'
!C             CALL APPENDEXT (FILENAME1, '.TXT', .FALSE.)
!C             CALL APPENDEXT (FILENAME2, '.TXT', .TRUE. )
!C
!C  after execution of this code segment, "FILENAME1" has the value
!C  'TEST.DAT', while "FILENAME2" has the value 'TEST.TXT'
!C
!C***********************************************************************
!C
SUBROUTINE appendext(filename, defext, overwrite)
  IMPLICIT NONE
  character(len=*), intent(inout) :: filename
  character(len=*), intent(inout) :: defext     
  LOGICAL, intent(inout) :: overwrite
  INTEGER :: idot
  INTEGER :: iend
  !C
  !C***********************************************************************
  !
  CALL chrbln(filename, iend)
  IF (filename(1:1)==' ' .AND. iend==1) RETURN ! Filename empty
  idot = index(filename, '.') ! Append extension ?
  IF (idot==0) filename = filename(1:iend) // defext
  IF (overwrite .AND. idot/=0) filename = filename(:idot-1) // defext
  RETURN
END SUBROUTINE appendext





!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE POLY3
! *** FINDS THE REAL ROOTS OF THE THIRD ORDER ALGEBRAIC EQUATION:
!     X**3 + A1*X**2 + A2*X + A3 = 0.0
!     THE EQUATION IS SOLVED ANALYTICALLY.
!
!     PARAMETERS A1, A2, A3 ARE SPECIFIED BY THE USER. THE MINIMUM
!     NONEGATIVE ROOT IS RETURNED IN VARIABLE 'ROOT'. IF NO ROOT IS
!     FOUND (WHICH IS GREATER THAN ZERO), ROOT HAS THE VALUE 1D30.
!     AND THE FLAG ISLV HAS A VALUE GREATER THAN ZERO.
!
!     SOLUTION FORMULA IS FOUND IN PAGE 32 OF:
!     MATHEMATICAL HANDBOOK OF FORMULAS AND TABLES
!     SCHAUM'S OUTLINE SERIES
!     MURRAY SPIEGER, McGRAW-HILL, NEW YORK, 1968
!     (GREEK TRANSLATION: BY SOTIRIOS PERSIDES, ESPI, ATHENS, 1976)
!
!     A SPECIAL CASE IS CONSIDERED SEPERATELY ; WHEN A3 = 0, THEN
!     ONE ROOT IS X=0.0, AND THE OTHER TWO FROM THE SOLUTION OF THE
!     QUADRATIC EQUATION X**2 + A1*X + A2 = 0.0
!     THIS SPECIAL CASE IS CONSIDERED BECAUSE THE ANALYTICAL FORMULA
!     DOES NOT YIELD ACCURATE RESULTS (DUE TO NUMERICAL ROUNDOFF ERRORS)
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE poly3 ( a1, a2, a3, root, islv )
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(in)    :: a1
  DOUBLE PRECISION, INTENT(in)    :: a2
  DOUBLE PRECISION, INTENT(in)    :: a3
  DOUBLE PRECISION, INTENT(inout) :: root
  INTEGER, INTENT(inout) :: islv
  DOUBLE PRECISION :: coef
  DOUBLE PRECISION :: d
  INTEGER :: i
  INTEGER :: ix
  DOUBLE PRECISION :: q
  DOUBLE PRECISION :: r
  DOUBLE PRECISION :: s
  DOUBLE PRECISION :: sqd
  DOUBLE PRECISION :: ssig
  DOUBLE PRECISION :: t
  DOUBLE PRECISION :: thet
  DOUBLE PRECISION :: tsig
  DOUBLE PRECISION :: x(3)
  !
  DOUBLE PRECISION, PARAMETER :: expon = 1.D0/3.D0
  DOUBLE PRECISION, PARAMETER :: zero=0.D0
  DOUBLE PRECISION, PARAMETER :: thet1=120.D0/180.D0              
  DOUBLE PRECISION, PARAMETER :: thet2=240.D0/180.D0
  DOUBLE PRECISION, PARAMETER :: pi=3.1415926535897932
  DOUBLE PRECISION, PARAMETER :: eps=1D-50
  !
  ! *** SPECIAL CASE : QUADRATIC*X EQUATION *****************************
  !
  IF (abs(a3)<=eps) THEN
    islv = 1
    ix = 1
    x(1) = zero
    d = a1*a1 - 4.D0*a2
    IF (d>=zero) THEN
      ix = 3
      sqd = sqrt(d)
      x(2) = 0.5*(-a1+sqd)
      x(3) = 0.5*(-a1-sqd)
    END IF
  ELSE
    !
    ! *** NORMAL CASE : CUBIC EQUATION ************************************
    !
    ! DEFINE PARAMETERS Q, R, S, T, D
    !
    islv = 1
    q = (3.D0*a2-a1*a1)/9.D0
    r = (9.D0*a1*a2-27.D0*a3-2.D0*a1*a1*a1)/54.D0
    d = q*q*q + r*r
    !
    ! *** CALCULATE ROOTS *************************************************
    !
    !  D < 0, THREE REAL ROOTS
    !
    IF (d<-eps) THEN ! D < -EPS  : D < ZERO
      ix = 3
      thet = expon*acos(r/sqrt(-q*q*q))
      coef = 2.D0*sqrt(-q)
      x(1) = coef*cos(thet) - expon*a1
      x(2) = coef*cos(thet+thet1*pi) - expon*a1
      x(3) = coef*cos(thet+thet2*pi) - expon*a1
      !
      !  D = 0, THREE REAL (ONE DOUBLE) ROOTS
      !
    ELSE IF (d<=eps) THEN ! -EPS <= D <= EPS  : D = ZERO
      ix = 2
      ssig = sign(1.D0, r)
      s = ssig*(abs(r))**expon
      x(1) = 2.D0*s - expon*a1
      x(2) = -s - expon*a1
      !
      !  D > 0, ONE REAL ROOT
      !
    ELSE ! D > EPS  : D > ZERO
      ix = 1
      sqd = sqrt(d)
      ssig = sign(1.D0, r+sqd) ! TRANSFER SIGN TO SSIG
      tsig = sign(1.D0, r-sqd)
      s = ssig*(abs(r+sqd))**expon ! EXPONENTIATE ABS()
      t = tsig*(abs(r-sqd))**expon
      x(1) = s + t - expon*a1
    END IF
  END IF
  !
  ! *** SELECT APPROPRIATE ROOT *****************************************
  !
  root = 1.D30
  DO i = 1, ix
    IF (x(i)>zero) THEN
      root = min(root, x(i))
      islv = 0
    END IF
  END DO
  !
  RETURN
END SUBROUTINE poly3




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE POLY3B
! *** FINDS A REAL ROOT OF THE THIRD ORDER ALGEBRAIC EQUATION:
!     X**3 + A1*X**2 + A2*X + A3 = 0.0
!     THE EQUATION IS SOLVED NUMERICALLY (BISECTION).
!
!     PARAMETERS A1, A2, A3 ARE SPECIFIED BY THE USER. THE MINIMUM
!     NONEGATIVE ROOT IS RETURNED IN VARIABLE 'ROOT'. IF NO ROOT IS
!     FOUND (WHICH IS GREATER THAN ZERO), ROOT HAS THE VALUE 1D30.
!     AND THE FLAG ISLV HAS A VALUE GREATER THAN ZERO.
!
!     RTLW, RTHI DEFINE THE INTERVAL WHICH THE ROOT IS LOOKED FOR.
!
!=======================================================================
!
SUBROUTINE poly3b ( a1, a2, a3, rtlw, rthi, root, islv )
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: a1
  DOUBLE PRECISION, INTENT(inout) :: a2
  DOUBLE PRECISION, INTENT(inout) :: a3
  DOUBLE PRECISION, INTENT(inout) :: rtlw
  DOUBLE PRECISION, INTENT(inout) :: rthi
  DOUBLE PRECISION, INTENT(inout) :: root
  INTEGER, INTENT(inout) :: islv
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: x
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  DOUBLE PRECISION, PARAMETER :: zero=0.D0
  DOUBLE PRECISION, PARAMETER :: eps=1D-15
  INTEGER, PARAMETER :: maxit=100
  INTEGER, PARAMETER :: ndiv=5
  DOUBLE PRECISION :: func
  !
  func(x) = x**3.D0 + a1*x**2.0 + a2*x + a3
  !
  ! *** INITIAL VALUES FOR BISECTION *************************************
  !
  x1 = rtlw
  y1 = func(x1)
  IF (abs(y1)<=eps) THEN ! Is low a root?
    root = rtlw
    GO TO 120
  END IF
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO ***********************
  !
  dx = (rthi-rtlw)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = func(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION FOUND
  !
  IF (abs(y2)<eps) THEN ! X2 is a root
    root = x2
  ELSE
    root = 1.D30
    islv = 1
  END IF
  GO TO 120
  !
  ! *** BISECTION *******************************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    y3 = func(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  !
  ! *** CONVERGED ; RETURN ***********************************************
  !
110 x3 = 0.5*(x1+x2)
  y3 = func(x3)
  root = x3
  islv = 0
  !
120 RETURN
  !
END SUBROUTINE poly3b



!cc      PROGRAM DRIVER
!cc      DOUBLE PRECISION ROOT
!ccC
!cc      CALL POLY3 (-1.d0, 1.d0, -1.d0, ROOT, ISLV)
!cc      IF (ISLV.NE.0) STOP 'Error in POLY3'
!cc      WRITE (*,*) 'Root=', ROOT
!ccC
!cc      CALL POLY3B (-1.d0, 1.d0, -1.d0, -10.d0, 10.d0, ROOT, ISLV)
!cc      IF (ISLV.NE.0) STOP 'Error in POLY3B'
!cc      WRITE (*,*) 'Root=', ROOT
!ccC
!cc      END
!=======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION EX10
! *** 10^X FUNCTION ; ALTERNATE OF LIBRARY ROUTINE ; USED BECAUSE IT IS
!     MUCH FASTER BUT WITHOUT GREAT LOSS IN ACCURACY. ,
!     MAXIMUM ERROR IS 2%, EXECUTION TIME IS 42% OF THE LIBRARY ROUTINE
!     (ON A 80286/80287 MACHINE, using Lahey FORTRAN 77 v.3.0).
!
!     EXPONENT RANGE IS BETWEEN -K AND K (K IS THE REAL ARGUMENT 'K')
!     MAX VALUE FOR K: 9.999
!     IF X < -K, X IS SET TO -K, IF X > K, X IS SET TO K
!
!     THE EXPONENT IS CALCULATED BY THE PRODUCT ADEC*AINT, WHERE ADEC
!     IS THE MANTISSA AND AINT IS THE MAGNITUDE (EXPONENT). BOTH
!     MANTISSA AND MAGNITUDE ARE PRE-CALCULATED AND STORED IN LOOKUP
!     TABLES ; THIS LEADS TO THE INCREASED SPEED.
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
REAL FUNCTION ex10 ( x, k )
  IMPLICIT NONE
  REAL, INTENT(inout) :: x
  REAL, INTENT(inout) :: k
  INTEGER :: k1
  INTEGER :: k2
  REAL :: y
  !
  ! *** integer part
  !
  REAL, PARAMETER :: aint10(20) = [                                &
       0.1000E-08, 0.1000E-07, 0.1000E-06, 0.1000E-05, 0.1000E-04, &
       0.1000E-03, 0.1000E-02, 0.1000E-01, 0.1000E+00, 0.1000E+01, &
       0.1000E+02, 0.1000E+03, 0.1000E+04, 0.1000E+05, 0.1000E+06, &
       0.1000E+07, 0.1000E+08, 0.1000E+09, 0.1000E+10, 0.1000E+11 ]
  !
  ! *** decimal part
  !
  REAL, PARAMETER :: adec10(200) = [                               &
       0.1023E+00, 0.1047E+00, 0.1072E+00,                         &
       0.1096E+00, 0.1122E+00, 0.1148E+00, 0.1175E+00, 0.1202E+00, &
       0.1230E+00, 0.1259E+00, 0.1288E+00, 0.1318E+00, 0.1349E+00, &
       0.1380E+00, 0.1413E+00, 0.1445E+00, 0.1479E+00, 0.1514E+00, &
       0.1549E+00, 0.1585E+00, 0.1622E+00, 0.1660E+00, 0.1698E+00, &
       0.1738E+00, 0.1778E+00, 0.1820E+00, 0.1862E+00, 0.1905E+00, &
       0.1950E+00, 0.1995E+00, 0.2042E+00, 0.2089E+00, 0.2138E+00, &
       0.2188E+00, 0.2239E+00, 0.2291E+00, 0.2344E+00, 0.2399E+00, &
       0.2455E+00, 0.2512E+00, 0.2570E+00, 0.2630E+00, 0.2692E+00, &
       0.2754E+00, 0.2818E+00, 0.2884E+00, 0.2951E+00, 0.3020E+00, &
       0.3090E+00, 0.3162E+00, 0.3236E+00, 0.3311E+00, 0.3388E+00, &
       0.3467E+00, 0.3548E+00, 0.3631E+00, 0.3715E+00, 0.3802E+00, &
       0.3890E+00, 0.3981E+00, 0.4074E+00, 0.4169E+00, 0.4266E+00, &
       0.4365E+00, 0.4467E+00, 0.4571E+00, 0.4677E+00, 0.4786E+00, &
       0.4898E+00, 0.5012E+00, 0.5129E+00, 0.5248E+00, 0.5370E+00, &
       0.5495E+00, 0.5623E+00, 0.5754E+00, 0.5888E+00, 0.6026E+00, &
       0.6166E+00, 0.6310E+00, 0.6457E+00, 0.6607E+00, 0.6761E+00, &
       0.6918E+00, 0.7079E+00, 0.7244E+00, 0.7413E+00, 0.7586E+00, &
       0.7762E+00, 0.7943E+00, 0.8128E+00, 0.8318E+00, 0.8511E+00, &
       0.8710E+00, 0.8913E+00, 0.9120E+00, 0.9333E+00, 0.9550E+00, &
       0.9772E+00, 0.1000E+01, 0.1023E+01, 0.1047E+01, 0.1072E+01, &
       0.1096E+01, 0.1122E+01, 0.1148E+01, 0.1175E+01, 0.1202E+01, &
       0.1230E+01, 0.1259E+01, 0.1288E+01, 0.1318E+01, 0.1349E+01, &
       0.1380E+01, 0.1413E+01, 0.1445E+01, 0.1479E+01, 0.1514E+01, &
       0.1549E+01, 0.1585E+01, 0.1622E+01, 0.1660E+01, 0.1698E+01, &
       0.1738E+01, 0.1778E+01, 0.1820E+01, 0.1862E+01, 0.1905E+01, &
       0.1950E+01, 0.1995E+01, 0.2042E+01, 0.2089E+01, 0.2138E+01, &
       0.2188E+01, 0.2239E+01, 0.2291E+01, 0.2344E+01, 0.2399E+01, &
       0.2455E+01, 0.2512E+01, 0.2570E+01, 0.2630E+01, 0.2692E+01, &
       0.2754E+01, 0.2818E+01, 0.2884E+01, 0.2951E+01, 0.3020E+01, &
       0.3090E+01, 0.3162E+01, 0.3236E+01, 0.3311E+01, 0.3388E+01, &
       0.3467E+01, 0.3548E+01, 0.3631E+01, 0.3715E+01, 0.3802E+01, &
       0.3890E+01, 0.3981E+01, 0.4074E+01, 0.4169E+01, 0.4266E+01, &
       0.4365E+01, 0.4467E+01, 0.4571E+01, 0.4677E+01, 0.4786E+01, &
       0.4898E+01, 0.5012E+01, 0.5129E+01, 0.5248E+01, 0.5370E+01, &
       0.5495E+01, 0.5623E+01, 0.5754E+01, 0.5888E+01, 0.6026E+01, &
       0.6166E+01, 0.6310E+01, 0.6457E+01, 0.6607E+01, 0.6761E+01, &
       0.6918E+01, 0.7079E+01, 0.7244E+01, 0.7413E+01, 0.7586E+01, &
       0.7762E+01, 0.7943E+01, 0.8128E+01, 0.8318E+01, 0.8511E+01, &
       0.8710E+01, 0.8913E+01, 0.9120E+01, 0.9333E+01, 0.9550E+01, &
       0.9772E+01, 0.1000E+02 ]
  !
  ! *** LIMIT X TO [-K, K] RANGE *****************************************
  !
  y = max(-k, min(x,k)) ! MIN: -9.999, MAX: 9.999
  !
  ! *** GET INTEGER AND DECIMAL PART *************************************
  !
  k1 = int(y)
  k2 = int(100*(y-k1))
  !
  ! *** CALCULATE EXP FUNCTION *******************************************
  !
  ex10 = aint10(k1+10)*adec10(k2+100)
  !
  ! *** END OF EXP FUNCTION **********************************************
  !
  RETURN
END FUNCTION ex10
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE PUSHERR
! *** THIS SUBROUTINE SAVES AN ERROR MESSAGE IN THE ERROR STACK
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE pusherr(ierr, errinf)
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER, INTENT(in) :: ierr
  CHARACTER(len=*), INTENT(in) :: errinf 
  !
  ! *** SAVE ERROR CODE IF THERE IS ANY SPACE ***************************
  !
  IF (nofer<nerrmx) THEN
    nofer = nofer + 1
    errstk(nofer) = ierr
    errmsg(nofer) = errinf
    stkofl = .FALSE.
  ELSE
    stkofl = .TRUE. ! STACK OVERFLOW
  END IF
  !
END SUBROUTINE pusherr

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ISERRINF
! *** THIS SUBROUTINE OBTAINS A COPY OF THE ERROR STACK (& MESSAGES)
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE iserrinf(errstki, errmsgi, noferi, stkofli)
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER, INTENT(inout) :: errstki(25)
  CHARACTER(len=40), INTENT(inout) :: errmsgi(25) 
  INTEGER, INTENT(inout) :: noferi
  LOGICAL, INTENT(inout) :: stkofli
  INTEGER :: i

  !
  ! *** OBTAIN WHOLE ERROR STACK ****************************************
  !
  DO i = 1, nofer ! Error messages & codes
    errstki(i) = errstk(i)
    errmsgi(i) = errmsg(i)
  END DO
  !
  stkofli = stkofl
  noferi = nofer
  !
  RETURN
  !
END SUBROUTINE iserrinf
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ERRSTAT
! *** THIS SUBROUTINE REPORTS ERROR MESSAGES TO UNIT 'IO'
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE errstat(io, ierr, errinf)
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER, INTENT(inout) :: io
  INTEGER, INTENT(inout) :: ierr
  CHARACTER(len=*), INTENT(inout) :: errinf
  character(len=4) :: cer
  INTEGER :: iend
  INTEGER :: iok
  character(len=27) :: ncif = "NO CONVERGENCE IN FUNCTION "
  character(len=29) :: ncis = "NO CONVERGENCE IN SUBROUTINE "
  character(len=24) :: nsif = "NO SOLUTION IN FUNCTION "
  character(len=26) :: nsis = "NO SOLUTION IN SUBROUTINE "
  !
  ! *** WRITE ERROR IN CHARACTER *****************************************
  !
  WRITE (cer, '(I4)') ierr
  CALL rplstr(cer, ' ', '0', iok) ! REPLACE BLANKS WITH ZEROS
  CALL chrbln(errinf, iend) ! LAST POSITION OF ERRINF CHAR
  !
  ! *** WRITE ERROR TYPE (FATAL, WARNING ) *******************************
  !
  IF (ierr==0) THEN
    WRITE (io, 110) 'NO ERRORS DETECTED '
    GO TO 100
    !
  ELSE IF (ierr<0) THEN
    WRITE (io, 110) 'ERROR STACK EXHAUSTED '
    GO TO 100
    !
  ELSE IF (ierr>1000) THEN
    WRITE (io, 120) 'FATAL', cer
    !
  ELSE
    WRITE (io, 120) 'WARNING', cer
  END IF
  !
  ! *** WRITE ERROR MESSAGE **********************************************
  !
  ! FATAL MESSAGES
  !
  IF (ierr==1001) THEN
    CALL chrbln(scase, iend)
    WRITE (io, 110) 'CASE NOT SUPPORTED IN CALCMR [' // scase(1:iend) //   &
         ']'
    !
  ELSE IF (ierr==1002) THEN
    CALL chrbln(scase, iend)
    WRITE (io, 110) 'CASE NOT SUPPORTED [' // scase(1:iend) // ']'
    !
    ! WARNING MESSAGES
    !
  ELSE IF (ierr==0001) THEN
    WRITE (io, 110) nsis, errinf
    !
  ELSE IF (ierr==0002) THEN
    WRITE (io, 110) ncis, errinf
    !
  ELSE IF (ierr==0003) THEN
    WRITE (io, 110) nsif, errinf
    !
  ELSE IF (ierr==0004) THEN
    WRITE (io, 110) ncif, errinf
    !
  ELSE IF (ierr==0019) THEN
    WRITE (io, 110) 'HNO3(aq) AFFECTS H+, WHICH ' //                       &
         'MIGHT AFFECT SO4/HSO4 RATIO'
    WRITE (io, 110) 'DIRECT INCREASE IN H+ [', errinf(1:iend), '] %'
    !
  ELSE IF (ierr==0020) THEN
    IF (w(4)>tiny .AND. w(5)>tiny) THEN
      WRITE (io, 110) 'HSO4-SO4 EQUILIBRIUM MIGHT AFFECT HNO3,' //         &
           'HCL DISSOLUTION'
    ELSE
      WRITE (io, 110) 'HSO4-SO4 EQUILIBRIUM MIGHT AFFECT NH3 ' //          &
           'DISSOLUTION'
    END IF
    WRITE (io, 110) 'DIRECT DECREASE IN H+ [', errinf(1:iend), '] %'
    !
  ELSE IF (ierr==0021) THEN
    WRITE (io, 110) 'HNO3(aq),HCL(aq) AFFECT H+, WHICH ' //                &
         'MIGHT AFFECT SO4/HSO4 RATIO'
    WRITE (io, 110) 'DIRECT INCREASE IN H+ [', errinf(1:iend), '] %'
    !
  ELSE IF (ierr==0022) THEN
    WRITE (io, 110) 'HCL(g) EQUILIBRIUM YIELDS NONPHYSICAL ' //            &
         'DISSOLUTION'
    WRITE (io, 110) 'A TINY AMOUNT [', errinf(1:iend),                     &
         '] IS ' // 'ASSUMED TO BE DISSOLVED'
    !
  ELSE IF (ierr==0033) THEN
    WRITE (io, 110) 'HCL(aq) AFFECTS H+, WHICH ' //                        &
         'MIGHT AFFECT SO4/HSO4 RATIO'
    WRITE (io, 110) 'DIRECT INCREASE IN H+ [', errinf(1:iend), '] %'
    !
  ELSE IF (ierr==0050) THEN
    WRITE (io, 110) 'TOO MUCH SODIUM GIVEN AS INPUT.'
    WRITE (io, 110) 'REDUCED TO COMPLETELY NEUTRALIZE SO4,Cl,NO3.'
    WRITE (io, 110) 'EXCESS SODIUM IS IGNORED.'
    !
  ELSE IF (ierr==0051) THEN
    WRITE (io, 110) 'TOO MUCH CALCIUM GIVEN AS INPUT.'
    WRITE (io, 110) 'REDUCED TO COMPLETELY NEUTRALIZE SO4,Cl,NO3.'
    WRITE (io, 110) 'EXCESS CALCIUM IS IGNORED.'
    !
  ELSE IF (ierr==0052) THEN
    WRITE (io, 110) 'TOO MUCH SODIUM (+Ca) GIVEN AS INPUT.'
    WRITE (io, 110) 'REDUCED TO COMPLETELY NEUTRALIZE SO4,Cl,NO3.'
    WRITE (io, 110) 'EXCESS SODIUM IS IGNORED.'
    !
  ELSE IF (ierr==0053) THEN
    WRITE (io, 110) 'TOO MUCH MAGNESIUM (+Ca,Na) GIVEN AS INPUT.'
    WRITE (io, 110) 'REDUCED TO COMPLETELY NEUTRALIZE SO4,Cl,NO3.'
    WRITE (io, 110) 'EXCESS MAGNESIUM IS IGNORED.'
    !
  ELSE IF (ierr==0054) THEN
    WRITE (io, 110) 'TOO MUCH POTASSIUM(+Ca,Na,Mg) GIVEN AS INPUT.'
    WRITE (io, 110) 'REDUCED TO COMPLETELY NEUTRALIZE SO4,Cl,NO3.'
    WRITE (io, 110) 'EXCESS POTASSIUM IS IGNORED.'
    !
  ELSE
    WRITE (io, 110) 'NO DIAGNOSTIC MESSAGE AVAILABLE'
  END IF
  !
100 RETURN
  !
  ! *** FORMAT STATEMENTS *************************************
  !
110 FORMAT (1X, A, :, A, :, A, :, A, :, A)
120 FORMAT (1X, A, ' ERROR [', A4, ']:')
  !
END SUBROUTINE errstat
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ISORINF
! *** THIS SUBROUTINE PROVIDES INFORMATION ABOUT ISORROPIA
!
! ======================== ARGUMENTS / USAGE ===========================
!
!  OUTPUT:
!  1. [VERSI]
!     CHARACTER*15 variable.
!     Contains version-date information of ISORROPIA
!
!  2. [NCMP]
!     INTEGER variable.
!     The number of components needed in input array WI
!     (or, the number of major species accounted for by ISORROPIA)
!
!  3. [NION]
!     INTEGER variable
!     The number of ions considered in the aqueous phase
!
!  4. [NAQGAS]
!     INTEGER variable
!     The number of undissociated species found in aqueous aerosol
!     phase
!
!  5. [NSOL]
!     INTEGER variable
!     The number of solids considered in the solid aerosol phase
!
!  6. [NERR]
!     INTEGER variable
!     The size of the error stack (maximum number of errors that can
!     be stored before the stack exhausts).
!
!  7. [TIN]
!     DOUBLE PRECISION variable
!     The value used for a very small number.
!
!  8. [GRT]
!     DOUBLE PRECISION variable
!     The value used for a very large number.
!
! *** COPYRIGHT 1996-2012, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
! *** UPDATE|ADJOINT BY SHANNON CAPPS
!
!=======================================================================
!
SUBROUTINE isorinf(versi, ncmp, nion, naqgas, nsol, nerr, tin, grt)
  USE mo_isrpia
  IMPLICIT NONE
  character(len=*), INTENT(out) :: versi
  INTEGER, INTENT(out) :: ncmp
  INTEGER, INTENT(out) :: nion
  INTEGER, INTENT(out) :: naqgas
  INTEGER, INTENT(out) :: nsol
  INTEGER, INTENT(out) :: nerr
  DOUBLE PRECISION, INTENT(out) :: tin
  DOUBLE PRECISION, INTENT(out) :: grt  
  !
  ! *** ASSIGN INFO *******************************************************
  !
  versi = version
  ncmp = ncomp
  nion = nions
  naqgas = ngasaq
  nsol = nslds
  nerr = nerrmx
  tin = tiny
  grt = great
  !
  RETURN
  !
END SUBROUTINE isorinf
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ISRP1F
! *** THIS SUBROUTINE IS THE DRIVER ROUTINE FOR THE FOREWARD PROBLEM OF
!     AN AMMONIUM-SULFATE AEROSOL SYSTEM.
!     THE COMPOSITION REGIME IS DETERMINED BY THE SULFATE RATIO AND BY
!     THE AMBIENT RELATIVE HUMIDITY.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!

SUBROUTINE isrp1f ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  DOUBLE PRECISION :: dc

  !
  ! *** INITIALIZE ALL VARIABLES IN COMMON BLOCK **************************
  !
  CALL init1(wi, rhi, tempi)
  !
  ! *** CALCULATE SULFATE RATIO *******************************************
  !
  sulrat = w(3)/w(2)
  !
  ! *** FIND CALCULATION REGIME FROM (SULRAT,RH) **************************
  !
  ! *** SULFATE POOR
  !
  IF (2.0<=sulrat) THEN
    dc = w(3) - 2.001D0*w(2) ! For numerical stability
    w(3) = w(3) + max(-dc, zero)
    !
    IF (metstbl==1) THEN
      scase = 'A2'
      CALL calca2 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh42s4) THEN
        scase = 'A1'
        CALL calca1 ! NH42SO4              ; case A1
        !
      ELSE IF (drnh42s4<=rh) THEN
        scase = 'A2'
        CALL calca2 ! Only liquid          ; case A2
      END IF
    END IF
    !
    ! *** SULFATE RICH (NO ACID)
    !
  ELSE IF (1.0<=sulrat .AND. sulrat<2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'B4'
      CALL calcb4 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'B1'
        CALL calcb1 ! NH4HSO4,LC,NH42SO4   ; case B1
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drlc) THEN
        scase = 'B2'
        CALL calcb2 ! LC,NH42S4            ; case B2
        !
      ELSE IF (drlc<=rh .AND. rh<drnh42s4) THEN
        scase = 'B3'
        CALL calcb3 ! NH42S4               ; case B3
        !
      ELSE IF (drnh42s4<=rh) THEN
        scase = 'B4'
        CALL calcb4 ! Only liquid          ; case B4
      END IF
    END IF
    CALL calcnh3
    !
    ! *** SULFATE RICH (FREE ACID)
    !
  ELSE IF (sulrat<1.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'C2'
      CALL calcc2 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'C1'
        CALL calcc1 ! NH4HSO4              ; case C1
        !
      ELSE IF (drnh4hs4<=rh) THEN
        scase = 'C2'
        CALL calcc2 ! Only liquid          ; case C2
        !
      END IF
    END IF
    CALL calcnh3
  END IF
  !
  ! *** RETURN POINT
  !
  RETURN
  !
END SUBROUTINE isrp1f
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ISRP2F
! *** THIS SUBROUTINE IS THE DRIVER ROUTINE FOR THE FOREWARD PROBLEM OF
!     AN AMMONIUM-SULFATE-NITRATE AEROSOL SYSTEM.
!     THE COMPOSITION REGIME IS DETERMINED BY THE SULFATE RATIO AND BY
!     THE AMBIENT RELATIVE HUMIDITY.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!

SUBROUTINE isrp2f ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  !
  ! *** INITIALIZE ALL VARIABLES IN COMMON BLOCK **************************
  !
  CALL init2(wi, rhi, tempi)
  !
  ! *** CALCULATE SULFATE RATIO *******************************************
  !
  sulrat = w(3)/w(2)
  !
  ! *** FIND CALCULATION REGIME FROM (SULRAT,RH) **************************
  !
  ! *** SULFATE POOR
  !
  IF (2.0<=sulrat) THEN
    !
    IF (metstbl==1) THEN
      scase = 'D3'
      CALL calcd3 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4no3) THEN
        scase = 'D1'
        CALL calcd1 ! NH42SO4,NH4NO3       ; case D1
        !
      ELSE IF (drnh4no3<=rh .AND. rh<drnh42s4) THEN
        scase = 'D2'
        CALL calcd2 ! NH42S4               ; case D2
        !
      ELSE IF (drnh42s4<=rh) THEN
        scase = 'D3'
        CALL calcd3 ! Only liquid          ; case D3
      END IF
    END IF
    !
    ! *** SULFATE RICH (NO ACID)
    !     FOR SOLVING THIS CASE, NITRIC ACID IS ASSUMED A MINOR SPECIES,
    !     THAT DOES NOT SIGNIFICANTLY PERTURB THE HSO4-SO4 EQUILIBRIUM.
    !     SUBROUTINES CALCB? ARE CALLED, AND THEN THE NITRIC ACID IS DISSOLVED
    !     FROM THE HNO3(G) -> (H+) + (NO3-) EQUILIBRIUM.
    !
  ELSE IF (1.0<=sulrat .AND. sulrat<2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'B4'
      CALL calcb4 ! Only liquid (metastable)
      scase = 'E4'
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'B1'
        CALL calcb1 ! NH4HSO4,LC,NH42SO4   ; case E1
        scase = 'E1'
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drlc) THEN
        scase = 'B2'
        CALL calcb2 ! LC,NH42S4            ; case E2
        scase = 'E2'
        !
      ELSE IF (drlc<=rh .AND. rh<drnh42s4) THEN
        scase = 'B3'
        CALL calcb3 ! NH42S4               ; case E3
        scase = 'E3'
        !
      ELSE IF (drnh42s4<=rh) THEN
        scase = 'B4'
        CALL calcb4 ! Only liquid          ; case E4
        scase = 'E4'
      END IF
    END IF
    !
    CALL calcna ! HNO3(g) DISSOLUTION
    !
    ! *** SULFATE RICH (FREE ACID)
    !     FOR SOLVING THIS CASE, NITRIC ACID IS ASSUMED A MINOR SPECIES,
    !     THAT DOES NOT SIGNIFICANTLY PERTURB THE HSO4-SO4 EQUILIBRIUM
    !     SUBROUTINE CALCC? IS CALLED, AND THEN THE NITRIC ACID IS DISSOLVED
    !     FROM THE HNO3(G) -> (H+) + (NO3-) EQUILIBRIUM.
    !
  ELSE IF (sulrat<1.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'C2'
      CALL calcc2 ! Only liquid (metastable)
      scase = 'F2'
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'C1'
        CALL calcc1 ! NH4HSO4              ; case F1
        scase = 'F1'
        !
      ELSE IF (drnh4hs4<=rh) THEN
        scase = 'C2'
        CALL calcc2 ! Only liquid          ; case F2
        scase = 'F2'
      END IF
    END IF
    !
    CALL calcna ! HNO3(g) DISSOLUTION
  END IF
  !
  ! *** RETURN POINT
  !
  RETURN
  !
END SUBROUTINE isrp2f
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ISRP3F
! *** THIS SUBROUTINE IS THE DRIVER ROUTINE FOR THE FORWARD PROBLEM OF
!     AN AMMONIUM-SULFATE-NITRATE-CHLORIDE-SODIUM AEROSOL SYSTEM.
!     THE COMPOSITION REGIME IS DETERMINED BY THE SULFATE & SODIUM
!     RATIOS AND BY THE AMBIENT RELATIVE HUMIDITY.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE isrp3f ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  DOUBLE PRECISION :: rest
  !
  ! *** ADJUST FOR TOO LITTLE AMMONIUM AND CHLORIDE ***********************
  !
  wi(3) = max(wi(3), 1.D-10) ! NH4+ : 1e-4 umoles/m3
  wi(5) = max(wi(5), 1.D-10) ! Cl-  : 1e-4 umoles/m3
  !
  ! *** ADJUST FOR TOO LITTLE SODIUM, SULFATE AND NITRATE COMBINED ********
  !
  IF (wi(1)+wi(2)+wi(4)<=1D-10) THEN
    wi(1) = 1.D-10 ! Na+  : 1e-4 umoles/m3
    wi(2) = 1.D-10 ! SO4- : 1e-4 umoles/m3
  END IF
  !
  ! *** INITIALIZE ALL VARIABLES IN COMMON BLOCK **************************
  !
  CALL isoinit3(wi, rhi, tempi)
  !
  ! *** CHECK IF TOO MUCH SODIUM ; ADJUST AND ISSUE ERROR MESSAGE *********
  !
  rest = 2.D0*w(2) + w(4) + w(5)
  IF (w(1)>rest) THEN ! NA > 2*SO4+CL+NO3 ?
    w(1) = (one-1D-6)*rest ! Adjust Na amount
    CALL pusherr(0050, 'ISRP3F') ! Warning error: Na adjusted
  END IF
  !
  ! *** CALCULATE SULFATE & SODIUM RATIOS *********************************
  !
  sulrat = (w(1)+w(3))/w(2)
  sodrat = w(1)/w(2)
  !
  ! *** FIND CALCULATION REGIME FROM (SULRAT,RH) **************************

  ! *** SULFATE POOR ; SODIUM POOR
  !
  IF (2.0<=sulrat .AND. sodrat<2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'G5'
      CALL calcg5 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4no3) THEN
        scase = 'G1'
        CALL calcg1 ! NH42SO4,NH4NO3,NH4CL,NA2SO4
        !
      ELSE IF (drnh4no3<=rh .AND. rh<drnh4cl) THEN
        scase = 'G2'
        CALL calcg2 ! NH42SO4,NH4CL,NA2SO4
        !
      ELSE IF (drnh4cl<=rh .AND. rh<drnh42s4) THEN
        scase = 'G3'
        CALL calcg3 ! NH42SO4,NA2SO4
        !
      ELSE IF (drnh42s4<=rh .AND. rh<drna2so4) THEN
        scase = 'G4'
        CALL calcg4 ! NA2SO4
        !
      ELSE IF (drna2so4<=rh) THEN
        scase = 'G5'
        CALL calcg5 ! Only liquid
      END IF
    END IF
    !
    ! *** SULFATE POOR ; SODIUM RICH
    !
  ELSE IF (sulrat>=2.0 .AND. sodrat>=2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'H6'
      CALL calch6 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4no3) THEN
        scase = 'H1'
        CALL calch1 ! NH4NO3,NH4CL,NA2SO4,NACL,NANO3
        !
      ELSE IF (drnh4no3<=rh .AND. rh<drnano3) THEN
        scase = 'H2'
        CALL calch2 ! NH4CL,NA2SO4,NACL,NANO3
        !
      ELSE IF (drnano3<=rh .AND. rh<drnacl) THEN
        scase = 'H3'
        CALL calch3 ! NH4CL,NA2SO4,NACL
        !
      ELSE IF (drnacl<=rh .AND. rh<drnh4cl) THEN
        scase = 'H4'
        CALL calch4 ! NH4CL,NA2SO4
        !
      ELSE IF (drnh4cl<=rh .AND. rh<drna2so4) THEN
        scase = 'H5'
        CALL calch5 ! NA2SO4
        !
      ELSE IF (drna2so4<=rh) THEN
        scase = 'H6'
        CALL calch6 ! NO SOLID
      END IF
    END IF
    !
    ! *** SULFATE RICH (NO ACID)
    !
  ELSE IF (1.0<=sulrat .AND. sulrat<2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'I6'
      CALL calci6 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'I1'
        CALL calci1 ! NA2SO4,(NH4)2SO4,NAHSO4,NH4HSO4,LC
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drnahso4) THEN
        scase = 'I2'
        CALL calci2 ! NA2SO4,(NH4)2SO4,NAHSO4,LC
        !
      ELSE IF (drnahso4<=rh .AND. rh<drlc) THEN
        scase = 'I3'
        CALL calci3 ! NA2SO4,(NH4)2SO4,LC
        !
      ELSE IF (drlc<=rh .AND. rh<drnh42s4) THEN
        scase = 'I4'
        CALL calci4 ! NA2SO4,(NH4)2SO4
        !
      ELSE IF (drnh42s4<=rh .AND. rh<drna2so4) THEN
        scase = 'I5'
        CALL calci5 ! NA2SO4
        !
      ELSE IF (drna2so4<=rh) THEN
        scase = 'I6'
        CALL calci6 ! NO SOLIDS
      END IF
    END IF
    !
    CALL calcnha ! MINOR SPECIES: HNO3, HCl
    CALL calcnh3 !                NH3
    !
    ! *** SULFATE RICH (FREE ACID)
    !
  ELSE IF (sulrat<1.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'J3'
      CALL calcj3 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'J1'
        CALL calcj1 ! NH4HSO4,NAHSO4
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drnahso4) THEN
        scase = 'J2'
        CALL calcj2 ! NAHSO4
        !
      ELSE IF (drnahso4<=rh) THEN
        scase = 'J3'
        CALL calcj3
      END IF
    END IF
    !
    CALL calcnha ! MINOR SPECIES: HNO3, HCl
    CALL calcnh3 !                NH3
  END IF
  !
  ! *** RETURN POINT
  !
  RETURN
  !
END SUBROUTINE isrp3f
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE ISRP4F
! *** THIS SUBROUTINE IS THE DRIVER ROUTINE FOR THE FORWARD PROBLEM OF
!     AN AMMONIUM-SULFATE-NITRATE-CHLORIDE-SODIUM-CALCIUM-POTASSIUM-MAGNESIUM
!     AEROSOL SYSTEM.
!     THE COMPOSITION REGIME IS DETERMINED BY THE SULFATE & SODIUM
!     RATIOS AND BY THE AMBIENT RELATIVE HUMIDITY.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE isrp4f ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  DOUBLE PRECISION :: cafri
  DOUBLE PRECISION :: ccacl2i
  DOUBLE PRECISION :: ccano32i
  DOUBLE PRECISION :: ccaso4i
  DOUBLE PRECISION :: clfri
  DOUBLE PRECISION :: cmgcl2i
  DOUBLE PRECISION :: cmgno32i
  DOUBLE PRECISION :: cmgso4i
  DOUBLE PRECISION :: cna2so4i
  DOUBLE PRECISION :: cnacli
  DOUBLE PRECISION :: cnano3i
  DOUBLE PRECISION :: frmgi
  DOUBLE PRECISION :: frso4i
  DOUBLE PRECISION :: nafri
  INTEGER :: no3fr
  DOUBLE PRECISION :: no3fri
  DOUBLE PRECISION :: rest
  DOUBLE PRECISION :: rest1
  DOUBLE PRECISION :: rest2
  DOUBLE PRECISION :: rest3

  !
  ! *** ADJUST FOR TOO LITTLE AMMONIUM AND CHLORIDE ***********************
  !
  !      WI(3) = MAX (WI(3), 1.D-10)  ! NH4+ : 1e-4 umoles/m3
  !      WI(5) = MAX (WI(5), 1.D-10)  ! Cl-  : 1e-4 umoles/m3
  !
  ! *** ADJUST FOR TOO LITTLE SODIUM, SULFATE AND NITRATE COMBINED ********
  !
  !      IF (WI(1)+WI(2)+WI(4) .LE. 1d-10) THEN
  !         WI(1) = 1.D-10  ! Na+  : 1e-4 umoles/m3
  !         WI(2) = 1.D-10  ! SO4- : 1e-4 umoles/m3
  !      ENDIF
  !
  ! *** INITIALIZE ALL VARIABLES IN COMMON BLOCK **************************
  !
  CALL init4(wi, rhi, tempi)
  !
  ! *** CHECK IF TOO MUCH SODIUM+CRUSTALS ; ADJUST AND ISSUE ERROR MESSAGE
  !
  rest = 2.D0*w(2) + w(4) + w(5)
  !
  IF (w(1)+w(6)+w(7)+w(8)>rest) THEN
    !
    ccaso4i = min(w(2), w(6))
    frso4i = max(w(2)-ccaso4i, zero)
    cafri = max(w(6)-ccaso4i, zero)
    ccano32i = min(cafri, 0.5D0*w(4))
    cafri = max(cafri-ccano32i, zero)
    no3fri = max(w(4)-2.D0*ccano32i, zero)
    ccacl2i = min(cafri, 0.5D0*w(5))
    clfri = max(w(5)-2.D0*ccacl2i, zero)
    rest1 = 2.D0*frso4i + no3fri + clfri
    !
    cna2so4i = min(frso4i, 0.5D0*w(1))
    frso4i = max(frso4i-cna2so4i, zero)
    nafri = max(w(1)-2.D0*cna2so4i, zero)
    cnacli = min(nafri, clfri)
    nafri = max(nafri-cnacli, zero)
    clfri = max(clfri-cnacli, zero)
    cnano3i = min(nafri, no3fri)
    no3fr = max(no3fri-cnano3i, zero)
    rest2 = 2.D0*frso4i + no3fri + clfri
    !
    cmgso4i = min(frso4i, w(8))
    frmgi = max(w(8)-cmgso4i, zero)
    frso4i = max(frso4i-cmgso4i, zero)
    cmgno32i = min(frmgi, 0.5D0*no3fri)
    frmgi = max(frmgi-cmgno32i, zero)
    no3fri = max(no3fri-2.D0*cmgno32i, zero)
    cmgcl2i = min(frmgi, 0.5D0*clfri)
    clfri = max(clfri-2.D0*cmgcl2i, zero)
    rest3 = 2.D0*frso4i + no3fri + clfri
    !
    IF (w(6)>rest) THEN ! Ca > 2*SO4+CL+NO3 ?
      w(6) = (one-1D-6)*rest ! Adjust Ca amount
      w(1) = zero ! Adjust Na amount
      w(7) = zero ! Adjust K amount
      w(8) = zero ! Adjust Mg amount
      CALL pusherr(0051, 'ISRP4F') ! Warning error: Ca, Na, K, Mg in excess
      !
    ELSE IF (w(1)>rest1) THEN ! Na > 2*FRSO4+FRCL+FRNO3 ?
      w(1) = (one-1D-6)*rest1 ! Adjust Na amount
      w(7) = zero ! Adjust K amount
      w(8) = zero ! Adjust Mg amount
      CALL pusherr(0052, 'ISRP4F') ! Warning error: Na, K, Mg in excess
      !
    ELSE IF (w(8)>rest2) THEN ! Mg > 2*FRSO4+FRCL+FRNO3 ?
      w(8) = (one-1D-6)*rest2 ! Adjust Mg amount
      w(7) = zero ! Adjust K amount
      CALL pusherr(0053, 'ISRP4F') ! Warning error: K, Mg in excess
      !
    ELSE IF (w(7)>rest3) THEN ! K > 2*FRSO4+FRCL+FRNO3 ?
      w(7) = (one-1D-6)*rest3 ! Adjust K amount
      CALL pusherr(0054, 'ISRP4F') ! Warning error: K in excess
    END IF
  END IF
  !
  ! *** CALCULATE RATIOS *************************************************
  !
  so4rat = (w(1)+w(3)+w(6)+w(7)+w(8))/w(2)
  crnarat = (w(1)+w(6)+w(7)+w(8))/w(2)
  crrat = (w(6)+w(7)+w(8))/w(2)
  !
  ! *** FIND CALCULATION REGIME FROM (SO4RAT, CRNARAT, CRRAT, RRH) ********
  !
  ! *** SULFATE POOR: Rso4>2; (DUST + SODIUM) POOR: R(Cr+Na)<2
  !
  IF (2.0<=so4rat .AND. crnarat<2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'O7'
      CALL calco7 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4no3) THEN
        scase = 'O1'
        CALL calco1 ! CaSO4, NH4NO3, NH4CL, (NH4)2SO4, MGSO4, NA2SO4, K2SO4
        !
      ELSE IF (drnh4no3<=rh .AND. rh<drnh4cl) THEN
        scase = 'O2'
        CALL calco2 ! CaSO4, NH4CL, (NH4)2SO4, MGSO4, NA2SO4, K2SO4
        !
      ELSE IF (drnh4cl<=rh .AND. rh<drnh42s4) THEN
        scase = 'O3'
        CALL calco3 ! CaSO4, (NH4)2SO4, MGSO4, NA2SO4, K2SO4
        !
      ELSE IF (drnh42s4<=rh .AND. rh<drmgso4) THEN
        scase = 'O4'
        CALL calco4 ! CaSO4, MGSO4, NA2SO4, K2SO4
        !
      ELSE IF (drmgso4<=rh .AND. rh<drna2so4) THEN
        scase = 'O5'
        CALL calco5 ! CaSO4, NA2SO4, K2SO4
        !
      ELSE IF (drna2so4<=rh .AND. rh<drk2so4) THEN
        scase = 'O6'
        CALL calco6 ! CaSO4, K2SO4
        !
      ELSE IF (drk2so4<=rh) THEN
        scase = 'O7'
        CALL calco7 ! CaSO4
      END IF
    END IF
    !
    ! *** SULFATE POOR: Rso4>2; (DUST + SODIUM) RICH: R(Cr+Na)>2; DUST POOR: Rcr<2.
    !
  ELSE IF (so4rat>=2.0 .AND. crnarat>=2.0) THEN
    !
    IF (crrat<=2.0) THEN
      !
      IF (metstbl==1) THEN
        scase = 'M8'
        CALL calcm8 ! Only liquid (metastable)
      ELSE
        !
        IF (rh<drnh4no3) THEN
          scase = 'M1'
          CALL calcm1 ! CaSO4, NH4NO3, NH4CL, MGSO4, NA2SO4, K2SO4, NACL, NANO3
          !
        ELSE IF (drnh4no3<=rh .AND. rh<drnano3) THEN
          scase = 'M2'
          CALL calcm2 ! CaSO4, NH4CL, MGSO4, NA2SO4, K2SO4, NACL, NANO3
          !
        ELSE IF (drnano3<=rh .AND. rh<drnacl) THEN
          scase = 'M3'
          CALL calcm3 ! CaSO4, NH4CL, MGSO4, NA2SO4, K2SO4, NACL
          !
        ELSE IF (drnacl<=rh .AND. rh<drnh4cl) THEN
          scase = 'M4'
          CALL calcm4 ! CaSO4, NH4CL, MGSO4, NA2SO4, K2SO4
          !
        ELSE IF (drnh4cl<=rh .AND. rh<drmgso4) THEN
          scase = 'M5'
          CALL calcm5 ! CaSO4, MGSO4, NA2SO4, K2SO4
          !
        ELSE IF (drmgso4<=rh .AND. rh<drna2so4) THEN
          scase = 'M6'
          CALL calcm6 ! CaSO4, NA2SO4, K2SO4
          !
        ELSE IF (drna2so4<=rh .AND. rh<drk2so4) THEN
          scase = 'M7'
          CALL calcm7 ! CaSO4, K2SO4
          !
        ELSE IF (drk2so4<=rh) THEN
          scase = 'M8'
          CALL calcm8 ! CaSO4
        END IF
      END IF
      !        CALL CALCHCO3
      !
      ! *** SULFATE POOR: Rso4>2; (DUST + SODIUM) RICH: R(Cr+Na)>2; DUST POOR: Rcr<2.
      !
    ELSE IF (crrat>2.0) THEN
      !
      IF (metstbl==1) THEN
        scase = 'P13'
        CALL calcp13 ! Only liquid (metastable)
      ELSE
        !
        IF (rh<drcacl2) THEN
          scase = 'P1'
          CALL calcp1 ! CaSO4, CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4,
          !                                    ! MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drcacl2<=rh .AND. rh<drmgcl2) THEN
          scase = 'P2'
          CALL calcp2 ! CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
          !                                   ! MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drmgcl2<=rh .AND. rh<drcano32) THEN
          scase = 'P3'
          CALL calcp3 ! CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
          !                                   ! MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drcano32<=rh .AND. rh<drmgno32) THEN
          scase = 'P4'
          CALL calcp4 ! CaSO4, K2SO4, KNO3, KCL, MGSO4,
          !                                   ! MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drmgno32<=rh .AND. rh<drnh4no3) THEN
          scase = 'P5'
          CALL calcp5 ! CaSO4, K2SO4, KNO3, KCL, MGSO4,
          !                                   ! NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drnh4no3<=rh .AND. rh<drnano3) THEN
          scase = 'P6'
          CALL calcp6 ! CaSO4, K2SO4, KNO3, KCL, MGSO4, NANO3, NACL, NH4CL
          !
        ELSE IF (drnano3<=rh .AND. rh<drnacl) THEN
          scase = 'P7'
          CALL calcp7 ! CaSO4, K2SO4, KNO3, KCL, MGSO4, NACL, NH4CL
          !
        ELSE IF (drnacl<=rh .AND. rh<drnh4cl) THEN
          scase = 'P8'
          CALL calcp8 ! CaSO4, K2SO4, KNO3, KCL, MGSO4, NH4CL
          !
        ELSE IF (drnh4cl<=rh .AND. rh<drkcl) THEN
          scase = 'P9'
          CALL calcp9 ! CaSO4, K2SO4, KNO3, KCL, MGSO4
          !
        ELSE IF (drkcl<=rh .AND. rh<drmgso4) THEN
          scase = 'P10'
          CALL calcp10 ! CaSO4, K2SO4, KNO3, MGSO4
          !
        ELSE IF (drmgso4<=rh .AND. rh<drkno3) THEN
          scase = 'P11'
          CALL calcp11 ! CaSO4, K2SO4, KNO3
          !
        ELSE IF (drkno3<=rh .AND. rh<drk2so4) THEN
          scase = 'P12'
          CALL calcp12 ! CaSO4, K2SO4
          !
        ELSE IF (drk2so4<=rh) THEN
          scase = 'P13'
          CALL calcp13 ! CaSO4
        END IF
      END IF
      !        CALL CALCHCO3
    END IF
    !
    ! *** SULFATE RICH (NO ACID): 1<Rso4<2;
    !
  ELSE IF (1.0<=so4rat .AND. so4rat<2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'L9'
      CALL calcl9 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'L1'
        CALL calcl1 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4,(NH4)2SO4,NAHSO4,NH4HSO4,LC
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drnahso4) THEN
        scase = 'L2'
        CALL calcl2 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4,(NH4)2SO4,NAHSO4,LC
        !
      ELSE IF (drnahso4<=rh .AND. rh<drlc) THEN
        scase = 'L3'
        CALL calcl3 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4,(NH4)2SO4,LC
        !
      ELSE IF (drlc<=rh .AND. rh<drnh42s4) THEN
        scase = 'L4'
        CALL calcl4 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4,(NH4)2SO4
        !
      ELSE IF (drnh42s4<=rh .AND. rh<drkhso4) THEN
        scase = 'L5'
        CALL calcl5 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4
        !
      ELSE IF (drkhso4<=rh .AND. rh<drmgso4) THEN
        scase = 'L6'
        CALL calcl6 ! CASO4,K2SO4,MGSO4,NA2SO4
        !
      ELSE IF (drmgso4<=rh .AND. rh<drna2so4) THEN
        scase = 'L7'
        CALL calcl7 ! CASO4,K2SO4,NA2SO4
        !
      ELSE IF (drna2so4<=rh .AND. rh<drk2so4) THEN
        scase = 'L8'
        CALL calcl8 ! CASO4,K2SO4
        !
      ELSE IF (drk2so4<=rh) THEN
        scase = 'L9'
        CALL calcl9 ! CaSO4
      END IF
    END IF
    !
    CALL calcnha ! MINOR SPECIES: HNO3, HCl
    CALL calcnh3 !                NH3
    !
    ! *** SULFATE SUPER RICH (FREE ACID): Rso4<1;
    !
  ELSE IF (so4rat<1.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'K4'
      CALL calck4 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4hs4) THEN ! RH < 0.4
        scase = 'K1'
        CALL calck1 ! NH4HSO4,NAHSO4,KHSO4,CASO4
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drnahso4) THEN
        scase = 'K2'
        CALL calck2 ! NAHSO4,KHSO4,CASO4
        !
      ELSE IF (drnahso4<=rh .AND. rh<drkhso4) THEN
        scase = 'K3'
        CALL calck3 ! KHSO4,CASO4    0.52 < RH < 0.86
        !
      ELSE IF (drkhso4<=rh) THEN
        scase = 'K4'
        CALL calck4 ! CASO4
      END IF
    END IF
    !
    CALL calcnha ! MINOR SPECIES: HNO3, HCl
    CALL calcnh3 !                NH3
    !
  END IF
  !
  RETURN
END SUBROUTINE isrp4f
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCA2
! *** CASE A2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT >= 2.0)
!     2. LIQUID AEROSOL PHASE ONLY POSSIBLE
!
!     FOR CALCULATIONS, A BISECTION IS PERFORMED TOWARDS X, THE
!     AMOUNT OF HYDROGEN IONS (H+) FOUND IN THE LIQUID PHASE.
!     FOR EACH ESTIMATION OF H+, FUNCTION FUNCB2A CALCULATES THE
!     CONCENTRATION OF IONS FROM THE NH3(GAS) - NH4+(LIQ) EQUILIBRIUM.
!     ELECTRONEUTRALITY IS USED AS THE OBJECTIVE FUNCTION.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calca2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: omehi
  DOUBLE PRECISION :: omelo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  omelo = tiny ! Low  limit: SOLUTION IS VERY BASIC
  omehi = 2.0D0*w(2) ! High limit: FROM NH4+ -> NH3(g) + H+(aq)
  !
  ! *** CALCULATE WATER CONTENT *****************************************
  !
  molal(5) = w(2)
  molal(6) = zero
  CALL calcmr
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = omehi
  y1 = funca2(x1)
  IF (abs(y1)<=eps) RETURN
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (omehi-omelo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, omelo)
    y2 = funca2(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  IF (abs(y2)<=eps) THEN
    RETURN
  ELSE
    CALL pusherr(0001, 'CALCA2') ! WARNING ERROR: NO SOLUTION
    RETURN
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    call rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funca2(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCA2') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  call rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funca2(x3)
  RETURN
  !
END SUBROUTINE calca2



!=======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION FUNCA2
! *** CASE A2
!     FUNCTION THAT SOLVES THE SYSTEM OF EQUATIONS FOR CASE A2 ;
!     AND RETURNS THE VALUE OF THE ZEROED FUNCTION IN FUNCA2.
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funca2 ( omegi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: omegi
  DOUBLE PRECISION :: a1
  DOUBLE PRECISION :: a2
  DOUBLE PRECISION :: a3
  DOUBLE PRECISION :: denom
  INTEGER :: i
  DOUBLE PRECISION :: lamda
  DOUBLE PRECISION :: psi
  DOUBLE PRECISION :: zeta
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  psi = w(2) ! INITIAL AMOUNT OF (NH4)2SO4 IN SOLUTION
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a1 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    a2 = xk2*r*temp/xkw*(gama(8)/gama(9))**2.
    a3 = xkw*rh*water*water
    !
    lamda = psi/(a1/omegi+one)
    zeta = a3/omegi
    !
    ! *** SPECIATION & WATER CONTENT ***************************************
    !
    molal(1) = omegi                                         ! HI
    ! slc.question
    molal (5) = max(psi-lamda,tiny)                          ! SO4I
    molal(3) = w(3)/(one/a2/omegi+one)                       ! NH4I
    molal(6) = lamda                                         ! HSO4I
    gnh3 = max(w(3)-molal(3), tiny)                          ! NH3GI
    coh = zeta                                               ! OHI
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 denom = (2.0*molal(5)+molal(6))
  funca2 = (molal(3)/denom-one) + molal(1)/denom
  RETURN
  !
  ! *** END OF FUNCTION FUNCA2 ********************************************
  !
END FUNCTION funca2
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCA1
! *** CASE A1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4
!
!     A SIMPLE MATERIAL BALANCE IS PERFORMED, AND THE SOLID (NH4)2SO4
!     IS CALCULATED FROM THE SULFATES. THE EXCESS AMMONIA REMAINS IN
!     THE GAS PHASE.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calca1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  cnh42s4 = w(2)
  gnh3 = max(w(3)-2.0*cnh42s4, zero)
  RETURN
  !
END SUBROUTINE calca1



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB4
! *** CASE B4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. LIQUID AEROSOL PHASE ONLY POSSIBLE
!
!     FOR CALCULATIONS, A BISECTION IS PERFORMED WITH RESPECT TO H+.
!     THE OBJECTIVE FUNCTION IS THE DIFFERENCE BETWEEN THE ESTIMATED H+
!     AND THAT CALCULATED FROM ELECTRONEUTRALITY.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb4 (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: ak1
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bet
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: gam
  INTEGER :: i
  !
  ! *** SOLVE EQUATIONS **************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  ! *** CALCULATE WATER CONTENT ******************************************
  !
  CALL calcb1a ! GET DRY SALT CONTENT, AND USE FOR WATER.
  molalr(13) = clc
  molalr(9) = cnh4hs4
  molalr(4) = cnh42s4
  clc = zero
  cnh4hs4 = zero
  cnh42s4 = zero
  water = molalr(13)/m0(13) + molalr(9)/m0(9) + molalr(4)/m0(4)
  !
  molal(3) = w(3) ! NH4I
  !
  DO i = 1, nsweep
    ak1 = xk1*((gama(8)/gama(7))**2.)*(water/gama(7))
    bet = w(2)
    gam = molal(3)
    !
    bb = bet + ak1 - gam
    cc = -ak1*bet
    dd = bb*bb - 4.D0*cc
    !
    ! *** SPECIATION & WATER CONTENT ***************************************
    !
    molal(5) = max(tiny, min(0.5*(-bb+sqrt(dd)),w(2))) ! SO4I
    molal(6) = max(tiny, min(w(2)-molal(5),w(2))) ! HSO4I
    molal(1) = max(tiny, min(ak1*molal(6)/molal(5),w(2))) ! HI
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (.NOT. calain) GO TO 100
    CALL calcact
  END DO
  !
100 RETURN
  !
END SUBROUTINE calcb4
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB3
! *** CASE B3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. BOTH LIQUID & SOLID PHASE IS POSSIBLE
!     3. SOLIDS POSSIBLE: (NH4)2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!

SUBROUTINE calcb3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: tlc
  DOUBLE PRECISION :: tnh42s4
  DOUBLE PRECISION :: tnh4hs4
  DOUBLE PRECISION :: x
  DOUBLE PRECISION :: y

  !
  ! *** CALCULATE EQUIVALENT AMOUNT OF HSO4 AND SO4 ***********************
  !
  x = max(2*w(2)-w(3), zero) ! Equivalent NH4HSO4
  y = max(w(3)-w(2), zero) ! Equivalent NH42SO4
  !
  ! *** CALCULATE SPECIES ACCORDING TO RELATIVE ABUNDANCE OF HSO4 *********
  !
  IF (x<y) THEN ! LC is the MIN (x,y)
    scase = 'B3 ; SUBCASE 1'
    tlc = x
    tnh42s4 = y - x
    CALL calcb3a(tlc, tnh42s4) ! LC + (NH4)2SO4
  ELSE
    scase = 'B3 ; SUBCASE 2'
    tlc = y
    tnh4hs4 = x - y
    CALL calcb3b(tlc, tnh4hs4) ! LC + NH4HSO4
  END IF
  !
  RETURN
  !
END SUBROUTINE calcb3


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB3A
! *** CASE B3 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH (1.0 < SULRAT < 2.0)
!     2. BOTH LIQUID & SOLID PHASE IS POSSIBLE
!     3. SOLIDS POSSIBLE: (NH4)2SO4
!
!     FOR CALCULATIONS, A BISECTION IS PERFORMED TOWARDS ZETA, THE
!     AMOUNT OF SOLID (NH4)2SO4 DISSOLVED IN THE LIQUID PHASE.
!     FOR EACH ESTIMATION OF ZETA, FUNCTION FUNCB3A CALCULATES THE
!     AMOUNT OF H+ PRODUCED (BASED ON THE SO4 RELEASED INTO THE
!     SOLUTION). THE SOLUBILITY PRODUCT OF (NH4)2SO4 IS USED AS THE
!     OBJECTIVE FUNCTION.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb3a ( tlc, tnh42s4 )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: tlc
  DOUBLE PRECISION, INTENT(inout) :: tnh42s4
  DOUBLE PRECISION :: dz
  INTEGER :: i
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  DOUBLE PRECISION :: z1
  DOUBLE PRECISION :: z2
  DOUBLE PRECISION :: z3
  DOUBLE PRECISION :: zhi
  DOUBLE PRECISION :: zk
  DOUBLE PRECISION :: zlo
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  zlo = zero ! MIN DISSOLVED (NH4)2SO4
  zhi = tnh42s4 ! MAX DISSOLVED (NH4)2SO4
  !
  ! *** INITIAL VALUES FOR BISECTION (DISSOLVED (NH4)2SO4 ****************
  !
  z1 = zlo
  y1 = funcb3a(z1, tlc, tnh42s4)
  IF (abs(y1)<=eps) RETURN
  ylo = y1
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO ***********************
  !
  dz = (zhi-zlo)/float(ndiv)
  DO i = 1, ndiv
    z2 = z1 + dz
    y2 = funcb3a(z2, tlc, tnh42s4)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    z1 = z2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION FOUND
  !
  yhi = y1 ! Save Y-value at HI position
  IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    RETURN
    !
    ! *** { YLO, YHI } < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH LC
    !
  ELSE IF (ylo<zero .AND. yhi<zero) THEN
    z1 = zhi
    z2 = zhi
    GO TO 110
    !
    ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH LC
    !
  ELSE IF (ylo>zero .AND. yhi>zero) THEN
    z1 = zlo
    z2 = zlo
    GO TO 110
  ELSE
    CALL pusherr(0001, 'CALCB3A') ! WARNING ERROR: NO SOLUTION
    RETURN
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    z3 = 0.5*(z1+z2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcb3a(z3, tlc, tnh42s4)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      z2 = z3
    ELSE
      y1 = y3
      z1 = z3
    END IF
    IF (abs(z2-z1)<=eps*z1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCB3A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN ************************************************
  !
110 zk = 0.5*(z1+z2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcb3a(zk, tlc, tnh42s4)
  !
  RETURN
  !
END SUBROUTINE calcb3a



!=======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION FUNCB3A
! *** CASE B3 ; SUBCASE 1
!     FUNCTION THAT SOLVES THE SYSTEM OF EQUATIONS FOR CASE B3
!     AND RETURNS THE VALUE OF THE ZEROED FUNCTION IN FUNCA3.
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcb3a ( zk, y, x )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: zk
  DOUBLE PRECISION, INTENT(inout) :: y
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: grat1
  INTEGER :: i
  DOUBLE PRECISION :: kk
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  frst = .TRUE.
  calain = .TRUE.
  DO i = 1, nsweep
    grat1 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    dd = sqrt((zk+grat1+y)**2.+4.0*y*grat1)
    kk = 0.5*(-(zk+grat1+y)+dd)
    !
    ! *** SPECIATION & WATER CONTENT ***************************************
    !
    molal(1) = kk ! HI
    molal(5) = kk + zk + y ! SO4I
    molal(6) = max(y-kk, tiny) ! HSO4I
    molal(3) = 3.0*y + 2*zk ! NH4I
    cnh42s4 = x - zk ! Solid (NH4)2SO4
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
  !CC30    FUNCB3A= ( SO4I*NH4I**2.0 )/( XK7*(WATER/GAMA(4))**3.0 )
100 funcb3a = molal(5)*molal(3)**2.0
  funcb3a = funcb3a/(xk7*(water/gama(4))**3.0) - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCB3A ********************************************
  !
END FUNCTION funcb3a



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB3B
! *** CASE B3 ; SUBCASE 2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH (1.0 < SULRAT < 2.0)
!     2. LIQUID PHASE ONLY IS POSSIBLE
!
!     SPECIATION CALCULATIONS IS BASED ON THE HSO4 <--> SO4 EQUILIBRIUM.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb3b ( y, x )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: y
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: grat1
  INTEGER :: i
  DOUBLE PRECISION :: kk
  !
  calaou = .FALSE. ! Outer loop activity calculation flag
  frst = .FALSE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    grat1 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    dd = sqrt((grat1+y)**2.+4.0*(x+y)*grat1)
    kk = 0.5*(-(grat1+y)+dd)
    !
    ! *** SPECIATION & WATER CONTENT ***************************************
    !
    molal(1) = kk ! HI
    molal(5) = y + kk ! SO4I
    molal(6) = max(x+y-kk, tiny) ! HSO4I
    molal(3) = 3.0*y + x ! NH4I
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (.NOT. calain) GO TO 100
    CALL calcact
  END DO
  !
100 RETURN
  !
END SUBROUTINE calcb3b
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB2
! *** CASE B2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : LC, (NH4)2SO4
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON THE SULFATE RATIO:
!     1. WHEN BOTH LC AND (NH4)2SO4 ARE POSSIBLE (SUBROUTINE CALCB2A)
!     2. WHEN ONLY LC IS POSSIBLE (SUBROUTINE CALCB2B).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: x
  DOUBLE PRECISION :: y

  !
  ! *** CALCULATE EQUIVALENT AMOUNT OF HSO4 AND SO4 ***********************
  !
  x = max(2*w(2)-w(3), tiny) ! Equivalent NH4HSO4
  y = max(w(3)-w(2), tiny) ! Equivalent NH42SO4
  !
  ! *** CALCULATE SPECIES ACCORDING TO RELATIVE ABUNDANCE OF HSO4 *********
  !
  IF (x<=y) THEN ! LC is the MIN (x,y)
    scase = 'B2 ; SUBCASE 1'
    CALL calcb2a(x, y-x) ! LC + (NH4)2SO4 POSSIBLE
  ELSE
    scase = 'B2 ; SUBCASE 2'
    CALL calcb2b(y, x-y) ! LC ONLY POSSIBLE
  END IF
  !
  RETURN
  !
END SUBROUTINE calcb2



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB2
! *** CASE B2 ; SUBCASE A.
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH (1.0 < SULRAT < 2.0)
!     2. SOLID PHASE ONLY POSSIBLE
!     3. SOLIDS POSSIBLE: LC, (NH4)2SO4
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE
!
!     FOR SOLID CALCULATIONS, A MATERIAL BALANCE BASED ON THE STOICHIMETRIC
!     PROPORTION OF AMMONIUM AND SULFATE IS DONE TO CALCULATE THE AMOUNT
!     OF LC AND (NH4)2SO4 IN THE SOLID PHASE.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb2a ( tlc, tnh42s4 )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(in) :: tlc
  DOUBLE PRECISION, INTENT(in) :: tnh42s4

  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmlcas) THEN
    scase = 'B2 ; SUBCASE A1' ! SOLIDS POSSIBLE ONLY
    clc = tlc
    cnh42s4 = tnh42s4
    scase = 'B2 ; SUBCASE A1'
  ELSE
    scase = 'B2 ; SUBCASE A2'
    CALL calcb2a2(tlc, tnh42s4) ! LIQUID & SOLID PHASE POSSIBLE
    scase = 'B2 ; SUBCASE A2'
  END IF
  !
  RETURN
  !
END SUBROUTINE calcb2a



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB2A2
! *** CASE B2 ; SUBCASE A2.
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH (1.0 < SULRAT < 2.0)
!     2. SOLID PHASE ONLY POSSIBLE
!     3. SOLIDS POSSIBLE: LC, (NH4)2SO4
!
!     THIS IS THE CASE WHERE THE RELATIVE HUMIDITY IS IN THE MUTUAL
!     DRH REGION. THE SOLUTION IS ASSUMED TO BE THE SUM OF TWO WEIGHTED
!     SOLUTIONS ; THE SOLID PHASE ONLY (SUBROUTINE CALCB2A1) AND THE
!     THE SOLID WITH LIQUID PHASE (SUBROUTINE CALCB3).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb2a2 ( tlc, tnh42s4 )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(in) :: tlc
  DOUBLE PRECISION, INTENT(in) :: tnh42s4
  DOUBLE PRECISION :: clco
  DOUBLE PRECISION :: cnh42so
  DOUBLE PRECISION :: onemwf
  DOUBLE PRECISION :: wf

  !
  ! *** FIND WEIGHT FACTOR **********************************************
  !
  IF (wftyp==0) THEN
    wf = zero
  ELSE IF (wftyp==1) THEN
    wf = 0.5D0
  ELSE
    wf = (drlc-rh)/(drlc-drmlcas)
  END IF
  onemwf = one - wf
  !
  ! *** FIND FIRST SECTION ; DRY ONE ************************************
  !
  clco = tlc ! FIRST (DRY) SOLUTION
  cnh42so = tnh42s4
  !
  ! *** FIND SECOND SECTION ; DRY & LIQUID ******************************
  !
  clc = zero
  cnh42s4 = zero
  CALL calcb3 ! SECOND (LIQUID) SOLUTION
  !
  ! *** FIND SOLUTION AT MDRH BY WEIGHTING DRY & LIQUID SOLUTIONS.
  !
  molal(1) = onemwf*molal(1) ! H+
  molal(3) = onemwf*(2.D0*(cnh42so-cnh42s4)+3.D0*(clco-clc)) ! NH4+
  molal(5) = onemwf*(cnh42so-cnh42s4+clco-clc) ! SO4--
  molal(6) = onemwf*(clco-clc) ! HSO4-
  !
  water = onemwf*water
  !
  clc = wf*clco + onemwf*clc
  cnh42s4 = wf*cnh42so + onemwf*cnh42s4
  !
  RETURN
  !
END SUBROUTINE calcb2a2



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB2
! *** CASE B2 ; SUBCASE B
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH (1.0 < SULRAT < 2.0)
!     2. BOTH LIQUID & SOLID PHASE IS POSSIBLE
!     3. SOLIDS POSSIBLE: LC
!
!     FOR CALCULATIONS, A BISECTION IS PERFORMED TOWARDS ZETA, THE
!     AMOUNT OF SOLID LC DISSOLVED IN THE LIQUID PHASE.
!     FOR EACH ESTIMATION OF ZETA, FUNCTION FUNCB2A CALCULATES THE
!     AMOUNT OF H+ PRODUCED (BASED ON THE HSO4, SO4 RELEASED INTO THE
!     SOLUTION). THE SOLUBILITY PRODUCT OF LC IS USED AS THE OBJECTIVE
!     FUNCTION.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb2b ( tlc, tnh4hs4 )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(in) :: tlc
  DOUBLE PRECISION, INTENT(in) :: tnh4hs4
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  DOUBLE PRECISION :: zhi
  DOUBLE PRECISION :: zlo
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  zlo = zero
  zhi = tlc ! High limit: all of it in liquid phase
  !
  ! *** INITIAL VALUES FOR BISECTION **************************************
  !
  x1 = zhi
  y1 = funcb2b(x1, tnh4hs4, tlc)
  IF (abs(y1)<=eps) RETURN
  yhi = y1 ! Save Y-value at Hi position
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO ************************
  !
  dx = (zhi-zlo)/ndiv
  DO i = 1, ndiv
    x2 = x1 - dx
    y2 = funcb2b(x2, tnh4hs4, tlc)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION FOUND
  !
  ylo = y1 ! Save Y-value at LO position
  IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    RETURN
    !
    ! *** { YLO, YHI } < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH LC
    !
  ELSE IF (ylo<zero .AND. yhi<zero) THEN
    x1 = zhi
    x2 = zhi
    GO TO 110
    !
    ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH LC
    !
  ELSE IF (ylo>zero .AND. yhi>zero) THEN
    x1 = zlo
    x2 = zlo
    GO TO 110
  ELSE
    CALL pusherr(0001, 'CALCB2B') ! WARNING ERROR: NO SOLUTION
    RETURN
  END IF
  !
  ! *** PERFORM BISECTION *************************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcb2b(x3, tnh4hs4, tlc)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCB2B') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN ************************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcb2b(x3, tnh4hs4, tlc)
  !
  RETURN
  !
END SUBROUTINE calcb2b



!=======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION FUNCB2B
! *** CASE B2 ;
!     FUNCTION THAT SOLVES THE SYSTEM OF EQUATIONS FOR CASE B2 ; SUBCASE 2
!     AND RETURNS THE VALUE OF THE ZEROED FUNCTION IN FUNCB2B.
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcb2b ( x, tnh4hs4, tlc )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(in) :: x
  DOUBLE PRECISION, INTENT(in) :: tnh4hs4
  DOUBLE PRECISION, INTENT(in) :: tlc
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: grat2
  INTEGER :: i
  DOUBLE PRECISION :: omega
  DOUBLE PRECISION :: parm

  !
  ! *** SOLVE EQUATIONS **************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  DO i = 1, nsweep
    grat2 = xk1*water*(gama(8)/gama(7))**2./gama(7)
    parm = x + grat2
    delta = parm*parm + 4.0*(x+tnh4hs4)*grat2 ! Diakrinousa
    omega = 0.5*(-parm+sqrt(delta)) ! Thetiki riza (ie:H+>0)
    !
    ! *** SPECIATION & WATER CONTENT ***************************************
    !
    molal(1) = omega ! HI
    molal(3) = 3.0*x + tnh4hs4 ! NH4I
    molal(5) = x + omega ! SO4I
    molal(6) = max(x+tnh4hs4-omega, tiny) ! HSO4I
    clc = max(tlc-x, zero) ! Solid LC
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP ******************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION **************************************
  !
  !CC30    FUNCB2B= ( NH4I**3.*SO4I*HSO4I )/( XK13*(WATER/GAMA(13))**5. )
100 funcb2b = (molal(3)**3.)*molal(5)*molal(6)
  funcb2b = funcb2b/(xk13*(water/gama(13))**5.) - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCB2B *******************************************
  !
END FUNCTION funcb2b


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB1
! *** CASE B1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : LC, (NH4)2SO4, NH4HSO4
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCB1A)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb1 (  )
  USE mo_isrpia
  IMPLICIT NONE

  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmlcab) THEN
    scase = 'B1 ; SUBCASE 1'
    CALL calcb1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'B1 ; SUBCASE 1'
  ELSE
    scase = 'B1 ; SUBCASE 2'
    CALL calcb1b ! LIQUID & SOLID PHASE POSSIBLE
    scase = 'B1 ; SUBCASE 2'
  END IF
  !
  RETURN
  !
END SUBROUTINE calcb1



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB1A
! *** CASE B1 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH
!     2. THERE IS NO LIQUID PHASE
!     3. SOLIDS POSSIBLE: LC, { (NH4)2SO4  XOR  NH4HSO4 } (ONE OF TWO
!                         BUT NOT BOTH)
!
!     A SIMPLE MATERIAL BALANCE IS PERFORMED, AND THE AMOUNT OF LC
!     IS CALCULATED FROM THE (NH4)2SO4 AND NH4HSO4 WHICH IS LEAST
!     ABUNDANT (STOICHIMETRICALLY). THE REMAINING EXCESS OF SALT
!     IS MIXED WITH THE LC.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb1a (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: x
  DOUBLE PRECISION :: y

  !
  ! *** SETUP PARAMETERS ************************************************
  !
  x = 2*w(2) - w(3) ! Equivalent NH4HSO4
  y = w(3) - w(2) ! Equivalent (NH4)2SO4
  !
  ! *** CALCULATE COMPOSITION *******************************************
  !
  IF (x<=y) THEN ! LC is the MIN (x,y)
    clc = x ! NH4HSO4 >= (NH4)2S04
    cnh4hs4 = zero
    cnh42s4 = y - x
  ELSE
    clc = y ! NH4HSO4 <  (NH4)2S04
    cnh4hs4 = x - y
    cnh42s4 = zero
  END IF
  RETURN
  !
END SUBROUTINE calcb1a




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCB1B
! *** CASE B1 ; SUBCASE 2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE: LC, { (NH4)2SO4  XOR  NH4HSO4 } (ONE OF TWO
!                         BUT NOT BOTH)
!
!     THIS IS THE CASE WHERE THE RELATIVE HUMIDITY IS IN THE MUTUAL
!     DRH REGION. THE SOLUTION IS ASSUMED TO BE THE SUM OF TWO WEIGHTED
!     SOLUTIONS ; THE SOLID PHASE ONLY (SUBROUTINE CALCB1A) AND THE
!     THE SOLID WITH LIQUID PHASE (SUBROUTINE CALCB2).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcb1b (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: clco
  DOUBLE PRECISION :: cnh42so
  DOUBLE PRECISION :: cnh4hso
  DOUBLE PRECISION :: onemwf
  DOUBLE PRECISION :: wf

  !
  ! *** FIND WEIGHT FACTOR **********************************************
  !
  IF (wftyp==0) THEN
    wf = zero
  ELSE IF (wftyp==1) THEN
    wf = 0.5D0
  ELSE
    wf = (drnh4hs4-rh)/(drnh4hs4-drmlcab)
  END IF
  onemwf = one - wf
  !
  ! *** FIND FIRST SECTION ; DRY ONE ************************************
  !
  CALL calcb1a
  clco = clc ! FIRST (DRY) SOLUTION
  cnh42so = cnh42s4
  cnh4hso = cnh4hs4
  !
  ! *** FIND SECOND SECTION ; DRY & LIQUID ******************************
  !
  clc = zero
  cnh42s4 = zero
  cnh4hs4 = zero
  CALL calcb2 ! SECOND (LIQUID) SOLUTION
  !
  ! *** FIND SOLUTION AT MDRH BY WEIGHTING DRY & LIQUID SOLUTIONS.
  !
  molal(1) = onemwf*molal(1) ! H+
  molal(3) = onemwf*(2.D0*(cnh42so-cnh42s4)+(cnh4hso-cnh4hs4)+3.D0*(clco-  &
       clc)) ! NH4+
  molal(5) = onemwf*(cnh42so-cnh42s4+clco-clc) ! SO4--
  molal(6) = onemwf*(cnh4hso-cnh4hs4+clco-clc) ! HSO4-
  !
  water = onemwf*water
  !
  clc = wf*clco + onemwf*clc
  cnh42s4 = wf*cnh42so + onemwf*cnh42s4
  cnh4hs4 = wf*cnh4hso + onemwf*cnh4hs4
  !
  RETURN
  !
END SUBROUTINE calcb1b


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCC2
! *** CASE C2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, FREE ACID (SULRAT < 1.0)
!     2. THERE IS ONLY A LIQUID PHASE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcc2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  INTEGER :: i
  DOUBLE PRECISION :: kapa
  DOUBLE PRECISION :: lamda
  DOUBLE PRECISION :: parm
  DOUBLE PRECISION :: psi
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS **************************************************
  !
  lamda = w(3) ! NH4HSO4 INITIALLY IN SOLUTION
  psi = w(2) - w(3) ! H2SO4 IN SOLUTION
  DO i = 1, nsweep
    parm = water*xk1/gama(7)*(gama(8)/gama(7))**2.
    bb = psi + parm
    cc = -parm*(lamda+psi)
    kapa = 0.5*(-bb+sqrt(bb*bb-4.0*cc))
    !
    ! *** SPECIATION & WATER CONTENT ***************************************
    !
    molal(1) = psi + kapa ! HI
    molal(3) = lamda ! NH4I
    molal(5) = kapa ! SO4I
    molal(6) = max(lamda+psi-kapa, tiny) ! HSO4I
    ch2so4 = max(molal(5)+molal(6)-molal(3), zero) ! Free H2SO4
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (.NOT. calain) GO TO 100
    CALL calcact
  END DO
  !
100 RETURN
  !
END SUBROUTINE calcc2



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCC1
! *** CASE C1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, FREE ACID (SULRAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE: NH4HSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcc1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: khi
  DOUBLE PRECISION :: klo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  klo = tiny
  khi = w(3)
  !
  ! *** INITIAL VALUES FOR BISECTION *************************************
  !
  x1 = klo
  y1 = funcc1(x1)
  IF (abs(y1)<=eps) GO TO 120
  ylo = y1
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO ***********************
  !
  dx = (khi-klo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcc1(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2 .LT. ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION FOUND
  !
  yhi = y2 ! Save Y-value at HI position
  IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
    !
    ! *** { YLO, YHI } < 0.0  SOLUTION IS ALWAYS UNDERSATURATED WITH NH4HS04
    !
  ELSE IF (ylo<zero .AND. yhi<zero) THEN
    GO TO 120
    !
    ! *** { YLO, YHI } > 0.0 SOLUTION IS ALWAYS SUPERSATURATED WITH NH4HS04
    !
  ELSE IF (ylo>zero .AND. yhi>zero) THEN
    x1 = klo
    x2 = klo
    GO TO 110
  ELSE
    CALL pusherr(0001, 'CALCC1') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION OF DISSOLVED NH4HSO4 **************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcc1(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCC1') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN ***********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcc1(x3)
  !
120 RETURN
  !
END SUBROUTINE calcc1



!=======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION FUNCC1
! *** CASE C1 ;
!     FUNCTION THAT SOLVES THE SYSTEM OF EQUATIONS FOR CASE C1
!     AND RETURNS THE VALUE OF THE ZEROED FUNCTION IN FUNCC1.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcc1 ( kapa )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: kapa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  INTEGER :: i
  DOUBLE PRECISION :: lamda
  DOUBLE PRECISION :: par1
  DOUBLE PRECISION :: par2
  DOUBLE PRECISION :: psi
  !
  ! *** SOLVE EQUATIONS **************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  psi = w(2) - w(3)
  DO i = 1, nsweep
    par1 = xk1*water/gama(7)*(gama(8)/gama(7))**2.0
    par2 = xk12*(water/gama(9))**2.0
    bb = psi + par1
    cc = -par1*(psi+kapa)
    lamda = 0.5*(-bb+sqrt(bb*bb-4*cc))
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY *******************************
    !
    molal(1) = psi + lamda ! HI
    molal(3) = kapa ! NH4I
    molal(5) = lamda ! SO4I
    molal(6) = max(zero, psi+kapa-lamda) ! HSO4I
    cnh4hs4 = max(w(3)-molal(3), zero) ! Solid NH4HSO4
    ch2so4 = max(psi, zero) ! Free H2SO4
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE ZERO FUNCTION *******************************************
  !
  !CC30    FUNCC1= (NH4I*HSO4I/PAR2) - ONE
100 funcc1 = (molal(3)*molal(6)/par2) - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCC1 ********************************************
  !
END FUNCTION funcc1

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCD3
! *** CASE D3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. THERE IS OLNY A LIQUID PHASE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcd3 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: p4
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  DOUBLE PRECISION :: yy
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcd1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4no3 ! Save from CALCD1 run
  chi2 = cnh42s4
  chi3 = ghno3
  chi4 = gnh3
  !
  psi1 = cnh4no3 ! ASSIGN INITIAL PSI's
  psi2 = chi2
  psi3 = zero
  psi4 = zero
  !
  molal(5) = psi2 ! Include initial amount in water calc
  molal(6) = zero
  molal(3) = psi1
  molal(7) = psi1
  CALL calcmr ! Initial water
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi4lo = tiny ! Low  limit
  psi4hi = chi4 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
100 x1 = psi4lo 
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y1 = funcd3(x1)
  IF (abs(y1)<=eps) RETURN
  ylo = y1 ! Save Y-value at HI position
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcd3(x2)
    IF (((y1) < zero) .and. ((y2) > zero)) goto 110  ! (Y1*Y2.LT.ZERO) (slc.1.2012)
  ! IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 110 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION FOUND
  !
  yhi = y1 ! Save Y-value at Hi position
  IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    RETURN
    !
    ! *** { YLO, YHI } < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NH3
    ! Physically I dont know when this might happen, but I have put this
    ! branch in for completeness. I assume there is no solution; all NO3 goes to the
    ! gas phase.
    !
  ELSE IF (ylo<zero .AND. yhi<zero) THEN
    p4 = tiny ! PSI4LO ! CHI4
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    yy = funcd3(p4)
    GO TO 130
    !
    ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NH3
    ! This happens when Sul.Rat. = 2.0, so some NH4+ from sulfate evaporates
    ! and goes to the gas phase ; so I redefine the LO and HI limits of PSI4
    ! and proceed again with root tracking.
    !
  ELSE IF (ylo>zero .AND. yhi>zero) THEN
    psi4hi = psi4lo
    psi4lo = psi4lo - 0.1*(psi1+psi2) ! No solution; some NH3 evaporates
    IF (psi4lo<-(psi1+psi2)) THEN
      CALL pusherr(0001, 'CALCD3') ! WARNING ERROR: NO SOLUTION
      RETURN
    ELSE
      molal(5) = psi2              ! Include sulfate in initial water calculation
      molal(6) = zero
      molal(3) = psi1
      molal(7) = psi1
      CALL calcmr ! Initial water
      GO TO 100 ! Redo root tracking
    END IF
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
110 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcd3(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*abs(x1)) GO TO 120
  END DO
  CALL pusherr(0002, 'CALCD3') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
120 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcd3(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
130 CONTINUE
  IF (molal(1)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  RETURN
  !
END SUBROUTINE calcd3



!=======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION FUNCD3
! *** CASE D3
!     FUNCTION THAT SOLVES THE SYSTEM OF EQUATIONS FOR CASE D3 ;
!     AND RETURNS THE VALUE OF THE ZEROED FUNCTION IN FUNCD3.
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcd3 ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: abb
  DOUBLE PRECISION :: ahi
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: denm
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  psi4 = p4
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a2 = xk7*(water/gama(4))**3.0
    a3 = xk4*r*temp*(water/gama(10))**2.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a7 = xkw*rh*water*water
    !
    psi3 = a3*a4*chi3*(chi4-psi4) - psi1*(2.D0*psi2+psi1+psi4)
    psi3 = psi3/(a3*a4*(chi4-psi4)+2.D0*psi2+psi1+psi4)
    psi3 = min(max(psi3,zero), chi3)
    !
    bb = psi4 - psi3
    !CCOLD         AHI  = 0.5*(-BB + SQRT(BB*BB + 4.d0*A7)) ! This is correct also
    !CC         AHI  =2.0*A7/(BB+SQRT(BB*BB + 4.d0*A7)) ! Avoid overflow when HI->0
    denm = bb + sqrt(bb*bb+4.D0*a7)
    IF (denm<=tiny) THEN ! Avoid overflow when HI->0
      abb = abs(bb)
      denm = (bb+abb) + 2.0*a7/abb ! Taylor expansion of SQRT
    END IF
    ahi = 2.0*a7/denm
    !
    ! *** SPECIATION & WATER CONTENT ***************************************
    !
    molal(1) = ahi ! HI
    molal(3) = psi1 + psi4 + 2.D0*psi2 ! NH4I
    molal(5) = psi2 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi3 + psi1 ! NO3I
    cnh42s4 = chi2 - psi2 ! Solid (NH4)2SO4
    cnh4no3 = zero ! Solid NH4NO3
    ghno3 = chi3 - psi3 ! Gas HNO3
    gnh3 = chi4 - psi4 ! Gas NH3
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 CONTINUE
  !CC      FUNCD3= NH4I/HI/MAX(GNH3,TINY)/A4 - ONE
  funcd3 = molal(3)/molal(1)/max(gnh3, tiny)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCD3 ********************************************
  !
END FUNCTION funcd3
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCD2
! *** CASE D2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcd2 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: p4
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  DOUBLE PRECISION :: yy
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcd1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4no3 ! Save from CALCD1 run
  chi2 = cnh42s4
  chi3 = ghno3
  chi4 = gnh3
  !
  psi1 = cnh4no3 ! ASSIGN INITIAL PSI's
  psi2 = cnh42s4
  psi3 = zero
  psi4 = zero
  !
  molal(5) = psi2              ! Include initial amount in water calc
  molal(6) = zero
  molal(3) = psi1
  molal(7) = psi1
  CALL calcmr ! Initial water
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi4lo = tiny ! Low  limit
  psi4hi = chi4 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
100 x1 = psi4lo
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y1 = funcd2(x1)
  IF (abs(y1)<=eps) RETURN
  ylo = y1 ! Save Y-value at HI position
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcd2(x2)
    IF (((y1) > zero) .and. ((y2) > zero)) goto 110  ! (Y1*Y2.LT.ZERO) slc.1.2012
 !       if (sign(1.D0,y1)*sign(1.D0,y2).lt.zero) then
 !!
 !! This is done, in case if Y(PSI4LO)>0, but Y(PSI4LO+DX) < 0 (i.e.undersat)
 !!
 !       IF (y1 .le. y2) goto 20  ! (y1*y2.lt.zero)
 !       END IF
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION FOUND
  !
  yhi = y1 ! Save Y-value at Hi position
  IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    RETURN
    !
    ! *** { YLO, YHI } < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NH3
    ! Physically I dont know when this might happen, but I have put this
    ! branch in for completeness. I assume there is no solution; all NO3 goes to the
    ! gas phase.
    !
  ELSE IF (ylo<zero .AND. yhi<zero) THEN
    p4 = tiny ! PSI4LO ! CHI4
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    yy = funcd2(p4)
    GO TO 130
    !
    ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NH3
    ! This happens when Sul.Rat. = 2.0, so some NH4+ from sulfate evaporates
    ! and goes to the gas phase ; so I redefine the LO and HI limits of PSI4
    ! and proceed again with root tracking.
    !
  ELSE IF (ylo>zero .AND. yhi>zero) THEN
    psi4hi = psi4lo
    psi4lo = psi4lo - 0.1*(psi1+psi2) ! No solution; some NH3 evaporates
    IF (psi4lo<-(psi1+psi2)) THEN
      CALL pusherr(0001, 'CALCD2') ! WARNING ERROR: NO SOLUTION
      RETURN
    ELSE
      molal(5) = psi2              ! Include initial amount in water calc
      molal(6) = zero
      molal(3) = psi1
      molal(7) = psi1
      CALL calcmr ! Initial water
      GO TO 100 ! Redo root tracking
    END IF
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
110 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcd2(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*abs(x1)) GO TO 120
  END DO
  CALL pusherr(0002, 'CALCD2') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
120 x3 = min(x1, x2) ! 0.5*(X1+X2)  ! Get "low" side, it's acidic soln.
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcd2(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
130 CONTINUE
  IF (molal(1)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  RETURN
  !
END SUBROUTINE calcd2



!=======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION FUNCD2
! *** CASE D2
!     FUNCTION THAT SOLVES THE SYSTEM OF EQUATIONS FOR CASE D2 ;
!     AND RETURNS THE VALUE OF THE ZEROED FUNCTION IN FUNCD2.
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcd2 ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: abb
  DOUBLE PRECISION :: ahi
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: denm
  INTEGER :: i
  INTEGER :: islv
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  CALL rstgam ! Reset activity coefficients to 0.1
  frst = .TRUE.
  calain = .TRUE.
  psi4 = p4
  psi2 = chi2
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a2 = xk7*(water/gama(4))**3.0
    a3 = xk4*r*temp*(water/gama(10))**2.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a7 = xkw*rh*water*water
    !
    IF (chi2>tiny .AND. water>tiny) THEN
      psi14 = psi1 + psi4
      CALL poly3(psi14, 0.25*psi14**2., -a2/4.D0, psi2, islv) ! PSI2
      IF (islv==0) THEN
        psi2 = min(psi2, chi2)
      ELSE
        psi2 = tiny
      END IF
    END IF
    !
    psi3 = a3*a4*chi3*(chi4-psi4) - psi1*(2.D0*psi2+psi1+psi4)
    psi3 = psi3/(a3*a4*(chi4-psi4)+2.D0*psi2+psi1+psi4)
    !cc         PSI3  = MIN(MAX(PSI3, ZERO), CHI3)
    !
    bb = psi4 - psi3 ! (BB > 0, acidic solution, <0 alkaline)
    !
    ! Do not change computation scheme for H+, all others did not work well.
    !
    denm = bb + sqrt(bb*bb+4.D0*a7)
    IF (denm<=tiny) THEN ! Avoid overflow when HI->0
      abb = abs(bb)
      denm = (bb+abb) + 2.D0*a7/abb ! Taylor expansion of SQRT
    END IF
    ahi = 2.D0*a7/denm
    !
    ! *** SPECIATION & WATER CONTENT ***************************************
    !
    molal(1) = ahi ! HI
    molal(3) = psi1 + psi4 + 2.D0*psi2 ! NH4
    molal(5) = psi2 ! SO4
    molal(6) = zero ! HSO4
    molal(7) = psi3 + psi1 ! NO3
    cnh42s4 = chi2 - psi2 ! Solid (NH4)2SO4
    cnh4no3 = zero ! Solid NH4NO3
    ghno3 = chi3 - psi3 ! Gas HNO3
    gnh3 = chi4 - psi4 ! Gas NH3
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 CONTINUE
  !CC      FUNCD2= NH4I/HI/MAX(GNH3,TINY)/A4 - ONE
  funcd2 = molal(3)/molal(1)/max(gnh3, tiny)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCD2 ********************************************
  !
END FUNCTION funcd2
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCD1
! *** CASE D1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3
!
!     THERE ARE TWO REGIMES DEFINED BY RELATIVE HUMIDITY:
!     1. RH < MDRH ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCD1A)
!     2. RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcd1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmasan) THEN
    scase = 'D1 ; SUBCASE 1' ! SOLID PHASE ONLY POSSIBLE
    CALL calcd1a
    scase = 'D1 ; SUBCASE 1'
  ELSE
    scase = 'D1 ; SUBCASE 2' ! LIQUID & SOLID PHASE POSSIBLE
    CALL calcmdrh(rh, drmasan, drnh4no3, calcd1a, calcd2)
    scase = 'D1 ; SUBCASE 2'
  END IF
  !
  RETURN
  !
END SUBROUTINE calcd1


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCD1A
! *** CASE D1 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3
!
!     THE SOLID (NH4)2SO4 IS CALCULATED FROM THE SULFATES, WHILE NH4NO3
!     IS CALCULATED FROM NH3-HNO3 EQUILIBRIUM. 'ZE' IS THE AMOUNT OF
!     NH4NO3 THAT VOLATIZES WHEN ALL POSSILBE NH4NO3 IS INITIALLY IN
!     THE SOLID PHASE.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcd1a (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: om
  DOUBLE PRECISION :: omps
  DOUBLE PRECISION :: parm
  DOUBLE PRECISION :: ps
  DOUBLE PRECISION :: x
  DOUBLE PRECISION :: ze
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  parm = xk10/(r*temp)/(r*temp)
  !
  ! *** CALCULATE NH4NO3 THAT VOLATIZES *********************************
  !
  cnh42s4 = w(2)
  x = max(zero, min(w(3)-2.0*cnh42s4,w(4))) ! MAX NH4NO3
  ps = max(w(3)-x-2.0*cnh42s4, zero)
  om = max(w(4)-x, zero)
  !
  omps = om + ps
  diak = sqrt(omps*omps+4.0*parm) ! DIAKRINOUSA
  ze = min(x, 0.5*(-omps+diak)) ! THETIKI RIZA
  !
  ! *** SPECIATION *******************************************************
  !
  cnh4no3 = x - ze ! Solid NH4NO3
  gnh3 = ps + ze ! Gas NH3
  ghno3 = om + ze ! Gas HNO3
  !
  RETURN
  !
END SUBROUTINE calcd1a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCG5
! *** CASE G5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcg5 (  )
  USE mo_isrpia
  USE mo_caseg
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi1 = 0.5*w(1)
  chi2 = max(w(2)-chi1, zero)
  chi3 = zero
  chi4 = max(w(3)-2.D0*chi2, zero)
  chi5 = w(4)
  chi6 = w(5)
  !
  psi1 = chi1
  psi2 = chi2
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  water = chi2/m0(4) + chi1/m0(2)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcg5a(x1)
  IF (chi6<=tiny) GO TO 120
  !cc      IF (ABS(Y1).LE.EPS .OR. CHI6.LE.TINY) GOTO 50
  !cc      IF (WATER .LE. TINY) RETURN                    ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcg5a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcg5a (psi6lo)
  END IF

  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcg5a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCG5') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcg5a(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN ! If quadrat.called
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcg5




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCG5A
! *** CASE G5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcg5a ( x )
  USE mo_isrpia
  USE mo_caseg
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: akk
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a2 = xk7*(water/gama(4))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    akk = a4*a6
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    IF (chi5>=tiny) THEN
      psi5 = psi6*chi5/(a6/a5*(chi6-psi6)+psi6)
    ELSE
      psi5 = tiny
    END IF
    !
    !CC      IF(CHI4.GT.TINY) THEN
    IF (w(2)>tiny) THEN ! Accounts for NH3 evaporation
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6) - 2.D0*psi2/a4
      dd = max(bb*bb-4.D0*cc, zero) ! Patch proposed by Uma Shankar, 19/11/01
      psi4 = 0.5D0*(-bb-sqrt(dd))
    ELSE
      psi4 = tiny
    END IF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = 2.0D0*psi1 ! NAI
    molal(3) = 2.0*psi2 + psi4 ! NH4I
    molal(4) = psi6 ! CLI
    molal(5) = psi2 + psi1 ! SO4I
    molal(6) = zero
    molal(7) = psi5 ! NO3I
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny) ! Gas NH3
    ghno3 = max(chi5-psi5, tiny) ! Gas HNO3
    ghcl = max(chi6-psi6, tiny) ! Gas HCl
    !
    cnh42s4 = zero ! Solid (NH4)2SO4
    cnh4no3 = zero ! Solid NH4NO3
    cnh4cl = zero ! Solid NH4Cl
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funcg5a = molal(1)*molal(4)/ghcl/a6 - one
  !CC         FUNCG5A = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCG5A *******************************************
  !
END FUNCTION funcg5a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCG4
! *** CASE G4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcg4 (  )
  USE mo_isrpia
  USE mo_caseg
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi1 = 0.5*w(1)
  chi2 = max(w(2)-chi1, zero)
  chi3 = zero
  chi4 = max(w(3)-2.D0*chi2, zero)
  chi5 = w(4)
  chi6 = w(5)
  !
  psi2 = chi2
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  water = chi2/m0(4) + chi1/m0(2)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcg4a(x1)
  IF (chi6<=tiny) GO TO 120
  !CC      IF (ABS(Y1).LE.EPS .OR. CHI6.LE.TINY .OR. WATER .LE. TINY) GOTO 50
  !CC      IF (WATER .LE. TINY) RETURN                    ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcg4a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    call rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcg4a (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    call rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcg4a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCG4') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  call rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcg4a(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcg4




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCG4A
! *** CASE G4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcg4a ( x )
  USE mo_isrpia
  USE mo_caseg
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: nai
  INTEGER :: nh4i
  INTEGER :: no3i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a2 = xk7*(water/gama(4))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    IF (chi5>=tiny) THEN
      psi5 = psi6*chi5/(a6/a5*(chi6-psi6)+psi6)
    ELSE
      psi5 = tiny
    END IF
    !
    !CC      IF(CHI4.GT.TINY) THEN
    IF (w(2)>tiny) THEN ! Accounts for NH3 evaporation
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6) - 2.D0*psi2/a4
      dd = max(bb*bb-4.D0*cc, zero) ! Patch proposed by Uma shankar, 19/11/2001
      psi4 = 0.5D0*(-bb-sqrt(dd))
    ELSE
      psi4 = tiny
    END IF
    !
    !  CALCULATE CONCENTRATIONS
    !
    nh4i = 2.0*psi2 + psi4
    cli = psi6
    so4i = psi2 + psi1
    no3i = psi5
    nai = 2.0D0*psi1
    !
    CALL calcph(2.D0*so4i+no3i+cli-nai-nh4i, hi, ohi)
    !
    ! *** Na2SO4 DISSOLUTION
    !
    IF (chi1>tiny .AND. water>tiny) THEN ! PSI1
      CALL poly3(psi2, zero, -a1/4.D0, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    ELSE
      psi1 = zero
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = zero
    molal(7) = no3i
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnh4cl = zero
    cna2so4 = max(chi1-psi1, zero)
    !
    ! *** CALCULATE MOLALR ARRAY, WATER AND ACTIVITIES **********************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funcg4a = molal(1)*molal(4)/ghcl/a6 - one
  !CC         FUNCG4A = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCG4A *******************************************
  !
END FUNCTION funcg4a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCG3
! *** CASE G3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. LIQUID & SOLID PHASE ARE BOTH POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcg3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i

  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  IF (w(4)>tiny .AND. w(5)>tiny) THEN ! NO3,CL EXIST, WATER POSSIBLE
    scase = 'G3 ; SUBCASE 1'
    CALL calcg3a
    scase = 'G3 ; SUBCASE 1'
  ELSE ! NO3, CL NON EXISTANT
    scase = 'G1 ; SUBCASE 1'
    CALL calcg1a
    scase = 'G1 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmg3) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcg1a
      scase = 'G3 ; SUBCASE 2'
      RETURN
    ELSE
      scase = 'G3 ; SUBCASE 3' ! MDRH REGION (NA2SO4, NH42S4)
      CALL calcmdrh(rh, drmg3, drnh42s4, calcg1a, calcg4)
      scase = 'G3 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcg3


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCG3A
! *** CASE G3 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcg3a (  )
  USE mo_isrpia
  USE mo_caseg

  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi1 = 0.5*w(1)
  chi2 = max(w(2)-chi1, zero)
  chi3 = zero
  chi4 = max(w(3)-2.D0*chi2, zero)
  chi5 = w(4)
  chi6 = w(5)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  water = tiny
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcg3a(x1)
  IF (chi6<=tiny) GO TO 120
  !CC      IF (ABS(Y1).LE.EPS .OR. CHI6.LE.TINY .OR. WATER .LE. TINY) GOTO 50
  !CC      IF (WATER .LE. TINY) RETURN                    ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcg3a(x2)
    !
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcg3a (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcg3a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCG3A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcg3a(x3)
  !
  ! *** FINAL CALCULATIONS *************************************************
  !
120 CONTINUE
  !
  ! *** Na2SO4 DISSOLUTION
  !
  IF (chi1>tiny .AND. water>tiny) THEN ! PSI1
    CALL poly3(psi2, zero, -a1/4.D0, psi1, islv)
    IF (islv==0) THEN
      psi1 = min(psi1, chi1)
    ELSE
      psi1 = zero
    END IF
  ELSE
    psi1 = zero
  END IF
  molal(2) = 2.0D0*psi1 ! Na+  EFFECT
  molal(5) = molal(5) + psi1 ! SO4  EFFECT
  cna2so4 = max(chi1-psi1, zero) ! NA2SO4(s) depletion
  !
  ! *** HSO4 equilibrium
  !
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcg3a




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCG3A
! *** CASE G3 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcg3a ( x )
  USE mo_isrpia
  USE mo_caseg
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi20
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi2 = chi2
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a2 = xk7*(water/gama(4))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    IF (chi5>=tiny) THEN
      psi5 = psi6*chi5/(a6/a5*(chi6-psi6)+psi6)
    ELSE
      psi5 = tiny
    END IF
    !
    !CC      IF(CHI4.GT.TINY) THEN
    IF (w(2)>tiny) THEN ! Accounts for NH3 evaporation
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6) - 2.D0*psi2/a4
      dd = max(bb*bb-4.D0*cc, zero) ! Patch proposed by Uma Shankar, 19/11/01
      psi4 = 0.5D0*(-bb-sqrt(dd))
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi2>tiny .AND. water>tiny) THEN
      CALL poly3(psi4, psi4*psi4/4.D0, -a2/4.D0, psi20, islv)
      IF (islv==0) psi2 = min(psi20, chi2)
    END IF
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    molal(2) = zero ! Na
    molal(3) = 2.0*psi2 + psi4 ! NH4I
    molal(4) = psi6 ! CLI
    molal(5) = psi2 ! SO4I
    molal(6) = zero ! HSO4
    molal(7) = psi5 ! NO3I
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny) ! Gas NH3
    ghno3 = max(chi5-psi5, tiny) ! Gas HNO3
    ghcl = max(chi6-psi6, tiny) ! Gas HCl
    !
    cnh42s4 = chi2 - psi2 ! Solid (NH4)2SO4
    cnh4no3 = zero ! Solid NH4NO3
    cnh4cl = zero ! Solid NH4Cl
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funcg3a = molal(1)*molal(4)/ghcl/a6 - one
  !CC         FUNCG3A = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCG3A *******************************************
  !
END FUNCTION funcg3a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCG2
! *** CASE G2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. LIQUID & SOLID PHASE ARE BOTH POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcg2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i

  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF NITRATES ***********************
  !
  IF (w(4)>tiny) THEN ! NO3 EXISTS, WATER POSSIBLE
    scase = 'G2 ; SUBCASE 1'
    CALL calcg2a
    scase = 'G2 ; SUBCASE 1'
  ELSE ! NO3 NON EXISTANT, WATER NOT POSSIBLE
    scase = 'G1 ; SUBCASE 1'
    CALL calcg1a
    scase = 'G1 ; SUBCASE 1'
  END IF
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  IF (water<=tiny) THEN
    IF (rh<drmg2) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcg1a
      scase = 'G2 ; SUBCASE 2'
    ELSE
      IF (w(5)>tiny) THEN
        scase = 'G2 ; SUBCASE 3' ! MDRH (NH4CL, NA2SO4, NH42S4)
        CALL calcmdrh(rh, drmg2, drnh4cl, calcg1a, calcg3a)
        scase = 'G2 ; SUBCASE 3'
      END IF
      IF (water<=tiny .AND. rh>=drmg3) THEN
        scase = 'G2 ; SUBCASE 4' ! MDRH (NA2SO4, NH42S4)
        CALL calcmdrh(rh, drmg3, drnh42s4, calcg1a, calcg4)
        scase = 'G2 ; SUBCASE 4'
      ELSE
        water = tiny
        DO i = 1, nions
          molal(i) = zero
        END DO
        CALL calcg1a
        scase = 'G2 ; SUBCASE 2'
      END IF
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcg2


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCG2A
! *** CASE G2 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcg2a (  )
  USE mo_isrpia
  USE mo_caseg
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi1 = 0.5*w(1)
  chi2 = max(w(2)-chi1, zero)
  chi3 = zero
  chi4 = max(w(3)-2.D0*chi2, zero)
  chi5 = w(4)
  chi6 = w(5)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny
  !
  water = tiny
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcg2a(x1)
  IF (chi6<=tiny) GO TO 120
  !CC      IF (ABS(Y1).LE.EPS .OR. CHI6.LE.TINY) GOTO 50
  !CC      IF (WATER .LE. TINY) GOTO 50               ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcg2a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2)>eps) water = tiny
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcg2a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCG2A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  IF (x3<=tiny2) THEN ! PRACTICALLY NO NITRATES, SO DRY SOLUTION
    water = tiny
  ELSE
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcg2a(x3)
  END IF
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  !
  ! *** Na2SO4 DISSOLUTION
  !
  IF (chi1>tiny .AND. water>tiny) THEN ! PSI1
    CALL poly3(psi2, zero, -a1/4.D0, psi1, islv)
    IF (islv==0) THEN
      psi1 = min(psi1, chi1)
    ELSE
      psi1 = zero
    END IF
  ELSE
    psi1 = zero
  END IF
  molal(2) = 2.0D0*psi1 ! Na+  EFFECT
  molal(5) = molal(5) + psi1 ! SO4  EFFECT
  cna2so4 = max(chi1-psi1, zero) ! NA2SO4(s) depletion
  !
  ! *** HSO4 equilibrium
  !
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   AFFECT
    molal(5) = molal(5) - delta ! SO4  AFFECT
    molal(6) = delta ! HSO4 AFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcg2a




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCG2A
! *** CASE G2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcg2a ( x )
  USE mo_isrpia
  USE mo_caseg
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: deno
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi20
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi2 = chi2
  psi3 = zero
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a2 = xk7*(water/gama(4))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    !
    deno = max(chi6-psi6-psi3, zero)
    psi5 = chi5/((a6/a5)*(deno/psi6)+one)
    !
    !!psi4 = min(psi5+psi6, chi4)
    IF(w(2)>tiny) THEN       ! Accounts for NH3 evaporation
      bb   =-(chi4 + psi6 + psi5 + 1.D0/a4)
      cc   = chi4*(psi5+psi6) - 2.D0*psi2/a4
      dd   = max(bb*bb-4.D0*cc,zerO)  ! Patch proposed by Uma Shankar, 19/11/01
      psi4 =0.5D0*(-bb - sqrt(dd))
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi2>tiny .AND. water>tiny) THEN
      CALL poly3(psi4, psi4*psi4/4.D0, -a2/4.D0, psi20, islv)
      IF (islv==0) psi2 = min(psi20, chi2)
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(2) = zero ! NA
    molal(3) = 2.0*psi2 + psi4 ! NH4I
    molal(4) = psi6 ! CLI
    molal(5) = psi2 ! SO4I
    molal(6) = zero ! HSO4
    molal(7) = psi5 ! NO3I
    !
    !CC      MOLAL (1) = MAX(CHI5 - PSI5, TINY)*A5/PSI5   ! HI
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = max(chi2-psi2, zero)
    cnh4no3 = zero
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    ! *** CALCULATE MOLALR ARRAY, WATER AND ACTIVITIES **********************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 IF (chi4<=tiny) THEN
    funcg2a = molal(1)*molal(4)/ghcl/a6 - one
  ELSE
    funcg2a = molal(3)*molal(4)/ghcl/gnh3/a6/a4 - one
  END IF
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCG2A *******************************************
  !
END FUNCTION funcg2a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCG1
! *** CASE G1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3, NH4CL, NA2SO4
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCG1A)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcg1 (  )
  USE mo_isrpia
  IMPLICIT NONE

  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmg1) THEN
    scase = 'G1 ; SUBCASE 1'
    CALL calcg1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'G1 ; SUBCASE 1'
  ELSE
    scase = 'G1 ; SUBCASE 2' ! LIQUID & SOLID PHASE POSSIBLE
    CALL calcmdrh(rh, drmg1, drnh4no3, calcg1a, calcg2a)
    scase = 'G1 ; SUBCASE 2'
  END IF
  !
  RETURN
  !
END SUBROUTINE calcg1


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCG1A
! *** CASE G1 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3
!
!     SOLID (NH4)2SO4 IS CALCULATED FROM THE SULFATES, WHILE NH4NO3
!     IS CALCULATED FROM NH3-HNO3 EQUILIBRIUM. 'ZE' IS THE AMOUNT OF
!     NH4NO3 THAT VOLATIZES WHEN ALL POSSILBE NH4NO3 IS INITIALLY IN
!     THE SOLID PHASE.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcg1a (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: a1
  DOUBLE PRECISION :: a2
  DOUBLE PRECISION :: alf
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bet
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: dd1
  DOUBLE PRECISION :: dd2
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: gam
  DOUBLE PRECISION :: kapa
  DOUBLE PRECISION :: kapa1
  DOUBLE PRECISION :: kapa2
  DOUBLE PRECISION :: lamda
  DOUBLE PRECISION :: lamda1
  DOUBLE PRECISION :: lamda2
  DOUBLE PRECISION :: rtsq
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: sqdd
  DOUBLE PRECISION :: sqdd1
  DOUBLE PRECISION :: sqdd2
  DOUBLE PRECISION :: theta1
  DOUBLE PRECISION :: theta2
  !
  ! *** CALCULATE NON VOLATILE SOLIDS ***********************************
  !
  cna2so4 = min(0.5*w(1), w(2))
  frna = max(w(1)-2.D0*cna2so4, zero)
  so4fr = max(w(2)-cna2so4, zero)
  !      CNH42S4 = W(2) - CNA2SO4
  cnh42s4 = max(so4fr, zero) ! CNH42S4
  !
  ! *** CALCULATE VOLATILE SPECIES **************************************
  !
  alf = w(3) - 2.0*cnh42s4
  bet = w(5)
  gam = w(4)
  !
  rtsq = r*temp*r*temp
  a1 = xk6/rtsq
  a2 = xk10/rtsq
  !
  theta1 = gam - bet*(a2/a1)
  theta2 = a2/a1
  !
  ! QUADRATIC EQUATION SOLUTION
  !
  bb = (theta1-alf-bet*(one+theta2))/(one+theta2)
  cc = (alf*bet-a1-bet*theta1)/(one+theta2)
  dd = bb*bb - 4.0D0*cc
  IF (dd<zero) GO TO 100 ! Solve each reaction seperately
  !
  ! TWO ROOTS FOR KAPA, CHECK AND SEE IF ANY VALID
  !
  sqdd = sqrt(dd)
  kapa1 = 0.5D0*(-bb+sqdd)
  kapa2 = 0.5D0*(-bb-sqdd)
  lamda1 = theta1 + theta2*kapa1
  lamda2 = theta1 + theta2*kapa2
  !
  IF (kapa1>=zero .AND. lamda1>=zero) THEN
    IF (alf-kapa1-lamda1>=zero .AND. bet-kapa1>=zero .AND.                 &
         gam-lamda1>=zero) THEN
      kapa = kapa1
      lamda = lamda1
      GO TO 110
    END IF
  END IF
  !
  IF (kapa2>=zero .AND. lamda2>=zero) THEN
    IF (alf-kapa2-lamda2>=zero .AND. bet-kapa2>=zero .AND.                 &
         gam-lamda2>=zero) THEN
      kapa = kapa2
      lamda = lamda2
      GO TO 110
    END IF
  END IF
  !
  ! SEPERATE SOLUTION OF NH4CL & NH4NO3 EQUILIBRIA
  !
100 kapa = zero
  lamda = zero
  dd1 = (alf+bet)*(alf+bet) - 4.0D0*(alf*bet-a1)
  dd2 = (alf+gam)*(alf+gam) - 4.0D0*(alf*gam-a2)
  !
  ! NH4CL EQUILIBRIUM
  !
  IF (dd1>=zero) THEN
    sqdd1 = sqrt(dd1)
    kapa1 = 0.5D0*(alf+bet+sqdd1)
    kapa2 = 0.5D0*(alf+bet-sqdd1)
    !
    IF (kapa1>=zero .AND. kapa1<=min(alf,bet)) THEN
      kapa = kapa1
    ELSE IF (kapa2>=zero .AND. kapa2<=min(alf,bet)) THEN
      kapa = kapa2
    ELSE
      kapa = zero
    END IF
  END IF
  !
  ! NH4NO3 EQUILIBRIUM
  !
  IF (dd2>=zero) THEN
    sqdd2 = sqrt(dd2)
    lamda1 = 0.5D0*(alf+gam+sqdd2)
    lamda2 = 0.5D0*(alf+gam-sqdd2)
    !
    IF (lamda1>=zero .AND. lamda1<=min(alf,gam)) THEN
      lamda = lamda1
    ELSE IF (lamda2>=zero .AND. lamda2<=min(alf,gam)) THEN
      lamda = lamda2
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! IF BOTH KAPA, LAMDA ARE > 0, THEN APPLY EXISTANCE CRITERION
  !
  IF (kapa>zero .AND. lamda>zero) THEN
    IF (bet<lamda/theta1) THEN
      kapa = zero
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! *** CALCULATE COMPOSITION OF VOLATILE SPECIES ***********************
  !
110 CONTINUE
  cnh4no3 = lamda
  cnh4cl = kapa
  !
  gnh3 = max(alf-kapa-lamda, zero)
  ghno3 = max(gam-lamda, zero)
  ghcl = max(bet-kapa, zero)
  !
  RETURN
  !
END SUBROUTINE calcg1a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCH6
! *** CASE H6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calch6 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi1 = w(2) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funch6a(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funch6a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funch6a (psi6lo)
  END IF

  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funch6a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCH6') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funch6a(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calch6




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCH6A
! *** CASE H6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funch6a ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    a9 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = max(psi5, tiny)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = bb*bb - 4.D0*cc
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(psi4, chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 ! NO3I
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    cna2so4 = max(chi1-psi1, zero)
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funch6a = molal(3)*molal(4)/ghcl/gnh3/a6/a4 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCH6A *******************************************
  !
END FUNCTION funch6a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCH5
! *** CASE H5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calch5 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF NITRATES ***********************
  !
  IF (w(4)<=tiny .AND. w(5)<=tiny) THEN
    scase = 'H5'
    CALL calch1a
    scase = 'H5'
    RETURN
  END IF
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi1 = w(2) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funch5a(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funch5a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funch5a (psi6lo)
  END IF

  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funch5a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCH5') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funch5a(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calch5




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCH5A
! *** CASE H5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NONE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funch5a ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    a9 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = max(psi5, tiny)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = bb*bb - 4.D0*cc
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(psi4, chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi1>tiny .AND. water>tiny) THEN ! NA2SO4 DISSOLUTION
      aa = psi7 + psi8
      bb = aa*aa
      cc = -a1/4.D0
      CALL poly3(aa, bb, cc, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    END IF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 ! SO4I
    molal(6) = zero
    molal(7) = psi5 + psi8 ! NO3I
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    cna2so4 = max(chi1-psi1, zero)
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funch5a = molal(3)*molal(4)/ghcl/gnh3/a6/a4 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCH5A *******************************************
  !
END FUNCTION funch5a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCH4
! *** CASE H4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calch4 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF NITRATES ***********************
  !
  IF (w(4)<=tiny .AND. w(5)<=tiny) THEN
    scase = 'H4'
    CALL calch1a
    scase = 'H4'
    RETURN
  END IF
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi1 = w(2) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funch4a(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funch4a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    y2 = funch4a (psi6lo)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funch4a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCH4') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funch4a(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calch4




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCH4A
! *** CASE H4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funch4a ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    a9 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = max(psi5, tiny)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = bb*bb - 4.D0*cc
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(psi4, chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi1>tiny .AND. water>tiny) THEN ! NA2SO4 DISSOLUTION
      aa = psi7 + psi8
      bb = aa*aa
      cc = -a1/4.D0
      CALL poly3(aa, bb, cc, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    END IF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 ! SO4I
    molal(6) = zero
    molal(7) = psi5 + psi8 ! NO3I
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    cna2so4 = max(chi1-psi1, zero)
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    delt = min(gnh3, ghcl)
    bb = -(gnh3+ghcl)
    cc = gnh3*ghcl - a3
    dd = bb*bb - 4.D0*cc
    psi31 = 0.5D0*(-bb+sqrt(dd))
    psi32 = 0.5D0*(-bb-sqrt(dd))
    IF (delt-psi31>zero .AND. psi31>zero) THEN
      psi3 = psi31
    ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
      psi3 = psi32
    ELSE
      psi3 = zero
    END IF
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funch4a = molal(3)*molal(4)/ghcl/gnh3/a6/a4 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCH4A *******************************************
  !
END FUNCTION funch4a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCH3
! *** CASE H3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calch3 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF NITRATES ***********************
  !
  IF (w(4)<=tiny) THEN ! NO3 NOT EXIST, WATER NOT POSSIBLE
    scase = 'H3'
    CALL calch1a
    scase = 'H3'
    RETURN
  END IF
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi1 = w(2) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funch3a(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funch3a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funch3a (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funch3a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCH3') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funch3a(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calch3




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCH3A
! *** CASE H3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funch3a ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    a9 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = max(psi5, tiny)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = bb*bb - 4.D0*cc
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(psi4, chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      diak = (psi8-psi6)**2.D0 + 4.D0*a7
      psi7 = 0.5D0*(-(psi8+psi6)+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !
    IF (chi1>tiny .AND. water>tiny) THEN ! NA2SO4 DISSOLUTION
      aa = psi7 + psi8
      bb = aa*aa
      cc = -a1/4.D0
      CALL poly3(aa, bb, cc, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    END IF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 ! SO4I
    molal(6) = zero
    molal(7) = psi5 + psi8 ! NO3I
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    cna2so4 = max(chi1-psi1, zero)
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    delt = min(gnh3, ghcl)
    bb = -(gnh3+ghcl)
    cc = gnh3*ghcl - a3
    dd = bb*bb - 4.D0*cc
    psi31 = 0.5D0*(-bb+sqrt(dd))
    psi32 = 0.5D0*(-bb-sqrt(dd))
    IF (delt-psi31>zero .AND. psi31>zero) THEN
      psi3 = psi31
    ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
      psi3 = psi32
    ELSE
      psi3 = zero
    END IF
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funch3a = molal(3)*molal(4)/ghcl/gnh3/a6/a4 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCH3A *******************************************
  !
END FUNCTION funch3a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCH2
! *** CASE H2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : NH4Cl, NA2SO4, NANO3, NACL
!
!     THERE ARE THREE REGIMES IN THIS CASE:
!     1. NH4NO3(s) POSSIBLE. LIQUID & SOLID AEROSOL (SUBROUTINE CALCH2A)
!     2. NH4NO3(s) NOT POSSIBLE, AND RH < MDRH. SOLID AEROSOL ONLY
!     3. NH4NO3(s) NOT POSSIBLE, AND RH >= MDRH. (MDRH REGION)
!
!     REGIMES 2. AND 3. ARE CONSIDERED TO BE THE SAME AS CASES H1A, H2B
!     RESPECTIVELY (BECAUSE MDRH POINTS COINCIDE).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calch2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF NITRATES ***********************
  !
  IF (w(4)>tiny) THEN ! NO3 EXISTS, WATER POSSIBLE
    scase = 'H2 ; SUBCASE 1'
    CALL calch2a
    scase = 'H2 ; SUBCASE 1'
  ELSE ! NO3 NON EXISTANT, WATER NOT POSSIBLE
    scase = 'H2 ; SUBCASE 1'
    CALL calch1a
    scase = 'H2 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny .AND. rh<drmh2) THEN ! DRY AEROSOL
    scase = 'H2 ; SUBCASE 2'
    !
  ELSE IF (water<=tiny .AND. rh>=drmh2) THEN ! MDRH OF H2
    scase = 'H2 ; SUBCASE 3'
    CALL calcmdrh(rh, drmh2, drnano3, calch1a, calch3)
    scase = 'H2 ; SUBCASE 3'
  END IF
  !
  RETURN
  !
END SUBROUTINE calch2




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCH2A
! *** CASE H2 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calch2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi1 = w(2) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funch2a(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funch2a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (y2 > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funch2a (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funch2a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCH2A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funch2a(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calch2a




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCH2A
! *** CASE H2 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funch2a ( x )
  USE mo_isrpia
  USE mo_solut 
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: a64
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    a64 = (xk3*xk2/xkw)*(gama(10)/gama(5)/gama(11))**2.0
    a64 = a64*(r*temp*water)**2.0
    a9 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = max(psi5, tiny)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = bb*bb - 4.D0*cc
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(psi4, chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      diak = (psi8-psi6)**2.D0 + 4.D0*a7
      psi7 = 0.5D0*(-(psi8+psi6)+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !
    IF (chi8>tiny .AND. water>tiny) THEN ! NANO3 DISSOLUTION
      diak = (psi7-psi5)**2.D0 + 4.D0*a8
      psi8 = 0.5D0*(-(psi7+psi5)+sqrt(diak))
      psi8 = max(min(psi8,chi8), zero)
    END IF
    !
    IF (chi1>tiny .AND. water>tiny) THEN ! NA2SO4 DISSOLUTION
      aa = psi7 + psi8
      bb = aa*aa
      cc = -a1/4.D0
      CALL poly3(aa, bb, cc, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    END IF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 ! NO3I
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    cna2so4 = max(chi1-psi1, zero)
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    delt = min(gnh3, ghcl)
    bb = -(gnh3+ghcl)
    cc = gnh3*ghcl - a3
    dd = bb*bb - 4.D0*cc
    psi31 = 0.5D0*(-bb+sqrt(dd))
    psi32 = 0.5D0*(-bb-sqrt(dd))
    IF (delt-psi31>zero .AND. psi31>zero) THEN
      psi3 = psi31
    ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
      psi3 = psi32
    ELSE
      psi3 = zero
    END IF
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funch2a = molal(3)*molal(4)/ghcl/gnh3/a64 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCH2A *******************************************
  !
END FUNCTION funch2a


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCH1
! *** CASE H1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, NA2SO4
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCH1A)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calch1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmh1) THEN
    scase = 'H1 ; SUBCASE 1'
    CALL calch1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'H1 ; SUBCASE 1'
  ELSE
    scase = 'H1 ; SUBCASE 2' ! LIQUID & SOLID PHASE POSSIBLE
    CALL calcmdrh(rh, drmh1, drnh4no3, calch1a, calch2a)
    scase = 'H1 ; SUBCASE 2'
  END IF
  !
  RETURN
  !
END SUBROUTINE calch1


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCH1A
! *** CASE H1 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, NANO3, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calch1a (  )
  USE mo_isrpia  
  IMPLICIT NONE
  DOUBLE PRECISION :: a1
  DOUBLE PRECISION :: a2
  DOUBLE PRECISION :: alf
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bet
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: clfr
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: dd1
  DOUBLE PRECISION :: dd2
  DOUBLE PRECISION :: gam
  DOUBLE PRECISION :: kapa
  DOUBLE PRECISION :: kapa1
  DOUBLE PRECISION :: kapa2
  DOUBLE PRECISION :: lamda
  DOUBLE PRECISION :: lamda1
  DOUBLE PRECISION :: lamda2
  DOUBLE PRECISION :: nafr
  DOUBLE PRECISION :: no3fr
  DOUBLE PRECISION :: rtsq
  DOUBLE PRECISION :: sqdd
  DOUBLE PRECISION :: sqdd1
  DOUBLE PRECISION :: sqdd2
  DOUBLE PRECISION :: theta1
  DOUBLE PRECISION :: theta2
  !
  ! *** CALCULATE NON VOLATILE SOLIDS ***********************************
  !
  cna2so4 = w(2)
  cnh42s4 = zero
  nafr = max(w(1)-2*cna2so4, zero)
  cnano3 = min(nafr, w(4))
  no3fr = max(w(4)-cnano3, zero)
  cnacl = min(max(nafr-cnano3,zero), w(5))
  clfr = max(w(5)-cnacl, zero)
  !
  ! *** CALCULATE VOLATILE SPECIES **************************************
  !
  alf = w(3) ! FREE NH3
  bet = clfr ! FREE CL
  gam = no3fr ! FREE NO3
  !
  rtsq = r*temp*r*temp
  a1 = xk6/rtsq
  a2 = xk10/rtsq
  !
  theta1 = gam - bet*(a2/a1)
  theta2 = a2/a1
  !
  ! QUADRATIC EQUATION SOLUTION
  !
  bb = (theta1-alf-bet*(one+theta2))/(one+theta2)
  cc = (alf*bet-a1-bet*theta1)/(one+theta2)
  dd = bb*bb - 4.0D0*cc
  IF (dd<zero) GO TO 100 ! Solve each reaction seperately
  !
  ! TWO ROOTS FOR KAPA, CHECK AND SEE IF ANY VALID
  !
  sqdd = sqrt(dd)
  kapa1 = 0.5D0*(-bb+sqdd)
  kapa2 = 0.5D0*(-bb-sqdd)
  lamda1 = theta1 + theta2*kapa1
  lamda2 = theta1 + theta2*kapa2
  !
  IF (kapa1>=zero .AND. lamda1>=zero) THEN
    IF (alf-kapa1-lamda1>=zero .AND. bet-kapa1>=zero .AND.                 &
         gam-lamda1>=zero) THEN
      kapa = kapa1
      lamda = lamda1
      GO TO 110
    END IF
  END IF
  !
  IF (kapa2>=zero .AND. lamda2>=zero) THEN
    IF (alf-kapa2-lamda2>=zero .AND. bet-kapa2>=zero .AND.                 &
         gam-lamda2>=zero) THEN
      kapa = kapa2
      lamda = lamda2
      GO TO 110
    END IF
  END IF
  !
  ! SEPERATE SOLUTION OF NH4CL & NH4NO3 EQUILIBRIA
  !
100 kapa = zero
  lamda = zero
  dd1 = (alf+bet)*(alf+bet) - 4.0D0*(alf*bet-a1)
  dd2 = (alf+gam)*(alf+gam) - 4.0D0*(alf*gam-a2)
  !
  ! NH4CL EQUILIBRIUM
  !
  IF (dd1>=zero) THEN
    sqdd1 = sqrt(dd1)
    kapa1 = 0.5D0*(alf+bet+sqdd1)
    kapa2 = 0.5D0*(alf+bet-sqdd1)
    !
    IF (kapa1>=zero .AND. kapa1<=min(alf,bet)) THEN
      kapa = kapa1
    ELSE IF (kapa2>=zero .AND. kapa2<=min(alf,bet)) THEN
      kapa = kapa2
    ELSE
      kapa = zero
    END IF
  END IF
  !
  ! NH4NO3 EQUILIBRIUM
  !
  IF (dd2>=zero) THEN
    sqdd2 = sqrt(dd2)
    lamda1 = 0.5D0*(alf+gam+sqdd2)
    lamda2 = 0.5D0*(alf+gam-sqdd2)
    !
    IF (lamda1>=zero .AND. lamda1<=min(alf,gam)) THEN
      lamda = lamda1
    ELSE IF (lamda2>=zero .AND. lamda2<=min(alf,gam)) THEN
      lamda = lamda2
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! IF BOTH KAPA, LAMDA ARE > 0, THEN APPLY EXISTANCE CRITERION
  !
  IF (kapa>zero .AND. lamda>zero) THEN
    IF (bet<lamda/theta1) THEN
      kapa = zero
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! *** CALCULATE COMPOSITION OF VOLATILE SPECIES ***********************
  !
110 CONTINUE
  cnh4no3 = lamda
  cnh4cl = kapa
  !
  gnh3 = alf - kapa - lamda
  ghno3 = gam - lamda
  ghcl = bet - kapa
  !
  RETURN
  !
END SUBROUTINE calch1a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCI6
! *** CASE I6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calci6 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calci1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCI1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = clc
  psi3 = cnahso4
  psi4 = cna2so4
  psi5 = cnh42s4
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a6 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    bb = psi2 + psi4 + psi5 + a6 ! PSI6
    cc = -a6*(psi2+psi3+psi1)
    dd = bb*bb - 4.D0*cc
    psi6 = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = psi6 ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 ! SO4I
    molal(6) = psi2 + psi3 + psi1 - psi6 ! HSO4I
    clc = zero
    cnahso4 = zero
    cna2so4 = chi4 - psi4
    cnh42s4 = zero
    cnh4hs4 = zero
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
100 RETURN
  !
END SUBROUTINE calci6

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCI5
! *** CASE I5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calci5 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calci1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCI1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = clc
  psi3 = cnahso4
  psi4 = zero
  psi5 = cnh42s4
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi4lo = zero ! Low  limit
  psi4hi = chi4 ! High limit
  !
  ! *** IF NA2SO4(S) =0, CALL FUNCI5B FOR Y4=0 ***************************
  !
  IF (chi4<=tiny) THEN
    y1 = funci5a(zero)
    GO TO 120
  END IF
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi4hi
  y1 = funci5a(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NA2SO4 **
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    y2 = funci5a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NH4CL
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funci5a(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCI5') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funci5a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCI5') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funci5a(x3)
  !
120 RETURN

END SUBROUTINE calci5




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCI5A
! *** CASE I5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funci5a ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi4 = p4 ! PSI3 already assigned in FUNCI5A
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = xk5*(water/gama(2))**3.0
    a5 = xk7*(water/gama(4))**3.0
    a6 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    bb = psi2 + psi4 + psi5 + a6 ! PSI6
    cc = -a6*(psi2+psi3+psi1)
    dd = bb*bb - 4.D0*cc
    psi6 = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = psi6 ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 ! SO4I
    molal(6) = psi2 + psi3 + psi1 - psi6 ! HSO4I
    clc = zero
    cnahso4 = zero
    cna2so4 = chi4 - psi4
    cnh42s4 = zero
    cnh4hs4 = zero
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a4 = xk5*(water/gama(2))**3.0
  funci5a = molal(5)*molal(2)*molal(2)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCI5A ********************************************
  !
END FUNCTION funci5a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCI4
! *** CASE I4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calci4 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calci1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCI1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = clc
  psi3 = cnahso4
  psi4 = zero
  psi5 = zero
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi4lo = zero ! Low  limit
  psi4hi = chi4 ! High limit
  !
  ! *** IF NA2SO4(S) =0, CALL FUNCI4B FOR Y4=0 ***************************
  !
  IF (chi4<=tiny) THEN
    y1 = funci4a(zero)
    GO TO 120
  END IF
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi4hi
  y1 = funci4a(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NA2SO4 **
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    y2 = funci4a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NH4CL
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funci4a(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCI4') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funci4a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCI4') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funci4a(x3)
  !
120 RETURN

END SUBROUTINE calci4




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCI4A
! *** CASE I4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funci4a ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi4 = p4 ! PSI3 already assigned in FUNCI4A
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = xk5*(water/gama(2))**3.0
    a5 = xk7*(water/gama(4))**3.0
    a6 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    a7 = sqrt(a4/a5)
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    bb = psi2 + psi4 + psi5 + a6 ! PSI6
    cc = -a6*(psi2+psi3+psi1)
    dd = bb*bb - 4.D0*cc
    psi6 = 0.5D0*(-bb+sqrt(dd))
    !
    psi5 = (psi3+2.D0*psi4-a7*(3.D0*psi2+psi1))/2.D0/a7
    psi5 = max (min (psi5, chi5), zero)
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = psi6 ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 ! SO4I
    molal(6) = psi2 + psi3 + psi1 - psi6 ! HSO4I
    clc = zero
    cnahso4 = zero
    cna2so4 = chi4 - psi4
    cnh42s4 = chi5 - psi5
    cnh4hs4 = zero
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a4 = xk5*(water/gama(2))**3.0
  funci4a = molal(5)*molal(2)*molal(2)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCI4A ********************************************
  !
END FUNCTION funci4a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCI3
! *** CASE I3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4, NAHSO4, LC
!
!     THERE ARE THREE REGIMES IN THIS CASE:
!     1.(NA,NH4)HSO4(s) POSSIBLE. LIQUID & SOLID AEROSOL (SUBROUTINE CALCI3A)
!     2.(NA,NH4)HSO4(s) NOT POSSIBLE, AND RH < MDRH. SOLID AEROSOL ONLY
!     3.(NA,NH4)HSO4(s) NOT POSSIBLE, AND RH >= MDRH. SOLID & LIQUID AEROSOL
!
!     REGIMES 2. AND 3. ARE CONSIDERED TO BE THE SAME AS CASES I1A, I2B
!     RESPECTIVELY
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calci3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calci1a
  !
  ! *** REGIME DEPENDS UPON THE POSSIBLE SOLIDS & RH **********************
  !
  IF (cnh4hs4>tiny .OR. cnahso4>tiny) THEN
    scase = 'I3 ; SUBCASE 1'
    CALL calci3a ! FULL SOLUTION
    scase = 'I3 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmi3) THEN ! SOLID SOLUTION
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calci1a
      scase = 'I3 ; SUBCASE 2'
      !
    ELSE IF (rh>=drmi3) THEN ! MDRH OF I3
      scase = 'I3 ; SUBCASE 3'
      CALL calcmdrh(rh, drmi3, drlc, calci1a, calci4)
      scase = 'I3 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calci3



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCI3A
! *** CASE I3 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4, LC
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calci3a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi2hi
  DOUBLE PRECISION :: psi2lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calci1a ! Needed when called from CALCMDRH
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCI1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = zero
  psi3 = cnahso4
  psi4 = zero
  psi5 = zero
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi2lo = zero ! Low  limit
  psi2hi = chi2 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi2hi
  y1 = funci3a(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH LC *********
  !
  IF (yhi<eps) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi2hi-psi2lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, psi2lo)
    y2 = funci3a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH LC
  !
  IF (y2 > eps) THEN 
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funci3a (zero)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funci3a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCI3A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funci3a(x3)
  !
120 RETURN

END SUBROUTINE calci3a
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCI3A
! *** CASE I3 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4, LC
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funci3a ( p2 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p2
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi2 = p2 ! Save PSI2 in COMMON BLOCK
  psi4lo = zero ! Low  limit for PSI4
  psi4hi = chi4 ! High limit for PSI4
  !
  ! *** IF NH3 =0, CALL FUNCI3B FOR Y4=0 ********************************
  !
  IF (chi4<=tiny) THEN
    funci3a = funci3b(zero)
    GO TO 120
  END IF
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi4hi
  y1 = funci3b(x1)
  IF (abs(y1)<=eps) GO TO 120
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NA2SO4 *****
  !
  IF (yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, psi4lo)
    y2 = funci3b(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NA2SO4
  !
  IF (y2 > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funci3b (psi4lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funci3b(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0004, 'FUNCI3A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** INNER LOOP CONVERGED **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funci3b(x3)
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
120 a2 = xk13*(water/gama(13))**5.0
  funci3a = molal(5)*molal(6)*molal(3)**3.D0/a2 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCI3A *******************************************
  !
END FUNCTION funci3a



!=======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION FUNCI3B
! *** CASE I3 ; SUBCASE 2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4, LC
!
!     SOLUTION IS SAVED IN COMMON BLOCK /CASE/
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funci3b ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi4 = p4
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = xk5*(water/gama(2))**3.0
    a5 = xk7*(water/gama(4))**3.0
    a6 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    a7 = sqrt(a4/a5)
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    bb = psi2 + psi4 + psi5 + a6 ! PSI6
    cc = -a6*(psi2+psi3+psi1)
    dd = bb*bb - 4.D0*cc
    psi6 = 0.5D0*(-bb+sqrt(dd))
    !
    psi5 = (psi3+2.D0*psi4-a7*(3.D0*psi2+psi1))/2.D0/a7
    psi5 = max (min (psi5, chi5), zero)
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = psi6 ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 ! SO4I
    molal(6) = max(psi2+psi3+psi1-psi6, tiny) ! HSO4I
    clc = max(chi2-psi2, zero)
    cnahso4 = zero
    cna2so4 = max(chi4-psi4, zero)
    cnh42s4 = max(chi5-psi5, zero)
    cnh4hs4 = zero
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a4 = xk5*(water/gama(2))**3.0
  funci3b = molal(5)*molal(2)*molal(2)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCI3B ********************************************
  !
END FUNCTION funci3b
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCI2
! *** CASE I2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4, NAHSO4, LC
!
!     THERE ARE THREE REGIMES IN THIS CASE:
!     1. NH4HSO4(s) POSSIBLE. LIQUID & SOLID AEROSOL (SUBROUTINE CALCI2A)
!     2. NH4HSO4(s) NOT POSSIBLE, AND RH < MDRH. SOLID AEROSOL ONLY
!     3. NH4HSO4(s) NOT POSSIBLE, AND RH >= MDRH. SOLID & LIQUID AEROSOL
!
!     REGIMES 2. AND 3. ARE CONSIDERED TO BE THE SAME AS CASES I1A, I2B
!     RESPECTIVELY
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calci2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calci1a
  !
  ! *** REGIME DEPENDS UPON THE POSSIBLE SOLIDS & RH **********************
  !
  IF (cnh4hs4>tiny) THEN
    scase = 'I2 ; SUBCASE 1'
    CALL calci2a
    scase = 'I2 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmi2) THEN ! SOLID SOLUTION ONLY
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calci1a
      scase = 'I2 ; SUBCASE 2'
      !
    ELSE IF (rh>=drmi2) THEN ! MDRH OF I2
      scase = 'I2 ; SUBCASE 3'
      CALL calcmdrh(rh, drmi2, drnahso4, calci1a, calci3a)
      scase = 'I2 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calci2


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCI2A
! *** CASE I2 ; SUBCASE A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4, NAHSO4, LC
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calci2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi2hi
  DOUBLE PRECISION :: psi2lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calci1a ! Needed when called from CALCMDRH
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCI1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = zero
  psi3 = zero
  psi4 = zero
  psi5 = zero
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi2lo = zero ! Low  limit
  psi2hi = chi2 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi2hi
  y1 = funci2a(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH LC *********
  !
  IF (yhi<eps) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi2hi-psi2lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, psi2lo)
    y2 = funci2a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH LC
  !
  IF (y2 >  eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funci2a (zero)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funci2a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCI2A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funci2a(x3)
  !
120 RETURN

END SUBROUTINE calci2a




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCI2A
! *** CASE I2 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4, NAHSO4, LC
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funci2a ( p2 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p2
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  INTEGER :: islv
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  psi2 = p2 ! Save PSI2 in COMMON BLOCK
  psi3 = chi3
  psi4 = chi4
  psi5 = chi5
  psi6 = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a3 = xk11*(water/gama(12))**2.0
    a4 = xk5*(water/gama(2))**3.0
    a5 = xk7*(water/gama(4))**3.0
    a6 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    a7 = sqrt(a4/a5)
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    IF (chi5>tiny .AND. water>tiny) THEN
      psi5 = (psi3+2.D0*psi4-a7*(3.D0*psi2+psi1))/2.D0/a7
      psi5 = max(min(psi5,chi5), tiny)
    END IF
    !
    IF (chi4>tiny .AND. water>tiny) THEN
      aa = psi2 + psi5 + psi6 + psi3
      bb = psi3*aa
      cc = 0.25D0*(psi3*psi3*(psi2+psi5+psi6)-a4)
      CALL poly3(aa, bb, cc, psi4, islv)
      IF (islv==0) THEN
        psi4 = min(psi4, chi4)
      ELSE
        psi4 = zero
      END IF
    END IF
    !
    IF (chi3>tiny .AND. water>tiny) THEN
      aa = 2.D0*psi4 + psi2 + psi1 - psi6
      bb = 2.D0*psi4*(psi2+psi1-psi6) - a3
      cc = zero
      CALL poly3(aa, bb, cc, psi3, islv)
      IF (islv==0) THEN
        psi3 = min(psi3, chi3)
      ELSE
        psi3 = zero
      END IF
    END IF
    !
    bb = psi2 + psi4 + psi5 + a6 ! PSI6
    cc = -a6*(psi2+psi3+psi1)
    dd = bb*bb - 4.D0*cc
    psi6 = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = psi6 ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 ! SO4I
    molal(6) = psi2 + psi3 + psi1 - psi6 ! HSO4I
    clc = chi2 - psi2
    cnahso4 = chi3 - psi3
    cna2so4 = chi4 - psi4
    cnh42s4 = chi5 - psi5
    cnh4hs4 = zero
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 a2 = xk13*(water/gama(13))**5.0
  funci2a = molal(5)*molal(6)*molal(3)**3.D0/a2 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCI2A *******************************************
  !
END FUNCTION funci2a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCI1
! *** CASE I1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, NA2SO4
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCI1A)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calci1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmi1) THEN
    scase = 'I1 ; SUBCASE 1'
    CALL calci1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'I1 ; SUBCASE 1'
  ELSE
    scase = 'I1 ; SUBCASE 2' ! LIQUID & SOLID PHASE POSSIBLE
    CALL calcmdrh(rh, drmi1, drnh4hs4, calci1a, calci2a)
    scase = 'I1 ; SUBCASE 2'
  END IF
  !
  ! *** AMMONIA IN GAS PHASE **********************************************
  !
  !      CALL CALCNH3
  !
  RETURN
  !
END SUBROUTINE calci1


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCI1A
! *** CASE I1 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4HSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calci1a (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: frnh4
  DOUBLE PRECISION :: frso4
  !
  ! *** CALCULATE NON VOLATILE SOLIDS ***********************************
  !
  cna2so4 = 0.5D0*w(1)
  cnh4hs4 = zero
  cnahso4 = zero
  cnh42s4 = zero
  frso4 = max(w(2)-cna2so4, zero)
  !
  clc = min(w(3)/3.D0, frso4/2.D0)
  frso4 = max(frso4-2.D0*clc, zero)
  frnh4 = max(w(3)-3.D0*clc, zero)
  !
  IF (frso4<=tiny) THEN
    clc = max(clc-frnh4, zero)
    cnh42s4 = 2.D0*frnh4

  ELSE IF (frnh4<=tiny) THEN
    cnh4hs4 = 3.D0*min(frso4, clc)
    clc = max(clc-frso4, zero)
    IF (cna2so4>tiny) THEN
      frso4 = max(frso4-cnh4hs4/3.D0, zero)
      cnahso4 = 2.D0*frso4
      cna2so4 = max(cna2so4-frso4, zero)
    END IF
  END IF
  !
  ! *** CALCULATE GAS SPECIES *********************************************
  !
  ghno3 = w(4)
  ghcl = w(5)
  gnh3 = zero
  !
  RETURN
  !
END SUBROUTINE calci1a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCJ3
! *** CASE J3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, FREE ACID (SULRAT < 1.0)
!     2. THERE IS ONLY A LIQUID PHASE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcj3 (  )
  USE mo_isrpia  
  IMPLICIT NONE
  DOUBLE PRECISION :: a3
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: chi1
  DOUBLE PRECISION :: chi2
  DOUBLE PRECISION :: dd
  INTEGER :: i
  DOUBLE PRECISION :: kapa
  DOUBLE PRECISION :: lamda
  DOUBLE PRECISION :: psi1
  DOUBLE PRECISION :: psi2
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  frst = .TRUE.
  calain = .TRUE.
  !
  lamda = max(w(2)-w(3)-w(1), tiny) ! FREE H2SO4
  chi1 = w(1) ! NA TOTAL as NaHSO4
  chi2 = w(3) ! NH4 TOTAL as NH4HSO4
  psi1 = chi1
  psi2 = chi2 ! ALL NH4HSO4 DELIQUESCED
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a3 = xk1*water/gama(7)*(gama(8)/gama(7))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    bb = a3 + lamda ! KAPA
    cc = -a3*(lamda+psi1+psi2)
    dd = bb*bb - 4.D0*cc
    kapa = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda + kapa ! HI
    molal(2) = psi1 ! NAI
    molal(3) = psi2 ! NH4I
    molal(4) = zero ! CLI
    molal(5) = kapa ! SO4I
    molal(6) = lamda + psi1 + psi2 - kapa ! HSO4I
    molal(7) = zero ! NO3I
    !
    cnahso4 = zero
    cnh4hs4 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
100 RETURN
  !
END SUBROUTINE calcj3
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCJ2
! *** CASE J2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, FREE ACID (SULRAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NAHSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcj2 (  )
  USE mo_isrpia
  USE mo_casej
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi1hi
  DOUBLE PRECISION :: psi1lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  chi1 = w(1) ! NA TOTAL
  chi2 = w(3) ! NH4 TOTAL
  psi1lo = tiny ! Low  limit
  psi1hi = chi1 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi1hi
  y1 = funcj2(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NH42SO4 ****
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi1hi-psi1lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    y2 = funcj2(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NH42SO4
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcj2(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCJ2') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcj2(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCJ2') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcj2(x3)
  !
120 RETURN
  !
END SUBROUTINE calcj2




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCJ2
! *** CASE J2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, FREE ACID (SULRAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcj2 ( p1 )
  USE mo_isrpia
  USE mo_casej
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p1
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  lamda = max(w(2)-w(3)-w(1), tiny) ! FREE H2SO4
  psi1 = p1
  psi2 = chi2 ! ALL NH4HSO4 DELIQUESCED
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk11*(water/gama(12))**2.0
    a3 = xk1*water/gama(7)*(gama(8)/gama(7))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    bb = a3 + lamda ! KAPA
    cc = -a3*(lamda+psi1+psi2)
    dd = bb*bb - 4.D0*cc
    kapa = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda + kapa ! HI
    molal(2) = psi1 ! NAI
    molal(3) = psi2 ! NH4I
    molal(4) = zero ! CLI
    molal(5) = kapa ! SO4I
    molal(6) = lamda + psi1 + psi2 - kapa ! HSO4I
    molal(7) = zero ! NO3I
    !
    cnahso4 = max(chi1-psi1, zero)
    cnh4hs4 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 funcj2 = molal(2)*molal(6)/a1 - one
  !
  ! *** END OF FUNCTION FUNCJ2 *******************************************
  !
END FUNCTION funcj2

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCJ1
! *** CASE J1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, FREE ACID (SULRAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NH4HSO4, NAHSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
SUBROUTINE calcj1 (  )
  USE mo_isrpia
  USE mo_casej
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi1hi
  DOUBLE PRECISION :: psi1lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  chi1 = w(1) ! Total NA initially as NaHSO4
  chi2 = w(3) ! Total NH4 initially as NH4HSO4
  !
  psi1lo = tiny ! Low  limit
  psi1hi = chi1 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi1hi
  y1 = funcj1(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NH42SO4 ****
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi1hi-psi1lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    y2 = funcj1(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NH42SO4
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcj1(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCJ1') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcj1(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCJ1') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcj1(x3)
  !
120 RETURN
  !
END SUBROUTINE calcj1




!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCJ1
! *** CASE J1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, FREE ACID (SULRAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
! *** UPDATED BY CHRISTOS FOUNTOUKIS
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcj1 ( p1 )
  USE mo_isrpia
  USE mo_casej
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p1
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  lamda = max(w(2)-w(3)-w(1), tiny) ! FREE H2SO4
  psi1 = p1
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk11*(water/gama(12))**2.0
    a2 = xk12*(water/gama(09))**2.0
    a3 = xk1*water/gama(7)*(gama(8)/gama(7))**2.0
    !
    psi2 = 0.5*(-(lamda+psi1)+sqrt((lamda+psi1)**2.D0+4.D0*a2)) ! PSI2
    psi2 = min(psi2, chi2)
    !
    bb = a3 + lamda ! KAPA
    cc = -a3*(lamda+psi2+psi1)
    dd = bb*bb - 4.D0*cc
    kapa = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = lamda + kapa ! HI
    molal(2) = psi1 ! NAI
    molal(3) = psi2 ! NH4I
    molal(4) = zero
    molal(5) = kapa ! SO4I
    molal(6) = lamda + psi1 + psi2 - kapa ! HSO4I
    molal(7) = zero
    !
    cnahso4 = max(chi1-psi1, zero)
    cnh4hs4 = max(chi2-psi2, zero)
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 funcj1 = molal(2)*molal(6)/a1 - one
  !
  ! *** END OF FUNCTION FUNCJ1 *******************************************
  !
END FUNCTION funcj1
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO7
! *** CASE O7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4, MgSO4, NA2SO4, K2SO4
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco7 (  )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  INTEGER :: i
  INTEGER :: nafr
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi9 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi9, zero)
  cafr = max(w(6)-chi9, zero)
  chi7 = min(0.5D0*w(7), so4fr) ! CK2SO4
  frk = max(w(7)-2.D0*chi7, zero)
  so4fr = max(so4fr-chi7, zero)
  chi1 = min(0.5D0*w(1), so4fr) ! NA2SO4
  nafr = max(w(1)-2.D0*chi1, zero)
  so4fr = max(so4fr-chi1, zero)
  chi8 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi8, zero)
  so4fr = max(so4fr-chi8, zero)
  chi3 = zero
  chi5 = w(4)
  chi6 = w(5)
  chi2 = max(so4fr, zero)
  chi4 = max(w(3)-2.D0*chi2, zero)
  !
  psi1 = chi1
  psi2 = chi2
  psi3 = zero
  psi4 = zero
  psi5 = zero
  psi6 = zero
  psi7 = chi7
  psi8 = chi8
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  water = chi2/m0(4) + chi1/m0(2) + chi7/m0(17) + chi8/m0(21)
  water = max(water, tiny)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funco7(x1)
  IF (chi6<=tiny) GO TO 120
  !cc      IF (ABS(Y1).LE.EPS .OR. CHI6.LE.TINY) GOTO 50
  !cc      IF (WATER .LE. TINY) RETURN                    ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funco7(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funco7 (psi6lo)
  END IF 
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funco7(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCO7') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funco7(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN ! If quadrat.called
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calco7
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCO7
! *** CASE O7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4, MgSO4, NA2SO4, K2SO4
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funco7 ( x )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    !
    !
    IF (chi5>=tiny) THEN
      psi5 = psi6*chi5/(a6/a5*(chi6-psi6)+psi6)
      psi5 = min(psi5, chi5)
    ELSE
      psi5 = tiny
    END IF
    !
    !CC      IF(CHI4.GT.TINY) THEN
    IF (w(2)>tiny) THEN ! Accounts for NH3 evaporation
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6) - 2.D0*psi2/a4
      dd = max(bb*bb-4.D0*cc, zero) ! Patch proposed by Uma Shankar, 19/11/01
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = max(min(psi4,chi4), zero)
    ELSE
      psi4 = tiny
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(2) = 2.0D0*psi1 ! Na+
    molal(3) = 2.0D0*psi2 + psi4 ! NH4I
    molal(4) = psi6 ! CLI
    molal(5) = psi1 + psi2 + psi7 + psi8 ! SO4I
    molal(6) = zero ! HSO4
    molal(7) = psi5 ! NO3I
    molal(8) = zero ! CaI
    molal(9) = 2.0D0*psi7 ! KI
    molal(10) = psi8 ! Mg
    !
    ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
    !
    !CC      MOLAL (1) = MAX(CHI5 - PSI5, TINY)*A5/PSI5   ! HI
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cna2so4 = zero
    cnh42s4 = zero
    cnh4no3 = zero
    cnh4cl = zero
    ck2so4 = zero
    cmgso4 = zero
    ccaso4 = chi9
    !
    ! *** CALCULATE MOLALR ARRAY, WATER AND ACTIVITIES **********************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funco7 = molal(1)*molal(4)/ghcl/a6 - one
  !CC         FUNCG5A = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCO7 *******************************************
  !
END FUNCTION funco7
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO6
! *** CASE O6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4, MGSO4, NA2SO4
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco6 (  )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  INTEGER :: i
  INTEGER :: nafr
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi9 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi9, zero)
  cafr = max(w(6)-chi9, zero)
  chi7 = min(0.5D0*w(7), so4fr) ! CK2SO4
  frk = max(w(7)-2.D0*chi7, zero)
  so4fr = max(so4fr-chi7, zero)
  chi1 = min(0.5D0*w(1), so4fr) ! NA2SO4
  nafr = max(w(1)-2.D0*chi1, zero)
  so4fr = max(so4fr-chi1, zero)
  chi8 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi8, zero)
  so4fr = max(so4fr-chi8, zero)
  chi3 = zero
  chi5 = w(4)
  chi6 = w(5)
  chi2 = max(so4fr, zero)
  chi4 = max(w(3)-2.D0*chi2, zero)
  !
  !
  psi1 = chi1
  psi2 = chi2
  psi3 = zero
  psi7 = zero
  psi8 = chi8
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  water = chi2/m0(4) + chi1/m0(2) + chi7/m0(17) + chi8/m0(21)
  water = max(water, tiny)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funco6(x1)
  IF (chi6<=tiny) GO TO 120
  !cc      IF (ABS(Y1).LE.EPS .OR. CHI6.LE.TINY) GOTO 50
  !cc      IF (WATER .LE. TINY) RETURN                    ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funco6(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funco6 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funco6(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCO6') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funco6(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN ! If quadrat.called
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calco6
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCO6
! *** CASE O6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4 , K2SO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4, MgSO4, NA2SO4
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funco6 ( x )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk17*(water/gama(17))**3.0
    !
    !
    IF (chi5>=tiny) THEN
      psi5 = psi6*chi5/(a6/a5*(chi6-psi6)+psi6)
      psi5 = min(psi5, chi5)
    ELSE
      psi5 = tiny
    END IF
    !
    !CC      IF(CHI4.GT.TINY) THEN
    IF (w(2)>tiny) THEN ! Accounts for NH3 evaporation
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6) - 2.D0*psi2/a4
      dd = max(bb*bb-4.D0*cc, zero) ! Patch proposed by Uma Shankar, 19/11/01
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = max(min(psi4,chi4), zero)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! PSI7
      CALL poly3(psi1+psi2+psi8, zero, -a7/4.D0, psi7, islv)
      IF (islv==0) THEN
        psi7 = max(min(psi7,chi7), zero)
      ELSE
        psi7 = zero
      END IF
    ELSE
      psi7 = zero
    END IF
    !
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(2) = 2.0D0*psi1 ! Na+
    molal(3) = 2.0D0*psi2 + psi4 ! NH4I
    molal(4) = psi6 ! CLI
    molal(5) = psi1 + psi2 + psi7 + psi8 ! SO4I
    molal(6) = zero ! HSO4
    molal(7) = psi5 ! NO3I
    molal(8) = zero ! CaI
    molal(9) = 2.0D0*psi7 ! KI
    molal(10) = psi8 ! Mg

    !
    ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************

    !
    !CC      MOLAL (1) = MAX(CHI5 - PSI5, TINY)*A5/PSI5   ! HI
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cna2so4 = zero
    cnh42s4 = zero
    cnh4no3 = zero
    cnh4cl = zero
    ck2so4 = max(chi7-psi7, tiny)
    cmgso4 = zero
    ccaso4 = chi9
    !
    ! *** CALCULATE MOLALR ARRAY, WATER AND ACTIVITIES **********************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funco6 = molal(1)*molal(4)/ghcl/a6 - one
  !CC         FUNCG5A = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCO6 *******************************************
  !
END FUNCTION funco6
!
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO5
! *** CASE O5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4, MGSO4
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco5 (  )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  INTEGER :: i
  INTEGER :: nafr
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi9 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi9, zero)
  cafr = max(w(6)-chi9, zero)
  chi7 = min(0.5D0*w(7), so4fr) ! CK2SO4
  frk = max(w(7)-2.D0*chi7, zero)
  so4fr = max(so4fr-chi7, zero)
  chi1 = min(0.5D0*w(1), so4fr) ! NA2SO4
  nafr = max(w(1)-2.D0*chi1, zero)
  so4fr = max(so4fr-chi1, zero)
  chi8 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi8, zero)
  so4fr = max(so4fr-chi8, zero)
  chi3 = zero
  chi5 = w(4)
  chi6 = w(5)
  chi2 = max(so4fr, zero)
  chi4 = max(w(3)-2.D0*chi2, zero)
  !
  psi1 = zero
  psi2 = chi2
  psi3 = zero
  psi7 = zero
  psi8 = chi8
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  water = chi2/m0(4) + chi1/m0(2) + chi7/m0(17) + chi8/m0(21)
  water = max(water, tiny)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funco5(x1)
  IF (chi6<=tiny) GO TO 120
  !cc      IF (ABS(Y1).LE.EPS .OR. CHI6.LE.TINY) GOTO 50
  !cc      IF (WATER .LE. TINY) RETURN                    ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funco5(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funco5 (psi6lo)
  END IF

  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    x3 = 0.5*(x1+x2)
    y3 = funco5(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCO5') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funco5(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN ! If quadrat.called
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calco5
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCO5
! *** CASE O5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4, MGSO4
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funco5 ( x )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk17*(water/gama(17))**3.0
    !
    !
    IF (chi5>=tiny) THEN
      psi5 = psi6*chi5/(a6/a5*(chi6-psi6)+psi6)
      psi5 = min(psi5, chi5)
    ELSE
      psi5 = tiny
    END IF
    !
    !CC      IF(CHI4.GT.TINY) THEN
    IF (w(2)>tiny) THEN ! Accounts for NH3 evaporation
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6) - 2.D0*psi2/a4
      dd = max(bb*bb-4.D0*cc, zero) ! Patch proposed by Uma Shankar, 19/11/01
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = max(min(psi4,chi4), zero)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! PSI7
      CALL poly3((psi2+psi8)/(sqrt(a1/a7)+1.D0), zero, -a7/4.D0/(sqrt(a1/  &
           a7)+1.D0), psi7, islv)
      IF (islv==0) THEN
        psi7 = max(min(psi7,chi7), zero)
      ELSE
        psi7 = zero
      END IF
    ELSE
      psi7 = zero
    END IF
    !
    IF (chi1>=tiny) THEN ! PSI1
      psi1 = sqrt(a1/a7)*psi7
      psi1 = min(psi1, chi1)
    ELSE
      psi1 = zero
    END IF
    !
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(2) = 2.0D0*psi1 ! NaI
    molal(3) = 2.0D0*psi2 + psi4 ! NH4I
    molal(4) = psi6 ! CLI
    molal(5) = psi1 + psi2 + psi7 + psi8 ! SO4I
    molal(6) = zero ! HSO4
    molal(7) = psi5 ! NO3I
    molal(8) = zero ! CaI
    molal(9) = 2.0D0*psi7 ! KI
    molal(10) = psi8 ! Mg

    !
    ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************

    !
    !CC      MOLAL (1) = MAX(CHI5 - PSI5, TINY)*A5/PSI5   ! HI
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cna2so4 = max(chi1-psi1, tiny)
    cnh42s4 = zero
    cnh4no3 = zero
    cnh4cl = zero
    ck2so4 = max(chi7-psi7, tiny)
    cmgso4 = zero
    ccaso4 = chi9
    !
    ! *** CALCULATE MOLALR ARRAY, WATER AND ACTIVITIES **********************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funco5 = molal(1)*molal(4)/ghcl/a6 - one
  !CC         FUNCG5A = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCO5 *******************************************
  !
END FUNCTION funco5
!
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO4
! *** CASE O4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NA2SO4, K2SO4, MGSO4, CASO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco4 (  )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: nafr
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi9 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi9, zero)
  cafr = max(w(6)-chi9, zero)
  chi7 = min(0.5D0*w(7), so4fr) ! CK2SO4
  frk = max(w(7)-2.D0*chi7, zero)
  so4fr = max(so4fr-chi7, zero)
  chi1 = min(0.5D0*w(1), so4fr) ! NA2SO4
  nafr = max(w(1)-2.D0*chi1, zero)
  so4fr = max(so4fr-chi1, zero)
  chi8 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi8, zero)
  so4fr = max(so4fr-chi8, zero)
  chi3 = zero
  chi5 = w(4)
  chi6 = w(5)
  chi2 = max(so4fr, zero)
  chi4 = max(w(3)-2.D0*chi2, zero)
  !
  psi2 = chi2
  psi3 = zero
  psi7 = zero
  psi8 = chi8
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  water = chi2/m0(4) + chi1/m0(2) + chi7/m0(17) + chi8/m0(21)
  water = max(water, tiny)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funco4(x1)
  IF (chi6<=tiny) GO TO 120
  !CC      IF (ABS(Y1).LE.EPS .OR. CHI6.LE.TINY) GOTO 50
  !CC      IF (WATER .LE. TINY) GOTO 50               ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funco4(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funco4 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funco4(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCO4') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funco4(x3)
  !
  ! *** FINAL CALCULATIONS **********************************************
  !
120 CONTINUE
  !
  ! *** Na2SO4 DISSOLUTION
  !
  IF (chi1>tiny .AND. water>tiny) THEN ! PSI1
    CALL poly3(psi2+psi7+psi8, zero, -a1/4.D0, psi1, islv)
    IF (islv==0) THEN
      psi1 = min(psi1, chi1)
    ELSE
      psi1 = zero
    END IF
  ELSE
    psi1 = zero
  END IF
  molal(2) = 2.0D0*psi1 ! Na+  EFFECT
  molal(5) = molal(5) + psi1 ! SO4  EFFECT
  cna2so4 = max(chi1-psi1, zero) ! NA2SO4(s) depletion
  !
  ! *** HSO4 equilibrium
  !

  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   AFFECT
    molal(5) = molal(5) - delta ! SO4  AFFECT
    molal(6) = delta ! HSO4 AFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calco4
!
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCO4
! *** CASE O4 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funco4 ( x )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi2 = chi2
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk17*(water/gama(17))**3.0
    !      A8  = XK23 *(WATER/GAMA(21))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    IF (chi5>=tiny) THEN
      psi5 = psi6*chi5/(a6/a5*(chi6-psi6)+psi6)
      psi5 = min(psi5, chi5)
    ELSE
      psi5 = tiny
    END IF
    !
    !CC      IF(CHI4.GT.TINY) THEN
    IF (w(2)>tiny) THEN ! Accounts for NH3 evaporation
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6) - 2.D0*psi2/a4
      dd = max(bb*bb-4.D0*cc, zero) ! Patch proposed by Uma Shankar, 19/11/01
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = max(min(psi4,chi4), zero)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! PSI7
      CALL poly3(psi2+psi8, zero, -a7/4.D0, psi7, islv)
      IF (islv==0) THEN
        psi7 = max(min(psi7,chi7), zero)
      ELSE
        psi7 = zero
      END IF
    ELSE
      psi7 = zero
    END IF
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    molal(2) = zero ! NAI
    molal(3) = 2.0D0*psi2 + psi4 ! NH4I
    molal(4) = psi6 ! CLI
    molal(5) = psi2 + psi7 + psi8 ! SO4I
    molal(6) = zero ! HSO4
    molal(7) = psi5 ! NO3I
    molal(8) = zero ! CAI
    molal(9) = 2.0D0*psi7 ! KI
    molal(10) = psi8 ! MGI

    !
    ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************

    !
    !CC      MOLAL (1) = MAX(CHI5 - PSI5, TINY)*A5/PSI5   ! HI
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnh4cl = zero
    ck2so4 = max(chi7-psi7, tiny)
    cmgso4 = zero
    ccaso4 = chi9
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funco4 = molal(1)*molal(4)/ghcl/a6 - one
  !CC         FUNCO4 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCO4 *******************************************
  !
END FUNCTION funco4
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO3
! *** CASE O3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3, NH4Cl, NA2SO4, K2SO4, MGSO4, CASO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  IF (w(4)>tiny .AND. w(5)>tiny) THEN ! NO3,CL EXIST, WATER POSSIBLE
    scase = 'O3 ; SUBCASE 1'
    CALL calco3a
    scase = 'O3 ; SUBCASE 1'
  ELSE ! NO3, CL NON EXISTANT
    scase = 'O1 ; SUBCASE 1'
    CALL calco1a
    scase = 'O1 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmo3) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calco1a
      scase = 'O3 ; SUBCASE 2'
      RETURN
    ELSE
      scase = 'O3 ; SUBCASE 3' ! MDRH REGION (NA2SO4, NH42S4, K2SO4, MGSO4, CASO4)
      CALL calcmdrh2(rh, drmo3, drnh42s4, calco1a, calco4)
      scase = 'O3 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calco3
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO3A
! *** CASE O3 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4, K2SO4, MGSO4, CASO4
!     4. Completely dissolved: NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco3a (  )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: nafr
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi9 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi9, zero)
  cafr = max(w(6)-chi9, zero)
  chi7 = min(0.5D0*w(7), so4fr) ! CK2SO4
  frk = max(w(7)-2.D0*chi7, zero)
  so4fr = max(so4fr-chi7, zero)
  chi1 = min(0.5D0*w(1), so4fr) ! NA2SO4
  nafr = max(w(1)-2.D0*chi1, zero)
  so4fr = max(so4fr-chi1, zero)
  chi8 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi8, zero)
  so4fr = max(so4fr-chi8, zero)
  chi3 = zero
  chi5 = w(4)
  chi6 = w(5)
  chi2 = max(so4fr, zero)
  chi4 = max(w(3)-2.D0*chi2, zero)
  !
  psi8 = chi8
  psi6lo = tiny
  psi6hi = chi6 - tiny
  !
  water = tiny
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funco3a(x1)
  IF (chi6<=tiny) GO TO 120
  !CC      IF (ABS(Y1).LE.EPS .OR. CHI7.LE.TINY) GOTO 50
  !CC      IF (WATER .LE. TINY) GOTO 50               ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funco3a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funco3a (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funco3a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCO3A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funco3a(x3)
  !
  ! *** FINAL CALCULATIONS *************************************************
  !
120 CONTINUE
  !
  ! *** Na2SO4 DISSOLUTION
  !
  IF (chi1>tiny .AND. water>tiny) THEN ! PSI1
    CALL poly3(psi2+psi7+psi8, zero, -a1/4.D0, psi1, islv)
    IF (islv==0) THEN
      psi1 = min(max(psi1,zero), chi1)
    ELSE
      psi1 = zero
    END IF
  ELSE
    psi1 = zero
  END IF
  molal(2) = 2.0D0*psi1 ! Na+  EFFECT
  molal(5) = molal(5) + psi1 ! SO4  EFFECT
  cna2so4 = max(chi1-psi1, zero) ! NA2SO4(s) depletion
  !
  ! *** HSO4 equilibrium
  !
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   AFFECT
    molal(5) = molal(5) - delta ! SO4  AFFECT
    molal(6) = delta ! HSO4 AFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calco3a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCO3A
! *** CASE O3; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NA2SO4, K2SO4, MgSO4, CaSO4
!     4. Completely dissolved: NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funco3a ( x )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: a65
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: deno
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi20
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi2 = chi2
  psi8 = chi8
  psi3 = zero
  psi6 = x
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0D0
    a2 = xk7*(water/gama(4))**3.0D0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0D0
    a5 = xk4*r*temp*(water/gama(10))**2.0D0
    a6 = xk3*r*temp*(water/gama(11))**2.0D0
    a7 = xk17*(water/gama(17))**3.0D0
    !      A8  = XK23 *(WATER/GAMA(21))**2.0D0
    a65 = a6/a5
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    deno = max(chi6-psi6-psi3, zero)
    psi5 = psi6*chi5/(a6/a5*deno+psi6)
    psi5 = min(max(psi5,zero), chi5)
    !
    !CC      IF(CHI4.GT.TINY) THEN                             ! PSI4
    IF (w(2)>tiny) THEN ! Accounts for NH3 evaporation
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6) - 2.D0*psi2/a4
      dd = max(bb*bb-4.D0*cc, zero) ! Patch proposed by Uma Shankar, 19/11/01
      psi4 = 0.5D0*(-bb-sqrt(dd))
    ELSE
      psi4 = tiny
    END IF
    psi4 = min(max(psi4,zero), chi4)
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! PSI7
      CALL poly3(psi2+psi8, zero, -a7/4.D0, psi7, islv)
      IF (islv==0) THEN
        psi7 = max(min(psi7,chi7), zero)
      ELSE
        psi7 = zero
      END IF
    ELSE
      psi7 = zero
    END IF
    !
    IF (chi2>tiny .AND. water>tiny) THEN
      CALL poly3(psi7+psi8+psi4, psi4*(psi7+psi8)+psi4*psi4/4.D0,          &
           (psi4*psi4*(psi7+psi8)-a2)/4.D0, psi20, islv)
      IF (islv==0) psi2 = min(max(psi20,zero), chi2)
    END IF
    !      PSI2 = 0.5D0*(2.0D0*SQRT(A2/A7)*PSI7 - PSI4)
    !      PSI2 = MIN (MAX(PSI2, ZERO), CHI2)
    !      ENDIF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(2) = zero ! NaI
    molal(3) = 2.0D0*psi2 + psi4 ! NH4I
    molal(4) = psi6 ! CLI
    molal(5) = psi2 + psi7 + psi8 ! SO4I
    molal(6) = zero ! HSO4
    molal(7) = psi5 ! NO3I
    molal(8) = zero ! CAI
    molal(9) = 2.0D0*psi7 ! KI
    molal(10) = psi8 ! MGI
    !
    !CC      MOLAL (1) = MAX(CHI5 - PSI5, TINY)*A5/PSI5   ! HI
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    !      CNA2SO4  = MAX(CHI1 - PSI1, ZERO)
    cnh42s4 = max(chi2-psi2, zero)
    cnh4no3 = zero
    cnh4cl = zero
    ck2so4 = max(chi7-psi7, zero)
    cmgso4 = zero
    ccaso4 = chi9
    !
    ! *** CALCULATE MOLALR ARRAY, WATER AND ACTIVITIES **********************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
100 funco3a = molal(1)*molal(4)/ghcl/a6 - one
  !
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCO3A *******************************************
  !
END FUNCTION funco3a

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO2
! *** CASE O2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3, NH4Cl, NA2SO4, K2SO4, MGSO4, CASO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF NITRATES ***********************
  !
  IF (w(4)>tiny) THEN ! NO3 EXISTS, WATER POSSIBLE
    scase = 'O2 ; SUBCASE 1'
    CALL calco2a
    scase = 'O2 ; SUBCASE 1'
  ELSE ! NO3 NON EXISTANT, WATER NOT POSSIBLE
    scase = 'O1 ; SUBCASE 1'
    CALL calco1a
    scase = 'O1 ; SUBCASE 1'
  END IF
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  IF (water<=tiny) THEN
    IF (rh<drmo2) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calco1a
      scase = 'O2 ; SUBCASE 2'
    ELSE
      IF (w(5)>tiny) THEN
        scase = 'O2 ; SUBCASE 3' ! MDRH (NH4CL, NA2SO4, NH42S4, K2SO4, MGSO4, CASO4)
        CALL calcmdrh2(rh, drmo2, drnh4cl, calco1a, calco3a)
        scase = 'O2 ; SUBCASE 3'
      END IF
      IF (water<=tiny .AND. rh>=drmo3) THEN
        scase = 'O2 ; SUBCASE 4' ! MDRH (NA2SO4, NH42S4, K2SO4, MGSO4, CASO4)
        CALL calcmdrh2(rh, drmo3, drnh42s4, calco1a, calco4)
        scase = 'O2 ; SUBCASE 4'
      ELSE
        water = tiny
        DO i = 1, nions
          molal(i) = zero
        END DO
        CALL calco1a
        scase = 'O2 ; SUBCASE 2'
      END IF
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calco2
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO2A
! *** CASE O2 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4, K2SO4, MgSO4, CaSO4
!     4. Completely dissolved: NH4NO3
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco2a (  )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: nafr
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS *************************************************
  !
  calaou = .TRUE.
  chi9 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi9, zero)
  cafr = max(w(6)-chi9, zero)
  chi7 = min(0.5D0*w(7), so4fr) ! CK2SO4
  frk = max(w(7)-2.D0*chi7, zero)
  so4fr = max(so4fr-chi7, zero)
  chi1 = min(0.5D0*w(1), so4fr) ! NA2SO4
  nafr = max(w(1)-2.D0*chi1, zero)
  so4fr = max(so4fr-chi1, zero)
  chi8 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi8, zero)
  so4fr = max(so4fr-chi8, zero)
  chi3 = zero
  chi5 = w(4)
  chi6 = w(5)
  chi2 = max(so4fr, zero)
  chi4 = max(w(3)-2.D0*chi2, zero)
  !
  psi8 = chi8
  psi6lo = tiny
  psi6hi = chi6 - tiny
  !
  water = tiny
  !
  ! *** INITIAL VALUES FOR BISECTION *************************************
  !
  x1 = psi6lo
  y1 = funco2a(x1)
  IF (chi6<=tiny) GO TO 120
  !CC      IF (ABS(Y1).LE.EPS .OR. CHI6.LE.TINY) GOTO 50
  !CC      IF (WATER .LE. TINY) GOTO 50               ! No water
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO ***********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funco2a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2)>eps) water = tiny
  GO TO 120
  !
  ! *** PERFORM BISECTION ************************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funco2a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCO2A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN ***********************************************
  !
110 x3 = 0.5*(x1+x2)
  IF (x3<=tiny2) THEN ! PRACTICALLY NO NITRATES, SO DRY SOLUTION
    water = tiny
  ELSE
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funco2a(x3)
  END IF
  !
  ! *** FINAL CALCULATIONS *************************************************
  !
120 CONTINUE
  !
  ! *** Na2SO4 DISSOLUTION
  !
  IF (chi1>tiny .AND. water>tiny) THEN ! PSI1
    CALL poly3(psi2+psi7+psi8, zero, -a1/4.D0, psi1, islv)
    IF (islv==0) THEN
      psi1 = min(psi1, chi1)
    ELSE
      psi1 = zero
    END IF
  ELSE
    psi1 = zero
  END IF
  molal(2) = 2.0D0*psi1 ! Na+  EFFECT
  molal(5) = molal(5) + psi1 ! SO4  EFFECT
  cna2so4 = max(chi1-psi1, zero) ! NA2SO4(s) depletion
  !
  ! *** HSO4 equilibrium
  !
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   AFFECT
    molal(5) = molal(5) - delta ! SO4  AFFECT
    molal(6) = delta ! HSO4 AFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calco2a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCO2A
! *** CASE O2; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. (Rsulfate > 2.0 ; R(Cr+Na) < 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4CL, NA2SO4, K2SO4, MgSO4, CaSO4
!     4. Completely dissolved: NH4NO3
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funco2a ( x )
  USE mo_isrpia
  USE mo_caseo
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: a65
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: deno
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi20
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi2 = chi2
  psi3 = zero
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0D0
    a2 = xk7*(water/gama(4))**3.0D0
    a3 = xk6/(r*temp*r*temp)
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0D0
    a5 = xk4*r*temp*(water/gama(10))**2.0D0
    a6 = xk3*r*temp*(water/gama(11))**2.0D0
    a65 = a6/a5
    a7 = xk17*(water/gama(17))**3.0D0
    !      A8  = XK23 *(WATER/GAMA(21))**2.0D0
    !
    deno = max(chi6-psi6-psi3, zero)
    psi5 = psi6*chi5/(a6/a5*deno+psi6)
    psi5 = min(psi5, chi5)
    !
    !!psi4 = min(psi5+psi6, chi4)
    IF (w(2)>tiny) THEN       ! Accounts for NH3 evaporation
       bb   =-(chi4 + psi6 + psi5 + 1.D0/a4)
       cc   = chi4*(psi5+psi6) - 2.D0*psi2/a4
       dd   = max(bb*bb-4.D0*cc,zero)  ! Patch proposed by Uma Shankar, 19/11/01
       psi4 =0.5D0*(-bb - sqrt(dd))
    ELSE
       psi4 = tiny
    END IF
    psi4 = min (max (psi4,zero), chi4)
    !
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! PSI7
      CALL poly3(psi2+psi8, zero, -a7/4.D0, psi7, islv)
      IF (islv==0) THEN
        psi7 = max(min(psi7,chi7), zero)
      ELSE
        psi7 = zero
      END IF
    ELSE
      psi7 = zero
    END IF
    !
    IF (chi2>tiny .AND. water>tiny) THEN
      CALL poly3(psi7+psi8+psi4, psi4*(psi7+psi8)+psi4*psi4/4.D0,          &
           (psi4*psi4*(psi7+psi8)-a2)/4.D0, psi20, islv)
      IF (islv==0) psi2 = min(max(psi20,zero), chi2)
    END IF
    !      PSI2 = 0.5D0*(2.0D0*SQRT(A2/A7)*PSI7 - PSI4)
    !      PSI2 = MIN (PSI2, CHI2)
    !      ENDIF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(2) = zero ! NaI
    molal(3) = 2.0D0*psi2 + psi4 ! NH4I
    molal(4) = psi6 ! CLI
    molal(5) = psi2 + psi7 + psi8 ! SO4I
    molal(6) = zero ! HSO4
    molal(7) = psi5 ! NO3I
    molal(8) = zero ! CAI
    molal(9) = 2.0D0*psi7 ! KI
    molal(10) = psi8 ! MGI
    !
    !CC      MOLAL (1) = MAX(CHI5 - PSI5, TINY)*A5/PSI5   ! HI
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    !      CNA2SO4  = MAX(CHI1 - PSI1, ZERO)
    cnh42s4 = max(chi2-psi2, zero)
    cnh4no3 = zero
    ck2so4 = max(chi7-psi7, zero)
    cmgso4 = zero
    ccaso4 = chi9

    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = min(min(psi3,chi4-psi4), chi6-psi6)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    ! *** CALCULATE MOLALR ARRAY, WATER AND ACTIVITIES *********************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP **************************
  !

  !20    IF (CHI4.LE.TINY) THEN
  !         FUNCO2A = MOLAL(1)*MOLAL(4)/GHCL/A6 - ONE
  !      ELSE
100 funco2a = molal(3)*molal(4)/ghcl/gnh3/a6/a4 - one
  !      ENDIF
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCO2A ****************************************
  !
END FUNCTION funco2a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO1
! *** CASE O1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3, NH4Cl, NA2SO4, K2SO4, MGSO4, CASO4
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCO1A)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmo1) THEN
    scase = 'O1 ; SUBCASE 1'
    CALL calco1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'O1 ; SUBCASE 1'
  ELSE
    scase = 'O1 ; SUBCASE 2' ! LIQUID & SOLID PHASE POSSIBLE
    CALL calcmdrh2(rh, drmo1, drnh4no3, calco1a, calco2a)
    scase = 'O1 ; SUBCASE 2'
  END IF
  !
  RETURN
  !
END SUBROUTINE calco1
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCO1A
! *** CASE O1A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3, NH4Cl, NA2SO4, K2SO4, MGSO4, CASO4
!
!     SOLID (NH4)2SO4 IS CALCULATED FROM THE SULFATES, WHILE NH4NO3
!     IS CALCULATED FROM NH3-HNO3 EQUILIBRIUM. 'ZE' IS THE AMOUNT OF
!     NH4NO3 THAT VOLATIZES WHEN ALL POSSILBE NH4NO3 IS INITIALLY IN
!     THE SOLID PHASE.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calco1a (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: a1
  DOUBLE PRECISION :: a2
  DOUBLE PRECISION :: alf
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bet
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: dd1
  DOUBLE PRECISION :: dd2
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: gam
  DOUBLE PRECISION :: kapa
  DOUBLE PRECISION :: kapa1
  DOUBLE PRECISION :: kapa2
  DOUBLE PRECISION :: lamda
  DOUBLE PRECISION :: lamda1
  DOUBLE PRECISION :: lamda2
  DOUBLE PRECISION :: rtsq
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: sqdd
  DOUBLE PRECISION :: sqdd1
  DOUBLE PRECISION :: sqdd2
  DOUBLE PRECISION :: theta1
  DOUBLE PRECISION :: theta2
  !
  ! *** CALCULATE NON VOLATILE SOLIDS ***********************************
  !
  ccaso4 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-ccaso4, zero)
  cafr = max(w(6)-ccaso4, zero)
  ck2so4 = min(0.5D0*w(7), so4fr) ! CK2S04
  frk = max(w(7)-2.D0*ck2so4, zero)
  so4fr = max(so4fr-ck2so4, zero)
  cna2so4 = min(0.5D0*w(1), so4fr) ! CNA2SO4
  frna = max(w(1)-2.D0*cna2so4, zero)
  so4fr = max(so4fr-cna2so4, zero)
  cmgso4 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-cmgso4, zero)
  so4fr = max(so4fr-cmgso4, zero)
  !
  cnh42s4 = max(so4fr, zero) ! CNH42S4
  !
  ! *** CALCULATE VOLATILE SPECIES **************************************
  !
  alf = w(3) - 2.0D0*cnh42s4
  bet = w(5)
  gam = w(4)
  !
  rtsq = r*temp*r*temp
  a1 = xk6/rtsq
  a2 = xk10/rtsq
  PRINT *, a2
  !
  theta1 = gam - bet*(a2/a1)
  theta2 = a2/a1
  !
  ! QUADRATIC EQUATION SOLUTION
  !
  bb = (theta1-alf-bet*(one+theta2))/(one+theta2)
  cc = (alf*bet-a1-bet*theta1)/(one+theta2)
  dd = bb*bb - 4.0D0*cc
  IF (dd<zero) GO TO 100 ! Solve each reaction seperately
  !
  ! TWO ROOTS FOR KAPA, CHECK AND SEE IF ANY VALID
  !
  sqdd = sqrt(dd)
  kapa1 = 0.5D0*(-bb+sqdd)
  kapa2 = 0.5D0*(-bb-sqdd)
  lamda1 = theta1 + theta2*kapa1
  lamda2 = theta1 + theta2*kapa2
  !
  IF (kapa1>=zero .AND. lamda1>=zero) THEN
    IF (alf-kapa1-lamda1>=zero .AND. bet-kapa1>=zero .AND.                 &
         gam-lamda1>=zero) THEN
      kapa = kapa1
      lamda = lamda1
      GO TO 110
    END IF
  END IF
  !
  IF (kapa2>=zero .AND. lamda2>=zero) THEN
    IF (alf-kapa2-lamda2>=zero .AND. bet-kapa2>=zero .AND.                 &
         gam-lamda2>=zero) THEN
      kapa = kapa2
      lamda = lamda2
      GO TO 110
    END IF
  END IF
  !
  ! SEPERATE SOLUTION OF NH4CL & NH4NO3 EQUILIBRIA
  !
100 kapa = zero
  lamda = zero
  dd1 = (alf+bet)*(alf+bet) - 4.0D0*(alf*bet-a1)
  dd2 = (alf+gam)*(alf+gam) - 4.0D0*(alf*gam-a2)
  !
  ! NH4CL EQUILIBRIUM
  !
  IF (dd1>=zero) THEN
    sqdd1 = sqrt(dd1)
    kapa1 = 0.5D0*(alf+bet+sqdd1)
    kapa2 = 0.5D0*(alf+bet-sqdd1)
    !
    IF (kapa1>=zero .AND. kapa1<=min(alf,bet)) THEN
      kapa = kapa1
    ELSE IF (kapa2>=zero .AND. kapa2<=min(alf,bet)) THEN
      kapa = kapa2
    ELSE
      kapa = zero
    END IF
  END IF
  !
  ! NH4NO3 EQUILIBRIUM
  !
  IF (dd2>=zero) THEN
    sqdd2 = sqrt(dd2)
    lamda1 = 0.5D0*(alf+gam+sqdd2)
    lamda2 = 0.5D0*(alf+gam-sqdd2)
    !
    IF (lamda1>=zero .AND. lamda1<=min(alf,gam)) THEN
      lamda = lamda1
    ELSE IF (lamda2>=zero .AND. lamda2<=min(alf,gam)) THEN
      lamda = lamda2
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! IF BOTH KAPA, LAMDA ARE > 0, THEN APPLY EXISTANCE CRITERION
  !
  IF (kapa>zero .AND. lamda>zero) THEN
    IF (bet<lamda/theta1) THEN
      kapa = zero
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! *** CALCULATE COMPOSITION OF VOLATILE SPECIES ************************
  !
110 CONTINUE
  cnh4no3 = lamda
  cnh4cl = kapa
  !
  gnh3 = max(alf-kapa-lamda, zero)
  ghno3 = max(gam-lamda, zero)
  ghcl = max(bet-kapa, zero)
  !
  RETURN
  !
END SUBROUTINE calco1a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM8
! *** CASE M8
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL, MgSO4, NA2SO4, K2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcm8 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi11, zero)
  cafr = max(w(6)-chi11, zero)
  chi9 = min(0.5D0*w(7), so4fr) ! CK2S04
  frk = max(w(7)-2.D0*chi9, zero)
  so4fr = max(so4fr-chi9, zero)
  chi10 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  so4fr = max(so4fr-chi10, zero)
  chi1 = max(so4fr, zero) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcm8(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcm8(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcm8 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcm8(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCM8') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcm8(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcm8




!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCM8
! *** CASE M8
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL, MgSO4, NA2SO4, K2SO4
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcm8 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = chi9
  psi10 = chi10
  psi11 = zero
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    !      A1  = XK5 *(WATER/GAMA(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    !      A7  = XK8 *(WATER/GAMA(1))**2.0
    !      A8  = XK9 *(WATER/GAMA(3))**2.0
    !      A11 = XK1*WATER/GAMA(7)*(GAMA(8)/GAMA(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 + psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 ! NO3I
    molal(8) = psi11 ! CAI
    molal(9) = 2.D0*psi9 ! KI
    molal(10) = psi10 ! MGI
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = zero
    cnano3 = zero
    cna2so4 = zero
    ck2so4 = zero
    cmgso4 = zero
    ccaso4 = chi11
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCM8 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcm8 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCM8 *******************************************
  !
END FUNCTION funcm8


!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM7
! *** CASE M7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL, MgSO4, NA2SO4
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcm7 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi11, zero)
  cafr = max(w(6)-chi11, zero)
  chi9 = min(0.5D0*w(7), so4fr) ! CK2S04
  frk = max(w(7)-2.D0*chi9, zero)
  so4fr = max(so4fr-chi9, zero)
  chi10 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  so4fr = max(so4fr-chi10, zero)
  chi1 = max(so4fr, zero) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcm7(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcm7(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcm7 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcm7(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCM7') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcm7(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcm7

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCM7
! *** CASE M7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL, MgSO4, NA2SO4
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcm7 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    !      A1  = XK5 *(WATER/GAMA(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    !      A7  = XK8 *(WATER/GAMA(1))**2.0
    !      A8  = XK9 *(WATER/GAMA(3))**2.0
    !      A11 = XK1*WATER/GAMA(7)*(GAMA(8)/GAMA(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      CALL poly3(psi1+psi10, zero, -a9/4.D0, psi9, islv)
      IF (islv==0) THEN
        psi9 = max(min(psi9,chi9), zero)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 + psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 ! NO3I
    molal(8) = psi11 ! CAI
    molal(9) = 2.D0*psi9 ! KI
    molal(10) = psi10 ! MGI
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = zero
    cnano3 = zero
    cna2so4 = zero
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCM7 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcm7 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCM7 *******************************************
  !
END FUNCTION funcm7
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM6
! *** CASE M6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL, MgSO4
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcm6 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi11, zero)
  cafr = max(w(6)-chi11, zero)
  chi9 = min(0.5D0*w(7), so4fr) ! CK2S04
  frk = max(w(7)-2.D0*chi9, zero)
  so4fr = max(so4fr-chi9, zero)
  chi10 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  so4fr = max(so4fr-chi10, zero)
  chi1 = max(so4fr, zero) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcm6(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcm6(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcm6 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcm6(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCM6') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcm6(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcm6
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCM6
! *** CASE M6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL, MgSO4
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcm6 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: riz
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    !      A7  = XK8 *(WATER/GAMA(1))**2.0
    !      A8  = XK9 *(WATER/GAMA(3))**2.0
    !      A11 = XK1*WATER/GAMA(7)*(GAMA(8)/GAMA(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi1>tiny .AND. water>tiny) THEN !NA2SO4
      riz = sqrt(a9/a1)
      aa = (0.5D0*riz*(psi7+psi8)+psi10+(1.D0+riz)*(psi7+psi8))/(1.D0+riz)
      bb = ((psi7+psi8)*(0.5D0*riz*(psi7+psi8)+psi10)+0.25D0*(psi7+psi8)** &
           2.0*(1.D0+riz))/(1.D0+riz)
      cc = (0.25D0*(psi7+psi8)**2.0*(0.5D0*riz*(psi7+                      &
           psi8)+psi10)-a1/4.D0)/(1.D0+riz)
      !      AA  = PSI7+PSI8+PSI9+PSI10
      !      BB  = (PSI7+PSI8)*(PSI9+PSI10)+0.25D0*(PSI7+PSI8)**2.
      !      CC  = ((PSI7+PSI8)**2.*(PSI9+PSI10)-A1)/4.0D0
      !
      CALL poly3(aa, bb, cc, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    END IF
    !
    !      IF (CHI9.GE.TINY .AND. WATER.GT.TINY) THEN
    !         PSI9  = 0.5D0*SQRT(A9/A1)*(2.0D0*PSI1+PSI7+PSI8)
    !         PSI9  = MAX (MIN (PSI9,CHI9), ZERO)
    !      ELSE
    !         PSI9  = ZERO
    !      ENDIF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      CALL poly3(psi1+psi10, zero, -a9/4.D0, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(psi9, chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 + psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 ! NO3I
    molal(8) = psi11 ! CAI
    molal(9) = 2.D0*psi9 ! KI
    molal(10) = psi10 ! MGI
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = zero
    cnano3 = zero
    cna2so4 = max(chi1-psi1, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCM6 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcm6 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCM6 *******************************************
  !
END FUNCTION funcm6

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM5
! *** CASE M5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcm5 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi11, zero)
  cafr = max(w(6)-chi11, zero)
  chi9 = min(0.5D0*w(7), so4fr) ! CK2S04
  frk = max(w(7)-2.D0*chi9, zero)
  so4fr = max(so4fr-chi9, zero)
  chi10 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  so4fr = max(so4fr-chi10, zero)
  chi1 = max(so4fr, zero) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcm5(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcm5(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcm5 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcm5(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCM5') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcm5(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcm5
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCM5
! *** CASE M5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcm5 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: riz
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    !      A7  = XK8 *(WATER/GAMA(1))**2.0
    !      A8  = XK9 *(WATER/GAMA(3))**2.0
    !      A11 = XK1*WATER/GAMA(7)*(GAMA(8)/GAMA(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi1>tiny .AND. water>tiny) THEN !NA2SO4
      riz = sqrt(a9/a1)
      aa = (0.5D0*riz*(psi7+psi8)+psi10+(1.D0+riz)*(psi7+psi8))/(1.D0+riz)
      bb = ((psi7+psi8)*(0.5D0*riz*(psi7+psi8)+psi10)+0.25D0*(psi7+psi8)** &
           2.0*(1.D0+riz))/(1.D0+riz)
      cc = (0.25D0*(psi7+psi8)**2.0*(0.5D0*riz*(psi7+                      &
           psi8)+psi10)-a1/4.D0)/(1.D0+riz)
      !      AA  = PSI7+PSI8+PSI9+PSI10
      !      BB  = (PSI7+PSI8)*(PSI9+PSI10)+0.25D0*(PSI7+PSI8)**2.
      !      CC  = ((PSI7+PSI8)**2.*(PSI9+PSI10)-A1)/4.0D0
      !
      CALL poly3(aa, bb, cc, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    END IF
    !
    IF (chi9>=tiny .AND. water>tiny) THEN
      psi9 = 0.5D0*sqrt(a9/a1)*(2.0D0*psi1+psi7+psi8)
      psi9 = max(min(psi9,chi9), zero)
    ELSE
      psi9 = zero
    END IF
    !
    !      IF (CHI9.GT.TINY .AND. WATER.GT.TINY) THEN   !K2SO4
    !      CALL POLY3 (PSI1+PSI10,ZERO,-A9/4.D0, PSI9, ISLV)
    !        IF (ISLV.EQ.0) THEN
    !            PSI9 = MIN (PSI9,CHI9)
    !        ELSE
    !            PSI9 = ZERO
    !        ENDIF
    !      ENDIF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 + psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 ! NO3I
    molal(8) = psi11 ! CAI
    molal(9) = 2.D0*psi9 ! KI
    molal(10) = psi10 ! MGI
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = zero
    cnano3 = zero
    cna2so4 = max(chi1-psi1, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCM5 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcm5 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCM5 *******************************************
  !
END FUNCTION funcm5

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM4
! *** CASE M4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4, NH4CL
!     4. Completely dissolved: NH4NO3, NANO3, NACL
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcm4 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF NITRATES ***********************
  !
  IF (w(4)<=tiny .AND. w(5)<=tiny) THEN
    scase = 'M4 ; SUBCASE 1'
    CALL calcm1a
    scase = 'M4 ; SUBCASE 1'
    RETURN
  END IF
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi11, zero)
  cafr = max(w(6)-chi11, zero)
  chi9 = min(0.5D0*w(7), so4fr) ! CK2S04
  frk = max(w(7)-2.D0*chi9, zero)
  so4fr = max(so4fr-chi9, zero)
  chi10 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  so4fr = max(so4fr-chi10, zero)
  chi1 = max(so4fr, zero) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcm4(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcm4(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcm4 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcm4(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCM4') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcm4(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcm4
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCM4
! *** CASE M4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4, NH4CL
!     4. Completely dissolved: NH4NO3, NANO3, NACL
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcm4 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: riz
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a3 = xk6/(r*temp*r*temp)
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    !      A7  = XK8 *(WATER/GAMA(1))**2.0
    !      A8  = XK9 *(WATER/GAMA(3))**2.0
    !      A11 = XK1*WATER/GAMA(7)*(GAMA(8)/GAMA(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,tiny), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi1>tiny .AND. water>tiny) THEN !NA2SO4
      riz = sqrt(a9/a1)
      aa = (0.5D0*riz*(psi7+psi8)+psi10+(1.D0+riz)*(psi7+psi8))/(1.D0+riz)
      bb = ((psi7+psi8)*(0.5D0*riz*(psi7+psi8)+psi10)+0.25D0*(psi7+psi8)** &
           2.0*(1.D0+riz))/(1.D0+riz)
      cc = (0.25D0*(psi7+psi8)**2.0*(0.5D0*riz*(psi7+                      &
           psi8)+psi10)-a1/4.D0)/(1.D0+riz)
      !      AA  = PSI7+PSI8+PSI9+PSI10
      !      BB  = (PSI7+PSI8)*(PSI9+PSI10)+0.25D0*(PSI7+PSI8)**2.
      !      CC  = ((PSI7+PSI8)**2.*(PSI9+PSI10)-A1)/4.0D0
      !
      CALL poly3(aa, bb, cc, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    END IF
    !
    IF (chi9>=tiny .AND. water>tiny) THEN
      psi9 = 0.5D0*sqrt(a9/a1)*(2.0D0*psi1+psi7+psi8)
      psi9 = max(min(psi9,chi9), zero)
    ELSE
      psi9 = zero
    END IF
    !
    !      IF (CHI9.GT.TINY .AND. WATER.GT.TINY) THEN   !K2SO4
    !      CALL POLY3 (PSI1+PSI10,ZERO,-A9/4.D0, PSI9, ISLV)
    !        IF (ISLV.EQ.0) THEN
    !            PSI9 = MIN (PSI9,CHI9)
    !        ELSE
    !            PSI9 = ZERO
    !        ENDIF
    !      ENDIF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 + psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 ! NO3I
    molal(8) = psi11 ! CAI
    molal(9) = 2.D0*psi9 ! KI
    molal(10) = psi10 ! MGI
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = zero
    cnano3 = zero
    cna2so4 = max(chi1-psi1, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCM4 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcm4 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCM4 *******************************************
  !
END FUNCTION funcm4
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM3
! *** CASE M3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4, NH4CL, NACL
!     4. Completely dissolved: NH4NO3, NANO3
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcm3 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF NITRATES ***********************
  !
  IF (w(4)<=tiny) THEN ! NO3 NOT EXIST, WATER NOT POSSIBLE
    scase = 'M3 ; SUBCASE 1'
    CALL calcm1a
    scase = 'M3 ; SUBCASE 1'
    RETURN
  END IF
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi11, zero)
  cafr = max(w(6)-chi11, zero)
  chi9 = min(0.5D0*w(7), so4fr) ! CK2S04
  frk = max(w(7)-2.D0*chi9, zero)
  so4fr = max(so4fr-chi9, zero)
  chi10 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  so4fr = max(so4fr-chi10, zero)
  chi1 = max(so4fr, zero) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcm3(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcm3(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcm3 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcm3(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCM3') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcm3(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcm3
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCM3
! *** CASE M3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4, NH4CL, NACL
!     4. Completely dissolved: NH4NO3, NANO3
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcm3 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: riz
  DOUBLE PRECISION :: smin
  !
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a3 = xk6/(r*temp*r*temp)
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a10 = xk23*(water/gama(21))**2.0
    !      A8  = XK9 *(WATER/GAMA(3))**2.0
    !      A11 = XK1*WATER/GAMA(7)*(GAMA(8)/GAMA(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,tiny), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    !      IF (CHI7.GT.TINY .AND. WATER.GT.TINY) THEN     ! NACL DISSOLUTION
    !         VITA = 2.D0*PSI1+PSI8+PSI6                 ! AN DE DOULEPSEI KALA VGALE PSI1 APO DW
    !         GKAMA= PSI6*(2.D0*PSI1+PSI8)-A7
    !         DIAK = MAX(VITA**2.0 - 4.0D0*GKAMA,ZERO)
    !         PSI7 = 0.5D0*( -VITA + SQRT(DIAK) )
    !         PSI7 = MAX(MIN(PSI7, CHI7), ZERO)
    !      ENDIF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      diak = (psi8-psi6)**2.D0 + 4.D0*a7
      psi7 = 0.5D0*(-(psi8+psi6)+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !C
    !
    IF (chi1>tiny .AND. water>tiny) THEN !NA2SO4
      riz = sqrt(a9/a1)
      aa = (0.5D0*riz*(psi7+psi8)+psi10+(1.D0+riz)*(psi7+psi8))/(1.D0+riz)
      bb = ((psi7+psi8)*(0.5D0*riz*(psi7+psi8)+psi10)+0.25D0*(psi7+psi8)** &
           2.0*(1.D0+riz))/(1.D0+riz)
      cc = (0.25D0*(psi7+psi8)**2.0*(0.5D0*riz*(psi7+                      &
           psi8)+psi10)-a1/4.D0)/(1.D0+riz)
      !      AA  = PSI7+PSI8+PSI9+PSI10
      !      BB  = (PSI7+PSI8)*(PSI9+PSI10)+0.25D0*(PSI7+PSI8)**2.
      !      CC  = ((PSI7+PSI8)**2.*(PSI9+PSI10)-A1)/4.0D0
      !
      CALL poly3(aa, bb, cc, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    END IF
    !
    IF (chi9>=tiny) THEN
      psi9 = 0.5D0*sqrt(a9/a1)*(2.0D0*psi1+psi7+psi8)
      psi9 = max(min(psi9,chi9), zero)
    ELSE
      psi9 = zero
    END IF
    !
    !      IF (CHI9.GT.TINY .AND. WATER.GT.TINY) THEN   !K2SO4
    !      CALL POLY3 (PSI1+PSI10,ZERO,-A9/4.D0, PSI9, ISLV)
    !        IF (ISLV.EQ.0) THEN
    !            PSI9 = MIN (PSI9,CHI9)
    !        ELSE
    !            PSI9 = ZERO
    !        ENDIF
    !      ENDIF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 + psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 ! NO3I
    molal(8) = psi11 ! CAI
    molal(9) = 2.D0*psi9 ! KI
    molal(10) = psi10 ! MGI
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = max(chi7-psi7, zero)
    cnano3 = zero
    cna2so4 = max(chi1-psi1, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCM3 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcm3 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCM3 *******************************************
  !
END FUNCTION funcm3


!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM2
! *** CASE M2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3
!
!     THERE ARE THREE REGIMES IN THIS CASE:
!     1. NH4NO3(s) POSSIBLE. LIQUID & SOLID AEROSOL (SUBROUTINE CALCH2A)
!     2. NH4NO3(s) NOT POSSIBLE, AND RH < MDRH. SOLID AEROSOL ONLY
!     3. NH4NO3(s) NOT POSSIBLE, AND RH >= MDRH. (MDRH REGION)
!
!     REGIMES 2. AND 3. ARE CONSIDERED TO BE THE SAME AS CASES M1A, M2B
!     RESPECTIVELY (BECAUSE MDRH POINTS COINCIDE).
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcm2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF NITRATES ***********************
  !
  CALL calcm1a
  !
  IF (cnh4no3>tiny) THEN ! NO3 EXISTS, WATER POSSIBLE
    scase = 'M2 ; SUBCASE 1'
    CALL calcm2a
    scase = 'M2 ; SUBCASE 1'
  ELSE ! NO3 NON EXISTANT, WATER NOT POSSIBLE
    scase = 'M2 ; SUBCASE 1'
    CALL calcm1a
    scase = 'M2 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny .AND. rh<drmm2) THEN ! DRY AEROSOL
    scase = 'M2 ; SUBCASE 2'
    !
  ELSE IF (water<=tiny .AND. rh>=drmm2) THEN ! MDRH OF M2
    scase = 'M2 ; SUBCASE 3'
    CALL calcmdrh2(rh, drmm2, drnano3, calcm1a, calcm3)
    scase = 'M2 ; SUBCASE 3'
  END IF
  !
  RETURN
  !
END SUBROUTINE calcm2
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM2A
! *** CASE M2A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3
!     4. Completely dissolved: NH4NO3
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcm2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-chi11, zero)
  cafr = max(w(6)-chi11, zero)
  chi9 = min(0.5D0*w(7), so4fr) ! CK2S04
  frk = max(w(7)-2.D0*chi9, zero)
  so4fr = max(so4fr-chi9, zero)
  chi10 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  so4fr = max(so4fr-chi10, zero)
  chi1 = max(so4fr, zero) ! CNA2SO4
  chi2 = zero ! CNH42S4
  chi3 = zero ! CNH4CL
  frna = max(w(1)-2.D0*chi1, zero)
  chi8 = min(frna, w(4)) ! CNANO3
  chi4 = w(3) ! NH3(g)
  chi5 = max(w(4)-chi8, zero) ! HNO3(g)
  chi7 = min(max(frna-chi8,zero), w(5)) ! CNACL
  chi6 = max(w(5)-chi7, zero) ! HCL(g)
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcm2a(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcm2a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcm2a (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcm2a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCM2A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcm2a(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcm2a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCM2A
! *** CASE M2A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3
!     4. Completely dissolved: NH4NO3
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY  CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcm2a ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: a64
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: riz
  DOUBLE PRECISION :: smin
  !
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = chi1
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk5*(water/gama(2))**3.0
    a3 = xk6/(r*temp*r*temp)
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a64 = (xk3*xk2/xkw)*(gama(10)/gama(5)/gama(11))**2.0
    a64 = a64*(r*temp*water)**2.0
    !      A11 = XK1*WATER/GAMA(7)*(GAMA(8)/GAMA(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7) - a6/a5*psi8*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,tiny), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    !      IF (CHI7.GT.TINY .AND. WATER.GT.TINY) THEN     ! NACL DISSOLUTION
    !         VITA = 2.D0*PSI1+PSI8+PSI6
    !         GKAMA= PSI6*(2.D0*PSI1+PSI8)-A7
    !         DIAK = MAX(VITA**2.0 - 4.0D0*GKAMA,ZERO)
    !         PSI7 = 0.5D0*( -VITA + SQRT(DIAK) )
    !         PSI7 = MAX(MIN(PSI7, CHI7), ZERO)
    !      ENDIF
    !C
    !      IF (CHI8.GT.TINY .AND. WATER.GT.TINY) THEN     ! NANO3 DISSOLUTION
    !         BIT  = 2.D0*PSI1+PSI7+PSI5
    !         GKAM = PSI5*(2.D0*PSI1+PSI8)-A8
    !         DIA  = BIT**2.0 - 4.0D0*GKAM
    !        PSI8 = 0.5D0*( -BIT + SQRT(DIA) )
    !        PSI8 = MAX(MIN(PSI8, CHI8), ZERO)
    !      ENDIF
    !C
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      diak = (psi8-psi6)**2.D0 + 4.D0*a7
      psi7 = 0.5D0*(-(psi8+psi6)+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !
    IF (chi8>tiny .AND. water>tiny) THEN ! NANO3 DISSOLUTION
      diak = (psi7-psi5)**2.D0 + 4.D0*a8
      psi8 = 0.5D0*(-(psi7+psi5)+sqrt(diak))
      psi8 = max(min(psi8,chi8), zero)
    END IF
    !
    IF (chi1>tiny .AND. water>tiny) THEN !NA2SO4
      riz = sqrt(a9/a1)
      aa = (0.5D0*riz*(psi7+psi8)+psi10+(1.D0+riz)*(psi7+psi8))/(1.D0+riz)
      bb = ((psi7+psi8)*(0.5D0*riz*(psi7+psi8)+psi10)+0.25D0*(psi7+psi8)** &
           2.0*(1.D0+riz))/(1.D0+riz)
      cc = (0.25D0*(psi7+psi8)**2.0*(0.5D0*riz*(psi7+                      &
           psi8)+psi10)-a1/4.D0)/(1.D0+riz)
      !
      !      AA  = PSI7+PSI8+PSI9+PSI10
      !      BB  = (PSI7+PSI8)*(PSI9+PSI10)+0.25D0*(PSI7+PSI8)**2.
      !      CC  = ((PSI7+PSI8)**2.*(PSI9+PSI10)-A1)/4.0D0
      !C
      CALL poly3(aa, bb, cc, psi1, islv)
      IF (islv==0) THEN
        psi1 = min(psi1, chi1)
      ELSE
        psi1 = zero
      END IF
    END IF
    !
    IF (chi9>=tiny .AND. water>tiny) THEN
      !         PSI9  = 0.5D0*SQRT(A9/A1)*(2.0D0*PSI1+PSI7+PSI8)
      psi9 = 0.5D0*sqrt(a9/a1)*(2.0D0*psi1+psi7+psi8)
      psi9 = max(min(psi9,chi9), zero)
    ELSE
      psi9 = zero
    END IF
    !
    !      IF (CHI9.GT.TINY .AND. WATER.GT.TINY) THEN   !K2SO4
    !      CALL POLY3 (PSI1+PSI10,ZERO,-A9/4.D0, PSI9, ISLV)
    !        IF (ISLV.EQ.0) THEN
    !            PSI9 = MAX (MIN (PSI9,CHI9), ZERO)
    !        ELSE
    !            PSI9 = ZERO
    !        ENDIF
    !      ENDIF
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 + 2.D0*psi1 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 ! CLI
    molal(5) = psi2 + psi1 + psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 ! NO3I
    molal(8) = psi11 ! CAI
    molal(9) = 2.D0*psi9 ! KI
    molal(10) = psi10 ! MGI
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    cna2so4 = max(chi1-psi1, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCM2A = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A64 - ONE
100 funcm2a = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCM2A *******************************************
  !
END FUNCTION funcm2a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM1
! *** CASE M1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3, NH4NO3
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCH1A)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcm1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmm1) THEN
    scase = 'M1 ; SUBCASE 1'
    CALL calcm1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'M1 ; SUBCASE 1'
  ELSE
    scase = 'M1 ; SUBCASE 2' ! LIQUID & SOLID PHASE POSSIBLE
    CALL calcmdrh2(rh, drmm1, drnh4no3, calcm1a, calcm2a)
    scase = 'M1 ; SUBCASE 2'
  END IF
  !
  RETURN
  !
END SUBROUTINE calcm1
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCM1A
! *** CASE M1A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr < 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3, NH4NO3
!
! *** COPYRIGHT 1996-2000, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================

SUBROUTINE calcm1a (  )
  USE mo_isrpia  
  IMPLICIT NONE
  DOUBLE PRECISION :: a1
  DOUBLE PRECISION :: a2
  DOUBLE PRECISION :: alf
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bet
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: clfr
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: dd1
  DOUBLE PRECISION :: dd2
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: gam
  DOUBLE PRECISION :: kapa
  DOUBLE PRECISION :: kapa1
  DOUBLE PRECISION :: kapa2
  DOUBLE PRECISION :: lamda
  DOUBLE PRECISION :: lamda1
  DOUBLE PRECISION :: lamda2
  DOUBLE PRECISION :: nafr
  DOUBLE PRECISION :: no3fr
  DOUBLE PRECISION :: rtsq
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: sqdd
  DOUBLE PRECISION :: sqdd1
  DOUBLE PRECISION :: sqdd2
  DOUBLE PRECISION :: theta1
  DOUBLE PRECISION :: theta2
  !
  ! *** CALCULATE NON VOLATILE SOLIDS ***********************************
  !
  ccaso4 = min(w(6), w(2)) ! CCASO4
  so4fr = max(w(2)-ccaso4, zero)
  cafr = max(w(6)-ccaso4, zero)
  ck2so4 = min(0.5D0*w(7), so4fr) ! CK2S04
  frk = max(w(7)-2.D0*ck2so4, zero)
  so4fr = max(so4fr-ck2so4, zero)
  cmgso4 = min(w(8), so4fr) ! CMGSO4
  frmg = max(w(8)-cmgso4, zero)
  so4fr = max(so4fr-cmgso4, zero)
  cna2so4 = max(so4fr, zero) ! CNA2SO4
  nafr = max(w(1)-2.D0*cna2so4, zero)
  cnano3 = min(nafr, w(4)) ! CNANO3
  no3fr = max(w(4)-cnano3, zero)
  cnacl = min(max(nafr-cnano3,zero), w(5)) ! CNACL
  clfr = max(w(5)-cnacl, zero)
  !
  ! *** CALCULATE VOLATILE SPECIES **************************************
  !
  alf = w(3) ! FREE NH3
  bet = clfr ! FREE CL
  gam = no3fr ! FREE NO3
  !
  rtsq = r*temp*r*temp
  a1 = xk6/rtsq
  a2 = xk10/rtsq
  !
  theta1 = gam - bet*(a2/a1)
  theta2 = a2/a1
  !
  ! QUADRATIC EQUATION SOLUTION
  !
  bb = (theta1-alf-bet*(one+theta2))/(one+theta2)
  cc = (alf*bet-a1-bet*theta1)/(one+theta2)
  dd = bb*bb - 4.0D0*cc
  IF (dd<zero) GO TO 100 ! Solve each reaction seperately
  !
  ! TWO ROOTS FOR KAPA, CHECK AND SEE IF ANY VALID
  !
  sqdd = sqrt(dd)
  kapa1 = 0.5D0*(-bb+sqdd)
  kapa2 = 0.5D0*(-bb-sqdd)
  lamda1 = theta1 + theta2*kapa1
  lamda2 = theta1 + theta2*kapa2
  !
  IF (kapa1>=zero .AND. lamda1>=zero) THEN
    IF (alf-kapa1-lamda1>=zero .AND. bet-kapa1>=zero .AND.                 &
         gam-lamda1>=zero) THEN
      kapa = kapa1
      lamda = lamda1
      GO TO 110
    END IF
  END IF
  !
  IF (kapa2>=zero .AND. lamda2>=zero) THEN
    IF (alf-kapa2-lamda2>=zero .AND. bet-kapa2>=zero .AND.                 &
         gam-lamda2>=zero) THEN
      kapa = kapa2
      lamda = lamda2
      GO TO 110
    END IF
  END IF
  !
  ! SEPERATE SOLUTION OF NH4CL & NH4NO3 EQUILIBRIA
  !
100 kapa = zero
  lamda = zero
  dd1 = (alf+bet)*(alf+bet) - 4.0D0*(alf*bet-a1)
  dd2 = (alf+gam)*(alf+gam) - 4.0D0*(alf*gam-a2)
  !
  ! NH4CL EQUILIBRIUM
  !
  IF (dd1>=zero) THEN
    sqdd1 = sqrt(dd1)
    kapa1 = 0.5D0*(alf+bet+sqdd1)
    kapa2 = 0.5D0*(alf+bet-sqdd1)
    !
    IF (kapa1>=zero .AND. kapa1<=min(alf,bet)) THEN
      kapa = kapa1
    ELSE IF (kapa2>=zero .AND. kapa2<=min(alf,bet)) THEN
      kapa = kapa2
    ELSE
      kapa = zero
    END IF
  END IF
  !
  ! NH4NO3 EQUILIBRIUM
  !
  IF (dd2>=zero) THEN
    sqdd2 = sqrt(dd2)
    lamda1 = 0.5D0*(alf+gam+sqdd2)
    lamda2 = 0.5D0*(alf+gam-sqdd2)
    !
    IF (lamda1>=zero .AND. lamda1<=min(alf,gam)) THEN
      lamda = lamda1
    ELSE IF (lamda2>=zero .AND. lamda2<=min(alf,gam)) THEN
      lamda = lamda2
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! IF BOTH KAPA, LAMDA ARE > 0, THEN APPLY EXISTANCE CRITERION
  !
  IF (kapa>zero .AND. lamda>zero) THEN
    IF (bet<lamda/theta1) THEN
      kapa = zero
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! *** CALCULATE COMPOSITION OF VOLATILE SPECIES ***********************
  !
110 CONTINUE
  cnh4no3 = lamda
  cnh4cl = kapa
  !
  gnh3 = alf - kapa - lamda
  ghno3 = gam - lamda
  ghcl = bet - kapa
  !
  RETURN
  !
END SUBROUTINE calcm1a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP13
! *** CASE P13
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4
!     4. Completely dissolved: CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp13 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp13(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp13(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp13 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp13(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP13') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp13(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp13

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP13
! *** CASE P13
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4
!     4. Completely dissolved: CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp13 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: hi
  INTEGER :: i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi4 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = chi9
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = chi13
  psi14 = chi14
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -                  &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6)
    psi5 = psi5/(a6/a5*(chi6-psi6)+psi6+psi7+psi14+2.D0*psi16+2.D0*psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    ! *** CALCULATE SPECIATION *********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !
    ! *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh4no3 = zero
    cnh4cl = zero
    cnacl = zero
    cnano3 = zero
    ck2so4 = zero
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = zero
    ckcl = zero
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP13 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp13 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP13 *******************************************
  !
END FUNCTION funcp13
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP12
! *** CASE P12
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4
!     4. Completely dissolved: CA(NO3)2, CACL2, KNO3, KCL, MGSO4,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp12 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp12(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp12(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp12 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp12(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP12') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp12(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp12

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP12
! *** CASE P12
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4
!     4. Completely dissolved: CA(NO3)2, CACL2, KNO3, KCL, MGSO4,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp12 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi4 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = chi13
  psi14 = chi14
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -                  &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh4no3 = zero
    cnh4cl = zero
    cnacl = zero
    cnano3 = zero
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = zero
    ckcl = zero
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP12 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp12 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP12 *******************************************
  !
END FUNCTION funcp12

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP11
! *** CASE P11
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3
!     4. Completely dissolved: CA(NO3)2, CACL2, KCL, MGSO4,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp11 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp11(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp11(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp11 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp11(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP11') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp11(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp11

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP11
! *** CASE P11
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3
!     4. Completely dissolved: CA(NO3)2, CACL2, KCL, MGSO4,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp11 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = chi14
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -                  &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh4no3 = zero
    cnh4cl = zero
    cnacl = zero
    cnano3 = zero
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = zero
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP11 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp11 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP11 *******************************************
  !
END FUNCTION funcp11

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP10
! *** CASE P10
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4
!     4. Completely dissolved: CA(NO3)2, CACL2, KCL,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp10 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp10(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp10(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp10 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp10(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP10') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp10(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp10

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP10
! *** CASE P10
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4
!     4. Completely dissolved: CA(NO3)2, CACL2, KCL,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp10 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = chi14
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -                  &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh4no3 = zero
    cnh4cl = zero
    cnacl = zero
    cnano3 = zero
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = zero
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP10 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp10 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP10 *******************************************
  !
END FUNCTION funcp10

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP9
! *** CASE P9
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp9 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp9(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp9(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp9 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp9(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP9') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp9(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp9

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP9
! *** CASE P9
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp9 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = zero
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    a14 = xk20*(water/gama(20))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -                  &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi14>tiny .AND. water>tiny) THEN !KCL
      psi14 = a14/a13*(psi5+psi8+2.D0*psi12+psi13+2.D0*psi15) - psi6 -     &
           psi7 - 2.D0*psi16 - 2.D0*psi17
      psi14 = min(max(psi14,zero), chi14)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh4no3 = zero
    cnh4cl = zero
    cnacl = zero
    cnano3 = zero
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = max(chi14-psi14, zero)
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP9 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp9 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP9 *******************************************
  !
END FUNCTION funcp9
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP8
! *** CASE P8
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp8 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp8(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp8(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp8 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp8(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP8') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp8(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp8

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP8
! *** CASE P8
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp8 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = chi7
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = zero
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    a14 = xk20*(water/gama(20))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -                  &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi14>tiny .AND. water>tiny) THEN !KCL
      psi14 = a14/a13*(psi5+psi8+2.D0*psi12+psi13+2.D0*psi15) - psi6 -     &
           psi7 - 2.D0*psi16 - 2.D0*psi17
      psi14 = min(max(psi14,zero), chi14)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !     ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !     ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh4no3 = zero
    !      CNH4CL    = ZERO
    cnacl = zero
    cnano3 = zero
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = max(chi14-psi14, zero)
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP8 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp8 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP8 *******************************************
  !
END FUNCTION funcp8
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP7
! *** CASE P7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NANO3, NH4NO3
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp7 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp7(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp7(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp7 (psi6lo)
  END IF

  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp7(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP7') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp7(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp7

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP7
! *** CASE P7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NANO3, NH4NO3
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp7 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  DOUBLE PRECISION :: vita
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = zero
  psi8 = chi8
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = zero
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    a14 = xk20*(water/gama(20))**2.0
    a7 = xk8*(water/gama(1))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -                  &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi14>tiny .AND. water>tiny) THEN !KCL
      psi14 = a14/a13*(psi5+psi8+2.D0*psi12+psi13+2.D0*psi15) - psi6 -     &
           psi7 - 2.D0*psi16 - 2.D0*psi17
      psi14 = min(max(psi14,zero), chi14)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      vita = psi6 + psi14 + psi8 + 2.D0*psi16 + 2.D0*psi17
      gkama = psi8*(2.D0*psi16+psi6+psi14+2.D0*psi17) - a7
      diak = max(vita*vita-4.0D0*gkama, zero)
      psi7 = 0.5D0*(-vita+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh4no3 = zero
    !      CNH4CL    = ZERO
    cnacl = max(chi7-psi7, zero)
    cnano3 = zero
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = max(chi14-psi14, zero)
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP7 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp7 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP7 *******************************************
  !
END FUNCTION funcp7
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP6
! *** CASE P6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL, NANO3
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NH4NO3
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp6 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp6(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp6(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp6 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp6(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP6') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp6(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp6

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP6
! *** CASE P6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL, NANO3
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NH4NO3
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp6 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  DOUBLE PRECISION :: vita
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = zero
  psi8 = zero
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = zero
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    a14 = xk20*(water/gama(20))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = chi5*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -                  &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi14>tiny .AND. water>tiny) THEN !KCL
      psi14 = a14/a13*(psi5+psi8+2.D0*psi12+psi13+2.D0*psi15) - psi6 -     &
           psi7 - 2.D0*psi16 - 2.D0*psi17
      psi14 = min(max(psi14,zero), chi14)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      vita = psi6 + psi14 + psi8 + 2.D0*psi16 + 2.D0*psi17
      gkama = psi8*(2.D0*psi16+psi6+psi14+2.D0*psi17) - a7
      diak = max(vita*vita-4.0D0*gkama, zero)
      psi7 = 0.5D0*(-vita+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !
    IF (chi8>tiny .AND. water>tiny) THEN ! NANO3 DISSOLUTION
      !         VIT  = PSI5+PSI13+PSI7+2.D0*PSI12+2.D0*PSI15
      !         GKAM = PSI7*(2.D0*PSI12+PSI5+PSI13+2.D0*PSI15)-A8
      !         DIA  = MAX(VIT*VIT - 4.0D0*GKAM,ZERO)
      !         PSI8 = 0.5D0*( -VIT + SQRT(DIA) )
      psi8 = a8/a7*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) - psi5 -        &
           2.D0*psi12 - psi13 - 2.D0*psi15
      psi8 = max(min(psi8,chi8), zero)
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    cnh4no3 = zero
    !      CNH4CL    = ZERO
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = max(chi14-psi14, zero)
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP6 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp6 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP6 *******************************************
  !
END FUNCTION funcp6
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP5
! *** CASE P5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, KCL, MGSO4,
!                          NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp5 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  IF (w(4)>tiny) THEN ! NO3 EXIST, WATER POSSIBLE
    scase = 'P5 ; SUBCASE 1'
    CALL calcp5a
    scase = 'P5 ; SUBCASE 1'
  ELSE ! NO3, CL NON EXISTANT
    scase = 'P1 ; SUBCASE 1'
    CALL calcp1a
    scase = 'P1 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmp5) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcp1a
      scase = 'P5 ; SUBCASE 2'
      RETURN
    ELSE
      scase = 'P5 ; SUBCASE 3' ! MDRH REGION (CaSO4, K2SO4, KNO3, KCL, MGSO4,
      !                                                    NANO3, NACL, NH4NO3, NH4CL)
      CALL calcmdrh2(rh, drmp5, drnh4no3, calcp1a, calcp6)
      scase = 'P5 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp5
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP5A
! *** CASE P5A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL, NANO3, NH4NO3
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp5a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp5(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp5(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp5 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp5(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP5') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp5(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp5a

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP5
! *** CASE P5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL, NANO3, NH4NO3
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp5 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi21
  DOUBLE PRECISION :: psi22
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  DOUBLE PRECISION :: vita
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = zero
  psi8 = zero
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = zero
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    a14 = xk20*(water/gama(20))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = (chi5-psi2)*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -           &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi14>tiny .AND. water>tiny) THEN !KCL
      psi14 = a14/a13*(psi5+psi8+2.D0*psi12+psi13+2.D0*psi15) - psi6 -     &
           psi7 - 2.D0*psi16 - 2.D0*psi17
      psi14 = min(max(psi14,zero), chi14)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      vita = psi6 + psi14 + psi8 + 2.D0*psi16 + 2.D0*psi17
      gkama = psi8*(2.D0*psi16+psi6+psi14+2.D0*psi17) - a7
      diak = max(vita*vita-4.0D0*gkama, zero)
      psi7 = 0.5D0*(-vita+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !
    IF (chi8>tiny .AND. water>tiny) THEN ! NANO3 DISSOLUTION
      !         VIT  = PSI5+PSI13+PSI7+2.D0*PSI12+2.D0*PSI15
      !         GKAM = PSI7*(2.D0*PSI12+PSI5+PSI13+2.D0*PSI15)-A8
      !         DIA  = MAX(VIT*VIT - 4.0D0*GKAM,ZERO)
      !         PSI8 = 0.5D0*( -VIT + SQRT(DIA) )
      psi8 = a8/a7*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) - psi5 -        &
           2.D0*psi12 - psi13 - 2.D0*psi15
      psi8 = max(min(psi8,chi8), zero)
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !C
    !C *** CALCULATE H+ *****************************************************
    !C
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    !      CNH4NO3   = ZERO
    !      CNH4CL    = ZERO
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = max(chi14-psi14, zero)
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    ! *** NH4NO3(s) calculations
    !
    a2 = xk10/(r*temp*r*temp)
    IF (gnh3*ghno3>a2) THEN
      delt = min(gnh3, ghno3)
      bb = -(gnh3+ghno3)
      cc = gnh3*ghno3 - a2
      dd = bb*bb - 4.D0*cc
      psi21 = 0.5D0*(-bb+sqrt(dd))
      psi22 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi21>zero .AND. psi21>zero) THEN
        psi2 = psi21
      ELSE IF (delt-psi22>zero .AND. psi22>zero) THEN
        psi2 = psi22
      ELSE
        psi2 = zero
      END IF
    ELSE
      psi2 = zero
    END IF
    psi2 = max(min(min(psi2,chi4-psi4-psi3),chi5-psi5), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi2, tiny)
    ghcl = max(ghno3-psi2, tiny)
    cnh4no3 = psi2
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP5 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp5 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP5 *******************************************
  !
END FUNCTION funcp5
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP4
! *** CASE P4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp4 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  IF (w(4)>tiny) THEN ! NO3 EXIST, WATER POSSIBLE
    scase = 'P4 ; SUBCASE 1'
    CALL calcp4a
    scase = 'P4 ; SUBCASE 1'
  ELSE ! NO3, CL NON EXISTANT
    scase = 'P1 ; SUBCASE 1'
    CALL calcp1a
    scase = 'P1 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmp4) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcp1a
      scase = 'P4 ; SUBCASE 2'
      RETURN
    ELSE
      scase = 'P4 ; SUBCASE 3' ! MDRH REGION (CaSO4, K2SO4, KNO3, KCL, MGSO4,
      !                                                    MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL)
      CALL calcmdrh2(rh, drmp4, drmgno32, calcp1a, calcp5a)
      scase = 'P4 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp4
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP4A
! *** CASE P4A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL, NANO3, NH4NO3, MG(NO3)2
!     4. Completely dissolved: CA(NO3)2, CACL2, MGCL2
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp4a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp4(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp4(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp4 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp4(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP4') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp4(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp4a

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP4
! *** CASE P4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL, NANO3, NH4NO3, MG(NO3)2
!     4. Completely dissolved: CA(NO3)2, CACL2, MGCL2
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp4 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi21
  DOUBLE PRECISION :: psi22
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  DOUBLE PRECISION :: vita
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = zero
  psi8 = zero
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = zero
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    a14 = xk20*(water/gama(20))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = (chi5-psi2)*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -           &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi14>tiny .AND. water>tiny) THEN !KCL
      psi14 = a14/a13*(psi5+psi8+2.D0*psi12+psi13+2.D0*psi15) - psi6 -     &
           psi7 - 2.D0*psi16 - 2.D0*psi17
      psi14 = min(max(psi14,zero), chi14)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      vita = psi6 + psi14 + psi8 + 2.D0*psi16 + 2.D0*psi17
      gkama = psi8*(2.D0*psi16+psi6+psi14+2.D0*psi17) - a7
      diak = max(vita*vita-4.0D0*gkama, zero)
      psi7 = 0.5D0*(-vita+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !
    IF (chi8>tiny .AND. water>tiny) THEN ! NANO3 DISSOLUTION
      !         VIT  = PSI5+PSI13+PSI7+2.D0*PSI12+2.D0*PSI15
      !         GKAM = PSI7*(2.D0*PSI12+PSI5+PSI13+2.D0*PSI15)-A8
      !         DIA  = MAX(VIT*VIT - 4.0D0*GKAM,ZERO)
      !         PSI8 = 0.5D0*( -VIT + SQRT(DIA) )
      psi8 = a8/a7*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) - psi5 -        &
           2.D0*psi12 - psi13 - 2.D0*psi15
      psi8 = max(min(psi8,chi8), zero)
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    !      CNH4CL    = ZERO
    !      CNH4NO3   = ZERO
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = max(chi14-psi14, zero)
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    ! *** NH4NO3(s) calculations
    !
    a2 = xk10/(r*temp*r*temp)
    IF (gnh3*ghno3>a2) THEN
      delt = min(gnh3, ghno3)
      bb = -(gnh3+ghno3)
      cc = gnh3*ghno3 - a2
      dd = bb*bb - 4.D0*cc
      psi21 = 0.5D0*(-bb+sqrt(dd))
      psi22 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi21>zero .AND. psi21>zero) THEN
        psi2 = psi21
      ELSE IF (delt-psi22>zero .AND. psi22>zero) THEN
        psi2 = psi22
      ELSE
        psi2 = zero
      END IF
    ELSE
      psi2 = zero
    END IF
    psi2 = max(min(min(psi2,chi4-psi4-psi3),chi5-psi5), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi2, tiny)
    ghcl = max(ghno3-psi2, tiny)
    cnh4no3 = psi2
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP4 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp4 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP4 *******************************************
  !
END FUNCTION funcp4

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP3
! *** CASE P3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  IF (w(4)>tiny .AND. w(5)>tiny) THEN ! NO3,CL EXIST, WATER POSSIBLE
    scase = 'P3 ; SUBCASE 1'
    CALL calcp3a
    scase = 'P3 ; SUBCASE 1'
  ELSE ! NO3, CL NON EXISTANT
    scase = 'P1 ; SUBCASE 1'
    CALL calcp1a
    scase = 'P1 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmp3) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcp1a
      scase = 'P3 ; SUBCASE 2'
      RETURN
    ELSE
      scase = 'P3 ; SUBCASE 3' ! MDRH REGION (CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
      !                                                    MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL)
      CALL calcmdrh2(rh, drmp3, drcano32, calcp1a, calcp4a)
      scase = 'P3 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp3
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP3A
! *** CASE P3A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL,
!                          NANO3, NH4NO3, MG(NO3)2, CA(NO3)2
!     4. Completely dissolved: CACL2, MGCL2
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp3a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp3(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp3(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp3 (psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp3(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP3') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp3(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp3a

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP3
! *** CASE P3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL,
!                          NANO3, NH4NO3, MG(NO3)2, CA(NO3)2
!     4. Completely dissolved: CACL2, MGCL2
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp3 ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi21
  DOUBLE PRECISION :: psi22
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  DOUBLE PRECISION :: vita
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = zero
  psi8 = zero
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = zero
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    a14 = xk20*(water/gama(20))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = (chi5-psi2)*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -           &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi14>tiny .AND. water>tiny) THEN !KCL
      psi14 = a14/a13*(psi5+psi8+2.D0*psi12+psi13+2.D0*psi15) - psi6 -     &
           psi7 - 2.D0*psi16 - 2.D0*psi17
      psi14 = min(max(psi14,zero), chi14)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      vita = psi6 + psi14 + psi8 + 2.D0*psi16 + 2.D0*psi17
      gkama = psi8*(2.D0*psi16+psi6+psi14+2.D0*psi17) - a7
      diak = max(vita*vita-4.0D0*gkama, zero)
      psi7 = 0.5D0*(-vita+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !
    IF (chi8>tiny .AND. water>tiny) THEN ! NANO3 DISSOLUTION
      !         VIT  = PSI5+PSI13+PSI7+2.D0*PSI12+2.D0*PSI15
      !         GKAM = PSI7*(2.D0*PSI12+PSI5+PSI13+2.D0*PSI15)-A8
      !         DIA  = MAX(VIT*VIT - 4.0D0*GKAM,ZERO)
      !         PSI8 = 0.5D0*( -VIT + SQRT(DIA) )
      psi8 = a8/a7*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) - psi5 -        &
           2.D0*psi12 - psi13 - 2.D0*psi15
      psi8 = max(min(psi8,chi8), zero)
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    !      CNH4CL    = ZERO
    !      CNH4NO3   = ZERO
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = max(chi14-psi14, zero)
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    ! *** NH4NO3(s) calculations
    !
    a2 = xk10/(r*temp*r*temp)
    IF (gnh3*ghno3>a2) THEN
      delt = min(gnh3, ghno3)
      bb = -(gnh3+ghno3)
      cc = gnh3*ghno3 - a2
      dd = bb*bb - 4.D0*cc
      psi21 = 0.5D0*(-bb+sqrt(dd))
      psi22 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi21>zero .AND. psi21>zero) THEN
        psi2 = psi21
      ELSE IF (delt-psi22>zero .AND. psi22>zero) THEN
        psi2 = psi22
      ELSE
        psi2 = zero
      END IF
    ELSE
      psi2 = zero
    END IF
    psi2 = max(min(min(psi2,chi4-psi4-psi3),chi5-psi5), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi2, tiny)
    ghcl = max(ghno3-psi2, tiny)
    cnh4no3 = psi2
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP3 = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp3 = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP3 *******************************************
  !
END FUNCTION funcp3

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP2
! *** CASE P2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
!     THERE ARE THREE REGIMES IN THIS CASE:
!     1. CACL2(s) POSSIBLE. LIQUID & SOLID AEROSOL (SUBROUTINE CALCL2A)
!     2. CACL2(s) NOT POSSIBLE, AND RH < MDRH. SOLID AEROSOL ONLY
!     3. CACL2(s) NOT POSSIBLE, AND RH >= MDRH. SOLID & LIQUID AEROSOL
!
!     REGIMES 2. AND 3. ARE CONSIDERED TO BE THE SAME AS CASES P1A, P2B
!     RESPECTIVELY
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
!
SUBROUTINE calcp2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcp1a
  !
  ! *** REGIME DEPENDS UPON THE POSSIBLE SOLIDS & RH **********************
  !
  IF (ccacl2>tiny) THEN
    scase = 'P2 ; SUBCASE 1'
    CALL calcp2a
    scase = 'P2 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmp2) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcp1a
      scase = 'P2 ; SUBCASE 2'
    ELSE
      IF (cmgcl2>tiny) THEN
        scase = 'P2 ; SUBCASE 3' ! MDRH (CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4, MGCL2,
        !                                                  MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL)
        CALL calcmdrh2(rh, drmp2, drmgcl2, calcp1a, calcp3a)
        scase = 'P2 ; SUBCASE 3'
      END IF
      IF (water<=tiny .AND. rh>=drmp3 .AND. rh<drmp4) THEN
        scase = 'P2 ; SUBCASE 4' ! MDRH (CaSO4, K2SO4, KNO3, KCL, MGSO4, CANO32,
        !                                                  MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL)
        CALL calcmdrh2(rh, drmp3, drcano32, calcp1a, calcp4a)
        scase = 'P2 ; SUBCASE 4'
      END IF
      IF (water<=tiny .AND. rh>=drmp4 .AND. rh<drmp5) THEN
        scase = 'P2 ; SUBCASE 5' ! MDRH (CaSO4, K2SO4, KNO3, KCL, MGSO4,
        !                                                  MGNO32, NANO3, NACL, NH4NO3, NH4CL)
        CALL calcmdrh2(rh, drmp4, drmgno32, calcp1a, calcp5a)
        scase = 'P2 ; SUBCASE 5'
      END IF
      IF (water<=tiny .AND. rh>=drmp5) THEN
        scase = 'P2 ; SUBCASE 6' ! MDRH (CaSO4, K2SO4, KNO3, KCL, MGSO4,
        !                                                  NANO3, NACL, NH4NO3, NH4CL)
        CALL calcmdrh2(rh, drmp5, drnh4no3, calcp1a, calcp6)
        scase = 'P2 ; SUBCASE 6'
      ELSE
        water = tiny
        DO i = 1, nions
          molal(i) = zero
        END DO
        CALL calcp1a
        scase = 'P2 ; SUBCASE 2'
      END IF
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp2
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP2A
! *** CASE P2A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL,
!                          NANO3, NH4NO3, MG(NO3)2, CA(NO3)2
!     4. Completely dissolved: CACL2
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: dx
  DOUBLE PRECISION :: frca
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE.
  chi11 = min(w(2), w(6)) ! CCASO4
  frca = max(w(6)-chi11, zero)
  frso4 = max(w(2)-chi11, zero)
  chi9 = min(frso4, 0.5D0*w(7)) ! CK2SO4
  frk = max(w(7)-2.D0*chi9, zero)
  frso4 = max(frso4-chi9, zero)
  chi10 = frso4 ! CMGSO4
  frmg = max(w(8)-chi10, zero)
  chi7 = min(w(1), w(5)) ! CNACL
  frna = max(w(1)-chi7, zero)
  frcl = max(w(5)-chi7, zero)
  chi12 = min(frca, 0.5D0*w(4)) ! CCANO32
  frca = max(frca-chi12, zero)
  frno3 = max(w(4)-2.D0*chi12, zero)
  chi17 = min(frca, 0.5D0*frcl) ! CCACL2
  frca = max(frca-chi17, zero)
  frcl = max(frcl-2.D0*chi17, zero)
  chi15 = min(frmg, 0.5D0*frno3) ! CMGNO32
  frmg = max(frmg-chi15, zero)
  frno3 = max(frno3-2.D0*chi15, zero)
  chi16 = min(frmg, 0.5D0*frcl) ! CMGCL2
  frmg = max(frmg-chi16, zero)
  frcl = max(frcl-2.D0*chi16, zero)
  chi8 = min(frna, frno3) ! CNANO3
  frna = max(frna-chi8, zero)
  frno3 = max(frno3-chi8, zero)
  chi14 = min(frk, frcl) ! CKCL
  frk = max(frk-chi14, zero)
  frcl = max(frcl-chi14, zero)
  chi13 = min(frk, frno3) ! CKNO3
  frk = max(frk-chi13, zero)
  frno3 = max(frno3-chi13, zero)
  !
  chi5 = frno3 ! HNO3(g)
  chi6 = frcl ! HCL(g)
  chi4 = w(3) ! NH3(g)
  !
  chi3 = zero ! CNH4CL
  chi1 = zero
  chi2 = zero
  !
  psi6lo = tiny
  psi6hi = chi6 - tiny ! MIN(CHI6-TINY, CHI4)
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi6lo
  y1 = funcp2a(x1)
  IF (abs(y1)<=eps .OR. chi6<=tiny) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 + dx
    y2 = funcp2a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION; IF ABS(Y2)<EPS SOLUTION IS ASSUMED
  !
  IF (abs(y2) > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcp2a(psi6lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcp2a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCP2A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcp2a(x3)
  !
  ! *** CALCULATE HSO4 SPECIATION AND RETURN *******************************
  !
120 CONTINUE
  IF (molal(1)>tiny .AND. molal(5)>tiny) THEN
    CALL calchs4(molal(1), molal(5), zero, delta)
    molal(1) = molal(1) - delta ! H+   EFFECT
    molal(5) = molal(5) - delta ! SO4  EFFECT
    molal(6) = delta ! HSO4 EFFECT
  END IF
  !
  RETURN
  !
END SUBROUTINE calcp2a

!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCP2A
! *** CASE P2A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL,
!                          NANO3, NH4NO3, MG(NO3)2, CA(NO3)2, MGCL2
!     4. Completely dissolved: CACL2
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcp2a ( x )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: x
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bbp
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: ccp
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: ddp
  DOUBLE PRECISION :: delt
  DOUBLE PRECISION :: delta
  DOUBLE PRECISION :: diak
  DOUBLE PRECISION :: gkama
  DOUBLE PRECISION :: hi
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: psi21
  DOUBLE PRECISION :: psi22
  DOUBLE PRECISION :: psi31
  DOUBLE PRECISION :: psi32
  DOUBLE PRECISION :: smin
  DOUBLE PRECISION :: vhta
  DOUBLE PRECISION :: vita
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = x
  psi1 = zero
  psi2 = zero
  psi3 = zero
  psi7 = zero
  psi8 = zero
  psi9 = zero
  psi10 = chi10
  psi11 = zero
  psi12 = chi12
  psi13 = zero
  psi14 = zero
  psi15 = chi15
  psi16 = chi16
  psi17 = chi17
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2.0
    a5 = xk4*r*temp*(water/gama(10))**2.0
    a6 = xk3*r*temp*(water/gama(11))**2.0
    a9 = xk17*(water/gama(17))**3.0
    a13 = xk19*(water/gama(19))**2.0
    a14 = xk20*(water/gama(20))**2.0
    a7 = xk8*(water/gama(1))**2.0
    a8 = xk9*(water/gama(3))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = (chi5-psi2)*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) -           &
         a6/a5*(psi8+2.D0*psi12+psi13+2.D0*psi15)*(chi6-psi6-psi3)
    psi5 = psi5/(a6/a5*(chi6-psi6-psi3)+psi6+psi7+psi14+2.D0*psi16+2.D0*   &
         psi17)
    psi5 = min(max(psi5,tiny), chi5)
    !
    IF (w(3)>tiny .AND. water>tiny) THEN ! First try 3rd order soln
      bb = -(chi4+psi6+psi5+1.D0/a4)
      cc = chi4*(psi5+psi6)
      dd = max(bb*bb-4.D0*cc, zero)
      psi4 = 0.5D0*(-bb-sqrt(dd))
      psi4 = min(max(psi4,zero), chi4)
    ELSE
      psi4 = tiny
    END IF
    !
    IF (chi13>tiny .AND. water>tiny) THEN !KNO3
      vhta = psi5 + psi8 + 2.D0*psi12 + 2.D0*psi15 + psi14 + 2.D0*psi9
      gkama = (psi5+psi8+2.D0*psi12+2.D0*psi15)*(2.D0*psi9+psi14) - a13
      delta = max(vhta*vhta-4.D0*gkama, zero)
      psi13 = 0.5D0*(-vhta+sqrt(delta))
      psi13 = min(max(psi13,zero), chi13)
    END IF
    !
    IF (chi14>tiny .AND. water>tiny) THEN !KCL
      psi14 = a14/a13*(psi5+psi8+2.D0*psi12+psi13+2.D0*psi15) - psi6 -     &
           psi7 - 2.D0*psi16 - 2.D0*psi17
      psi14 = min(max(psi14,zero), chi14)
    END IF
    !
    IF (chi9>tiny .AND. water>tiny) THEN !K2SO4
      bbp = psi10 + psi13 + psi14
      ccp = (psi13+psi14)*(0.25D0*(psi13+psi14)+psi10)
      ddp = 0.25D0*(psi13+psi14)**2.0*psi10 - a9/4.D0
      CALL poly3(bbp, ccp, ddp, psi9, islv)
      IF (islv==0) THEN
        psi9 = min(max(psi9,zero), chi9)
      ELSE
        psi9 = zero
      END IF
    END IF
    !
    IF (chi7>tiny .AND. water>tiny) THEN ! NACL DISSOLUTION
      vita = psi6 + psi14 + psi8 + 2.D0*psi16 + 2.D0*psi17
      gkama = psi8*(2.D0*psi16+psi6+psi14+2.D0*psi17) - a7
      diak = max(vita*vita-4.0D0*gkama, zero)
      psi7 = 0.5D0*(-vita+sqrt(diak))
      psi7 = max(min(psi7,chi7), zero)
    END IF
    !
    IF (chi8>tiny .AND. water>tiny) THEN ! NANO3 DISSOLUTION
      !         VIT  = PSI5+PSI13+PSI7+2.D0*PSI12+2.D0*PSI15
      !         GKAM = PSI7*(2.D0*PSI12+PSI5+PSI13+2.D0*PSI15)-A8
      !         DIA  = MAX(VIT*VIT - 4.0D0*GKAM,ZERO)
      !         PSI8 = 0.5D0*( -VIT + SQRT(DIA) )
      psi8 = a8/a7*(psi6+psi7+psi14+2.D0*psi16+2.D0*psi17) - psi5 -        &
           2.D0*psi12 - psi13 - 2.D0*psi15
      psi8 = max(min(psi8,chi8), zero)
    END IF
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(2) = psi8 + psi7 ! NAI
    molal(3) = psi4 ! NH4I
    molal(4) = psi6 + psi7 + psi14 + 2.D0*psi16 + 2.D0*psi17 ! CLI
    molal(5) = psi9 + psi10 ! SO4I
    molal(6) = zero ! HSO4I
    molal(7) = psi5 + psi8 + 2.D0*psi12 + psi13 + 2.D0*psi15 ! NO3I
    molal(8) = psi11 + psi12 + psi17 ! CAI
    molal(9) = 2.D0*psi9 + psi13 + psi14 ! KI
    molal(10) = psi10 + psi15 + psi16 ! MGI
    !
    ! *** CALCULATE H+ *****************************************************
    !
    !      REST  = 2.D0*W(2) + W(4) + W(5)
    !C
    !      DELT1 = 0.0d0
    !      DELT2 = 0.0d0
    !      IF (W(1)+W(6)+W(7)+W(8).GT.REST) THEN
    !C
    !C *** CALCULATE EQUILIBRIUM CONSTANTS **********************************
    !C
    !      ALFA1 = XK26*RH*(WATER/1.0)                   ! CO2(aq) + H2O
    !      ALFA2 = XK27*(WATER/1.0)                      ! HCO3-
    !C
    !      X     = W(1)+W(6)+W(7)+W(8) - REST            ! EXCESS OF CRUSTALS EQUALS CO2(aq)
    !C
    !      DIAK  = SQRT( (ALFA1)**2.0 + 4.0D0*ALFA1*X)
    !      DELT1 = 0.5*(-ALFA1 + DIAK)
    !      DELT1 = MIN ( MAX (DELT1, ZERO), X)
    !      DELT2 = ALFA2
    !      DELT2 = MIN ( DELT2, DELT1)
    !      MOLAL(1) = DELT1 + DELT2                      ! H+
    !      ELSE
    !C
    !C *** NO EXCESS OF CRUSTALS CALCULATE H+ *******************************
    !C
    smin = 2.D0*molal(5) + molal(7) + molal(4) - molal(2) - molal(3) -     &
         molal(9) - 2.D0*molal(10) - 2.D0*molal(8)
    CALL calcph(smin, hi, ohi)
    molal(1) = hi
    !      ENDIF
    !
    gnh3 = max(chi4-psi4, tiny)
    ghno3 = max(chi5-psi5, tiny)
    ghcl = max(chi6-psi6, tiny)
    !
    !      CNH4CL    = ZERO
    !      CNH4NO3   = ZERO
    cnacl = max(chi7-psi7, zero)
    cnano3 = max(chi8-psi8, zero)
    ck2so4 = max(chi9-psi9, zero)
    cmgso4 = zero
    ccaso4 = chi11
    ccano32 = zero
    ckno3 = max(chi13-psi13, zero)
    ckcl = max(chi14-psi14, zero)
    cmgno32 = zero
    cmgcl2 = zero
    ccacl2 = zero
    !
    ! *** NH4Cl(s) calculations
    !
    a3 = xk6/(r*temp*r*temp)
    IF (gnh3*ghcl>a3) THEN
      delt = min(gnh3, ghcl)
      bb = -(gnh3+ghcl)
      cc = gnh3*ghcl - a3
      dd = bb*bb - 4.D0*cc
      psi31 = 0.5D0*(-bb+sqrt(dd))
      psi32 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi31>zero .AND. psi31>zero) THEN
        psi3 = psi31
      ELSE IF (delt-psi32>zero .AND. psi32>zero) THEN
        psi3 = psi32
      ELSE
        psi3 = zero
      END IF
    ELSE
      psi3 = zero
    END IF
    psi3 = max(min(min(psi3,chi4-psi4),chi6-psi6), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi3, tiny)
    ghcl = max(ghcl-psi3, tiny)
    cnh4cl = psi3
    !
    ! *** NH4NO3(s) calculations
    !
    a2 = xk10/(r*temp*r*temp)
    IF (gnh3*ghno3>a2) THEN
      delt = min(gnh3, ghno3)
      bb = -(gnh3+ghno3)
      cc = gnh3*ghno3 - a2
      dd = bb*bb - 4.D0*cc
      psi21 = 0.5D0*(-bb+sqrt(dd))
      psi22 = 0.5D0*(-bb-sqrt(dd))
      IF (delt-psi21>zero .AND. psi21>zero) THEN
        psi2 = psi21
      ELSE IF (delt-psi22>zero .AND. psi22>zero) THEN
        psi2 = psi22
      ELSE
        psi2 = zero
      END IF
    ELSE
      psi2 = zero
    END IF
    psi2 = max(min(min(psi2,chi4-psi4-psi3),chi5-psi5), zero)
    !
    ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
    !
    gnh3 = max(gnh3-psi2, tiny)
    ghcl = max(ghno3-psi2, tiny)
    cnh4no3 = psi2
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
  !20    FUNCP2A = MOLAL(3)*MOLAL(4)/GHCL/GNH3/A6/A4 - ONE
100 funcp2a = molal(1)*molal(4)/ghcl/a6 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCP2A *******************************************
  !
END FUNCTION funcp2a


!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP1
! *** CASE P1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCP1A)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcp1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmp1) THEN
    scase = 'P1 ; SUBCASE 1'
    CALL calcp1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'P1 ; SUBCASE 1'
  ELSE
    scase = 'P1 ; SUBCASE 2' ! LIQUID & SOLID PHASE POSSIBLE
    CALL calcmdrh2(rh, drmp1, drcacl2, calcp1a, calcp2a)
    scase = 'P1 ; SUBCASE 2'
  END IF
  !
  !
  RETURN
  !
END SUBROUTINE calcp1
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCP1A
! *** CASE P1A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================

SUBROUTINE calcp1a (  )
  USE mo_isrpia  
  IMPLICIT NONE
  DOUBLE PRECISION :: a1
  DOUBLE PRECISION :: a2
  DOUBLE PRECISION :: alf
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bet
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: clfr
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: dd1
  DOUBLE PRECISION :: dd2
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: gam
  DOUBLE PRECISION :: kapa
  DOUBLE PRECISION :: kapa1
  DOUBLE PRECISION :: kapa2
  DOUBLE PRECISION :: lamda
  DOUBLE PRECISION :: lamda1
  DOUBLE PRECISION :: lamda2
  DOUBLE PRECISION :: nafr
  DOUBLE PRECISION :: no3fr
  DOUBLE PRECISION :: rtsq
  DOUBLE PRECISION :: so4fr
  DOUBLE PRECISION :: sqdd
  DOUBLE PRECISION :: sqdd1
  DOUBLE PRECISION :: sqdd2
  DOUBLE PRECISION :: theta1
  DOUBLE PRECISION :: theta2
  !
  ! *** CALCULATE NON VOLATILE SOLIDS ***********************************
  !
  ccaso4 = min(w(2), w(6)) !SOLID CASO4
  cafr = max(w(6)-ccaso4, zero)
  so4fr = max(w(2)-ccaso4, zero)
  ck2so4 = min(so4fr, 0.5D0*w(7)) !SOLID K2SO4
  frk = max(w(7)-2.D0*ck2so4, zero)
  so4fr = max(so4fr-ck2so4, zero)
  cmgso4 = so4fr !SOLID MGSO4
  frmg = max(w(8)-cmgso4, zero)
  cnacl = min(w(1), w(5)) !SOLID NACL
  nafr = max(w(1)-cnacl, zero)
  clfr = max(w(5)-cnacl, zero)
  ccano32 = min(cafr, 0.5D0*w(4)) !SOLID CA(NO3)2
  cafr = max(cafr-ccano32, zero)
  no3fr = max(w(4)-2.D0*ccano32, zero)
  ccacl2 = min(cafr, 0.5D0*clfr) !SOLID CACL2
  cafr = max(cafr-ccacl2, zero)
  clfr = max(clfr-2.D0*ccacl2, zero)
  cmgno32 = min(frmg, 0.5D0*no3fr) !SOLID MG(NO3)2
  frmg = max(frmg-cmgno32, zero)
  no3fr = max(no3fr-2.D0*cmgno32, zero)
  cmgcl2 = min(frmg, 0.5D0*clfr) !SOLID MGCL2
  frmg = max(frmg-cmgcl2, zero)
  clfr = max(clfr-2.D0*cmgcl2, zero)
  cnano3 = min(nafr, no3fr) !SOLID NANO3
  nafr = max(nafr-cnano3, zero)
  no3fr = max(no3fr-cnano3, zero)
  ckcl = min(frk, clfr) !SOLID KCL
  frk = max(frk-ckcl, zero)
  clfr = max(clfr-ckcl, zero)
  ckno3 = min(frk, no3fr) !SOLID KNO3
  frk = max(frk-ckno3, zero)
  no3fr = max(no3fr-ckno3, zero)
  !
  ! *** CALCULATE VOLATILE SPECIES **************************************
  !
  alf = w(3) ! FREE NH3
  bet = clfr ! FREE CL
  gam = no3fr ! FREE NO3
  !
  rtsq = r*temp*r*temp
  a1 = xk6/rtsq
  a2 = xk10/rtsq
  !
  theta1 = gam - bet*(a2/a1)
  theta2 = a2/a1
  !
  ! QUADRATIC EQUATION SOLUTION
  !
  bb = (theta1-alf-bet*(one+theta2))/(one+theta2)
  cc = (alf*bet-a1-bet*theta1)/(one+theta2)
  dd = bb*bb - 4.0D0*cc
  IF (dd<zero) GO TO 100 ! Solve each reaction seperately
  !
  ! TWO ROOTS FOR KAPA, CHECK AND SEE IF ANY VALID
  !
  sqdd = sqrt(dd)
  kapa1 = 0.5D0*(-bb+sqdd)
  kapa2 = 0.5D0*(-bb-sqdd)
  lamda1 = theta1 + theta2*kapa1
  lamda2 = theta1 + theta2*kapa2
  !
  IF (kapa1>=zero .AND. lamda1>=zero) THEN
    IF (alf-kapa1-lamda1>=zero .AND. bet-kapa1>=zero .AND.                 &
         gam-lamda1>=zero) THEN
      kapa = kapa1
      lamda = lamda1
      GO TO 110
    END IF
  END IF
  !
  IF (kapa2>=zero .AND. lamda2>=zero) THEN
    IF (alf-kapa2-lamda2>=zero .AND. bet-kapa2>=zero .AND.                 &
         gam-lamda2>=zero) THEN
      kapa = kapa2
      lamda = lamda2
      GO TO 110
    END IF
  END IF
  !
  ! SEPERATE SOLUTION OF NH4CL & NH4NO3 EQUILIBRIA
  !
100 kapa = zero
  lamda = zero
  dd1 = (alf+bet)*(alf+bet) - 4.0D0*(alf*bet-a1)
  dd2 = (alf+gam)*(alf+gam) - 4.0D0*(alf*gam-a2)
  !
  ! NH4CL EQUILIBRIUM
  !
  IF (dd1>=zero) THEN
    sqdd1 = sqrt(dd1)
    kapa1 = 0.5D0*(alf+bet+sqdd1)
    kapa2 = 0.5D0*(alf+bet-sqdd1)
    !
    IF (kapa1>=zero .AND. kapa1<=min(alf,bet)) THEN
      kapa = kapa1
    ELSE IF (kapa2>=zero .AND. kapa2<=min(alf,bet)) THEN
      kapa = kapa2
    ELSE
      kapa = zero
    END IF
  END IF
  !
  ! NH4NO3 EQUILIBRIUM
  !
  IF (dd2>=zero) THEN
    sqdd2 = sqrt(dd2)
    lamda1 = 0.5D0*(alf+gam+sqdd2)
    lamda2 = 0.5D0*(alf+gam-sqdd2)
    !
    IF (lamda1>=zero .AND. lamda1<=min(alf,gam)) THEN
      lamda = lamda1
    ELSE IF (lamda2>=zero .AND. lamda2<=min(alf,gam)) THEN
      lamda = lamda2
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! IF BOTH KAPA, LAMDA ARE > 0, THEN APPLY EXISTANCE CRITERION
  !
  IF (kapa>zero .AND. lamda>zero) THEN
    IF (bet<lamda/theta1) THEN
      kapa = zero
    ELSE
      lamda = zero
    END IF
  END IF
  !
  ! *** CALCULATE COMPOSITION OF VOLATILE SPECIES ***********************
  !
110 CONTINUE
  cnh4no3 = lamda
  cnh4cl = kapa
  !
  gnh3 = alf - kapa - lamda
  ghno3 = gam - lamda
  ghcl = bet - kapa
  !
  RETURN
  !
END SUBROUTINE calcp1a
!
!======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCL9
! *** CASE L9
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : CASO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC, (NH4)2SO4, KHSO4, MGSO4, NA2SO4, K2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl9 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  DOUBLE PRECISION :: lamda
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcl1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCL1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  chi6 = ck2so4
  chi7 = cmgso4
  chi8 = ckhso4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = clc
  psi3 = cnahso4
  psi4 = cna2so4
  psi5 = cnh42s4
  psi6 = ck2so4
  psi7 = cmgso4
  psi8 = ckhso4
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk1*water/gama(7)*(gama(8)/gama(7))**2.
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    bb = psi7 + psi6 + psi5 + psi4 + psi2 + a9 ! LAMDA
    cc = -a9*(psi8+psi1+psi2+psi3)
    dd = max(bb*bb-4.D0*cc, zero)
    lamda = 0.5D0*(-bb+sqrt(dd))
    lamda = min(max(lamda,tiny), psi8+psi3+psi2+psi1)
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 + psi7 + lamda ! SO4I
    molal(6) = psi2 + psi3 + psi1 + psi8 - lamda ! HSO4I
    molal(9) = psi8 + 2.0D0*psi6 ! KI
    molal(10) = psi7 ! MGI
    !
    clc = zero
    cnahso4 = zero
    cna2so4 = zero
    cnh42s4 = zero
    cnh4hs4 = zero
    ck2so4 = zero
    cmgso4 = zero
    ckhso4 = zero
    !
    CALL calcmr ! Water content

    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
100 RETURN
  !
END SUBROUTINE calcl9
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCL8
! *** CASE L8
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC, (NH4)2SO4, KHSO4, MGSO4, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl8 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi6hi
  DOUBLE PRECISION :: psi6lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcl1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCL1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  chi6 = ck2so4
  chi7 = cmgso4
  chi8 = ckhso4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = clc
  psi3 = cnahso4
  psi4 = cna2so4
  psi5 = cnh42s4
  psi6 = zero
  psi7 = cmgso4
  psi8 = ckhso4
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi6lo = zero ! Low  limit
  psi6hi = chi6 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  IF (chi6<=tiny) THEN
    y1 = funcl8(zero)
    GO TO 120
  END IF
  !
  x1 = psi6hi
  y1 = funcl8(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH K2SO4 *********
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi6hi-psi6lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    y2 = funcl8(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH K2SO4
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl8(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCL8') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl8(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCL8') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcl8(x3)
  !
120 RETURN
  !
END SUBROUTINE calcl8
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** FUNCTION FUNCL8
! *** CASE L8
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC, (NH4)2SO4, KHSO4, MGSO4, NA2SO4
!
!     SOLUTION IS SAVED IN COMMON BLOCK /CASE/
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcl8 ( p6 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p6
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  DOUBLE PRECISION :: lamda
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi6 = p6
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk1*(water)*(gama(8)**2.0)/(gama(7)**3.0)
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    bb = psi7 + psi6 + psi5 + psi4 + psi2 + a9 ! LAMDA
    cc = -a9*(psi8+psi1+psi2+psi3)
    dd = bb*bb - 4.D0*cc
    lamda = 0.5D0*(-bb+sqrt(dd))
    lamda = min(max(lamda,tiny), psi8+psi3+psi2+psi1)
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 + psi7 + lamda ! SO4I
    molal(6) = max(psi2+psi3+psi1+psi8-lamda, tiny) ! HSO4I
    molal(9) = psi8 + 2.0*psi6 ! KI
    molal(10) = psi7 ! MGI
    !
    clc = zero
    cnahso4 = zero
    cna2so4 = zero
    cnh42s4 = zero
    cnh4hs4 = zero
    ck2so4 = max(chi6-psi6, zero)
    cmgso4 = zero
    ckhso4 = zero
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a6 = xk17*(water/gama(17))**3.0
  funcl8 = molal(9)*molal(9)*molal(5)/a6 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCL8 ****************************************
  !
END FUNCTION funcl8
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL7
! *** CASE L7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC, (NH4)2SO4, KHSO4, MGSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl7 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcl1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCL1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  chi6 = ck2so4
  chi7 = cmgso4
  chi8 = ckhso4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = clc
  psi3 = cnahso4
  psi4 = zero
  psi5 = cnh42s4
  psi6 = zero
  psi7 = cmgso4
  psi8 = ckhso4
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi4lo = zero ! Low  limit
  psi4hi = chi4 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  IF (chi4<=tiny) THEN
    y1 = funcl7(zero)
    GO TO 120
  END IF
  !
  x1 = psi4hi
  y1 = funcl7(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH K2SO4 *********
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl7(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH K2SO4
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl7(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCL7') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl7(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCL7') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcl7(x3)
  !
120 RETURN
  !
END SUBROUTINE calcl7
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** FUNCTION FUNCL7
! *** CASE L7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC, (NH4)2SO4, KHSO4, MGSO4
!
!     SOLUTION IS SAVED IN COMMON BLOCK /CASE/
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcl7 ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: lamda
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi4 = p4
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = xk5*(water/gama(2))**3.0
    a6 = xk17*(water/gama(17))**3.0
    a9 = xk1*(water)*(gama(8)**2.0)/(gama(7)**3.0)
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    !      PSI6 = 0.5*(SQRT(A6/A4)*(2.D0*PSI4+PSI3)-PSI8)             ! PSI6
    !      PSI6 = MIN (MAX (PSI6, ZERO), CHI6)
    !
    IF (chi6>tiny .AND. water>tiny) THEN
      aa = psi5 + psi4 + psi2 + psi7 + psi8 + lamda
      bb = psi8*(psi5+psi4+psi2+psi7+0.25D0*psi8+lamda)
      cc = 0.25D0*(psi8*psi8*(psi5+psi4+psi2+psi7+lamda)-a6)
      CALL poly3(aa, bb, cc, psi6, islv)
      IF (islv==0) THEN
        psi6 = min(psi6, chi6)
      ELSE
        psi6 = zero
      END IF
    END IF
    !
    bb = psi7 + psi6 + psi5 + psi4 + psi2 + a9 ! LAMDA
    cc = -a9*(psi8+psi1+psi2+psi3)
    dd = bb*bb - 4.D0*cc
    lamda = 0.5D0*(-bb+sqrt(dd))
    lamda = min(max(lamda,tiny), psi8+psi3+psi2+psi1)
    !
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 + psi7 + lamda ! SO4I
    molal(6) = max(psi2+psi3+psi1+psi8-lamda, tiny) ! HSO4I
    molal(9) = psi8 + 2.0*psi6 ! KI
    molal(10) = psi7 ! MGI
    !
    clc = zero
    cnahso4 = zero
    cna2so4 = max(chi4-psi4, zero)
    cnh42s4 = zero
    cnh4hs4 = zero
    ck2so4 = max(chi6-psi6, zero)
    cmgso4 = zero
    ckhso4 = zero
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a4 = xk5*(water/gama(2))**3.0
  funcl7 = molal(5)*molal(2)*molal(2)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCL7 ****************************************
  !
END FUNCTION funcl7
!
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL6
! *** CASE L6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, NA2SO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC, (NH4)2SO4, KHSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl6 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcl1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCL1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  chi6 = ck2so4
  chi7 = cmgso4
  chi8 = ckhso4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = clc
  psi3 = cnahso4
  psi4 = zero
  psi5 = cnh42s4
  psi6 = zero
  psi7 = zero
  psi8 = ckhso4
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi4lo = zero ! Low  limit
  psi4hi = chi4 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  IF (chi4<=tiny) THEN
    y1 = funcl6(zero)
    GO TO 120
  END IF
  !
  x1 = psi4hi
  y1 = funcl6(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH K2SO4 *********
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl6(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH K2SO4
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl6(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCL6') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl6(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCL6') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcl6(x3)
  !
120 RETURN
  !
END SUBROUTINE calcl6
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** FUNCTION FUNCL6
! *** CASE L6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, NA2SO4
!
!     SOLUTION IS SAVED IN COMMON BLOCK /CASE/
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcl6 ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: lamda
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi4 = p4
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = xk5*(water/gama(2))**3.0
    a6 = xk17*(water/gama(17))**3.0
    a9 = xk1*(water)*(gama(8)**2.0)/(gama(7)**3.0)
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    !      PSI6 = 0.5*(SQRT(A6/A4)*(2.D0*PSI4+PSI3)-PSI8)             ! PSI6
    !      PSI6 = MIN (MAX (PSI6, ZERO), CHI6)
    !
    IF (chi6>tiny .AND. water>tiny) THEN
      aa = psi5 + psi4 + psi2 + psi7 + psi8 + lamda
      bb = psi8*(psi5+psi4+psi2+psi7+0.25D0*psi8+lamda)
      cc = 0.25D0*(psi8*psi8*(psi5+psi4+psi2+psi7+lamda)-a6)
      CALL poly3(aa, bb, cc, psi6, islv)
      IF (islv==0) THEN
        psi6 = min(psi6, chi6)
      ELSE
        psi6 = zero
      END IF
    END IF
    !
    psi7 = chi7
    !
    bb = psi7 + psi6 + psi5 + psi4 + psi2 + a9 ! LAMDA
    cc = -a9*(psi8+psi1+psi2+psi3)
    dd = bb*bb - 4.D0*cc
    lamda = 0.5D0*(-bb+sqrt(dd))
    lamda = min(max(lamda,tiny), psi8+psi3+psi2+psi1)
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 + psi7 + lamda ! SO4I
    molal(6) = max(psi2+psi3+psi1+psi8-lamda, tiny) ! HSO4I
    molal(9) = psi8 + 2.0*psi6 ! KI
    molal(10) = psi7 ! MGI
    !
    clc = zero
    cnahso4 = zero
    cna2so4 = max(chi4-psi4, zero)
    cnh42s4 = zero
    cnh4hs4 = zero
    ck2so4 = max(chi6-psi6, zero)
    cmgso4 = zero
    ckhso4 = zero
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a4 = xk5*(water/gama(2))**3.0
  funcl6 = molal(5)*molal(2)*molal(2)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCL6 ****************************************
  !
END FUNCTION funcl6
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL5
! *** CASE L5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, NA2SO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC, (NH4)2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl5 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcl1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCL1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  chi6 = ck2so4
  chi7 = cmgso4
  chi8 = ckhso4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = clc
  psi3 = cnahso4
  psi4 = zero
  psi5 = cnh42s4
  psi6 = zero
  psi7 = zero
  psi8 = zero
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi4lo = zero ! Low  limit
  psi4hi = chi4 ! High limit

  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  IF (chi4<=tiny) THEN
    y1 = funcl5(zero)
    GO TO 120
  END IF
  !
  x1 = psi4hi
  y1 = funcl5(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NA2SO4 *********
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !

  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, psi4lo)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl5(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NA2SO4
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    y3 = funcl5(zero)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCL5') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl5(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCL5') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcl5(x3)
  !
120 RETURN
  !
END SUBROUTINE calcl5
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** FUNCTION FUNCL5
! *** CASE L5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, NA2SO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC, (NH4)2SO4
!
!     SOLUTION IS SAVED IN COMMON BLOCK /CASE/
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcl5 ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bita
  DOUBLE PRECISION :: cama
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: lamda
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi4 = p4
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = xk5*(water/gama(2))**3.0
    a6 = xk17*(water/gama(17))**3.0
    a8 = xk18*(water/gama(18))**2.0
    a9 = xk1*(water)*(gama(8)**2.0)/(gama(7)**3.0)
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    !      PSI6 = 0.5*(SQRT(A6/A4)*(2.D0*PSI4+PSI3)-PSI8)             ! PSI6
    !      PSI6 = MIN (MAX (PSI6, ZERO), CHI6)
    !
    IF (chi6>tiny .AND. water>tiny) THEN
      aa = psi5 + psi4 + psi2 + psi7 + psi8 + lamda
      bb = psi8*(psi5+psi4+psi2+psi7+0.25D0*psi8+lamda)
      cc = 0.25D0*(psi8*psi8*(psi5+psi4+psi2+psi7+lamda)-a6)
      CALL poly3(aa, bb, cc, psi6, islv)
      IF (islv==0) THEN
        psi6 = min(psi6, chi6)
      ELSE
        psi6 = zero
      END IF
    END IF
    !
    psi7 = chi7
    !
    bb = psi7 + psi6 + psi5 + psi4 + psi2 + a9 ! LAMDA
    cc = -a9*(psi8+psi1+psi2+psi3)
    dd = max(bb*bb-4.D0*cc, zero)
    lamda = 0.5D0*(-bb+sqrt(dd))
    lamda = min(max(lamda,tiny), psi8+psi3+psi2+psi1)
    !
    bita = psi3 + psi2 + psi1 + 2.D0*psi6 - lamda
    cama = 2.D0*psi6*(psi3+psi2+psi1-lamda) - a8
    delt = max(bita*bita-4.D0*cama, zero)
    psi8 = 0.5D0*(-bita+sqrt(delt))
    psi8 = min(max(psi8,zero), chi8)
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 + psi7 + lamda ! SO4I
    molal(6) = max(psi2+psi3+psi1+psi8-lamda, tiny) ! HSO4I
    molal(9) = psi8 + 2.0D0*psi6 ! KI
    molal(10) = psi7 ! MGI
    !
    clc = zero
    cnahso4 = zero
    cna2so4 = max(chi4-psi4, zero)
    cnh42s4 = zero
    cnh4hs4 = zero
    ck2so4 = max(chi6-psi6, zero)
    cmgso4 = zero
    ckhso4 = max(chi8-psi8, zero)
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !

    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a4 = xk5*(water/gama(2))**3.0
  funcl5 = molal(5)*molal(2)*molal(2)/a4 - one
  !
  RETURN
  !
  ! *** END OF FUNCTION FUNCL5 ****************************************
  !
END FUNCTION funcl5
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL4
! *** CASE L4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, (NH4)2SO4, NA2SO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl4 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcl1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCL1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  chi6 = ck2so4
  chi7 = cmgso4
  chi8 = ckhso4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = clc
  psi3 = cnahso4
  psi4 = zero
  psi5 = zero
  psi6 = zero
  psi7 = zero
  psi8 = zero
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi4lo = zero ! Low  limit
  psi4hi = chi4 ! High limit
  !
  IF (chi4<=tiny) THEN
    y1 = funcl4(zero)
    GO TO 120
  END IF
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi4hi
  y1 = funcl4(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NA2SO4 *********
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl4(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NA2SO4 **
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl4(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCL4') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl4(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCL4') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcl4(x3)
  !
120 RETURN
  !
END SUBROUTINE calcl4
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** FUNCTION FUNCL4
! *** CASE L4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, (NH4)2SO4, NA2SO4
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4, LC
!
!     SOLUTION IS SAVED IN COMMON BLOCK /CASE/
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcl4 ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bita
  DOUBLE PRECISION :: cama
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: lamda
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi4 = p4
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = xk5*(water/gama(2))**3.0
    a5 = xk7*(water/gama(4))**3.0
    a6 = xk17*(water/gama(17))**3.0
    a8 = xk18*(water/gama(18))**2.0
    a9 = xk1*water/gama(7)*(gama(8)/gama(7))**2.0
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = (psi3+2.D0*psi4-sqrt(a4/a5)*(3.D0*psi2+psi1))/ & ! psi5
         2.D0/sqrt(a4/a5)
    psi5 = max(min(psi5,chi5), zero)
    !
    psi7 = chi7
    !
    bb = psi7 + psi6 + psi5 + psi4 + psi2 + a9 ! LAMDA
    cc = -a9*(psi8+psi1+psi2+psi3)
    dd = max(bb*bb-4.D0*cc, zero)
    lamda = 0.5D0*(-bb+sqrt(dd))
    lamda = min(max(lamda,tiny), psi8+psi3+psi2+psi1)
    !
    !      PSI6 = 0.5*(SQRT(A6/A4)*(2.D0*PSI4+PSI3)-PSI8)             ! PSI6
    !      PSI6 = MIN (MAX (PSI6, ZERO), CHI6)
    !
    IF (chi6>tiny .AND. water>tiny) THEN
      aa = psi5 + psi4 + psi2 + psi7 + psi8 + lamda
      bb = psi8*(psi5+psi4+psi2+psi7+0.25D0*psi8+lamda)
      cc = 0.25D0*(psi8*psi8*(psi5+psi4+psi2+psi7+lamda)-a6)
      CALL poly3(aa, bb, cc, psi6, islv)
      IF (islv==0) THEN
        psi6 = min(psi6, chi6)
      ELSE
        psi6 = zero
      END IF
    END IF
    !
    bita = psi3 + psi2 + psi1 + 2.D0*psi6 - lamda
    cama = 2.D0*psi6*(psi3+psi2+psi1-lamda) - a8
    delt = max(bita*bita-4.D0*cama, zero)
    psi8 = 0.5D0*(-bita+sqrt(delt))
    psi8 = min(max(psi8,zero), chi8)
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 + psi7 + lamda ! SO4I
    molal(6) = max(psi2+psi3+psi1+psi8-lamda, tiny) ! HSO4I
    molal(9) = psi8 + 2.0D0*psi6 ! KI
    molal(10) = psi7 ! MGI
    !
    clc = zero
    cnahso4 = zero
    cna2so4 = max(chi4-psi4, zero)
    cnh42s4 = max(chi5-psi5, zero)
    cnh4hs4 = zero
    ck2so4 = max(chi6-psi6, zero)
    cmgso4 = zero
    ckhso4 = max(chi8-psi8, zero)
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a4 = xk5*(water/gama(2))**3.0
  funcl4 = molal(5)*molal(2)*molal(2)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCL4 ****************************************
  !
END FUNCTION funcl4
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL3
! *** CASE L3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, NH4HSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
!
!     THERE ARE THREE REGIMES IN THIS CASE:
!     1.(NA,NH4)HSO4(s) POSSIBLE. LIQUID & SOLID AEROSOL (SUBROUTINE CALCI3A)
!     2.(NA,NH4)HSO4(s) NOT POSSIBLE, AND RH < MDRH. SOLID AEROSOL ONLY
!     3.(NA,NH4)HSO4(s) NOT POSSIBLE, AND RH >= MDRH. SOLID & LIQUID AEROSOL
!
!     REGIMES 2. AND 3. ARE CONSIDERED TO BE THE SAME AS CASES I1A, I2B
!     RESPECTIVELY
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** FIND DRY COMPOSITION *********************************************
  !
  CALL calcl1a
  !
  ! *** REGIME DEPENDS UPON THE POSSIBLE SOLIDS & RH *********************
  !
  IF (cnh4hs4>tiny .OR. cnahso4>tiny) THEN
    scase = 'L3 ; SUBCASE 1'
    CALL calcl3a ! FULL SOLUTION
    scase = 'L3 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drml3) THEN ! SOLID SOLUTION
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcl1a
      scase = 'L3 ; SUBCASE 2'
      !
    ELSE IF (rh>=drml3) THEN ! MDRH OF L3
      scase = 'L3 ; SUBCASE 3'
      CALL calcmdrh2(rh, drml3, drlc, calcl1a, calcl4)
      scase = 'L3 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcl3
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL3A
! *** CASE L3 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, (NH4)2SO4, NA2SO4, LC
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl3a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi2hi
  DOUBLE PRECISION :: psi2lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcl1a
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCL1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  chi6 = ck2so4
  chi7 = cmgso4
  chi8 = ckhso4
  !
  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = zero
  psi3 = cnahso4
  psi4 = zero
  psi5 = zero
  psi6 = zero
  psi7 = zero
  psi8 = zero
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi2lo = zero ! Low  limit
  psi2hi = chi2 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi2hi
  y1 = funcl3a(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH LC *********
  !
  IF (yhi<eps) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi2hi-psi2lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, psi2lo)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl3a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH LC
  !
  IF (y2 > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl3a (zero)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl3a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCL3A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcl3a(x3)
  !
120 RETURN
  !
END SUBROUTINE calcl3a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCL3A
! *** CASE L3 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, (NH4)2SO4, NA2SO4, LC
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcl3a ( p2 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p2
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  !
  ! *** SETUP PARAMETERS ************************************************
  !

  psi2 = p2 ! Save PSI2 in COMMON BLOCK
  psi4lo = zero ! Low  limit for PSI4
  psi4hi = chi4 ! High limit for PSI4
  !
  ! *** IF NH3 =0, CALL FUNCL3B FOR Y4=0 ********************************
  !
  IF (chi4<=tiny) THEN
    funcl3a = funcl3b(zero)
    GO TO 120
  END IF
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi4hi
  y1 = funcl3b(x1)
  IF (abs(y1)<=eps) GO TO 120
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NA2SO4 *********
  !
  IF (yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, psi4lo)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl3b(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NA2SO4
  !
  IF (y2 > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl3b (psi4lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl3b(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0004, 'FUNCL3A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** INNER LOOP CONVERGED **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcl3b(x3)
  !
  ! *** CALCULATE FUNCTION VALUE FOR INTERNAL LOOP ***************************
  !
120 a2 = xk13*(water/gama(13))**5.0
  funcl3a = molal(5)*molal(6)*molal(3)**3.0/a2 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCL3A *******************************************
  !
END FUNCTION funcl3a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** FUNCTION FUNCL3B
! *** CASE L3 ; SUBCASE 2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SULRAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, (NH4)2SO4, NA2SO4, LC
!     4. COMPLETELY DISSOLVED: NH4HSO4, NAHSO4
!
!     SOLUTION IS SAVED IN COMMON BLOCK /CASE/
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcl3b ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bita
  DOUBLE PRECISION :: cama
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: lamda
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi4 = p4
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = xk5*(water/gama(2))**3.0
    a5 = xk7*(water/gama(4))**3.0
    a6 = xk17*(water/gama(17))**3.0
    a8 = xk18*(water/gama(18))**2.0
    a9 = xk1*(water)*(gama(8)**2.0)/(gama(7)**3.0)
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = (psi3+2.D0*psi4-sqrt(a4/a5)*(3.D0*psi2+psi1))/ & ! psi5
         2.D0/sqrt(a4/a5)
    psi5 = max(min(psi5,chi5), zero)
    !
    psi7 = chi7
    !
    bb = psi7 + psi6 + psi5 + psi4 + psi2 + a9 ! LAMDA
    cc = -a9*(psi8+psi1+psi2+psi3)
    dd = max(bb*bb-4.D0*cc, zero)
    lamda = 0.5D0*(-bb+sqrt(dd))
    lamda = min(max(lamda,tiny), psi8+psi3+psi2+psi1)
    !
    !      PSI6 = 0.5*(SQRT(A6/A4)*(2.D0*PSI4+PSI3)-PSI8)             ! PSI6
    !      PSI6 = MIN (MAX (PSI6, ZERO), CHI6)
    !
    IF (chi6>tiny .AND. water>tiny) THEN
      aa = psi5 + psi4 + psi2 + psi7 + psi8 + lamda
      bb = psi8*(psi5+psi4+psi2+psi7+0.25D0*psi8+lamda)
      cc = 0.25D0*(psi8*psi8*(psi5+psi4+psi2+psi7+lamda)-a6)
      CALL poly3(aa, bb, cc, psi6, islv)
      IF (islv==0) THEN
        psi6 = min(psi6, chi6)
      ELSE
        psi6 = zero
      END IF
    END IF
    !
    bita = psi3 + psi2 + psi1 + 2.D0*psi6 - lamda
    cama = 2.D0*psi6*(psi3+psi2+psi1-lamda) - a8
    delt = max(bita*bita-4.D0*cama, zero)
    psi8 = 0.5D0*(-bita+sqrt(delt))
    psi8 = min(max(psi8,zero), chi8)
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 + psi7 + lamda ! SO4I
    molal(6) = max(psi2+psi3+psi1+psi8-lamda, tiny) ! HSO4I
    molal(9) = psi8 + 2.0D0*psi6 ! KI
    molal(10) = psi7 ! MGI
    !
    clc = max(chi2-psi2, zero)
    cnahso4 = zero
    cna2so4 = max(chi4-psi4, zero)
    cnh42s4 = max(chi5-psi5, zero)
    cnh4hs4 = zero
    ck2so4 = max(chi6-psi6, zero)
    cmgso4 = max(chi7-psi7, zero)
    ckhso4 = max(chi8-psi8, zero)
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a4 = xk5*(water/gama(2))**3.0
  funcl3b = molal(5)*molal(2)*molal(2)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCL3B ****************************************
  !
END FUNCTION funcl3b

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL2
! *** CASE L2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, NH4HSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
!
!     THERE ARE THREE REGIMES IN THIS CASE:
!     1. NH4HSO4(s) POSSIBLE. LIQUID & SOLID AEROSOL (SUBROUTINE CALCL2A)
!     2. NH4HSO4(s) NOT POSSIBLE, AND RH < MDRH. SOLID AEROSOL ONLY
!     3. NH4HSO4(s) NOT POSSIBLE, AND RH >= MDRH. SOLID & LIQUID AEROSOL
!
!     REGIMES 2. AND 3. ARE CONSIDERED TO BE THE SAME AS CASES L1A, L2B
!     RESPECTIVELY
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcl1a
  !
  ! *** REGIME DEPENDS UPON THE POSSIBLE SOLIDS & RH **********************
  !
  IF (cnh4hs4>tiny) THEN
    scase = 'L2 ; SUBCASE 1'
    CALL calcl2a
    scase = 'L2 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drml2) THEN ! SOLID SOLUTION ONLY
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcl1a
      scase = 'L2 ; SUBCASE 2'
      !
    ELSE IF (rh>=drml2) THEN ! MDRH OF L2
      scase = 'L2 ; SUBCASE 3'
      CALL calcmdrh2(rh, drml2, drnahso4, calcl1a, calcl3a)
      scase = 'L2 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcl2
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL2A
! *** CASE L2 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
!     4. COMPLETELY DISSOLVED: NH4HSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi2hi
  DOUBLE PRECISION :: psi2lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = cnh4hs4 ! Save from CALCL1 run
  chi2 = clc
  chi3 = cnahso4
  chi4 = cna2so4
  chi5 = cnh42s4
  chi6 = ck2so4
  chi7 = cmgso4
  chi8 = ckhso4
  !

  psi1 = cnh4hs4 ! ASSIGN INITIAL PSI's
  psi2 = zero
  psi3 = zero
  psi4 = zero
  psi5 = zero
  psi6 = zero
  psi7 = zero
  psi8 = zero
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi2lo = zero ! Low  limit
  psi2hi = chi2 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi2hi
  y1 = funcl2a(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NA2SO4 *********
  !
  IF (yhi<eps) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi2hi-psi2lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, psi2lo)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl2a(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NA2SO4
  !
  IF (y2 > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl2a (zero)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funcl2a(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCL2A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funcl2a(x3)
  !
120 RETURN
  !
END SUBROUTINE calcl2a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCL2A
! *** CASE L2 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
!     4. COMPLETELY DISSOLVED: NH4HSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcl2a ( p2 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p2
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi4hi
  DOUBLE PRECISION :: psi4lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  !
  ! *** SETUP PARAMETERS ************************************************
  !

  psi2 = p2 ! Save PSI3 in COMMON BLOCK
  psi4lo = zero ! Low  limit for PSI4
  psi4hi = chi4 ! High limit for PSI4
  !
  ! *** IF NH3 =0, CALL FUNCL3B FOR Y4=0 ********************************
  !

  IF (chi4<=tiny) THEN
    funcl2a = funcl2b(zero)
    GO TO 120
  END IF
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !

  x1 = psi4hi
  y1 = funcl2b(x1)

  IF (abs(y1)<=eps) GO TO 120
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH LC *********
  !
  IF (yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi4hi-psi4lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, psi4lo)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funcl2b(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH LC
  !
  IF (y2 > eps) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)      
    y2 = funcl2b (psi4lo)
  END IF
  GO TO 120
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)      
    y3 = funcl2b(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0004, 'FUNCL2A') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** INNER LOOP CONVERGED **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)      
  y3 = funcl2b(x3)
  !
  ! *** CALCULATE FUNCTION VALUE FOR OUTER LOOP ***************************
  !
120 a2 = xk13*(water/gama(13))**5.0
  funcl2a = molal(5)*molal(6)*molal(3)**3.0/a2 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCL2A *******************************************
  !
END FUNCTION funcl2a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCL2B
! *** CASE L2 ; SUBCASE 2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
!     4. COMPLETELY DISSOLVED: NH4HSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcl2b ( p4 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p4
  DOUBLE PRECISION :: aa
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: bita
  DOUBLE PRECISION :: cama
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: delt
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: lamda
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  psi4 = p4 ! Save PSI4 in COMMON BLOCK
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  psi3 = chi3
  psi5 = chi5
  lamda = zero
  psi6 = chi6
  psi8 = chi8
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a3 = xk11*(water/gama(12))**2.0
    a4 = xk5*(water/gama(2))**3.0
    a5 = xk7*(water/gama(4))**3.0
    a6 = xk17*(water/gama(17))**3.0
    a8 = xk18*(water/gama(18))**2.0
    a9 = xk1*(water)*(gama(8)**2.0)/(gama(7)**3.0)
    !
    !  CALCULATE DISSOCIATION QUANTITIES
    !
    psi5 = (psi3+2.D0*psi4-sqrt(a4/a5)*(3.D0*psi2+psi1))/ & ! psi5
         2.D0/sqrt(a4/a5)
    psi5 = max(min(psi5,chi5), zero)
    !
    IF (chi3>tiny .AND. water>tiny) THEN
      aa = 2.D0*psi4 + psi2 + psi1 + psi8 - lamda
      bb = 2.D0*psi4*(psi2+psi1+psi8-lamda) - a3
      cc = zero
      CALL poly3(aa, bb, cc, psi3, islv)
      IF (islv==0) THEN
        psi3 = min(psi3, chi3)
      ELSE
        psi3 = zero
      END IF
    END IF
    !
    psi7 = chi7
    !
    bb = psi7 + psi6 + psi5 + psi4 + psi2 + a9 ! LAMDA
    cc = -a9*(psi8+psi1+psi2+psi3)
    dd = max(bb*bb-4.D0*cc, zero)
    lamda = 0.5D0*(-bb+sqrt(dd))
    lamda = min(max(lamda,tiny), psi8+psi3+psi2+psi1)
    !
    !      PSI6 = 0.5*(SQRT(A6/A4)*(2.D0*PSI4+PSI3)-PSI8)             ! PSI6
    !      PSI6 = MIN (MAX (PSI6, ZERO), CHI6)
    !
    IF (chi6>tiny .AND. water>tiny) THEN
      aa = psi5 + psi4 + psi2 + psi7 + psi8 + lamda
      bb = psi8*(psi5+psi4+psi2+psi7+0.25D0*psi8+lamda)
      cc = 0.25D0*(psi8*psi8*(psi5+psi4+psi2+psi7+lamda)-a6)
      CALL poly3(aa, bb, cc, psi6, islv)
      IF (islv==0) THEN
        psi6 = min(psi6, chi6)
      ELSE
        psi6 = zero
      END IF
    END IF
    !
    bita = psi3 + psi2 + psi1 + 2.D0*psi6 - lamda ! PSI8
    cama = 2.D0*psi6*(psi3+psi2+psi1-lamda) - a8
    delt = max(bita*bita-4.D0*cama, zero)
    psi8 = 0.5D0*(-bita+sqrt(delt))
    psi8 = min(max(psi8,zero), chi8)
    !
    ! *** CALCULATE SPECIATION ********************************************
    !
    molal(1) = lamda ! HI
    molal(2) = 2.D0*psi4 + psi3 ! NAI
    molal(3) = 3.D0*psi2 + 2.D0*psi5 + psi1 ! NH4I
    molal(5) = psi2 + psi4 + psi5 + psi6 + psi7 + lamda ! SO4I
    molal(6) = max(psi2+psi3+psi1+psi8-lamda, tiny) ! HSO4I
    molal(9) = psi8 + 2.0D0*psi6 ! KI
    molal(10) = psi7 ! MGI
    !
    clc = max(chi2-psi2, zero)
    cnahso4 = max(chi3-psi3, zero)
    cna2so4 = max(chi4-psi4, zero)
    cnh42s4 = max(chi5-psi5, zero)
    cnh4hs4 = zero
    ck2so4 = max(chi6-psi6, zero)
    cmgso4 = max(chi7-psi7, zero)
    ckhso4 = max(chi8-psi8, zero)
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 a4 = xk5*(water/gama(2))**3.0
  funcl2b = molal(5)*molal(2)*molal(2)/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCL2B ****************************************
  !
END FUNCTION funcl2b

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL1
! *** CASE L1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, NH4HSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCI1A)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drml1) THEN
    scase = 'L1 ; SUBCASE 1'
    CALL calcl1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'L1 ; SUBCASE 1'
  ELSE
    scase = 'L1 ; SUBCASE 2' ! LIQUID & SOLID PHASE POSSIBLE
    CALL calcmdrh2(rh, drml1, drnh4hs4, calcl1a, calcl2a)
    scase = 'L1 ; SUBCASE 2'
  END IF
  !
  ! *** AMMONIA IN GAS PHASE **********************************************
  !
  !      CALL CALCNH3
  !
  RETURN
  !
END SUBROUTINE calcl1
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCL1A
! *** CASE L1A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE RICH, NO FREE ACID (1.0 <= SO4RAT < 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, MGSO4, KHSO4, NH4HSO4, NAHSO4, (NH4)2SO4, NA2SO4, LC
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcl1a (  )
  USE mo_isrpia  
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frnh4
  DOUBLE PRECISION :: frso4
  !
  ! *** CALCULATE NON VOLATILE SOLIDS ***********************************
  !
  ccaso4 = min(w(6), w(2)) ! CCASO4
  frso4 = max(w(2)-ccaso4, zero)
  cafr = max(w(6)-ccaso4, zero)
  ck2so4 = min(0.5D0*w(7), frso4) ! CK2SO4
  frk = max(w(7)-2.D0*ck2so4, zero)
  frso4 = max(frso4-ck2so4, zero)
  cna2so4 = min(0.5D0*w(1), frso4) ! CNA2SO4
  frna = max(w(1)-2.D0*cna2so4, zero)
  frso4 = max(frso4-cna2so4, zero)
  cmgso4 = min(w(8), frso4) ! CMGSO4
  frmg = max(w(8)-cmgso4, zero)
  frso4 = max(frso4-cmgso4, zero)
  !
  cnh4hs4 = zero
  cnahso4 = zero
  cnh42s4 = zero
  ckhso4 = zero
  !
  clc = min(w(3)/3.D0, frso4/2.D0)
  frso4 = max(frso4-2.D0*clc, zero)
  frnh4 = max(w(3)-3.D0*clc, zero)
  !
  IF (frso4<=tiny) THEN
    clc = max(clc-frnh4, zero)
    cnh42s4 = 2.D0*frnh4

  ELSE IF (frnh4<=tiny) THEN
    cnh4hs4 = 3.D0*min(frso4, clc)
    clc = max(clc-frso4, zero)
    !         IF (CK2SO4.GT.TINY) THEN
    !            FRSO4  = MAX(FRSO4-CNH4HS4/3.D0, ZERO)
    !           CKHSO4 = 2.D0*FRSO4
    !            CK2SO4 = MAX(CK2SO4-FRSO4, ZERO)
    !         ENDIF
    !         IF (CNA2SO4.GT.TINY) THEN
    !            FRSO4   = MAX(FRSO4-CKHSO4/2.D0, ZERO)
    !            CNAHSO4 = 2.D0*FRSO4
    !            CNA2SO4 = MAX(CNA2SO4-FRSO4, ZERO)
    !         ENDIF
    !
    IF (cna2so4>tiny) THEN
      frso4 = max(frso4-cnh4hs4/3.D0, zero)
      cnahso4 = 2.D0*frso4
      cna2so4 = max(cna2so4-frso4, zero)
    END IF
    IF (ck2so4>tiny) THEN
      frso4 = max(frso4-cnh4hs4/3.D0, zero)
      ckhso4 = 2.D0*frso4
      ck2so4 = max(ck2so4-frso4, zero)
    END IF
  END IF
  !
  ! *** CALCULATE GAS SPECIES ********************************************
  !
  ghno3 = w(4)
  ghcl = w(5)
  gnh3 = zero
  !
  RETURN
  !
END SUBROUTINE calcl1a
!
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCK4
! *** CASE K4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE SUPER RICH, FREE ACID (SO4RAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CASO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calck4 (  )
  USE mo_isrpia
  USE mo_casek
  IMPLICIT NONE
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  frst = .TRUE.
  calain = .TRUE.
  !
  chi1 = w(3) !  Total NH4 initially as NH4HSO4
  chi2 = w(1) !  Total NA initially as NaHSO4
  chi3 = w(7) !  Total K initially as KHSO4
  !
  lamda = max(w(2)-w(3)-w(1)-w(6)-w(7)-w(8), tiny) ! FREE H2SO4
  psi1 = chi1 ! ALL NH4HSO4 DELIQUESCED
  psi2 = chi2 ! ALL NaHSO4 DELIQUESCED
  psi3 = chi3 ! ALL KHSO4 DELIQUESCED
  psi4 = chi4 ! ALL MgSO4 DELIQUESCED
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a4 = xk1*water/gama(7)*(gama(8)/gama(7))**2.0
    !
    bb   = a4+lamda+psi4                               ! KAPA
    cc   =-a4*(lamda + psi3 + psi2 + psi1) + lamda*psi4
    dd   = max(bb*bb-4.d0*cc, zero)
    kapa = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal (1) = max(lamda + kapa, tiny)                         ! HI
    molal (2) = psi2                                            ! NAI
    molal (3) = psi1                                            ! NH4I
    molal (5) = max(kapa + psi4, zero)                          ! SO4I
    molal (6) = max(lamda + psi1 + psi2 + psi3 - kapa, zero)    ! HSO4I
    molal (9) = psi3                                            ! KI
    molal (10)= psi4
    !
    cnh4hs4 = zero
    cnahso4 = zero
    ckhso4 = zero
    ccaso4 = w(6)
    cmgso4 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
100 RETURN
  !
END SUBROUTINE calck4
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCK3
! *** CASE K3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE SUPER RICH, FREE ACID (SO4RAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : KHSO4, CASO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calck3 (  )
  USE mo_isrpia
  USE mo_casek
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi3hi
  DOUBLE PRECISION :: psi3lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  chi1 = w(3) !  Total NH4 initially as NH4HSO4
  chi2 = w(1) !  Total NA initially as NaHSO4
  chi3 = w(7) !  Total K initially as KHSO4
  chi4 = w(8) !  Total Mg initially as MgSO4
  !
  psi3lo = tiny ! Low  limit
  psi3hi = chi3 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi3hi
  y1 = funck3(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH KHSO4 ****
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi3hi-psi3lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funck3(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH KHSO4
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funck3(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCK3') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funck3(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCK3') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funck3(x3)
  !
120 RETURN
  !
END SUBROUTINE calck3
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE FUNCK3
! *** CASE K3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE SUPER RICH, FREE ACID (SO4RAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : KHSO4, CaSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funck3 ( p1 )
  USE mo_isrpia
  USE mo_casek
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p1
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  lamda = max(w(2)-w(3)-w(1)-w(6)-w(7)-w(8), tiny) ! FREE H2SO4
  psi3 = p1
  psi1 = chi1 ! ALL NH4HSO4 DELIQUESCED
  psi2 = chi2 ! ALL NaHSO4 DELIQUESCED

  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a3 = xk18*(water/gama(18))**2.0
    a4 = xk1*water/gama(7)*(gama(8)/gama(7))**2.0
    !
    !
    bb = a4 + lamda + psi4 ! KAPA
    cc = -a4*(lamda+psi3+psi2+psi1) + lamda*psi4
    dd = max(bb*bb - 4.D0*cc, zero)
    kapa = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1)  = max(lamda + kapa, zero) ! HI
    molal(2)  = psi2 ! NAI
    molal(3)  = psi1 ! NH4I
    molal(4)  = zero
    molal(5)  = max(kapa + psi4, zero) ! SO4I
    molal(6)  = max(lamda + psi1 + psi2 + psi3 - kapa, zero) ! HSO4I
    molal(7)  = zero
    molal(8)  = zero
    molal(9)  = psi3 ! KI
    molal(10) = psi4 ! KI
    !
    cnh4hs4 = zero
    cnahso4 = zero
    ckhso4 = chi3 - psi3
    ccaso4 = w(6)
    cmgso4 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 funck3 = molal(9)*molal(6)/a3 - one
  !
  ! *** END OF FUNCTION FUNCK3 *******************************************
  !
END FUNCTION funck3
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCK2
! *** CASE K2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE SUPER RICH, FREE ACID (SO4RAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NAHSO4, KHSO4, CaSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calck2 (  )
  USE mo_isrpia
  USE mo_casek
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi3hi
  DOUBLE PRECISION :: psi3lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  chi1 = w(3) !  Total NH4 initially as NH4HSO4
  chi2 = w(1) !  Total NA initially as NaHSO4
  chi3 = w(7) !  Total K initially as KHSO4
  chi4 = w(8) !  Total Mg initially as MgSO4
  !
  psi3lo = tiny ! Low  limit
  psi3hi = chi3 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi3hi
  y1 = funck2(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH KHSO4 ****
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi3hi-psi3lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funck2(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH KHSO4
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funck2(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCK2') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funck2(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCK2') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  y3 = funck2(x3)
  !
120 RETURN
  !
END SUBROUTINE calck2
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCK2
! *** CASE K2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE SUPER RICH, FREE ACID (SO4RAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NAHSO4, KHSO4, CaSO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funck2 ( p1 )
  USE mo_isrpia
  USE mo_casek
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p1
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  lamda = max(w(2)-w(3)-w(1)-w(6)-w(7)-w(8), tiny) ! FREE H2SO4
  psi3 = p1
  psi1 = chi1 ! ALL NH4HSO4 DELIQUESCED
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a2 = xk11*(water/gama(12))**2.0
    a3 = xk18*(water/gama(18))**2.0
    a4 = xk1*water/gama(7)*(gama(8)/gama(7))**2.0
    !
    psi2 = a2/a3*psi3 ! PSI2
    psi2 = min(max(psi2,zero), chi2)
    !
    bb = a4 + lamda + psi4 ! KAPA
    cc = -a4*(lamda+psi3+psi2+psi1) + lamda*psi4
    dd = max(bb*bb - 4.D0*cc, zero)
    kapa = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = max(lamda + kapa, zero) ! HI
    molal(2) = psi2 ! NAI
    molal(3) = psi1 ! NH4I
    molal(4) = zero 
    molal(5) = max(kapa + psi4, zero) ! SO4I
    molal(6) = max(lamda + psi1 + psi2 + psi3 - kapa, zero) ! HSO4I
    molal(7) = zero
    molal(8) = zero
    molal(9) = psi3 ! KI
    molal(10)= psi4 
    !
    cnh4hs4 = zero
    cnahso4 = chi2 - psi2
    ckhso4 = chi3 - psi3
    ccaso4 = w(6)
    cmgso4 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 funck2 = molal(9)*molal(6)/a3 - one
  !
  ! *** END OF FUNCTION FUNCK2 *******************************************
  !
END FUNCTION funck2
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCK1
! *** CASE K1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE SUPER RICH, FREE ACID (SO4RAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NH4HSO4, NAHSO4, KHSO4, CASO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calck1 (  )
  USE mo_isrpia
  USE mo_casek
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: psi3hi
  DOUBLE PRECISION :: psi3lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  !
  ! *** SETUP PARAMETERS ************************************************
  !

  calaou = .TRUE. ! Outer loop activity calculation flag
  chi1 = w(3) !  Total NH4 initially as NH4HSO4
  chi2 = w(1) !  Total NA initially as NaHSO4
  chi3 = w(7) !  Total K initially as KHSO4
  chi4 = w(8) !  Total Mg initially as MgSO4
  !
  psi3lo = tiny ! Low  limit
  psi3hi = chi3 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi3hi
  y1 = funck1(x1)
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** YHI < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH KHSO4 ****
  !
  IF (abs(y1)<=eps .OR. yhi<zero) GO TO 120
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi3hi-psi3lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = x1 - dx
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y2 = funck1(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH KHSO4
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (ylo>zero .AND. yhi>zero) THEN
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funck1(zero)
    GO TO 120
  ELSE IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCK1') ! WARNING ERROR: NO SOLUTION
    GO TO 120
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
    y3 = funck1(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCK1') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  CALL rstgamp            ! reinitialize activity coefficients (slc.1.2012)
  y3 = funck1(x3)
  !
120 RETURN
  !
END SUBROUTINE calck1
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE FUNCK1
! *** CASE K1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE super RICH, FREE ACID (SO4RAT < 1.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : NH4HSO4, NAHSO4, KHSO4, CASO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funck1 ( p1 )
  USE mo_isrpia
  USE mo_casek
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p1
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: dd
  INTEGER :: i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  !
  lamda = max(w(2)-w(3)-w(1)-w(6)-w(7)-w(8), tiny) ! FREE H2SO4
  psi3 = p1
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a1 = xk12*(water/gama(09))**2.0
    a2 = xk11*(water/gama(12))**2.0
    a3 = xk18*(water/gama(18))**2.0
    a4 = xk1*water/gama(7)*(gama(8)/gama(7))**2.0
    !
    psi1 = a1/a3*psi3 ! PSI1
    psi1 = min(max(psi1,zero), chi1)
    !
    psi2 = a2/a3*psi3 ! PSI2
    psi2 = min(max(psi2,zero), chi2)
    !
    bb = a4 + lamda + psi4! KAPA
    cc = -a4*(lamda+psi3+psi2+psi1) + lamda*psi4
    dd = max(bb*bb - 4.D0*cc, zero)
    kapa = 0.5D0*(-bb+sqrt(dd))
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = max(lamda + kapa, zero) ! HI
    molal(2) = psi2 ! NAI
    molal(3) = psi1 ! NH4I
    molal(4) = zero ! CLI
    molal(5) = max(kapa + psi4, zero) ! SO4I
    molal(6) = max(lamda + psi1 + psi2 + psi3 - kapa, zero) ! HSO4I
    molal(7) = zero ! NO3I
    molal(8) = zero ! CAI
    molal(9) = psi3 ! KI
    molal(10) = psi4 ! MGI
    !
    cnh4hs4 = chi1 - psi1
    cnahso4 = chi2 - psi2
    ckhso4 = chi3 - psi3
    ccaso4 = w(6)
    cmgso4 = zero
    !
    CALL calcmr ! Water content
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact

    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 funck1 = molal(9)*molal(6)/a3 - one
  !
  ! *** END OF FUNCTION FUNCK1 ****************************************
  !
END FUNCTION funck1

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ISRP1R
! *** THIS SUBROUTINE IS THE DRIVER ROUTINE FOR THE REVERSE PROBLEM OF
!     AN AMMONIUM-SULFATE AEROSOL SYSTEM.
!     THE COMPOSITION REGIME IS DETERMINED BY THE SULFATE RATIO AND BY
!     THE AMBIENT RELATIVE HUMIDITY.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE isrp1r ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  !DOUBLE PRECISION :: getasr
  !
  ! *** INITIALIZE COMMON BLOCK VARIABLES *********************************
  !
  CALL init1(wi, rhi, tempi)
  !
  ! *** CALCULATE SULFATE RATIO *******************************************
  !
  IF (rh>=drnh42s4) THEN ! WET AEROSOL, NEED NH4 AT SRATIO=2.0
    sulratw = getasr(waer(2), rhi) ! AEROSOL SULFATE RATIO
  ELSE
    sulratw = 2.0D0 ! DRY AEROSOL SULFATE RATIO
  END IF
  sulrat = waer(3)/waer(2) ! SULFATE RATIO
  !
  ! *** FIND CALCULATION REGIME FROM (SULRAT,RH) **************************
  !
  ! *** SULFATE POOR
  !
  IF (sulratw<=sulrat) THEN
    !
    IF (metstbl==1) THEN
      scase = 'S2'
      CALL calcs2 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh42s4) THEN
        scase = 'S1'
        CALL calcs1 ! NH42SO4              ; case K1
        !
      ELSE IF (drnh42s4<=rh) THEN
        scase = 'S2'
        CALL calcs2 ! Only liquid          ; case K2
      END IF
    END IF
    !
    ! *** SULFATE RICH (NO ACID)
    !
  ELSE IF (1.0<=sulrat .AND. sulrat<sulratw) THEN
    w(2) = waer(2)
    w(3) = waer(3)
    !
    IF (metstbl==1) THEN
      scase = 'B4'
      CALL calcb4 ! Only liquid (metastable)
      scase = 'B4'
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'B1'
        CALL calcb1 ! NH4HSO4,LC,NH42SO4   ; case B1
        scase = 'B1'
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drlc) THEN
        scase = 'B2'
        CALL calcb2 ! LC,NH42S4            ; case B2
        scase = 'B2'
        !
      ELSE IF (drlc<=rh .AND. rh<drnh42s4) THEN
        scase = 'B3'
        CALL calcb3 ! NH42S4               ; case B3
        scase = 'B3'
        !
      ELSE IF (drnh42s4<=rh) THEN
        scase = 'B4'
        CALL calcb4 ! Only liquid          ; case B4
        scase = 'B4'
      END IF
    END IF
    !
    CALL calcnh3p ! Compute NH3(g)
    !
    ! *** SULFATE RICH (FREE ACID)
    !
  ELSE IF (sulrat<1.0) THEN
    w(2) = waer(2)
    w(3) = waer(3)
    !
    IF (metstbl==1) THEN
      scase = 'C2'
      CALL calcc2 ! Only liquid (metastable)
      scase = 'C2'
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'C1'
        CALL calcc1 ! NH4HSO4              ; case C1
        scase = 'C1'
        !
      ELSE IF (drnh4hs4<=rh) THEN
        scase = 'C2'
        CALL calcc2 ! Only liquid          ; case C2
        scase = 'C2'
      END IF
    END IF
    !
    CALL calcnh3p
    !
  END IF
  RETURN
  !
END SUBROUTINE isrp1r

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ISRP2R
! *** THIS SUBROUTINE IS THE DRIVER ROUTINE FOR THE REVERSE PROBLEM OF
!     AN AMMONIUM-SULFATE-NITRATE AEROSOL SYSTEM.
!     THE COMPOSITION REGIME IS DETERMINED BY THE SULFATE RATIO AND BY
!     THE AMBIENT RELATIVE HUMIDITY.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE isrp2r ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  !DOUBLE PRECISION :: getasr
  LOGICAL :: tryliq
  !
  ! *** INITIALIZE ALL VARIABLES IN COMMON BLOCK **************************
  !
  tryliq = .TRUE. ! Assume liquid phase, sulfate poor limit
  !
100 CALL init2(wi, rhi, tempi)
  !
  ! *** CALCULATE SULFATE RATIO *******************************************
  !
  IF (tryliq .AND. rh>=drnh4no3) THEN ! *** WET AEROSOL
    sulratw = getasr(waer(2), rhi) ! LIMITING SULFATE RATIO
  ELSE
    sulratw = 2.0D0 ! *** DRY AEROSOL
  END IF
  sulrat = waer(3)/waer(2)
  !
  ! *** FIND CALCULATION REGIME FROM (SULRAT,RH) **************************
  !
  ! *** SULFATE POOR
  !
  IF (sulratw<=sulrat) THEN
    !
    IF (metstbl==1) THEN
      scase = 'N3'
      CALL calcn3 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4no3) THEN
        scase = 'N1'
        CALL calcn1 ! NH42SO4,NH4NO3       ; case N1
        !
      ELSE IF (drnh4no3<=rh .AND. rh<drnh42s4) THEN
        scase = 'N2'
        CALL calcn2 ! NH42S4               ; case N2
        !
      ELSE IF (drnh42s4<=rh) THEN
        scase = 'N3'
        CALL calcn3 ! Only liquid          ; case N3
      END IF
    END IF
    !
    ! *** SULFATE RICH (NO ACID)
    !
    !     FOR SOLVING THIS CASE, NITRIC ACID AND AMMONIA IN THE GAS PHASE ARE
    !     ASSUMED A MINOR SPECIES, THAT DO NOT SIGNIFICANTLY AFFECT THE
    !     AEROSOL EQUILIBRIUM.
    !
  ELSE IF (1.0<=sulrat .AND. sulrat<sulratw) THEN
    w(2) = waer(2)
    w(3) = waer(3)
    w(4) = waer(4)
    !
    IF (metstbl==1) THEN
      scase = 'B4'
      CALL calcb4 ! Only liquid (metastable)
      scase = 'B4'
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'B1'
        CALL calcb1 ! NH4HSO4,LC,NH42SO4   ; case O1
        scase = 'B1'
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drlc) THEN
        scase = 'B2'
        CALL calcb2 ! LC,NH42S4            ; case O2
        scase = 'B2'
        !
      ELSE IF (drlc<=rh .AND. rh<drnh42s4) THEN
        scase = 'B3'
        CALL calcb3 ! NH42S4               ; case O3
        scase = 'B3'
        !
      ELSE IF (drnh42s4<=rh) THEN
        scase = 'B4'
        CALL calcb4 ! Only liquid          ; case O4
        scase = 'B4'
      END IF
    END IF
    !
    ! *** Add the NO3 to the solution now and calculate partitioning.
    !
    molal(7) = waer(4) ! There is always water, so NO3(aer) is NO3-
    molal(1) = molal(1) + waer(4) ! Add H+ to balance out
    CALL calcnap ! HNO3, NH3 dissolved
    CALL calcnh3p
    !
    ! *** SULFATE RICH (FREE ACID)
    !
    !     FOR SOLVING THIS CASE, NITRIC ACID AND AMMONIA IN THE GAS PHASE ARE
    !     ASSUMED A MINOR SPECIES, THAT DO NOT SIGNIFICANTLY AFFECT THE
    !     AEROSOL EQUILIBRIUM.
    !
  ELSE IF (sulrat<1.0) THEN
    w(2) = waer(2)
    w(3) = waer(3)
    w(4) = waer(4)
    !
    IF (metstbl==1) THEN
      scase = 'C2'
      CALL calcc2 ! Only liquid (metastable)
      scase = 'C2'
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'C1'
        CALL calcc1 ! NH4HSO4              ; case P1
        scase = 'C1'
        !
      ELSE IF (drnh4hs4<=rh) THEN
        scase = 'C2'
        CALL calcc2 ! Only liquid          ; case P2
        scase = 'C2'
      END IF
    END IF
    !
    ! *** Add the NO3 to the solution now and calculate partitioning.
    !
    molal(7) = waer(4) ! There is always water, so NO3(aer) is NO3-
    molal(1) = molal(1) + waer(4) ! Add H+ to balance out
    !
    CALL calcnap ! HNO3, NH3 dissolved
    CALL calcnh3p
  END IF
  !
  ! *** IF SULRATW < SULRAT < 2.0 and WATER = 0 => SULFATE RICH CASE.
  !
  IF (sulratw<=sulrat .AND. sulrat<2.0 .AND. water<=tiny) THEN
    tryliq = .FALSE.
    GO TO 100
  END IF
  !
  RETURN
  !
END SUBROUTINE isrp2r
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE ISRP3R
! *** THIS SUBROUTINE IS THE DRIVER ROUTINE FOR THE REVERSE PROBLEM OF
!     AN AMMONIUM-SULFATE-NITRATE-CHLORIDE-SODIUM AEROSOL SYSTEM.
!     THE COMPOSITION REGIME IS DETERMINED BY THE SULFATE & SODIUM
!     RATIOS AND BY THE AMBIENT RELATIVE HUMIDITY.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE isrp3r ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: sri
  LOGICAL :: tryliq
  !DOUBLE PRECISION :: getasr
  !cC
  !cC *** ADJUST FOR TOO LITTLE AMMONIUM AND CHLORIDE ***********************
  !cC
  !c      WI(3) = MAX (WI(3), 1.D-10)  ! NH4+ : 1e-4 umoles/m3
  !c      WI(5) = MAX (WI(5), 1.D-10)  ! Cl-  : 1e-4 umoles/m3
  !
  ! *** INITIALIZE ALL VARIABLES ******************************************
  !
  tryliq = .TRUE. ! Use liquid phase sulfate poor limit
  !
100 CALL isoinit3(wi, rhi, tempi) ! COMMON block variables
  !cC
  !cC *** CHECK IF TOO MUCH SODIUM ; ADJUST AND ISSUE ERROR MESSAGE *********
  !cC
  !c      REST = 2.D0*WAER(2) + WAER(4) + WAER(5)
  !c      IF (WAER(1).GT.REST) THEN            ! NA > 2*SO4+CL+NO3 ?
  !c         WAER(1) = (ONE-1D-6)*REST         ! Adjust Na amount
  !c         CALL PUSHERR (0050, 'ISRP3R')     ! Warning error: Na adjusted
  !c      ENDIF
  !
  ! *** CALCULATE SULFATE & SODIUM RATIOS *********************************
  !
  IF (tryliq .AND. rh>=drnh4no3) THEN ! ** WET AEROSOL
    frso4 = waer(2) - waer(1)/2.0D0 ! SULFATE UNBOUND BY SODIUM
    frso4 = max(frso4, tiny)
    sri = getasr(frso4, rhi) ! SULFATE RATIO FOR NH4+
    sulratw = (waer(1)+frso4*sri)/waer(2) ! LIMITING SULFATE RATIO
    sulratw = min(sulratw, 2.0D0)
  ELSE
    sulratw = 2.0D0 ! ** DRY AEROSOL
  END IF
  sulrat = (waer(1)+waer(3))/waer(2)
  sodrat = waer(1)/waer(2)
  !
  ! *** FIND CALCULATION REGIME FROM (SULRAT,RH) **************************
  !
  ! *** SULFATE POOR ; SODIUM POOR
  !
  IF (sulratw<=sulrat .AND. sodrat<2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'Q5'
      CALL calcq5 ! Only liquid (metastable)
      scase = 'Q5'
    ELSE
      !
      IF (rh<drnh4no3) THEN
        scase = 'Q1'
        CALL calcq1 ! NH42SO4,NH4NO3,NH4CL,NA2SO4
        !
      ELSE IF (drnh4no3<=rh .AND. rh<drnh4cl) THEN
        scase = 'Q2'
        CALL calcq2 ! NH42SO4,NH4CL,NA2SO4
        !
      ELSE IF (drnh4cl<=rh .AND. rh<drnh42s4) THEN
        scase = 'Q3'
        CALL calcq3 ! NH42SO4,NA2SO4
        !
      ELSE IF (drnh42s4<=rh .AND. rh<drna2so4) THEN
        scase = 'Q4'
        CALL calcq4 ! NA2SO4
        scase = 'Q4'
        !
      ELSE IF (drna2so4<=rh) THEN
        scase = 'Q5'
        CALL calcq5 ! Only liquid
        scase = 'Q5'
      END IF
    END IF
    !
    ! *** SULFATE POOR ; SODIUM RICH
    !
  ELSE IF (sulrat>=sulratw .AND. sodrat>=2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'R6'
      CALL calcr6 ! Only liquid (metastable)
      scase = 'R6'
    ELSE
      !
      IF (rh<drnh4no3) THEN
        scase = 'R1'
        CALL calcr1 ! NH4NO3,NH4CL,NA2SO4,NACL,NANO3
        !
      ELSE IF (drnh4no3<=rh .AND. rh<drnano3) THEN
        scase = 'R2'
        CALL calcr2 ! NH4CL,NA2SO4,NACL,NANO3
        !
      ELSE IF (drnano3<=rh .AND. rh<drnacl) THEN
        scase = 'R3'
        CALL calcr3 ! NH4CL,NA2SO4,NACL
        !
      ELSE IF (drnacl<=rh .AND. rh<drnh4cl) THEN
        scase = 'R4'
        CALL calcr4 ! NH4CL,NA2SO4
        !
      ELSE IF (drnh4cl<=rh .AND. rh<drna2so4) THEN
        scase = 'R5'
        CALL calcr5 ! NA2SO4
        scase = 'R5'
        !
      ELSE IF (drna2so4<=rh) THEN
        scase = 'R6'
        CALL calcr6 ! NO SOLID
        scase = 'R6'
      END IF
    END IF
    !
    ! *** SULFATE RICH (NO ACID)
    !
  ELSE IF (1.0<=sulrat .AND. sulrat<sulratw) THEN
    DO i = 1, ncomp
      w(i) = waer(i)
    END DO
    !
    IF (metstbl==1) THEN
      scase = 'I6'
      CALL calci6 ! Only liquid (metastable)
      scase = 'I6'
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'I1'
        CALL calci1 ! NA2SO4,(NH4)2SO4,NAHSO4,NH4HSO4,LC
        scase = 'I1'
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drnahso4) THEN
        scase = 'I2'
        CALL calci2 ! NA2SO4,(NH4)2SO4,NAHSO4,LC
        scase = 'I2'
        !
      ELSE IF (drnahso4<=rh .AND. rh<drlc) THEN
        scase = 'I3'
        CALL calci3 ! NA2SO4,(NH4)2SO4,LC
        scase = 'I3'
        !
      ELSE IF (drlc<=rh .AND. rh<drnh42s4) THEN
        scase = 'I4'
        CALL calci4 ! NA2SO4,(NH4)2SO4
        scase = 'I4'
        !
      ELSE IF (drnh42s4<=rh .AND. rh<drna2so4) THEN
        scase = 'I5'
        CALL calci5 ! NA2SO4
        scase = 'I5'
        !
      ELSE IF (drna2so4<=rh) THEN
        scase = 'I6'
        CALL calci6 ! NO SOLIDS
        scase = 'I6'
      END IF
    END IF
    !
    CALL calcnhp ! HNO3, NH3, HCL in gas phase
    CALL calcnh3p
    !
    ! *** SULFATE RICH (FREE ACID)
    !
  ELSE IF (sulrat<1.0) THEN
    DO i = 1, ncomp
      w(i) = waer(i)
    END DO
    !
    IF (metstbl==1) THEN
      scase = 'J3'
      CALL calcj3 ! Only liquid (metastable)
      scase = 'J3'
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'J1'
        CALL calcj1 ! NH4HSO4,NAHSO4
        scase = 'J1'
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drnahso4) THEN
        scase = 'J2'
        CALL calcj2 ! NAHSO4
        scase = 'J2'
        !
      ELSE IF (drnahso4<=rh) THEN
        scase = 'J3'
        CALL calcj3
        scase = 'J3'
      END IF
    END IF
    !
    CALL calcnhp ! HNO3, NH3, HCL in gas phase
    CALL calcnh3p
    !
  END IF
  !
  ! *** IF AFTER CALCULATIONS, SULRATW < SULRAT < 2.0
  !                            and WATER = 0          => SULFATE RICH CASE.
  !
  IF (sulratw<=sulrat .AND. sulrat<2.0 .AND. water<=tiny) THEN
    tryliq = .FALSE.
    GO TO 100
  END IF
  !
  RETURN
  !
END SUBROUTINE isrp3r
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE ISRP4R
! *** THIS SUBROUTINE IS THE DRIVER ROUTINE FOR THE REVERSE PROBLEM OF
!     AN AMMONIUM-SULFATE-NITRATE-CHLORIDE-SODIUM-CALCIUM-POTTASIUM-MAGNESIUM AEROSOL SYSTEM.
!     THE COMPOSITION REGIME IS DETERMINED BY THE SULFATE & SODIUM
!     RATIOS AND BY THE AMBIENT RELATIVE HUMIDITY.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE isrp4r ( wi, rhi, tempi )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: wi(ncomp)
  DOUBLE PRECISION, INTENT(inout) :: rhi
  DOUBLE PRECISION, INTENT(inout) :: tempi
  DOUBLE PRECISION :: frso4
  INTEGER :: i
  DOUBLE PRECISION :: sri
  LOGICAL :: tryliq
  !DOUBLE PRECISION :: getasr
  !cC
  !cC *** ADJUST FOR TOO LITTLE AMMONIUM AND CHLORIDE ***********************
  !cC
  !c      WI(3) = MAX (WI(3), 1.D-10)  ! NH4+ : 1e-4 umoles/m3
  !c      WI(5) = MAX (WI(5), 1.D-10)  ! Cl-  : 1e-4 umoles/m3
  !
  ! *** INITIALIZE ALL VARIABLES ******************************************
  !
  tryliq = .TRUE. ! Use liquid phase sulfate poor limit
  iprob = 1 ! SOLVE REVERSE PROBLEM
  !      METSTBL = 1
  !
100 CALL init4(wi, rhi, tempi) ! COMMON block variables
  !cC
  !cC *** CHECK IF TOO MUCH SODIUM ; ADJUST AND ISSUE ERROR MESSAGE *********
  !cC
  !c      REST = 2.D0*WAER(2) + WAER(4) + WAER(5)
  !c      IF (WAER(1).GT.REST) THEN            ! NA > 2*SO4+CL+NO3 ?
  !c         WAER(1) = (ONE-1D-6)*REST         ! Adjust Na amount
  !c         CALL PUSHERR (0050, 'ISRP3R')     ! Warning error: Na adjusted
  !c      ENDIF
  !
  ! *** CALCULATE SULFATE, CRUSTAL & SODIUM RATIOS ***********************
  !
  IF (tryliq) THEN ! ** WET AEROSOL
    frso4 = waer(2) - waer(1)/2.0D0 - waer(6) - waer(7)/2.0D0 - waer(8) ! SULFATE UNBOUND BY SODIUM,CALCIUM,POTTASIUM,MAGNESIUM
    frso4 = max(frso4, tiny)
    sri = getasr(frso4, rhi) ! SULFATE RATIO FOR NH4+
    sulratw = (waer(1)+frso4*sri+waer(6)+waer(7)+waer(8))/waer(2) ! LIMITING SULFATE RATIO
    sulratw = min(sulratw, 2.0D0)
  ELSE
    sulratw = 2.0D0 ! ** DRY AEROSOL
  END IF
  so4rat = (waer(1)+waer(3)+waer(6)+waer(7)+waer(8))/waer(2)
  crnarat = (waer(1)+waer(6)+waer(7)+waer(8))/waer(2)
  crrat = (waer(6)+waer(7)+waer(8))/waer(2)
  !
  ! *** FIND CALCULATION REGIME FROM (SULRAT,RH) **************************
  !
  ! *** SULFATE POOR ; SODIUM+CRUSTALS POOR
  !
  IF (sulratw<=so4rat .AND. crnarat<2.0) THEN
    !
    IF (metstbl==1) THEN
      scase = 'V7'
      CALL calcv7 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4no3) THEN
        scase = 'V1'
        CALL calcv1 ! CaSO4, NH4NO3, NH4CL, (NH4)2SO4, MGSO4, NA2SO4, K2SO4
        !
      ELSE IF (drnh4no3<=rh .AND. rh<drnh4cl) THEN
        scase = 'V2'
        CALL calcv2 ! CaSO4, NH4CL, (NH4)2SO4, MGSO4, NA2SO4, K2SO4
        !
      ELSE IF (drnh4cl<=rh .AND. rh<drnh42s4) THEN
        scase = 'V3'
        CALL calcv3 ! CaSO4, (NH4)2SO4, MGSO4, NA2SO4, K2SO4
        !
      ELSE IF (drnh42s4<=rh .AND. rh<drmgso4) THEN
        scase = 'V4'
        CALL calcv4 ! CaSO4, MGSO4, NA2SO4, K2SO4
        !
      ELSE IF (drmgso4<=rh .AND. rh<drna2so4) THEN
        scase = 'V5'
        CALL calcv5 ! CaSO4, NA2SO4, K2SO4
        !
      ELSE IF (drna2so4<=rh .AND. rh<drk2so4) THEN
        scase = 'V6'
        CALL calcv6 ! CaSO4, K2SO4
        !
      ELSE IF (drk2so4<=rh) THEN
        scase = 'V7'
        CALL calcv7 ! CaSO4
      END IF
    END IF
    !
    ! *** SULFATE POOR: Rso4>2; (DUST + SODIUM) RICH: R(Cr+Na)>2; DUST POOR: Rcr<2.
    !
  ELSE IF (so4rat>=sulratw .AND. crnarat>=2.0) THEN
    !
    IF (crrat<=2.0) THEN
      !
      IF (metstbl==1) THEN
        scase = 'U8'
        CALL calcu8 ! Only liquid (metastable)
      ELSE
        !
        IF (rh<drnh4no3) THEN
          scase = 'U1'
          CALL calcu1 ! CaSO4, NH4NO3, NH4CL, MGSO4, NA2SO4, K2SO4, NACL, NANO3
          !
        ELSE IF (drnh4no3<=rh .AND. rh<drnano3) THEN
          scase = 'U2'
          CALL calcu2 ! CaSO4, NH4CL, MGSO4, NA2SO4, K2SO4, NACL, NANO3
          !
        ELSE IF (drnano3<=rh .AND. rh<drnacl) THEN
          scase = 'U3'
          CALL calcu3 ! CaSO4, NH4CL, MGSO4, NA2SO4, K2SO4, NACL
          !
        ELSE IF (drnacl<=rh .AND. rh<drnh4cl) THEN
          scase = 'U4'
          CALL calcu4 ! CaSO4, NH4CL, MGSO4, NA2SO4, K2SO4
          !
        ELSE IF (drnh4cl<=rh .AND. rh<drmgso4) THEN
          scase = 'U5'
          CALL calcu5 ! CaSO4, MGSO4, NA2SO4, K2SO4
          !
        ELSE IF (drmgso4<=rh .AND. rh<drna2so4) THEN
          scase = 'U6'
          CALL calcu6 ! CaSO4, NA2SO4, K2SO4
          !
        ELSE IF (drna2so4<=rh .AND. rh<drk2so4) THEN
          scase = 'U7'
          CALL calcu7 ! CaSO4, K2SO4
          !
        ELSE IF (drk2so4<=rh) THEN
          scase = 'U8'
          CALL calcu8 ! CaSO4
        END IF
      END IF
      !
      ! *** SULFATE POOR: Rso4>2; (DUST + SODIUM) RICH: R(Cr+Na)>2; DUST POOR: Rcr<2.
      !
    ELSE IF (crrat>2.0) THEN
      !
      IF (metstbl==1) THEN
        scase = 'W13'
        CALL calcw13 ! Only liquid (metastable)
      ELSE
        !
        IF (rh<drcacl2) THEN
          scase = 'W1'
          CALL calcw1 ! CaSO4, CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4,
          !                                    ! MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drcacl2<=rh .AND. rh<drmgcl2) THEN
          scase = 'W2'
          CALL calcw2 ! CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
          !                                   ! MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drmgcl2<=rh .AND. rh<drcano32) THEN
          scase = 'W3'
          CALL calcw3 ! CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
          !                                   ! MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drcano32<=rh .AND. rh<drmgno32) THEN
          scase = 'W4'
          CALL calcw4 ! CaSO4, K2SO4, KNO3, KCL, MGSO4,
          !                                   ! MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drmgno32<=rh .AND. rh<drnh4no3) THEN
          scase = 'W5'
          CALL calcw5 ! CaSO4, K2SO4, KNO3, KCL, MGSO4,
          !                                   ! NANO3, NACL, NH4NO3, NH4CL
          !
        ELSE IF (drnh4no3<=rh .AND. rh<drnano3) THEN
          scase = 'W6'
          CALL calcw6 ! CaSO4, K2SO4, KNO3, KCL, MGSO4, NANO3, NACL, NH4CL
          !
        ELSE IF (drnano3<=rh .AND. rh<drnacl) THEN
          scase = 'W7'
          CALL calcw7 ! CaSO4, K2SO4, KNO3, KCL, MGSO4, NACL, NH4CL
          !
        ELSE IF (drnacl<=rh .AND. rh<drnh4cl) THEN
          scase = 'W8'
          CALL calcw8 ! CaSO4, K2SO4, KNO3, KCL, MGSO4, NH4CL
          !
        ELSE IF (drnh4cl<=rh .AND. rh<drkcl) THEN
          scase = 'W9'
          CALL calcw9 ! CaSO4, K2SO4, KNO3, KCL, MGSO4
          !
        ELSE IF (drkcl<=rh .AND. rh<drmgso4) THEN
          scase = 'W10'
          CALL calcw10 ! CaSO4, K2SO4, KNO3, MGSO4
          !
        ELSE IF (drmgso4<=rh .AND. rh<drkno3) THEN
          scase = 'W11'
          CALL calcw11 ! CaSO4, K2SO4, KNO3
          !
        ELSE IF (drkno3<=rh .AND. rh<drk2so4) THEN
          scase = 'W12'
          CALL calcw12 ! CaSO4, K2SO4
          !
        ELSE IF (drk2so4<=rh) THEN
          scase = 'W13'
          CALL calcw13 ! CaSO4
        END IF
      END IF
      !        CALL CALCNH3
    END IF
    !
    ! *** SULFATE RICH (NO ACID): 1<Rso4<2;
    !
  ELSE IF (1.0<=so4rat .AND. so4rat<sulratw) THEN
    DO i = 1, ncomp
      w(i) = waer(i)
    END DO
    !
    IF (metstbl==1) THEN
      scase = 'L9'
      CALL calcl9 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4hs4) THEN
        scase = 'L1'
        CALL calcl1 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4,(NH4)2SO4,NAHSO4,NH4HSO4,LC
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drnahso4) THEN
        scase = 'L2'
        CALL calcl2 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4,(NH4)2SO4,NAHSO4,LC
        !
      ELSE IF (drnahso4<=rh .AND. rh<drlc) THEN
        scase = 'L3'
        CALL calcl3 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4,(NH4)2SO4,LC
        !
      ELSE IF (drlc<=rh .AND. rh<drnh42s4) THEN
        scase = 'L4'
        CALL calcl4 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4,(NH4)2SO4
        !
      ELSE IF (drnh42s4<=rh .AND. rh<drkhso4) THEN
        scase = 'L5'
        CALL calcl5 ! CASO4,K2SO4,MGSO4,KHSO4,NA2SO4
        !
      ELSE IF (drkhso4<=rh .AND. rh<drmgso4) THEN
        scase = 'L6'
        CALL calcl6 ! CASO4,K2SO4,MGSO4,NA2SO4
        !
      ELSE IF (drmgso4<=rh .AND. rh<drna2so4) THEN
        scase = 'L7'
        CALL calcl7 ! CASO4,K2SO4,NA2SO4
        !
      ELSE IF (drna2so4<=rh .AND. rh<drk2so4) THEN
        scase = 'L8'
        CALL calcl8 ! CASO4,K2SO4
        !
      ELSE IF (drk2so4<=rh) THEN
        scase = 'L9'
        CALL calcl9 ! CaSO4
      END IF
    END IF
    !
    CALL calcnhp ! MINOR SPECIES: HNO3, HCl
    CALL calcnh3p !                NH3
    !
    ! *** SULFATE SUPER RICH (FREE ACID): Rso4<1;
    !
  ELSE IF (so4rat<1.0) THEN
    DO i = 1, ncomp
      w(i) = waer(i)
    END DO
    !
    IF (metstbl==1) THEN
      scase = 'K4'
      CALL calck4 ! Only liquid (metastable)
    ELSE
      !
      IF (rh<drnh4hs4) THEN ! RH < 0.4
        scase = 'K1'
        CALL calck1 ! NH4HSO4,NAHSO4,KHSO4,CASO4
        !
      ELSE IF (drnh4hs4<=rh .AND. rh<drnahso4) THEN
        scase = 'K2'
        CALL calck2 ! NAHSO4,KHSO4,CASO4
        !
      ELSE IF (drnahso4<=rh .AND. rh<drkhso4) THEN
        scase = 'K3'
        CALL calck3 ! KHSO4,CASO4    0.52 < RH < 0.86
        !
      ELSE IF (drkhso4<=rh) THEN
        scase = 'K4'
        CALL calck4 ! CASO4
      END IF
    END IF
    !
    CALL calcnhp ! MINOR SPECIES: HNO3, HCl
    CALL calcnh3p !                NH3
    !
  END IF
  !
  ! *** IF AFTER CALCULATIONS, SULRATW < SO4RAT < 2.0
  !                            and WATER = 0          => SULFATE RICH CASE.
  !
  IF (sulratw<=so4rat .AND. so4rat<2.0 .AND. water<=tiny) THEN
    tryliq = .FALSE.
    GO TO 100
  END IF
  !
  RETURN
  !
END SUBROUTINE isrp4r
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCS2
! *** CASE S2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. LIQUID AEROSOL PHASE ONLY POSSIBLE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcs2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: a2
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh3gi
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** CALCULATE WATER CONTENT *****************************************
  !
  molalr(4) = min(waer(2), 0.5D0*waer(3))
  water = molalr(4)/m0(4) ! ZSR correlation
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !C         A21  = XK21*WATER*R*TEMP
    a2 = xk2*r*temp/xkw/rh*(gama(8)/gama(9))**2.
    akw = xkw*rh*water*water
    !
    nh4i = waer(3)
    so4i = waer(2)
    hso4i = zero
    !
    CALL calcph(2.D0*so4i-nh4i, hi, ohi) ! Get pH
    !
    nh3aq = zero ! AMMONIA EQUILIBRIUM
    IF (hi<ohi) THEN
      CALL calcamaq(nh4i, ohi, del)
      nh4i = max(nh4i-del, zero)
      ohi = max(ohi-del, tiny)
      nh3aq = del
      hi = akw/ohi
    END IF
    !
    CALL calchs4(hi, so4i, zero, del) ! SULFATE EQUILIBRIUM
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !
    nh3gi = nh4i/hi/a2 !    NH3AQ/A21
    !
    ! *** SPECIATION & WATER CONTENT ***************************************
    !
    molal(1) = hi
    molal(3) = nh4i
    molal(5) = so4i
    molal(6) = hso4i
    coh = ohi
    gasaq(1) = nh3aq
    gnh3 = nh3gi
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
100 RETURN
  !
END SUBROUTINE calcs2
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCS1
! *** CASE S1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4
!
!     A SIMPLE MATERIAL BALANCE IS PERFORMED, AND THE SOLID (NH4)2SO4
!     IS CALCULATED FROM THE SULFATES. THE EXCESS AMMONIA REMAINS IN
!     THE GAS PHASE.
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcs1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  cnh42s4 = min(waer(2), 0.5D0*waer(3)) ! For bad input problems
  gnh3 = zero
  !
  w(2) = cnh42s4
  w(3) = 2.D0*cnh42s4 + gnh3
  !
  RETURN
  !
END SUBROUTINE calcs1


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCN3
! *** CASE N3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. THERE IS ONLY A LIQUID PHASE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcn3 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: aml5
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  frst = .TRUE.
  calain = .TRUE.
  !
  ! *** AEROSOL WATER CONTENT
  !
  molalr(4) = min(waer(2), 0.5D0*waer(3)) ! (NH4)2SO4
  aml5 = max(waer(3)-2.D0*molalr(4), zero) ! "free" NH4
  molalr(5) = max(min(aml5,waer(4)), zero) ! NH4NO3=MIN("free",NO3)
  water = molalr(4)/m0(4) + molalr(5)/m0(5)
  water = max(water, tiny)
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a2 = xk2*r*temp/xkw/rh*(gama(8)/gama(9))**2.
    !C         A21   = XK21*WATER*R*TEMP
    a3 = xk4*r*temp*(water/gama(10))**2.0
    a4 = xk7*(water/gama(4))**3.0
    akw = xkw*rh*water*water
    !
    ! ION CONCENTRATIONS
    !
    nh4i = waer(3)
    no3i = waer(4)
    so4i = waer(2)
    hso4i = zero
    !
    CALL calcph(2.D0*so4i+no3i-nh4i, hi, ohi)
    !
    ! AMMONIA ASSOCIATION EQUILIBRIUM
    !
    nh3aq = zero
    no3aq = zero
    gg = 2.D0*so4i + no3i - nh4i
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
    ELSE
      hi = zero
      CALL calcniaq2(gg, no3i, hi, no3aq) ! HNO3
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(3) = nh4i
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    coh = ohi
    !
    cnh42s4 = zero
    cnh4no3 = zero
    !
    gasaq(1) = nh3aq
    gasaq(3) = no3aq
    !
    ghno3 = hi*no3i/a3
    gnh3 = nh4i/hi/a2 !   NH3AQ/A21
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP ******************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** RETURN ***********************************************************
  !
100 RETURN
  !
END SUBROUTINE calcn3
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCN2
! *** CASE N2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : (NH4)2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcn2 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: dx
  INTEGER :: i
  DOUBLE PRECISION :: p4
  DOUBLE PRECISION :: psi1hi
  DOUBLE PRECISION :: psi1lo
  DOUBLE PRECISION :: x1
  DOUBLE PRECISION :: x2
  DOUBLE PRECISION :: x3
  DOUBLE PRECISION :: y1
  DOUBLE PRECISION :: y2
  DOUBLE PRECISION :: y3
  DOUBLE PRECISION :: yhi
  DOUBLE PRECISION :: ylo
  DOUBLE PRECISION :: yy
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  chi1 = min(waer(2), 0.5D0*waer(3)) ! (NH4)2SO4
  chi2 = max(waer(3)-2.D0*chi1, zero) ! "Free" NH4+
  chi3 = max(waer(4)-chi2, zero) ! "Free" NO3
  !
  psi2 = chi2
  psi3 = chi3
  !
  calaou = .TRUE. ! Outer loop activity calculation flag
  psi1lo = tiny ! Low  limit
  psi1hi = chi1 ! High limit
  !
  ! *** INITIAL VALUES FOR BISECTION ************************************
  !
  x1 = psi1hi
  y1 = funcn2(x1)
  IF (y1<=eps) RETURN ! IF (ABS(Y1).LE.EPS .OR. Y1.LE.ZERO) RETURN
  yhi = y1 ! Save Y-value at HI position
  !
  ! *** ROOT TRACKING ; FOR THE RANGE OF HI AND LO **********************
  !
  dx = (psi1hi-psi1lo)/float(ndiv)
  DO i = 1, ndiv
    x2 = max(x1-dx, zero)
    y2 = funcn2(x2)
    IF (sign(1.D0,y1)*sign(1.D0,y2)<zero) GO TO 100 ! (Y1*Y2.LT.ZERO)
    x1 = x2
    y1 = y2
  END DO
  !
  ! *** NO SUBDIVISION WITH SOLUTION FOUND
  !
  ylo = y1 ! Save Y-value at Hi position
  IF (abs(y2)<eps) THEN ! X2 IS A SOLUTION
    RETURN
    !
    ! *** { YLO, YHI } < 0.0 THE SOLUTION IS ALWAYS UNDERSATURATED WITH NH3
    !
  ELSE IF (ylo<zero .AND. yhi<zero) THEN
    p4 = chi4
    yy = funcn2(p4)
    GO TO 120
    !
    ! *** { YLO, YHI } > 0.0 THE SOLUTION IS ALWAYS SUPERSATURATED WITH NH3
    !
  ELSE IF (ylo>zero .AND. yhi>zero) THEN
    p4 = tiny
    yy = funcn2(p4)
    GO TO 120
  ELSE
    CALL pusherr(0001, 'CALCN2') ! WARNING ERROR: NO SOLUTION
    RETURN
  END IF
  !
  ! *** PERFORM BISECTION ***********************************************
  !
100 DO i = 1, maxit
    x3 = 0.5*(x1+x2)
    y3 = funcn2(x3)
    IF (sign(1.D0,y1)*sign(1.D0,y3)<=zero) THEN ! (Y1*Y3 .LE. ZERO)
      y2 = y3
      x2 = x3
    ELSE
      y1 = y3
      x1 = x3
    END IF
    IF (abs(x2-x1)<=eps*x1) GO TO 110
  END DO
  CALL pusherr(0002, 'CALCN2') ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CONVERGED ; RETURN **********************************************
  !
110 x3 = 0.5*(x1+x2)
  y3 = funcn2(x3)
120 CONTINUE
  RETURN
  !
END SUBROUTINE calcn2



!======================================================================
!
! *** ISORROPIA CODE
! *** FUNCTION FUNCN2
! *** CASE D2
!     FUNCTION THAT SOLVES THE SYSTEM OF EQUATIONS FOR CASE D2 ;
!     AND RETURNS THE VALUE OF THE ZEROED FUNCTION IN FUNCN2.
!
!=======================================================================
!
DOUBLE PRECISION FUNCTION funcn2 ( p1 )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION, INTENT(inout) :: p1
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  psi1 = p1
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a2 = xk2*r*temp/xkw/rh*(gama(8)/gama(9))**2.
    !C         A21   = XK21*WATER*R*TEMP
    a3 = xk4*r*temp*(water/gama(10))**2.0
    a4 = xk7*(water/gama(4))**3.0
    akw = xkw*rh*water*water
    !
    ! ION CONCENTRATIONS
    !
    nh4i = 2.D0*psi1 + psi2
    no3i = psi2 + psi3
    so4i = psi1
    hso4i = zero
    !
    CALL calcph(2.D0*so4i+no3i-nh4i, hi, ohi)
    !
    ! AMMONIA ASSOCIATION EQUILIBRIUM
    !
    nh3aq = zero
    no3aq = zero
    gg = 2.D0*so4i + no3i - nh4i
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
    ELSE
      hi = zero
      CALL calcniaq2(gg, no3i, hi, no3aq) ! HNO3
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(3) = nh4i
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    coh = ohi
    !
    cnh42s4 = chi1 - psi1
    cnh4no3 = zero
    !
    gasaq(1) = nh3aq
    gasaq(3) = no3aq
    !
    ghno3 = hi*no3i/a3
    gnh3 = nh4i/hi/a2 !   NH3AQ/A21
    !
    ! *** CALCULATE MOLALR ARRAY, WATER AND ACTIVITIES **********************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !
  ! *** CALCULATE OBJECTIVE FUNCTION ************************************
  !
100 funcn2 = nh4i*nh4i*so4i/a4 - one
  RETURN
  !
  ! *** END OF FUNCTION FUNCN2 ********************************************
  !
END FUNCTION funcn2
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCN1
! *** CASE N1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3
!
!     THERE ARE TWO REGIMES DEFINED BY RELATIVE HUMIDITY:
!     1. RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCN1A)
!     2. RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcn1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmasan) THEN
    scase = 'N1 ; SUBCASE 1'
    CALL calcn1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'N1 ; SUBCASE 1'
  ELSE
    scase = 'N1 ; SUBCASE 2'
    CALL calcmdrp(rh, drmasan, drnh4no3, calcn1a, calcn2)
    scase = 'N1 ; SUBCASE 2'
  END IF
  !
  RETURN
  !
END SUBROUTINE calcn1



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCN1A
! *** CASE N1 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcn1a (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: psi1
  DOUBLE PRECISION :: psi2
  !
  ! *** SETUP PARAMETERS *************************************************
  !
  !CC      A1      = XK10/R/TEMP/R/TEMP
  !
  ! *** CALCULATE AEROSOL COMPOSITION ************************************
  !
  !CC      CHI1    = 2.D0*WAER(4)        ! Free parameter ; arbitrary value.
  psi1 = waer(4)
  !
  ! *** The following statment is here to avoid negative NH4+ values in
  !     CALCN? routines that call CALCN1A
  !
  psi2 = max(min(waer(2),0.5D0*(waer(3)-psi1)), tiny)
  !
  cnh4no3 = psi1
  cnh42s4 = psi2
  !
  !CC      GNH3    = CHI1 + PSI1 + 2.0*PSI2
  !CC      GHNO3   = A1/(CHI1-PSI1) + PSI1
  gnh3 = zero
  ghno3 = zero
  !
  w(2) = psi2
  w(3) = gnh3 + psi1 + 2.0*psi2
  w(4) = ghno3 + psi1
  !
  RETURN
  !
END SUBROUTINE calcn1a

!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCQ5
! *** CASE Q5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); SODIUM POOR (SODRAT < 2.0)
!     2. LIQUID AND SOLID PHASES ARE POSSIBLE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcq5 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: ggcl
  DOUBLE PRECISION :: ggno3
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcq1a
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  !
  CALL calcmr ! WATER
  !
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! ION CONCENTRATIONS
    !
    nai = waer(1)
    so4i = waer(2)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
      hso4i = zero
    ELSE
      ggno3 = max(2.D0*so4i+no3i-nai-nh4i, zero)
      ggcl = max(gg-ggno3, zero)
      IF (ggcl>tiny) CALL calcclaq2(ggcl, cli, hi, claq) ! HCl
      IF (ggno3>tiny) THEN
        IF (ggcl<=tiny) hi = zero
        CALL calcniaq2(ggno3, no3i, hi, no3aq) ! HNO3
      END IF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCQ5')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cna2so4 = zero
  !
  RETURN
  !
END SUBROUTINE calcq5
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCQ4
! *** CASE Q4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); SODIUM POOR (SODRAT < 2.0)
!     2. LIQUID AND SOLID PHASES ARE POSSIBLE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcq4 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: ggcl
  DOUBLE PRECISION :: ggno3
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: root3
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv1 = .TRUE.
  psi1o = -great
  root3 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcq1a
  !
  chi1 = cna2so4 ! SALTS
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = waer(2)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a5 = xk5*(water/gama(2))**3. ! Na2SO4    <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a5) THEN
      bb = -(waer(2)+waer(1))
      cc = waer(1)*waer(2) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*waer(2)-a5)
      CALL poly3(bb, cc, dd, root3, islv)
      IF (islv/=0) root3 = tiny
      root3 = min(root3, waer(1)/2.0, waer(2), chi1)
      root3 = max(root3, zero)
      psi1 = chi1 - root3
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    nai = waer(1) - 2.D0*root3
    so4i = waer(2) - root3
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
      hso4i = zero
    ELSE
      ggno3 = max(2.D0*so4i+no3i-nai-nh4i, zero)
      ggcl = max(gg-ggno3, zero)
      IF (ggcl>tiny) CALL calcclaq2(ggcl, cli, hi, claq) ! HCl
      IF (ggno3>tiny) THEN
        IF (ggcl<=tiny) hi = zero
        CALL calcniaq2(ggno3, no3i, hi, no3aq) ! HNO3
      END IF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv1) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCQ4')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cna2so4 = chi1 - psi1
  !
  RETURN
  !
END SUBROUTINE calcq4
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCQ3
! *** CASE Q3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : NH4CL, NA2SO4, NANO3, NACL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcq3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: excl
  LOGICAL :: exno
  !
  ! *** REGIME DEPENDS ON AMBIENT RELATIVE HUMIDITY & POSSIBLE SPECIES ***
  !
  exno = waer(4) > tiny
  excl = waer(5) > tiny
  !
  IF (exno .OR. excl) THEN ! *** NITRATE OR CHLORIDE EXISTS
    scase = 'Q3 ; SUBCASE 1'
    CALL calcq3a
    scase = 'Q3 ; SUBCASE 1'
    !
  ELSE ! *** NO CHLORIDE AND NITRATE
    IF (rh<drmg3) THEN
      scase = 'Q3 ; SUBCASE 2'
      CALL calcq1a ! SOLID
      scase = 'Q3 ; SUBCASE 2'
    ELSE
      scase = 'Q3 ; SUBCASE 3' ! MDRH (NH4)2SO4, NA2SO4
      CALL calcmdrp(rh, drmg3, drnh42s4, calcq1a, calcq4)
      scase = 'Q3 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcq3



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCQ3A
! *** CASE Q3 ; SUBCASE A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); SODIUM POOR (SODRAT < 2.0)
!     2. LIQUID AND SOLID PHASES ARE POSSIBLE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcq3a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: ggcl
  DOUBLE PRECISION :: ggno3
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv6
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi6o
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root3
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv1 = .TRUE.
  psconv6 = .TRUE.
  !
  psi1o = -great
  psi6o = -great
  !
  root1 = zero
  root3 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcq1a
  !
  chi1 = cna2so4 ! SALTS
  chi4 = cnh4cl
  chi6 = cnh42s4
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = waer(2)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a5 = xk5*(water/gama(2))**3. ! Na2SO4    <==> Na+
    a7 = xk7*(water/gama(4))**3. ! (NH4)2SO4 <==> NH4+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a5) THEN
      bb = -(waer(2)+waer(1)-root1)
      cc = waer(1)*(waer(2)-root1) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*(waer(2)-root1)-a5)
      CALL poly3(bb, cc, dd, root3, islv)
      IF (islv/=0) root3 = tiny
      root3 = min(root3, waer(1)/2.0, waer(2)-root1, chi1)
      root3 = max(root3, zero)
      psi1 = chi1 - root3
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! AMMONIUM SULFATE
    !
    IF (nh4i*nh4i*so4i>a7) THEN
      bb = -(waer(2)+waer(3)-root3)
      cc = waer(3)*(waer(2)-root3+0.5D0*waer(3))
      dd = -((waer(2)-root3)*waer(3)**2.D0+a7)/4.D0
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = min(root1, waer(3), waer(2)-root3, chi6)
      root1 = max(root1, zero)
      psi6 = chi6 - root1
    END IF
    psconv6 = abs(psi6-psi6o) <= eps*psi6o
    psi6o = psi6
    !
    ! ION CONCENTRATIONS
    !
    nai = waer(1) - 2.D0*root3
    so4i = waer(2) - root1 - root3
    nh4i = waer(3) - 2.D0*root1
    no3i = waer(4)
    cli = waer(5)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
      hso4i = zero
    ELSE
      ggno3 = max(2.D0*so4i+no3i-nai-nh4i, zero)
      ggcl = max(gg-ggno3, zero)
      IF (ggcl>tiny) CALL calcclaq2(ggcl, cli, hi, claq) ! HCl
      IF (ggno3>tiny) THEN
        IF (ggcl<=tiny) hi = zero
        CALL calcniaq2(ggno3, no3i, hi, no3aq) ! HNO3
      END IF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv1 .AND. psconv6) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCQ3A')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = chi6 - psi6
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cna2so4 = chi1 - psi1
  !
  RETURN
  !
END SUBROUTINE calcq3a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCQ2
! *** CASE Q2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID & LIQUID AEROSOL POSSIBLE
!     3. SOLIDS POSSIBLE : NH4CL, NA2SO4, NANO3, NACL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcq2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: excl
  LOGICAL :: exno
  !
  ! *** REGIME DEPENDS ON AMBIENT RELATIVE HUMIDITY & POSSIBLE SPECIES ***
  !
  exno = waer(4) > tiny
  excl = waer(5) > tiny
  !
  IF (exno) THEN ! *** NITRATE EXISTS
    scase = 'Q2 ; SUBCASE 1'
    CALL calcq2a
    scase = 'Q2 ; SUBCASE 1'
    !
  ELSE IF (.NOT. exno .AND. excl) THEN ! *** ONLY CHLORIDE EXISTS
    IF (rh<drmg2) THEN
      scase = 'Q2 ; SUBCASE 2'
      CALL calcq1a ! SOLID
      scase = 'Q2 ; SUBCASE 2'
    ELSE
      scase = 'Q2 ; SUBCASE 3' ! MDRH (NH4)2SO4, NA2SO4, NH4CL
      CALL calcmdrp(rh, drmg2, drnh4cl, calcq1a, calcq3a)
      scase = 'Q2 ; SUBCASE 3'
    END IF
    !
  ELSE ! *** NO CHLORIDE AND NITRATE
    IF (rh<drmg3) THEN
      scase = 'Q2 ; SUBCASE 2'
      CALL calcq1a ! SOLID
      scase = 'Q2 ; SUBCASE 2'
    ELSE
      scase = 'Q2 ; SUBCASE 4' ! MDRH (NH4)2SO4, NA2SO4
      CALL calcmdrp(rh, drmg3, drnh42s4, calcq1a, calcq4)
      scase = 'Q2 ; SUBCASE 4'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcq2


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCQ2A
! *** CASE Q2 ; SUBCASE A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); SODIUM POOR (SODRAT < 2.0)
!     2. LIQUID AND SOLID PHASES ARE POSSIBLE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcq2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: ggcl
  DOUBLE PRECISION :: ggno3
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv4
  LOGICAL :: psconv6
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi4o
  DOUBLE PRECISION :: psi6o
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root2
  DOUBLE PRECISION :: root2a
  DOUBLE PRECISION :: root2b
  DOUBLE PRECISION :: root3
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv1 = .TRUE.
  psconv4 = .TRUE.
  psconv6 = .TRUE.
  !
  psi1o = -great
  psi4o = -great
  psi6o = -great
  !
  root1 = zero
  root2 = zero
  root3 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcq1a
  !
  chi1 = cna2so4 ! SALTS
  chi4 = cnh4cl
  chi6 = cnh42s4
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = waer(2)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a5 = xk5*(water/gama(2))**3. ! Na2SO4    <==> Na+
    a14 = xk14*(water/gama(6))**2. ! NH4Cl     <==> NH4+
    a7 = xk7*(water/gama(4))**3. ! (NH4)2SO4 <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! AMMONIUM CHLORIDE
    !
    IF (nh4i*cli>a14) THEN
      bb = -(waer(3)+waer(5)-2.D0*root1)
      cc = waer(5)*(waer(3)-2.D0*root1) - a14
      dd = bb*bb - 4.D0*cc
      IF (dd<zero) THEN
        root2 = zero
      ELSE
        dd = sqrt(dd)
        root2a = 0.5D0*(-bb+dd)
        root2b = 0.5D0*(-bb-dd)
        IF (zero<=root2a) THEN
          root2 = root2a
        ELSE
          root2 = root2b
        END IF
        root2 = min(root2, waer(5), waer(3)-2.D0*root1, chi4)
        root2 = max(root2, zero)
        psi4 = chi4 - root2
      END IF
    END IF
    psconv4 = abs(psi4-psi4o) <= eps*psi4o
    psi4o = psi4
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a5) THEN
      bb = -(waer(2)+waer(1)-root1)
      cc = waer(1)*(waer(2)-root1) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*(waer(2)-root1)-a5)
      CALL poly3(bb, cc, dd, root3, islv)
      IF (islv/=0) root3 = tiny
      root3 = min(root3, waer(1)/2.0, waer(2)-root1, chi1)
      root3 = max(root3, zero)
      psi1 = chi1 - root3
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! AMMONIUM SULFATE
    !
    IF (nh4i*nh4i*so4i>a7) THEN
      bb = -(waer(2)+waer(3)-root2-root3)
      cc = (waer(3)-root2)*(waer(2)-root3+0.5D0*(waer(3)-root2))
      dd = -((waer(2)-root3)*(waer(3)-root2)**2.D0+a7)/4.D0
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = min(root1, waer(3)-root2, waer(2)-root3, chi6)
      root1 = max(root1, zero)
      psi6 = chi6 - root1
    END IF
    psconv6 = abs(psi6-psi6o) <= eps*psi6o
    psi6o = psi6
    !
    ! ION CONCENTRATIONS
    !
    nai = waer(1) - 2.D0*root3
    so4i = waer(2) - root1 - root3
    nh4i = waer(3) - root2 - 2.D0*root1
    no3i = waer(4)
    cli = waer(5) - root2
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
      hso4i = zero
    ELSE
      ggno3 = max(2.D0*so4i+no3i-nai-nh4i, zero)
      ggcl = max(gg-ggno3, zero)
      IF (ggcl>tiny) CALL calcclaq2(ggcl, cli, hi, claq) ! HCl
      IF (ggno3>tiny) THEN
        IF (ggcl<=tiny) hi = zero
        CALL calcniaq2(ggno3, no3i, hi, no3aq) ! HNO3
      END IF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv1 .AND. psconv4 .AND. psconv6) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCQ2A')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = chi6 - psi6
  cnh4no3 = zero
  cnh4cl = chi4 - psi4
  cnacl = zero
  cnano3 = zero
  cna2so4 = chi1 - psi1
  !
  RETURN
  !
END SUBROUTINE calcq2a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCQ1
! *** CASE Q1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, (NH4)2SO4, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcq1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: excl
  LOGICAL :: exno
  !
  ! *** REGIME DEPENDS ON AMBIENT RELATIVE HUMIDITY & POSSIBLE SPECIES ***
  !
  exno = waer(4) > tiny
  excl = waer(5) > tiny
  !
  IF (exno .AND. excl) THEN ! *** NITRATE & CHLORIDE EXIST
    IF (rh<drmg1) THEN
      scase = 'Q1 ; SUBCASE 1'
      CALL calcq1a ! SOLID
      scase = 'Q1 ; SUBCASE 1'
    ELSE
      scase = 'Q1 ; SUBCASE 2' ! MDRH (NH4)2SO4, NA2SO4, NH4CL, NH4NO3
      CALL calcmdrp(rh, drmg1, drnh4no3, calcq1a, calcq2a)
      scase = 'Q1 ; SUBCASE 2'
    END IF
    !
  ELSE IF (exno .AND. .NOT. excl) THEN ! *** ONLY NITRATE EXISTS
    IF (rh<drmq1) THEN
      scase = 'Q1 ; SUBCASE 1'
      CALL calcq1a ! SOLID
      scase = 'Q1 ; SUBCASE 1'
    ELSE
      scase = 'Q1 ; SUBCASE 3' ! MDRH (NH4)2SO4, NA2SO4, NH4NO3
      CALL calcmdrp(rh, drmq1, drnh4no3, calcq1a, calcq2a)
      scase = 'Q1 ; SUBCASE 3'
    END IF
    !
  ELSE IF (.NOT. exno .AND. excl) THEN ! *** ONLY CHLORIDE EXISTS
    IF (rh<drmg2) THEN
      scase = 'Q1 ; SUBCASE 1'
      CALL calcq1a ! SOLID
      scase = 'Q1 ; SUBCASE 1'
    ELSE
      scase = 'Q1 ; SUBCASE 4' ! MDRH (NH4)2SO4, NA2SO4, NH4CL
      CALL calcmdrp(rh, drmg2, drnh4cl, calcq1a, calcq3a)
      scase = 'Q1 ; SUBCASE 4'
    END IF
    !
  ELSE ! *** NO CHLORIDE AND NITRATE
    IF (rh<drmg3) THEN
      scase = 'Q1 ; SUBCASE 1'
      CALL calcq1a ! SOLID
      scase = 'Q1 ; SUBCASE 1'
    ELSE
      scase = 'Q1 ; SUBCASE 5' ! MDRH (NH4)2SO4, NA2SO4
      CALL calcmdrp(rh, drmg3, drnh42s4, calcq1a, calcq4)
      scase = 'Q1 ; SUBCASE 5'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcq1


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCQ1A
! *** CASE Q1 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM POOR (SODRAT < 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, (NH4)2SO4, NA2SO4
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcq1a (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: frnh3
  DOUBLE PRECISION :: frso4
  !
  ! *** CALCULATE SOLIDS **************************************************
  !
  cna2so4 = 0.5D0*waer(1)
  frso4 = max(waer(2)-cna2so4, zero)
  !
  cnh42s4 = max(min(frso4,0.5D0*waer(3)), tiny)
  frnh3 = max(waer(3)-2.D0*cnh42s4, zero)
  !
  cnh4no3 = min(frnh3, waer(4))
  !CC      FRNO3   = MAX (WAER(4)-CNH4NO3, ZERO)
  frnh3 = max(frnh3-cnh4no3, zero)
  !
  cnh4cl = min(frnh3, waer(5))
  !CC      FRCL    = MAX (WAER(5)-CNH4CL, ZERO)
  frnh3 = max(frnh3-cnh4cl, zero)
  !
  ! *** OTHER PHASES ******************************************************
  !
  water = zero
  !
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  RETURN
  !
END SUBROUTINE calcq1a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR6
! *** CASE R6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); SODIUM RICH (SODRAT >= 2.0)
!     2. THERE IS ONLY A LIQUID PHASE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr6 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: ggcl
  DOUBLE PRECISION :: ggno3
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  CALL calcr1a
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  ! *** CALCULATE WATER **************************************************
  !
  CALL calcmr
  !
  ! *** SETUP LIQUID CONCENTRATIONS **************************************
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    akw = xkw*rh*water*water ! H2O    <==> H+
    !
    nai = waer(1)
    so4i = waer(2)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*waer(2) + no3i + cli - nai - nh4i
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
    ELSE
      ggno3 = max(2.D0*so4i+no3i-nai-nh4i, zero)
      ggcl = max(gg-ggno3, zero)
      IF (ggcl>tiny) CALL calcclaq2(ggcl, cli, hi, claq) ! HCl
      IF (ggno3>tiny) THEN
        IF (ggcl<=tiny) hi = zero
        CALL calcniaq2(ggno3, no3i, hi, no3aq) ! HNO3
      END IF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCR6')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cna2so4 = zero
  !
  RETURN
  !
END SUBROUTINE calcr6
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR5
! *** CASE R5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); SODIUM RICH (SODRAT >= 2.0)
!     2. LIQUID AND SOLID PHASES ARE POSSIBLE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr5 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: ggcl
  DOUBLE PRECISION :: ggno3
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: nai
  LOGICAL :: neac
  LOGICAL :: nean
  LOGICAL :: nesc
  LOGICAL :: nesn
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv
  DOUBLE PRECISION :: psio
  DOUBLE PRECISION :: root
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  CALL calcr1a ! DRY SOLUTION
  !
  nean = cnh4no3 <= tiny ! NH4NO3       ! Water exists?
  neac = cnh4cl <= tiny ! NH4CL
  nesn = cnano3 <= tiny ! NANO3
  nesc = cnacl <= tiny ! NACL
  IF (nean .AND. neac .AND. nesn .AND. nesc) RETURN
  !
  chi1 = cna2so4
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  !
  psio = -great
  !
  ! *** CALCULATE WATER **************************************************
  !
  CALL calcmr
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  psconv = .FALSE.
  !
  ! *** SETUP LIQUID CONCENTRATIONS **************************************
  !
  nai = waer(1)
  so4i = waer(2)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a5 = xk5*(water/gama(2))**3. ! Na2SO4 <==> Na+
    akw = xkw*rh*water*water ! H2O    <==> H+
    !
    ! SODIUM SULFATE
    !
    root = zero
    IF (nai*nai*so4i>a5) THEN
      bb = -3.D0*chi1
      cc = 3.D0*chi1**2.0
      dd = -chi1**3.0 + 0.25D0*a5
      CALL poly3(bb, cc, dd, root, islv)
      IF (islv/=0) root = tiny
      root = min(max(root,zero), chi1)
      psi1 = chi1 - root
    END IF
    psconv = abs(psi1-psio) <= eps*psio
    psio = psi1
    !
    ! ION CONCENTRATIONS
    !
    nai = waer(1) - 2.D0*root
    so4i = waer(2) - root
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
    ELSE
      ggno3 = max(2.D0*so4i+no3i-nai-nh4i, zero)
      ggcl = max(gg-ggno3, zero)
      IF (ggcl>tiny) CALL calcclaq2(ggcl, cli, hi, claq) ! HCl
      IF (ggno3>tiny) THEN
        IF (ggcl<=tiny) hi = zero
        CALL calcniaq2(ggno3, no3i, hi, no3aq) ! HNO3
      END IF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCR5')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  !C      A21      = XK21*WATER*R*TEMP
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2 ! NH4I*OHI/A2/AKW
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cna2so4 = chi1 - psi1
  !
  RETURN
  !
END SUBROUTINE calcr5
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR4
! *** CASE R4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, NA2SO4, NANO3, NACL
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr4 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: exac
  LOGICAL :: exan
  LOGICAL :: exsc
  LOGICAL :: exsn
  !
  ! *** SOLVE FOR DRY CASE AND SEE WHICH SOLIDS ARE POSSIBLE **************
  !
  scase = 'R4 ; SUBCASE 2'
  CALL calcr1a ! SOLID
  scase = 'R4 ; SUBCASE 2'
  !
  exan = cnh4no3 > tiny ! NH4NO3
  exac = cnh4cl > tiny ! NH4CL
  exsn = cnano3 > tiny ! NANO3
  exsc = cnacl > tiny ! NACL
  !
  ! *** REGIME DEPENDS ON RELATIVE HUMIDITY AND POSSIBLE SPECIES **********
  !
  IF (exan .OR. exsn .OR. exsc) THEN ! *** NH4NO3,NANO3 EXIST
    IF (rh>=drmh1) THEN
      scase = 'R4 ; SUBCASE 1'
      CALL calcr4a
      scase = 'R4 ; SUBCASE 1'
    END IF
    !
  ELSE IF (exac) THEN ! *** NH4CL EXISTS ONLY
    IF (rh>=drmr5) THEN
      scase = 'R4 ; SUBCASE 3'
      CALL calcmdrp(rh, drmr5, drnh4cl, calcr1a, calcr5)
      scase = 'R4 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcr4



!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR4A
! *** CASE R4A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); SODIUM RICH (SODRAT >= 2.0)
!     2. LIQUID AND SOLID PHASES ARE POSSIBLE
!
! *** COPYRIGHT 1996-2006, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr4a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: ggcl
  DOUBLE PRECISION :: ggno3
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv4
  DOUBLE PRECISION :: psio1
  DOUBLE PRECISION :: psio4
  DOUBLE PRECISION :: root
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  psconv1 = .FALSE.
  psconv4 = .FALSE.
  psio1 = -great
  psio4 = -great
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcr1a
  !
  chi1 = cna2so4 ! SALTS
  chi4 = cnh4cl
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = waer(2)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a5 = xk5*(water/gama(2))**3. ! Na2SO4 <==> Na+
    a14 = xk14*(water/gama(6))**2. ! NH4Cl  <==> NH4+
    akw = xkw*rh*water*water ! H2O    <==> H+
    !
    ! SODIUM SULFATE
    !
    root = zero
    IF (nai*nai*so4i>a5) THEN
      bb = -3.D0*chi1
      cc = 3.D0*chi1**2.0
      dd = -chi1**3.0 + 0.25D0*a5
      CALL poly3(bb, cc, dd, root, islv)
      IF (islv/=0) root = tiny
      root = min(max(root,zero), chi1)
      psi1 = chi1 - root
      nai = waer(1) - 2.D0*root
      so4i = waer(2) - root
    END IF
    psconv1 = abs(psi1-psio1) <= eps*psio1
    psio1 = psi1
    !
    ! AMMONIUM CHLORIDE
    !
    root = zero
    IF (nh4i*cli>a14) THEN
      bb = -(nh4i+cli)
      cc = -a14 + nh4i*cli
      dd = bb*bb - 4.D0*cc
      root = 0.5D0*(-bb-sqrt(dd))
      IF (root>tiny) THEN
        root = min(root, chi4)
        psi4 = chi4 - root
        nh4i = waer(3) - root
        cli = waer(5) - root
      END IF
    END IF
    psconv4 = abs(psi4-psio4) <= eps*psio4
    psio4 = psi4
    !
    no3i = waer(4)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
    ELSE
      ggno3 = max(2.D0*so4i+no3i-nai-nh4i, zero)
      ggcl = max(gg-ggno3, zero)
      IF (ggcl>tiny) CALL calcclaq2(ggcl, cli, hi, claq) ! HCl
      IF (ggno3>tiny) THEN
        IF (ggcl<=tiny) hi = zero
        CALL calcniaq2(ggno3, no3i, hi, no3aq) ! HNO3
      END IF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv1 .AND. psconv4) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCR4A')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = chi4 - psi4
  cnacl = zero
  cnano3 = zero
  cna2so4 = chi1 - psi1
  !
  RETURN
  !
END SUBROUTINE calcr4a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR3
! *** CASE R3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, NA2SO4, NANO3, NACL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: exac
  LOGICAL :: exan
  LOGICAL :: exsc
  LOGICAL :: exsn
  !
  ! *** SOLVE FOR DRY CASE AND SEE WHICH SOLIDS ARE POSSIBLE **************
  !
  scase = 'R3 ; SUBCASE 2'
  CALL calcr1a ! SOLID
  scase = 'R3 ; SUBCASE 2'
  !
  exan = cnh4no3 > tiny ! NH4NO3
  exac = cnh4cl > tiny ! NH4CL
  exsn = cnano3 > tiny ! NANO3
  exsc = cnacl > tiny ! NACL
  !
  ! *** REGIME DEPENDS ON RELATIVE HUMIDITY AND POSSIBLE SPECIES **********
  !
  IF (exan .OR. exsn) THEN ! *** NH4NO3,NANO3 EXIST
    IF (rh>=drmh1) THEN
      scase = 'R3 ; SUBCASE 1'
      CALL calcr3a
      scase = 'R3 ; SUBCASE 1'
    END IF
    !
  ELSE IF (.NOT. exan .AND. .NOT. exsn) THEN ! *** NH4NO3,NANO3 = 0
    IF (exac .AND. exsc) THEN
      IF (rh>=drmr4) THEN
        scase = 'R3 ; SUBCASE 3'
        CALL calcmdrp(rh, drmr4, drnacl, calcr1a, calcr4a)
        scase = 'R3 ; SUBCASE 3'
      END IF

    ELSE IF (.NOT. exac .AND. exsc) THEN
      IF (rh>=drmr2) THEN
        scase = 'R3 ; SUBCASE 4'
        CALL calcmdrp(rh, drmr2, drnacl, calcr1a, calcr4a)
        scase = 'R3 ; SUBCASE 4'
      END IF

    ELSE IF (exac .AND. .NOT. exsc) THEN
      IF (rh>=drmr5) THEN
        scase = 'R3 ; SUBCASE 5'
        CALL calcmdrp(rh, drmr5, drnacl, calcr1a, calcr5)
        scase = 'R3 ; SUBCASE 5'
      END IF
    END IF
    !
  END IF
  !
  RETURN
  !
END SUBROUTINE calcr3


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR3A
! *** CASE R3A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); SODIUM RICH (SODRAT >= 2.0)
!     2. LIQUID AND SOLID PHASES ARE POSSIBLE
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr3a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: ggcl
  DOUBLE PRECISION :: ggno3
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv3
  LOGICAL :: psconv4
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi3o
  DOUBLE PRECISION :: psi4o
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root2
  DOUBLE PRECISION :: root2a
  DOUBLE PRECISION :: root2b
  DOUBLE PRECISION :: root3
  DOUBLE PRECISION :: root3a
  DOUBLE PRECISION :: root3b
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  psconv1 = .TRUE.
  psconv3 = .TRUE.
  psconv4 = .TRUE.
  psi1o = -great
  psi3o = -great
  psi4o = -great
  root1 = zero
  root2 = zero
  root3 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcr1a
  !
  chi1 = cna2so4 ! SALTS
  chi4 = cnh4cl
  chi3 = cnacl
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = waer(2)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  molal(1) = zero
  molal(2) = nai
  molal(3) = nh4i
  molal(4) = cli
  molal(5) = so4i
  molal(6) = hso4i
  molal(7) = no3i
  !
  CALL calcact ! CALCULATE ACTIVITY COEFFICIENTS
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a5 = xk5*(water/gama(2))**3. ! Na2SO4 <==> Na+
    a8 = xk8*(water/gama(1))**2. ! NaCl   <==> Na+
    a14 = xk14*(water/gama(6))**2. ! NH4Cl  <==> NH4+
    akw = xkw*rh*water*water ! H2O    <==> H+
    !
    ! AMMONIUM CHLORIDE
    !
    IF (nh4i*cli>a14) THEN
      bb = -(waer(3)+waer(5)-root3)
      cc = -a14 + nh4i*(waer(5)-root3)
      dd = max(bb*bb-4.D0*cc, zero)
      root2a = 0.5D0*(-bb+sqrt(dd))
      root2b = 0.5D0*(-bb-sqrt(dd))
      IF (zero<=root2a) THEN
        root2 = root2a
      ELSE
        root2 = root2b
      END IF
      root2 = min(max(zero,root2), max(waer(5)-root3,zero), chi4, waer(3))
      psi4 = chi4 - root2
    END IF
    psconv4 = abs(psi4-psi4o) <= eps*psi4o
    psi4o = psi4
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a5) THEN
      bb = -(chi1+waer(1)-root3)
      cc = 0.25D0*(waer(1)-root3)*(4.D0*chi1+waer(1)-root3)
      dd = -0.25D0*(chi1*(waer(1)-root3)**2.D0-a5)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = min(max(root1,zero), max(waer(1)-root3,zero), chi1, waer(2))
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! ION CONCENTRATIONS
    !
    nai = waer(1) - (2.D0*root1+root3)
    so4i = waer(2) - root1
    nh4i = waer(3) - root2
    cli = waer(5) - (root3+root2)
    no3i = waer(4)
    !
    ! SODIUM CHLORIDE  ; To obtain new value for ROOT3
    !
    IF (nai*cli>a8) THEN
      bb = -((chi1-2.D0*root1)+(waer(5)-root2))
      cc = (chi1-2.D0*root1)*(waer(5)-root2) - a8
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root3a = 0.5D0*(-bb-sqrt(dd))
      root3b = 0.5D0*(-bb+sqrt(dd))
      IF (zero<=root3a) THEN
        root3 = root3a
      ELSE
        root3 = root3b
      END IF
      root3 = min(max(root3,zero), chi3)
      psi3 = chi3 - root3
    END IF
    psconv3 = abs(psi3-psi3o) <= eps*psi3o
    psi3o = psi3
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
    ELSE
      ggno3 = max(2.D0*so4i+no3i-nai-nh4i, zero)
      ggcl = max(gg-ggno3, zero)
      IF (ggcl>tiny) CALL calcclaq2(ggcl, cli, hi, claq) ! HCl
      IF (ggno3>tiny) THEN
        IF (ggcl<=tiny) hi = zero
        CALL calcniaq2(ggno3, no3i, hi, no3aq) ! HNO3
      END IF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv1 .AND. psconv3 .AND. psconv4) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCR3A')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 IF (cli<=tiny .AND. waer(5)>tiny) THEN !No disslv Cl-;solid only
    DO i = 1, nions
      molal(i) = zero
    END DO
    DO i = 1, ngasaq
      gasaq(i) = zero
    END DO
    CALL calcr1a
  ELSE
    a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
    a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
    a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
    !
    gnh3 = nh4i/hi/a2
    ghno3 = hi*no3i/a3
    ghcl = hi*cli/a4
    !
    gasaq(1) = nh3aq
    gasaq(2) = claq
    gasaq(3) = no3aq
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnh4cl = chi4 - psi4
    cnacl = chi3 - psi3
    cnano3 = zero
    cna2so4 = chi1 - psi1
  END IF
  !
  RETURN
  !
END SUBROUTINE calcr3a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR2
! *** CASE R2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, NA2SO4, NANO3, NACL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: exac
  LOGICAL :: exan
  LOGICAL :: exsc
  LOGICAL :: exsn
  !
  ! *** SOLVE FOR DRY CASE AND SEE WHICH SOLIDS ARE POSSIBLE **************
  !
  scase = 'R2 ; SUBCASE 2'
  CALL calcr1a ! SOLID
  scase = 'R2 ; SUBCASE 2'
  !
  exan = cnh4no3 > tiny ! NH4NO3
  exac = cnh4cl > tiny ! NH4CL
  exsn = cnano3 > tiny ! NANO3
  exsc = cnacl > tiny ! NACL
  !
  ! *** REGIME DEPENDS ON RELATIVE HUMIDITY AND POSSIBLE SPECIES **********
  !
  IF (exan) THEN ! *** NH4NO3 EXISTS
    IF (rh>=drmh1) THEN
      scase = 'R2 ; SUBCASE 1'
      CALL calcr2a
      scase = 'R2 ; SUBCASE 1'
    END IF
    !
  ELSE IF (.NOT. exan) THEN ! *** NH4NO3 = 0
    IF (exac .AND. exsn .AND. exsc) THEN
      IF (rh>=drmh2) THEN
        scase = 'R2 ; SUBCASE 3'
        CALL calcmdrp(rh, drmh2, drnano3, calcr1a, calcr3a)
        scase = 'R2 ; SUBCASE 3'
      END IF

    ELSE IF (.NOT. exac .AND. exsn .AND. exsc) THEN
      IF (rh>=drmr1) THEN
        scase = 'R2 ; SUBCASE 4'
        CALL calcmdrp(rh, drmr1, drnano3, calcr1a, calcr3a)
        scase = 'R2 ; SUBCASE 4'
      END IF

    ELSE IF (.NOT. exac .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr2) THEN
        scase = 'R2 ; SUBCASE 5'
        CALL calcmdrp(rh, drmr2, drnacl, calcr1a, calcr4a)
        scase = 'R2 ; SUBCASE 5'
      END IF

    ELSE IF (.NOT. exac .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr3) THEN
        scase = 'R2 ; SUBCASE 6'
        CALL calcmdrp(rh, drmr3, drnano3, calcr1a, calcr3a)
        scase = 'R2 ; SUBCASE 6'
      END IF

    ELSE IF (exac .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr4) THEN
        scase = 'R2 ; SUBCASE 7'
        CALL calcmdrp(rh, drmr4, drnacl, calcr1a, calcr4a)
        scase = 'R2 ; SUBCASE 7'
      END IF

    ELSE IF (exac .AND. .NOT. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr5) THEN
        scase = 'R2 ; SUBCASE 8'
        CALL calcmdrp(rh, drmr5, drnh4cl, calcr1a, calcr5)
        scase = 'R2 ; SUBCASE 8'
      END IF

    ELSE IF (exac .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr6) THEN
        scase = 'R2 ; SUBCASE 9'
        CALL calcmdrp(rh, drmr6, drnano3, calcr1a, calcr3a)
        scase = 'R2 ; SUBCASE 9'
      END IF
    END IF
    !
  END IF
  !
  RETURN
  !
END SUBROUTINE calcr2


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR2A
! *** CASE R2A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); SODIUM RICH (SODRAT >= 2.0)
!     2. LIQUID AND SOLID PHASES ARE POSSIBLE
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: ggcl
  DOUBLE PRECISION :: ggno3
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv2
  LOGICAL :: psconv3
  LOGICAL :: psconv4
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi2o
  DOUBLE PRECISION :: psi3o
  DOUBLE PRECISION :: psi4o
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root2
  DOUBLE PRECISION :: root2a
  DOUBLE PRECISION :: root2b
  DOUBLE PRECISION :: root3
  DOUBLE PRECISION :: root3a
  DOUBLE PRECISION :: root3b
  DOUBLE PRECISION :: root4
  DOUBLE PRECISION :: root4a
  DOUBLE PRECISION :: root4b
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv1 = .TRUE.
  psconv2 = .TRUE.
  psconv3 = .TRUE.
  psconv4 = .TRUE.
  !
  psi1o = -great
  psi2o = -great
  psi3o = -great
  psi4o = -great
  !
  root1 = zero
  root2 = zero
  root3 = zero
  root4 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcr1a
  !
  chi1 = cna2so4 ! SALTS
  chi2 = cnano3
  chi3 = cnacl
  chi4 = cnh4cl
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = waer(2)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  molal(1) = zero
  molal(2) = nai
  molal(3) = nh4i
  molal(4) = cli
  molal(5) = so4i
  molal(6) = hso4i
  molal(7) = no3i
  !
  CALL calcact ! CALCULATE ACTIVITY COEFFICIENTS
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a5 = xk5*(water/gama(2))**3. ! Na2SO4 <==> Na+
    a8 = xk8*(water/gama(1))**2. ! NaCl   <==> Na+
    a9 = xk9*(water/gama(3))**2. ! NaNO3  <==> Na+
    a14 = xk14*(water/gama(6))**2. ! NH4Cl  <==> NH4+
    akw = xkw*rh*water*water ! H2O    <==> H+
    !
    ! AMMONIUM CHLORIDE
    !
    IF (nh4i*cli>a14) THEN
      bb = -(waer(3)+waer(5)-root3)
      cc = nh4i*(waer(5)-root3) - a14
      dd = max(bb*bb-4.D0*cc, zero)
      dd = sqrt(dd)
      root2a = 0.5D0*(-bb+dd)
      root2b = 0.5D0*(-bb-dd)
      IF (zero<=root2a) THEN
        root2 = root2a
      ELSE
        root2 = root2b
      END IF
      root2 = min(max(root2,zero), chi4)
      psi4 = chi4 - root2
    END IF
    psconv4 = abs(psi4-psi4o) <= eps*psi4o
    psi4o = psi4
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a5) THEN
      bb = -(waer(2)+waer(1)-root3-root4)
      cc = waer(1)*(2.D0*root3+2.D0*root4-4.D0*waer(2)-one) -              &
           (root3+root4)**2.0 + 4.D0*waer(2)*(root3+root4)
      cc = -0.25*cc
      dd = waer(1)*waer(2)*(one-2.D0*root3-2.D0*root4) +                   &
           waer(2)*(root3+root4)**2.0 - a5
      dd = -0.25*dd
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = min(max(root1,zero), chi1)
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! SODIUM NITRATE
    !
    IF (nai*no3i>a9) THEN
      bb = -(waer(4)+waer(1)-2.D0*root1-root3)
      cc = waer(4)*(waer(1)-2.D0*root1-root3) - a9
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root4a = 0.5D0*(-bb-dd)
      root4b = 0.5D0*(-bb+dd)
      IF (zero<=root4a) THEN
        root4 = root4a
      ELSE
        root4 = root4b
      END IF
      root4 = min(max(root4,zero), chi2)
      psi2 = chi2 - root4
    END IF
    psconv2 = abs(psi2-psi2o) <= eps*psi2o
    psi2o = psi2
    !
    ! ION CONCENTRATIONS
    !
    nai = waer(1) - (2.D0*root1+root3+root4)
    so4i = waer(2) - root1
    nh4i = waer(3) - root2
    no3i = waer(4) - root4
    cli = waer(5) - (root3+root2)
    !
    ! SODIUM CHLORIDE  ; To obtain new value for ROOT3
    !
    IF (nai*cli>a8) THEN
      bb = -(waer(1)-2.D0*root1+waer(5)-root2-root4)
      cc = (waer(5)+root2)*(waer(1)-2.D0*root1-root4) - a8
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root3a = 0.5D0*(-bb-dd)
      root3b = 0.5D0*(-bb+dd)
      IF (zero<=root3a) THEN
        root3 = root3a
      ELSE
        root3 = root3b
      END IF
      root3 = min(max(root3,zero), chi3)
      psi3 = chi3 - root3
    END IF
    psconv3 = abs(psi3-psi3o) <= eps*psi3o
    psi3o = psi3
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi<ohi) THEN
      CALL calcamaq2(-gg, nh4i, ohi, nh3aq)
      hi = akw/ohi
    ELSE
      ggno3 = max(2.D0*so4i+no3i-nai-nh4i, zero)
      ggcl = max(gg-ggno3, zero)
      IF (ggcl>tiny) CALL calcclaq2(ggcl, cli, hi, claq) ! HCl
      IF (ggno3>tiny) THEN
        IF (ggcl<=tiny) hi = zero
        CALL calcniaq2(ggno3, no3i, hi, no3aq) ! HNO3
      END IF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
      so4i = so4i - del
      hi = hi - del
      hso4i = del
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv1 .AND. psconv2 .AND. psconv3 .AND. psconv4) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCR2A')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 IF (cli<=tiny .AND. waer(5)>tiny) THEN !No disslv Cl-;solid only
    DO i = 1, nions
      molal(i) = zero
    END DO
    DO i = 1, ngasaq
      gasaq(i) = zero
    END DO
    CALL calcr1a
  ELSE ! OK, aqueous phase present
    a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
    a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
    a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
    !
    gnh3 = nh4i/hi/a2
    ghno3 = hi*no3i/a3
    ghcl = hi*cli/a4
    !
    gasaq(1) = nh3aq
    gasaq(2) = claq
    gasaq(3) = no3aq
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnh4cl = chi4 - psi4
    cnacl = chi3 - psi3
    cnano3 = chi2 - psi2
    cna2so4 = chi1 - psi1
  END IF
  !
  RETURN
  !
END SUBROUTINE calcr2a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR1
! *** CASE R1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, NA2SO4, NANO3, NACL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: exac
  LOGICAL :: exan
  LOGICAL :: exsc
  LOGICAL :: exsn
  !
  ! *** SOLVE FOR DRY CASE AND SEE WHICH SOLIDS ARE POSSIBLE **************
  !
  scase = 'R1 ; SUBCASE 1'
  CALL calcr1a ! SOLID
  scase = 'R1 ; SUBCASE 1'
  !
  exan = cnh4no3 > tiny ! NH4NO3
  exac = cnh4cl > tiny ! NH4CL
  exsn = cnano3 > tiny ! NANO3
  exsc = cnacl > tiny ! NACL
  !
  ! *** REGIME DEPENDS ON RELATIVE HUMIDITY AND POSSIBLE SPECIES **********
  !
  IF (exan .AND. exac .AND. exsc .AND. exsn) THEN ! *** ALL EXIST
    IF (rh>=drmh1) THEN
      scase = 'R1 ; SUBCASE 2' ! MDRH
      CALL calcmdrp(rh, drmh1, drnh4no3, calcr1a, calcr2a)
      scase = 'R1 ; SUBCASE 2'
    END IF
    !
  ELSE IF (.NOT. exan) THEN ! *** NH4NO3 = 0
    IF (exac .AND. exsn .AND. exsc) THEN
      IF (rh>=drmh2) THEN
        scase = 'R1 ; SUBCASE 3'
        CALL calcmdrp(rh, drmh2, drnano3, calcr1a, calcr3a)
        scase = 'R1 ; SUBCASE 3'
      END IF

    ELSE IF (.NOT. exac .AND. exsn .AND. exsc) THEN
      IF (rh>=drmr1) THEN
        scase = 'R1 ; SUBCASE 4'
        CALL calcmdrp(rh, drmr1, drnano3, calcr1a, calcr3a)
        scase = 'R1 ; SUBCASE 4'
      END IF

    ELSE IF (.NOT. exac .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr2) THEN
        scase = 'R1 ; SUBCASE 5'
        CALL calcmdrp(rh, drmr2, drnacl, calcr1a, calcr3a) !, CALCR4A)
        scase = 'R1 ; SUBCASE 5'
      END IF

    ELSE IF (.NOT. exac .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr3) THEN
        scase = 'R1 ; SUBCASE 6'
        CALL calcmdrp(rh, drmr3, drnano3, calcr1a, calcr3a)
        scase = 'R1 ; SUBCASE 6'
      END IF

    ELSE IF (exac .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr4) THEN
        scase = 'R1 ; SUBCASE 7'
        CALL calcmdrp(rh, drmr4, drnacl, calcr1a, calcr3a) !, CALCR4A)
        scase = 'R1 ; SUBCASE 7'
      END IF

    ELSE IF (exac .AND. .NOT. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr5) THEN
        scase = 'R1 ; SUBCASE 8'
        CALL calcmdrp(rh, drmr5, drnh4cl, calcr1a, calcr3a) !, CALCR5)
        scase = 'R1 ; SUBCASE 8'
      END IF

    ELSE IF (exac .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr6) THEN
        scase = 'R1 ; SUBCASE 9'
        CALL calcmdrp(rh, drmr6, drnano3, calcr1a, calcr3a)
        scase = 'R1 ; SUBCASE 9'
      END IF
    END IF
    !
  ELSE IF (.NOT. exac) THEN ! *** NH4CL  = 0
    IF (exan .AND. exsn .AND. exsc) THEN
      IF (rh>=drmr7) THEN
        scase = 'R1 ; SUBCASE 10'
        CALL calcmdrp(rh, drmr7, drnh4no3, calcr1a, calcr2a)
        scase = 'R1 ; SUBCASE 10'
      END IF

    ELSE IF (exan .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr8) THEN
        scase = 'R1 ; SUBCASE 11'
        CALL calcmdrp(rh, drmr8, drnh4no3, calcr1a, calcr2a)
        scase = 'R1 ; SUBCASE 11'
      END IF

    ELSE IF (exan .AND. .NOT. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr9) THEN
        scase = 'R1 ; SUBCASE 12'
        CALL calcmdrp(rh, drmr9, drnh4no3, calcr1a, calcr2a)
        scase = 'R1 ; SUBCASE 12'
      END IF

    ELSE IF (exan .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr10) THEN
        scase = 'R1 ; SUBCASE 13'
        CALL calcmdrp(rh, drmr10, drnh4no3, calcr1a, calcr2a)
        scase = 'R1 ; SUBCASE 13'
      END IF
    END IF
    !
  ELSE IF (.NOT. exsn) THEN ! *** NANO3  = 0
    IF (exan .AND. exac .AND. exsc) THEN
      IF (rh>=drmr11) THEN
        scase = 'R1 ; SUBCASE 14'
        CALL calcmdrp(rh, drmr11, drnh4no3, calcr1a, calcr2a)
        scase = 'R1 ; SUBCASE 14'
      END IF

    ELSE IF (exan .AND. exac .AND. .NOT. exsc) THEN
      IF (rh>=drmr12) THEN
        scase = 'R1 ; SUBCASE 15'
        CALL calcmdrp(rh, drmr12, drnh4no3, calcr1a, calcr2a)
        scase = 'R1 ; SUBCASE 15'
      END IF
    END IF
    !
  ELSE IF (.NOT. exsc) THEN ! *** NACL   = 0
    IF (exan .AND. exac .AND. exsn) THEN
      IF (rh>=drmr13) THEN
        scase = 'R1 ; SUBCASE 16'
        CALL calcmdrp(rh, drmr13, drnh4no3, calcr1a, calcr2a)
        scase = 'R1 ; SUBCASE 16'
      END IF
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcr1


!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCR1A
! *** CASE R1 ; SUBCASE 1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; SODIUM RICH (SODRAT >= 2.0)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : NH4NO3, NH4CL, NANO3, NA2SO4, NANO3, NACL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcr1a (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frnh3
  DOUBLE PRECISION :: frno3
  !
  ! *** CALCULATE SOLIDS **************************************************
  !
  cna2so4 = waer(2)
  frna = max(waer(1)-2*cna2so4, zero)
  !
  cnh42s4 = zero
  !
  cnano3 = min(frna, waer(4))
  frno3 = max(waer(4)-cnano3, zero)
  frna = max(frna-cnano3, zero)
  !
  cnacl = min(frna, waer(5))
  frcl = max(waer(5)-cnacl, zero)
  frna = max(frna-cnacl, zero)
  !
  cnh4no3 = min(frno3, waer(3))
  frno3 = max(frno3-cnh4no3, zero)
  frnh3 = max(waer(3)-cnh4no3, zero)
  !
  cnh4cl = min(frcl, frnh3)
  frcl = max(frcl-cnh4cl, zero)
  frnh3 = max(frnh3-cnh4cl, zero)
  !
  ! *** OTHER PHASES ******************************************************
  !
  water = zero
  !
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  RETURN
  !
END SUBROUTINE calcr1a
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCV7
! *** CASE V7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv7 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcv1a
  !
  chi9 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! ION CONCENTRATIONS
    !
    nai = waer(1)
    so4i = max(waer(2)-waer(6), zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    ki = waer(7)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF

    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCV7')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cna2so4 = zero
  cmgso4 = zero
  ck2so4 = zero
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcv7
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCV6
! *** CASE V6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4, MGSO4, NA2SO4
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv6 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .TRUE.
  psi70 = -great ! GREAT = 1.D10
  root7 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcv1a
  !
  chi9 = ccaso4
  chi7 = ck2so4 ! SALTS
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7))
      cc = waer(7)*(waer(2)-waer(6)) + 0.25D0*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*waer(2)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = min(root7, waer(7)/2.0, max(waer(2)-waer(6),zero), chi7)
      root7 = max(root7, zero)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    so4i = max(waer(2)-waer(6)-root7, zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    nai = waer(1)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCV6')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cna2so4 = zero
  cmgso4 = zero
  ck2so4 = chi7 - psi7
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcv6
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCV5
! *** CASE V5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4, MGSO4
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv5 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .TRUE.
  psconv1 = .TRUE.
  !
  psi70 = -great ! GREAT = 1.D10
  psi1o = -great
  !
  root7 = zero
  root1 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcv1a
  !
  chi9 = ccaso4
  chi7 = ck2so4 ! SALTS
  chi1 = cna2so4
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a1 = xk5*(water/gama(2))**3.0 ! NA2S04    <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root1)
      cc = waer(7)*((waer(2)-waer(6))-root1) + 0.25*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*((waer(2)-waer(6))-root1)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max(waer(2)-waer(6)-root1,zero),     &
           chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a1) THEN
      bb = -((waer(2)-waer(6))+waer(1)-root7)
      cc = waer(1)*((waer(2)-waer(6))-root7) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*((waer(2)-waer(6))-root7)-a1)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = max(root1, zero)
      root1 = min(root1, waer(1)/2.0, max((waer(2)-                        &
           waer(6))-root7,zero), chi1)
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    nai = max(waer(1)-2.D0*root1, zero)
    so4i = max((waer(2)-waer(6))-root7-root1, zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7 .AND. psconv1) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCV5')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cna2so4 = chi1 - psi1
  cmgso4 = zero
  ck2so4 = chi7 - psi7
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcv5

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCV4
! *** CASE V4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4
!     4. Completely dissolved: NH4NO3, NH4CL, (NH4)2SO4
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv4 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .TRUE.
  psconv1 = .TRUE.
  !
  psi70 = -great ! GREAT = 1.D10
  psi1o = -great
  !
  root7 = zero
  root1 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcv1a
  !
  chi9 = ccaso4
  chi7 = ck2so4 ! SALTS
  chi1 = cna2so4
  chi8 = cmgso4
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = waer(2)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a1 = xk5*(water/gama(2))**3.0 ! NA2S04    <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root1)
      cc = waer(7)*((waer(2)-waer(6))-root1) + 0.25*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*((waer(2)-waer(6))-root1)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max((waer(2)-                        &
           waer(6))-root1,zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a1) THEN
      bb = -((waer(2)-waer(6))+waer(1)-root7)
      cc = waer(1)*((waer(2)-waer(6))-root7) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*((waer(2)-waer(6))-root7)-a1)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = max(root1, zero)
      root1 = min(root1, waer(1)/2.0, max((waer(2)-                        &
           waer(6))-root7,zero), chi1)
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    nai = max(waer(1)-2.D0*root1, zero)
    so4i = max((waer(2)-waer(6))-root7-root1, zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7 .AND. psconv1) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCV4')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cna2so4 = chi1 - psi1
  cmgso4 = zero
  ck2so4 = chi7 - psi7
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcv4
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCV3
! *** CASE V3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, (NH4)2SO4
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: excl
  LOGICAL :: exno
  !
  ! *** REGIME DEPENDS ON AMBIENT RELATIVE HUMIDITY & POSSIBLE SPECIES ***
  !
  exno = waer(4) > tiny
  excl = waer(5) > tiny
  !
  IF (exno .OR. excl) THEN ! *** NITRATE OR CHLORIDE EXISTS
    scase = 'V3 ; SUBCASE 1'
    CALL calcv3a
    scase = 'V3 ; SUBCASE 1'
    !
  ELSE ! *** NO CHLORIDE AND NITRATE
    IF (rh<drmo3) THEN
      scase = 'V3 ; SUBCASE 2'
      CALL calcv1a ! SOLID
      scase = 'V3 ; SUBCASE 2'
    ELSE
      scase = 'V3 ; SUBCASE 3' ! MDRH (CaSO4, (NH4)2SO4, MGSO4, NA2SO4, K2SO4)
      CALL calcmdrpii(rh, drmo3, drnh42s4, calcv1a, calcv4)
      scase = 'V3 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcv3
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCV3A
! *** CASE V3A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, (NH4)2SO4
!     4. Completely dissolved: NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv3a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv6
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi60
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root6
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .TRUE.
  psconv1 = .TRUE.
  psconv6 = .TRUE.
  !
  psi70 = -great ! GREAT = 1.D10
  psi1o = -great
  psi60 = -great
  !
  root7 = zero
  root1 = zero
  root6 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcv1a
  !
  chi9 = ccaso4
  chi7 = ck2so4 ! SALTS
  chi1 = cna2so4
  chi8 = cmgso4
  chi6 = cnh42s4
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = waer(2)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a1 = xk5*(water/gama(2))**3.0 ! NA2S04    <==> Na+
    a6 = xk7*(water/gama(4))**3.0 !(NH4)2SO4  <==> NH4+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root1-root6)
      cc = waer(7)*((waer(2)-waer(6))-root1-root6) + 0.25*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*((waer(2)-waer(6))-root1-root6)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max(waer(2)-waer(                    &
           6)-root1-root6,zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a1) THEN
      bb = -((waer(2)-waer(6))+waer(1)-root7-root6)
      cc = waer(1)*((waer(2)-waer(6))-root7-root6) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*((waer(2)-waer(6))-root7-root6)-a1)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = max(root1, zero)
      root1 = min(root1, waer(1)/2.0, max(waer(2)-waer(                    &
           6)-root7-root6,zero), chi1)
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! AMMONIUM SULFATE
    !
    IF (nh4i*nh4i*so4i>a6) THEN
      bb = -((waer(2)-waer(6))+waer(3)-root7-root1)
      cc = waer(3)*((waer(2)-waer(6))-root7-root1) + 0.25*waer(3)*waer(3)
      dd = -0.25*(waer(3)*waer(3)*((waer(2)-waer(6))-root7-root1)-a6)
      CALL poly3(bb, cc, dd, root6, islv)
      IF (islv/=0) root6 = tiny
      root6 = max(root6, zero)
      root6 = min(root6, waer(3)/2.0, max(waer(2)-waer(                    &
           6)-root7-root1,zero), chi6)
      psi6 = chi6 - root6
    END IF
    psconv6 = abs(psi6-psi60) <= eps*psi60
    psi60 = psi6
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    nai = max(waer(1)-2.D0*root1, zero)
    so4i = max(waer(2)-waer(6)-root7-root1-root6, zero)
    nh4i = max(waer(3)-2.D0*root6, zero)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7 .AND. psconv1 .AND. psconv6) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCV3')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = chi6 - psi6
  cnh4no3 = zero
  cnh4cl = zero
  cna2so4 = chi1 - psi1
  cmgso4 = zero
  ck2so4 = chi7 - psi7
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcv3a
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCV2
! *** CASE V2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, (NH4)2SO4, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: excl
  LOGICAL :: exno
  !
  ! *** REGIME DEPENDS ON AMBIENT RELATIVE HUMIDITY & POSSIBLE SPECIES ***
  !
  exno = waer(4) > tiny
  excl = waer(5) > tiny
  !
  IF (exno) THEN ! *** NITRATE EXISTS
    scase = 'V2 ; SUBCASE 1'
    CALL calcv2a
    scase = 'V2 ; SUBCASE 1'
    !
  ELSE IF (.NOT. exno .AND. excl) THEN ! *** ONLY CHLORIDE EXISTS
    IF (rh<drmo2) THEN
      scase = 'V2 ; SUBCASE 2'
      CALL calcv1a ! SOLID
      scase = 'V2 ; SUBCASE 2'
    ELSE
      scase = 'V2 ; SUBCASE 3' ! MDRH CaSO4, NH4CL, (NH4)2SO4, MGSO4, NA2SO4, K2SO4
      CALL calcmdrpii(rh, drmo2, drnh4cl, calcv1a, calcv3a)
      scase = 'V2 ; SUBCASE 3'
    END IF
    !
  ELSE ! *** NO CHLORIDE AND NITRATE
    IF (rh<drmo3) THEN
      scase = 'V2 ; SUBCASE 2'
      CALL calcv1a ! SOLID
      scase = 'V2 ; SUBCASE 2'
    ELSE
      scase = 'V2 ; SUBCASE 4' ! MDRH CaSO4, (NH4)2SO4, MGSO4, NA2SO4, K2SO4
      CALL calcmdrpii(rh, drmo3, drnh42s4, calcv1a, calcv4)
      scase = 'V2 ; SUBCASE 4'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcv2
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCV2A
! *** CASE V2A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, (NH4)2SO4, NH4CL
!     4. Completely dissolved: NH4NO3
!
! *** COPYRIGHT 1996-2008 UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv4
  LOGICAL :: psconv6
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi40
  DOUBLE PRECISION :: psi60
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root4
  DOUBLE PRECISION :: root4a
  DOUBLE PRECISION :: root4b
  DOUBLE PRECISION :: root6
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .TRUE.
  psconv1 = .TRUE.
  psconv6 = .TRUE.
  psconv4 = .TRUE.
  !
  psi70 = -great ! GREAT = 1.D10
  psi1o = -great
  psi60 = -great
  psi40 = -great
  !
  root7 = zero
  root1 = zero
  root6 = zero
  root4 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcv1a
  !
  chi9 = ccaso4
  chi7 = ck2so4 ! SALTS
  chi1 = cna2so4
  chi8 = cmgso4
  chi6 = cnh42s4
  chi4 = cnh4cl
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi6 = cnh42s4
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a1 = xk5*(water/gama(2))**3.0 ! NA2S04    <==> Na+
    a6 = xk7*(water/gama(4))**3.0 ! (NH4)2SO4 <==> NH4+
    a14 = xk14*(water/gama(6))**2. ! NH4Cl     <==> NH4+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! AMMONIUM CHLORIDE
    !
    IF (nh4i*cli>a14) THEN
      bb = -(waer(3)+waer(5)-2.D0*root6)
      cc = waer(5)*(waer(3)-2.D0*root6) - a14
      dd = bb*bb - 4.D0*cc
      IF (dd<zero) THEN
        root4 = zero
      ELSE
        dd = sqrt(dd)
        root4a = 0.5D0*(-bb+dd)
        root4b = 0.5D0*(-bb-dd)
        IF (zero<=root4a) THEN
          root4 = root4a
        ELSE
          root4 = root4b
        END IF
        root4 = max(root4, zero)
        root4 = min(root4, waer(5), max(waer(3)-2.D0*root6,zero), chi4)
        psi4 = chi4 - root4
      END IF
    END IF
    psconv4 = abs(psi4-psi40) <= eps*psi40
    psi40 = psi4
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root1-root6)
      cc = waer(7)*((waer(2)-waer(6))-root1-root6) + 0.25*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*((waer(2)-waer(6))-root1-root6)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max(waer(2)-waer(                    &
           6)-root1-root6,zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a1) THEN
      bb = -((waer(2)-waer(6))+waer(1)-root7-root6)
      cc = waer(1)*((waer(2)-waer(6))-root7-root6) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*((waer(2)-waer(6))-root7-root6)-a1)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = max(root1, zero)
      root1 = min(root1, waer(1)/2.0, max(waer(2)-waer(                    &
           6)-root7-root6,zero), chi1)
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! AMMONIUM SULFATE
    !
    IF (nh4i*nh4i*so4i>a6) THEN
      bb = -((waer(2)-waer(6))+waer(3)-root7-root1-root4)
      cc = waer(3)*((waer(2)-waer(6))-root7-root1) +                       &
           0.25*(waer(3)-root4)**2.0 + root4*(root1+root7-(waer(2)-waer(6)))
      dd = -0.25*((waer(3)-root4)**2.0*((waer(2)-waer(6))-root7-root1)-a6)
      CALL poly3(bb, cc, dd, root6, islv)
      IF (islv/=0) root6 = tiny
      root6 = max(root6, zero)
      root6 = min(root6, waer(3)/2.0, max(waer(2)-waer(                    &
           6)-root7-root1,zero), chi6)
      psi6 = chi6 - root6
    END IF
    psconv6 = abs(psi6-psi60) <= eps*psi60
    psi60 = psi6
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    nai = max(waer(1)-2.D0*root1, zero)
    so4i = max(waer(2)-waer(6)-root7-root1-root6, zero)
    nh4i = max(waer(3)-2.D0*root6, zero)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7 .AND. psconv1 .AND. psconv6 .AND. psconv4) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCV2')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = chi6 - psi6
  cnh4no3 = zero
  cnh4cl = chi4 - psi4
  cna2so4 = chi1 - psi1
  cmgso4 = zero
  ck2so4 = chi7 - psi7
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcv2a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCV1
! *** CASE V1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3, NH4Cl, NA2SO4, K2SO4, MGSO4, CASO4
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: excl
  LOGICAL :: exno
  !
  ! *** REGIME DEPENDS ON AMBIENT RELATIVE HUMIDITY & POSSIBLE SPECIES ***
  !
  exno = waer(4) > tiny
  excl = waer(5) > tiny
  !
  IF (exno .AND. excl) THEN ! *** NITRATE & CHLORIDE EXIST
    IF (rh<drmo1) THEN
      scase = 'V1 ; SUBCASE 1'
      CALL calcv1a ! SOLID
      scase = 'V1 ; SUBCASE 1'
    ELSE
      scase = 'V1 ; SUBCASE 2' ! MDRH (NH4)2SO4, NH4NO3, NH4Cl, NA2SO4, K2SO4, MGSO4, CASO4
      CALL calcmdrpii(rh, drmo1, drnh4no3, calcv1a, calcv2a)
      scase = 'V1 ; SUBCASE 2'
    END IF
    !
  ELSE IF (exno .AND. .NOT. excl) THEN ! *** ONLY NITRATE EXISTS
    IF (rh<drmv1) THEN
      scase = 'V1 ; SUBCASE 1'
      CALL calcv1a ! SOLID
      scase = 'V1 ; SUBCASE 1'
    ELSE
      scase = 'V1 ; SUBCASE 3' ! MDRH (NH4)2SO4, NH4NO3, NA2SO4, K2SO4, MGSO4, CASO4
      CALL calcmdrpii(rh, drmv1, drnh4no3, calcv1a, calcv2a)
      scase = 'V1 ; SUBCASE 3'
    END IF
    !
  ELSE IF (.NOT. exno .AND. excl) THEN ! *** ONLY CHLORIDE EXISTS
    IF (rh<drmo2) THEN
      scase = 'V1 ; SUBCASE 1'
      CALL calcv1a ! SOLID
      scase = 'V1 ; SUBCASE 1'
    ELSE
      scase = 'V1 ; SUBCASE 4' ! MDRH (NH4)2SO4, NH4Cl, NA2SO4, K2SO4, MGSO4, CASO4
      CALL calcmdrpii(rh, drmo2, drnh4cl, calcv1a, calcv3a)
      scase = 'V1 ; SUBCASE 4'
    END IF
    !
  ELSE ! *** NO CHLORIDE AND NITRATE
    IF (rh<drmo3) THEN
      scase = 'V1 ; SUBCASE 1'
      CALL calcv1a ! SOLID
      scase = 'V1 ; SUBCASE 1'
    ELSE
      scase = 'V1 ; SUBCASE 5' ! MDRH (NH4)2SO4, NA2SO4, K2SO4, MGSO4, CASO4
      CALL calcmdrpii(rh, drmo3, drnh42s4, calcv1a, calcv4)
      scase = 'V1 ; SUBCASE 5'
    END IF
  END IF
  !
  RETURN
  !
  !      IF (RH.LT.DRMO1) THEN
  !         SCASE = 'V1 ; SUBCASE 1'
  !         CALL CALCV1A              ! SOLID PHASE ONLY POSSIBLE
  !         SCASE = 'V1 ; SUBCASE 1'
  !      ELSE
  !         SCASE = 'V1 ; SUBCASE 2'  ! LIQUID & SOLID PHASE POSSIBLE
  !         CALL CALCMDRPII (RH, DRMO1, DRNH4NO3, CALCV1A, CALCV2A)
  !         SCASE = 'V1 ; SUBCASE 2'
  !         ENDIF
  !
  !      RETURN
  !
END SUBROUTINE calcv1
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCV1A
! *** CASE V1A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : (NH4)2SO4, NH4NO3, NH4Cl, NA2SO4, K2SO4, MGSO4, CASO4
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcv1a (  )
  USE mo_isrpia  
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frnh3
  INTEGER :: nafr
  DOUBLE PRECISION :: so4fr
  !
  ! *** CALCULATE SOLIDS **************************************************
  !
  ccaso4 = min(waer(6), waer(2)) ! CCASO4
  so4fr = max(waer(2)-ccaso4, zero)
  cafr = max(waer(6)-ccaso4, zero)
  ck2so4 = min(0.5D0*waer(7), so4fr) ! CK2SO4
  frk = max(waer(7)-2.D0*ck2so4, zero)
  so4fr = max(so4fr-ck2so4, zero)
  cna2so4 = min(0.5D0*waer(1), so4fr) ! CNA2SO4
  nafr = max(waer(1)-2.D0*cna2so4, zero)
  so4fr = max(so4fr-cna2so4, zero)
  cmgso4 = min(waer(8), so4fr) ! CMGSO4
  frmg = max(waer(8)-cmgso4, zero)
  so4fr = max(so4fr-cmgso4, zero)
  cnh42s4 = max(min(so4fr,0.5D0*waer(3)), tiny)
  frnh3 = max(waer(3)-2.D0*cnh42s4, zero)
  !
  cnh4no3 = min(frnh3, waer(4))
  !CC      FRNO3   = MAX (WAER(4) - CNH4NO3, ZERO)
  frnh3 = max(frnh3-cnh4no3, zero)
  !
  cnh4cl = min(frnh3, waer(5))
  !CC      FRCL    = MAX (WAER(5) - CNH4CL, ZERO)
  frnh3 = max(frnh3-cnh4cl, zero)
  !
  ! *** OTHER PHASES ******************************************************
  !
  water = zero
  !
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  RETURN
  !
END SUBROUTINE calcv1a
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCU8
! *** CASE U8
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); CRUSTAL+SODIUM RICH (CRNARAT >= 2.0); CRUSTAL POOR (CRRAT<2)
!     2. THERE IS ONLY A LIQUID PHASE
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu8 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  CALL calcu1a
  !
  chi9 = ccaso4 ! SALTS
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  ! *** CALCULATE WATER **************************************************
  !
  CALL calcmr
  !
  ! *** SETUP LIQUID CONCENTRATIONS **************************************
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    akw = xkw*rh*water*water ! H2O    <==> H+
    !
    nai = waer(1)
    so4i = max(waer(2)-waer(6), zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    ki = waer(7)
    mgi = waer(8)

    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    IF (hi<=tiny) hi = sqrt(akw)
    !      OHI   = AKW/HI
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCU8')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cna2so4 = zero
  cmgso4 = zero
  ck2so4 = zero
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcu8
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU7
! *** CASE U7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), CRUSTAL+SODIUM RICH (CRNARAT >= 2.0); CRUSTAL POOR (CRRAT<2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL, MGSO4, NA2SO4
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu7 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .TRUE.
  psi70 = -great ! GREAT = 1.D10
  root7 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcu1a
  !
  chi7 = ck2so4 ! SALTS
  chi9 = ccaso4
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  molal(1) = zero
  molal(2) = nai
  molal(3) = nh4i
  molal(4) = cli
  molal(5) = so4i
  molal(6) = hso4i
  molal(7) = no3i
  molal(8) = cai
  molal(9) = ki
  molal(10) = mgi
  !
  CALL calcact ! CALCULATE ACTIVITY COEFFICIENTS
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7))
      cc = waer(7)*(waer(2)-waer(6)) + 0.25D0*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*(waer(2)-waer(6))-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max(waer(2)-waer(6),zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    so4i = max(waer(2)-waer(6)-root7, zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    nai = waer(1)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !      IF (HI.LE.TINY) HI = SQRT(AKW)
    !      OHI   = AKW/HI
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCU7')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cna2so4 = zero
  cmgso4 = zero
  ck2so4 = chi7 - psi7
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcu7
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU6
! *** CASE U6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL, MGSO4
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu6 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .TRUE.
  psconv1 = .TRUE.
  !
  psi70 = -great ! GREAT = 1.D10
  psi1o = -great
  !
  root7 = zero
  root1 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcu1a
  !
  chi1 = cna2so4 ! SALTS
  chi7 = ck2so4
  chi9 = ccaso4
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  molal(1) = zero
  molal(2) = nai
  molal(3) = nh4i
  molal(4) = cli
  molal(5) = so4i
  molal(6) = hso4i
  molal(7) = no3i
  molal(8) = cai
  molal(9) = ki
  molal(10) = mgi
  !
  CALL calcact ! CALCULATE ACTIVITY COEFFICIENTS
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a1 = xk5*(water/gama(2))**3.0 ! NA2S04    <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root1)
      cc = waer(7)*((waer(2)-waer(6))-root1) + 0.25*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*((waer(2)-waer(6))-root1)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max((waer(2)-                        &
           waer(6))-root1,zero), chi7)
      psi7 = chi7 - root7

    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a1) THEN
      bb = -((waer(2)-waer(6))+waer(1)-root7)
      cc = waer(1)*((waer(2)-waer(6))-root7) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*((waer(2)-waer(6))-root7)-a1)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = max(root1, zero)
      root1 = min(root1, waer(1)/2.0, max((waer(2)-                        &
           waer(6))-root7,zero), chi1)
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    nai = max(waer(1)-2.D0*root1, zero)
    so4i = max(waer(2)-waer(6)-root7-root1, zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !      IF (HI.LE.TINY) HI = SQRT(AKW)
    !      OHI   = AKW/HI
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7 .AND. psconv1) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCU6')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cna2so4 = chi1 - psi1
  cmgso4 = zero
  ck2so4 = chi7 - psi7
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcu6
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU5
! *** CASE U5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4
!     4. Completely dissolved: NH4NO3, NH4CL, NANO3, NACL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu5 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .TRUE.
  psconv1 = .TRUE.
  !
  psi70 = -great ! GREAT = 1.D10
  psi1o = -great
  !
  root7 = zero
  root1 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcu1a
  !
  chi1 = cna2so4 ! SALTS
  chi7 = ck2so4
  chi8 = cmgso4
  chi9 = ccaso4
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  molal(1) = zero
  molal(2) = nai
  molal(3) = nh4i
  molal(4) = cli
  molal(5) = so4i
  molal(6) = hso4i
  molal(7) = no3i
  molal(8) = cai
  molal(9) = ki
  molal(10) = mgi
  !
  CALL calcact ! CALCULATE ACTIVITY COEFFICIENTS
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a1 = xk5*(water/gama(2))**3.0 ! NA2S04    <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root1)
      cc = waer(7)*((waer(2)-waer(6))-root1) + 0.25*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*((waer(2)-waer(6))-root1)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max(waer(2)-waer(6)-root1,zero),     &
           chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a1) THEN
      bb = -((waer(2)-waer(6))+waer(1)-root7)
      cc = waer(1)*((waer(2)-waer(6))-root7) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*((waer(2)-waer(6))-root7)-a1)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = max(root1, zero)
      root1 = min(root1, waer(1)/2.0, max(waer(2)-waer(6)-root7,zero),     &
           chi1)
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    nai = max(waer(1)-2.D0*root1, zero)
    so4i = max(waer(2)-waer(6)-root7-root1, zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !      IF (HI.LE.TINY) HI = SQRT(AKW)
    !      OHI   = AKW/HI
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7 .AND. psconv1) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCU5')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cna2so4 = chi1 - psi1
  cmgso4 = zero
  ck2so4 = chi7 - psi7
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcu5
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU4
! *** CASE U4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), (DUST + SODIUM) RICH: R(Cr+Na)>2; DUST POOR: Rcr<2.
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu4 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: exac
  LOGICAL :: exan
  LOGICAL :: exsc
  LOGICAL :: exsn
  !
  ! *** SOLVE FOR DRY CASE AND SEE WHICH SOLIDS ARE POSSIBLE **************
  !
  scase = 'U4 ; SUBCASE 2'
  CALL calcu1a ! SOLID
  scase = 'U4 ; SUBCASE 2'
  !
  exan = cnh4no3 > tiny ! NH4NO3
  exac = cnh4cl > tiny ! NH4CL
  exsn = cnano3 > tiny ! NANO3
  exsc = cnacl > tiny ! NACL
  !
  ! *** REGIME DEPENDS ON RELATIVE HUMIDITY AND POSSIBLE SPECIES **********
  !
  IF (exan .OR. exsn .OR. exsc) THEN ! *** NH4NO3,NANO3 EXIST
    IF (rh>=drmm1) THEN
      scase = 'U4 ; SUBCASE 1'
      CALL calcu4a
      scase = 'U4 ; SUBCASE 1'
    END IF
    !
  ELSE IF (exac) THEN ! *** NH4CL EXISTS ONLY
    IF (rh>=drmr5) THEN
      scase = 'U4 ; SUBCASE 3'
      CALL calcmdrpii(rh, drmr5, drnh4cl, calcu1a, calcu5)
      scase = 'U4 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcu4
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU4A
! *** CASE U4A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, NH4CL
!     4. Completely dissolved: NH4NO3, NANO3, NACL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu4a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv4
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi40
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root4
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .FALSE.
  psconv1 = .FALSE.
  psconv4 = .FALSE.
  !
  psi70 = -great ! GREAT = 1.D10
  psi1o = -great
  psi40 = -great
  !
  root7 = zero
  root1 = zero
  root4 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcu1a
  !
  chi1 = cna2so4 ! SALTS
  chi4 = cnh4cl
  chi7 = ck2so4
  chi8 = cmgso4
  chi9 = ccaso4
  !
  psi1 = cna2so4
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  molal(1) = zero
  molal(2) = nai
  molal(3) = nh4i
  molal(4) = cli
  molal(5) = so4i
  molal(6) = hso4i
  molal(7) = no3i
  molal(8) = cai
  molal(9) = ki
  molal(10) = mgi
  !
  CALL calcact ! CALCULATE ACTIVITY COEFFICIENTS
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a1 = xk5*(water/gama(2))**3.0 ! NA2S04    <==> Na+
    a14 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root1)
      cc = waer(7)*((waer(2)-waer(6))-root1) + 0.25*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*((waer(2)-waer(6))-root1)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max(waer(2)-waer(6)-root1,zero),     &
           chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a1) THEN
      bb = -((waer(2)-waer(6))+waer(1)-root7)
      cc = waer(1)*((waer(2)-waer(6))-root7) + 0.25*waer(1)*waer(1)
      dd = -0.25*(waer(1)*waer(1)*((waer(2)-waer(6))-root7)-a1)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = max(root1, zero)
      root1 = min(root1, waer(1)/2.0, max(waer(2)-waer(6)-root7,zero),     &
           chi1)
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! AMMONIUM CHLORIDE
    !
    IF (nh4i*cli>a14) THEN
      bb = -(nh4i+cli)
      cc = -a14 + nh4i*cli
      dd = bb*bb - 4.D0*cc
      root4 = 0.5D0*(-bb-sqrt(dd))
      IF (root4>tiny) THEN
        root4 = min(max(root4,zero), chi4)
        psi4 = chi4 - root4
      END IF
    END IF
    psconv4 = abs(psi4-psi40) <= eps*psi40
    psi40 = psi4
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    nai = max(waer(1)-2.D0*root1, zero)
    so4i = max(waer(2)-waer(6)-root7-root1, zero)
    nh4i = max(waer(3)-root4, zero)
    no3i = waer(4)
    cli = max(waer(5)-root4, zero)
    cai = zero
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !      IF (HI.LE.TINY) HI = SQRT(AKW)
    !      OHI   = AKW/HI
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7 .AND. psconv1 .AND. psconv4) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCU4')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = chi4 - psi4
  cnacl = zero
  cnano3 = zero
  cna2so4 = chi1 - psi1
  cmgso4 = zero
  ck2so4 = chi7 - psi7
  ccaso4 = min(waer(6), waer(2))
  !
  RETURN
  !
END SUBROUTINE calcu4a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU3
! *** CASE U3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), (DUST + SODIUM) RICH: R(Cr+Na)>2; DUST POOR: Rcr<2.
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, NH4CL, NANO3
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: exac
  LOGICAL :: exan
  LOGICAL :: exsc
  LOGICAL :: exsn
  !
  ! *** SOLVE FOR DRY CASE AND SEE WHICH SOLIDS ARE POSSIBLE **************
  !
  scase = 'U3 ; SUBCASE 2'
  CALL calcu1a ! SOLID
  scase = 'U3 ; SUBCASE 2'
  !
  exan = cnh4no3 > tiny ! NH4NO3
  exac = cnh4cl > tiny ! NH4CL
  exsn = cnano3 > tiny ! NANO3
  exsc = cnacl > tiny ! NACL
  !
  ! *** REGIME DEPENDS ON RELATIVE HUMIDITY AND POSSIBLE SPECIES **********
  !
  IF (exan .OR. exsn) THEN ! *** NH4NO3,NANO3 EXIST
    IF (rh>=drmm1) THEN
      scase = 'U3 ; SUBCASE 1'
      CALL calcu3a
      scase = 'U3 ; SUBCASE 1'
    END IF
    !
  ELSE IF (.NOT. exan .AND. .NOT. exsn) THEN ! *** NH4NO3,NANO3 = 0
    IF (exac .AND. exsc) THEN
      IF (rh>=drmr4) THEN
        scase = 'U3 ; SUBCASE 3'
        CALL calcmdrpii(rh, drmr4, drnacl, calcu1a, calcu4a)
        scase = 'U3 ; SUBCASE 3'
      END IF

    ELSE IF (.NOT. exac .AND. exsc) THEN
      IF (rh>=drmr2) THEN
        scase = 'U3 ; SUBCASE 4'
        CALL calcmdrpii(rh, drmr2, drnacl, calcu1a, calcu4a)
        scase = 'U3 ; SUBCASE 4'
      END IF

    ELSE IF (exac .AND. .NOT. exsc) THEN
      IF (rh>=drmr5) THEN
        scase = 'U3 ; SUBCASE 5'
        CALL calcmdrpii(rh, drmr5, drnacl, calcu1a, calcu5)
        scase = 'U3 ; SUBCASE 5'
      END IF
    END IF
    !
  END IF
  !
  RETURN
  !
END SUBROUTINE calcu3
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU3A
! *** CASE U3A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, NH4CL, NACL
!     4. Completely dissolved: NH4NO3, NANO3
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu3a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv3
  LOGICAL :: psconv4
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi30
  DOUBLE PRECISION :: psi40
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root3
  DOUBLE PRECISION :: root3a
  DOUBLE PRECISION :: root3b
  DOUBLE PRECISION :: root4
  DOUBLE PRECISION :: root4a
  DOUBLE PRECISION :: root4b
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .FALSE.
  psconv1 = .FALSE.
  psconv4 = .FALSE.
  psconv3 = .FALSE.
  !
  psi70 = -great ! GREAT = 1.D10
  psi1o = -great
  psi40 = -great
  psi30 = -great
  !
  root7 = zero
  root1 = zero
  root4 = zero
  root3 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcu1a
  !
  chi1 = cna2so4 ! SALTS
  chi3 = cnacl
  chi4 = cnh4cl
  chi7 = ck2so4
  chi8 = cmgso4
  chi9 = ccaso4
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  molal(1) = zero
  molal(2) = nai
  molal(3) = nh4i
  molal(4) = cli
  molal(5) = so4i
  molal(6) = hso4i
  molal(7) = no3i
  molal(8) = cai
  molal(9) = ki
  molal(10) = mgi
  !
  CALL calcact ! CALCULATE ACTIVITY COEFFICIENTS
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a1 = xk5*(water/gama(2))**3.0 ! NA2S04    <==> Na+
    a14 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    akw = xkw*rh*water*water ! H2O       <==> H+
    a8 = xk8*(water/gama(1))**2.0 ! NaCl      <==> Na+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root1)
      cc = waer(7)*((waer(2)-waer(6))-root1) + 0.25*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*((waer(2)-waer(6))-root1)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max(waer(2)-waer(6)-root1,zero),     &
           chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a1) THEN
      bb = -(((waer(2)-waer(6))-root7)*(waer(1)-root3))
      cc = ((waer(2)-waer(6))-root7)*(waer(1)-root3) +                     &
           0.25D0*(waer(1)-root3)**2.
      dd = -0.25D0*(((waer(2)-waer(6))-root7)*(waer(1)-root3)**2.D0-a1)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = min(max(root1,zero), max(waer(1)-root3,zero), chi1,          &
           max(waer(2)-waer(6),zero))
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! AMMONIUM CHLORIDE
    !
    IF (nh4i*cli>a14) THEN
      bb = -(waer(3)+waer(5)-root4)
      cc = -a14 + nh4i*(waer(5)-root4)
      dd = max(bb*bb-4.D0*cc, zero)
      root4a = 0.5D0*(-bb+sqrt(dd))
      root4b = 0.5D0*(-bb-sqrt(dd))
      IF (zero<=root4a) THEN
        root4 = root4a
      ELSE
        root4 = root4b
      END IF
      root4 = min(max(zero,root4), max(waer(5)-root3,zero), chi4, waer(3))
      psi4 = chi4 - root4
    END IF
    psconv4 = abs(psi4-psi40) <= eps*psi40
    psi40 = psi4
    !
    ! SODIUM CHLORIDE  ; To obtain new value for ROOT3
    !
    IF (nai*cli>a8) THEN
      bb = -((chi1-2.D0*root1)+(waer(5)-root4))
      cc = (chi1-2.D0*root1)*(waer(5)-root4) - a8
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root3a = 0.5D0*(-bb-sqrt(dd))
      root3b = 0.5D0*(-bb+sqrt(dd))
      IF (zero<=root3a) THEN
        root3 = root3a
      ELSE
        root3 = root3b
      END IF
      root3 = min(max(root3,zero), chi3)
      psi3 = chi3 - root3
    END IF
    psconv3 = abs(psi3-psi30) <= eps*psi30
    psi30 = psi3
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root7, zero)
    nai = max(waer(1)-2.D0*root1-root3, zero)
    so4i = max(waer(2)-waer(6)-root7-root1, zero)
    nh4i = max(waer(3)-root4, zero)
    no3i = waer(4)
    cli = max(waer(5)-root4-root3, zero)
    cai = zero
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !      IF (HI.LE.TINY) HI = SQRT(AKW)
    !      OHI   = AKW/HI
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7 .AND. psconv1 .AND. psconv4 .AND. psconv3) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCU3A')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 IF (cli<=tiny .AND. waer(5)>tiny) THEN !No disslv Cl-;solid only
    DO i = 1, nions
      molal(i) = zero
    END DO
    DO i = 1, ngasaq
      gasaq(i) = zero
    END DO
    CALL calcu1a
  ELSE
    a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
    a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
    a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
    !
    gnh3 = nh4i/hi/a2
    ghno3 = hi*no3i/a3
    ghcl = hi*cli/a4
    !
    gasaq(1) = nh3aq
    gasaq(2) = claq
    gasaq(3) = no3aq
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnh4cl = chi4 - psi4
    cnacl = chi3 - psi3
    cnano3 = zero
    cna2so4 = chi1 - psi1
    cmgso4 = zero
    ck2so4 = chi7 - psi7
    ccaso4 = min(waer(6), waer(2))
  END IF
  !
  RETURN
  !
END SUBROUTINE calcu3a

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU2
! *** CASE U2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), (DUST + SODIUM) RICH: R(Cr+Na)>2; DUST POOR: Rcr<2.
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, NH4CL, NANO3, NACL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: exac
  LOGICAL :: exan
  LOGICAL :: exsc
  LOGICAL :: exsn
  !
  ! *** SOLVE FOR DRY CASE AND SEE WHICH SOLIDS ARE POSSIBLE **************
  !
  scase = 'U2 ; SUBCASE 2'
  CALL calcu1a ! SOLID
  scase = 'U2 ; SUBCASE 2'
  !
  exan = cnh4no3 > tiny ! NH4NO3
  exac = cnh4cl > tiny ! NH4CL
  exsn = cnano3 > tiny ! NANO3
  exsc = cnacl > tiny ! NACL
  !
  ! *** REGIME DEPENDS ON RELATIVE HUMIDITY AND POSSIBLE SPECIES **********
  !
  IF (exan) THEN ! *** NH4NO3 EXISTS
    IF (rh>=drmm1) THEN
      scase = 'U2 ; SUBCASE 1'
      CALL calcu2a
      scase = 'U2 ; SUBCASE 1'
    END IF
    !
  ELSE IF (.NOT. exan) THEN ! *** NH4NO3 = 0
    IF (exac .AND. exsn .AND. exsc) THEN
      IF (rh>=drmm2) THEN
        scase = 'U2 ; SUBCASE 3'
        CALL calcmdrpii(rh, drmm2, drnano3, calcu1a, calcu3a)
        scase = 'U2 ; SUBCASE 3'
      END IF

    ELSE IF (.NOT. exac .AND. exsn .AND. exsc) THEN
      IF (rh>=drmr1) THEN
        scase = 'U2 ; SUBCASE 4'
        CALL calcmdrpii(rh, drmr1, drnano3, calcu1a, calcu3a)
        scase = 'U2 ; SUBCASE 4'
      END IF

    ELSE IF (.NOT. exac .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr2) THEN
        scase = 'U2 ; SUBCASE 5'
        CALL calcmdrpii(rh, drmr2, drnacl, calcu1a, calcu4a)
        scase = 'U2 ; SUBCASE 5'
      END IF

    ELSE IF (.NOT. exac .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr3) THEN
        scase = 'U2 ; SUBCASE 6'
        CALL calcmdrpii(rh, drmr3, drnano3, calcu1a, calcu3a)
        scase = 'U2 ; SUBCASE 6'
      END IF

    ELSE IF (exac .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr4) THEN
        scase = 'U2 ; SUBCASE 7'
        CALL calcmdrpii(rh, drmr4, drnacl, calcu1a, calcu4a)
        scase = 'U2 ; SUBCASE 7'
      END IF

    ELSE IF (exac .AND. .NOT. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr5) THEN
        scase = 'U2 ; SUBCASE 8'
        CALL calcmdrpii(rh, drmr5, drnh4cl, calcu1a, calcu5)
        scase = 'U2 ; SUBCASE 8'
      END IF

    ELSE IF (exac .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr6) THEN
        scase = 'U2 ; SUBCASE 9'
        CALL calcmdrpii(rh, drmr6, drnano3, calcu1a, calcu3a)
        scase = 'U2 ; SUBCASE 9'
      END IF
    END IF
    !
  END IF
  !
  RETURN

  !      IF (W(4).GT.TINY) THEN        ! NO3 EXISTS, WATER POSSIBLE
  !         SCASE = 'U2 ; SUBCASE 1'
  !         CALL CALCU2A
  !         SCASE = 'U2 ; SUBCASE 1'
  !      ELSE                          ! NO3 NON EXISTANT, WATER NOT POSSIBLE
  !         SCASE = 'U2 ; SUBCASE 1'
  !         CALL CALCU1A
  !         SCASE = 'U2 ; SUBCASE 1'
  !      ENDIF
  !C
  !      IF (WATER.LE.TINY .AND. RH.LT.DRMM2) THEN      ! DRY AEROSOL
  !         SCASE = 'U2 ; SUBCASE 2'
  !         CALL CALCU2A
  !         SCASE = 'U2 ; SUBCASE 1'
  !C
  !      ELSEIF (WATER.LE.TINY .AND. RH.GE.DRMM2) THEN  ! MDRH OF M2
  !         SCASE = 'U2 ; SUBCASE 3'
  !         CALL CALCMDRPII (RH, DRMM2, DRNANO3, CALCU1A, CALCU3A)
  !         SCASE = 'U2 ; SUBCASE 3'
  !      ENDIF
  !C
  !      RETURN
  !
END SUBROUTINE calcu2
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU2A
! *** CASE U2A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), Cr+NA poor (CRNARAT < 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, NH4CL, NACL, NANO3
!     4. Completely dissolved: NH4NO3
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv1
  LOGICAL :: psconv3
  LOGICAL :: psconv4
  LOGICAL :: psconv5
  LOGICAL :: psconv7
  DOUBLE PRECISION :: psi1o
  DOUBLE PRECISION :: psi20
  DOUBLE PRECISION :: psi30
  DOUBLE PRECISION :: psi40
  DOUBLE PRECISION :: psi50
  DOUBLE PRECISION :: psi70
  DOUBLE PRECISION :: root1
  DOUBLE PRECISION :: root3
  DOUBLE PRECISION :: root3a
  DOUBLE PRECISION :: root3b
  DOUBLE PRECISION :: root4
  DOUBLE PRECISION :: root4a
  DOUBLE PRECISION :: root4b
  DOUBLE PRECISION :: root5
  DOUBLE PRECISION :: root5a
  DOUBLE PRECISION :: root5b
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv7 = .FALSE.
  psconv1 = .FALSE.
  psconv4 = .FALSE.
  psconv3 = .FALSE.
  psconv5 = .FALSE.
  !
  psi70 = -great ! GREAT = 1.D10
  psi1o = -great
  psi40 = -great
  psi30 = -great
  psi50 = -great
  !
  root7 = zero
  root1 = zero
  root4 = zero
  root3 = zero
  root5 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcu1a
  !
  chi1 = cna2so4 ! SALTS
  chi2 = cnano3
  chi3 = cnacl
  chi4 = cnh4cl
  chi7 = ck2so4
  chi8 = cmgso4
  chi9 = ccaso4
  !
  psi1 = cna2so4 ! AMOUNT DISSOLVED
  psi2 = cnano3
  psi3 = cnacl
  psi4 = cnh4cl
  psi5 = cnh4no3
  psi7 = ck2so4
  psi8 = cmgso4
  psi9 = ccaso4
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  molal(1) = zero
  molal(2) = nai
  molal(3) = nh4i
  molal(4) = cli
  molal(5) = so4i
  molal(6) = hso4i
  molal(7) = no3i
  molal(8) = cai
  molal(9) = ki
  molal(10) = mgi
  !
  CALL calcact ! CALCULATE ACTIVITY COEFFICIENTS
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a7 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a1 = xk5*(water/gama(2))**3.0 ! NA2S04    <==> Na+
    a14 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    a8 = xk8*(water/gama(1))**2.0 ! NaCl      <==> Na+
    a9 = xk9*(water/gama(3))**2.0 ! NaNO3     <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a7) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root1)
      cc = waer(7)*((waer(2)-waer(6))-root1) + 0.25*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*((waer(2)-waer(6))-root1)-a7)
      CALL poly3(bb, cc, dd, root7, islv)
      IF (islv/=0) root7 = tiny
      root7 = max(root7, zero)
      root7 = min(root7, waer(7)/2.0, max(waer(2)-waer(6)-root1,zero),     &
           chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi70) <= eps*psi70
    psi70 = psi7
    !
    ! SODIUM SULFATE
    !
    IF (nai*nai*so4i>a1) THEN
      bb = -(((waer(2)-waer(6))-root7)*(waer(1)-root3-root5))
      cc = ((waer(2)-waer(6))-root7)*(waer(1)-root3-root5) +               &
           0.25D0*(waer(1)-root3-root5)**2.0
      dd = -0.25D0*(((waer(2)-waer(6))-root7)*(waer(1)-root3-root5)**2.D0- &
           a1)
      CALL poly3(bb, cc, dd, root1, islv)
      IF (islv/=0) root1 = tiny
      root1 = min(max(root1,zero), max(waer(1)-root3-root5,zero), chi1,    &
           max(waer(2)-waer(6),zero))
      psi1 = chi1 - root1
    END IF
    psconv1 = abs(psi1-psi1o) <= eps*psi1o
    psi1o = psi1
    !
    ! AMMONIUM CHLORIDE
    !
    IF (nh4i*cli>a14) THEN
      bb = -(waer(3)+waer(5)-root4)
      cc = -a14 + nh4i*(waer(5)-root4)
      dd = max(bb*bb-4.D0*cc, zero)
      root4a = 0.5D0*(-bb+sqrt(dd))
      root4b = 0.5D0*(-bb-sqrt(dd))
      IF (zero<=root4a) THEN
        root4 = root4a
      ELSE
        root4 = root4b
      END IF
      root4 = min(max(zero,root4), max(waer(5)-root3,zero), chi4, waer(3))
      psi4 = chi4 - root4
    END IF
    psconv4 = abs(psi4-psi40) <= eps*psi40
    psi40 = psi4
    !
    ! SODIUM CHLORIDE  ; To obtain new value for ROOT3
    !
    IF (nai*cli>a8) THEN
      bb = -((chi1-2.D0*root1-root5)+(waer(5)-root4))
      cc = (chi1-2.D0*root1-root5)*(waer(5)-root4) - a8
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root3a = 0.5D0*(-bb-sqrt(dd))
      root3b = 0.5D0*(-bb+sqrt(dd))
      IF (zero<=root3a) THEN
        root3 = root3a
      ELSE
        root3 = root3b
      END IF
      root3 = min(max(root3,zero), chi3)
      psi3 = chi3 - root3
    END IF
    psconv3 = abs(psi3-psi30) <= eps*psi30
    psi30 = psi3
    !
    ! SODIUM NITRATE
    !
    IF (nai*no3i>a9) THEN
      bb = -(waer(4)+waer(1)-2.D0*root1-root3)
      cc = waer(4)*(waer(1)-2.D0*root1-root3) - a9
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root5a = 0.5D0*(-bb-dd)
      root5b = 0.5D0*(-bb+dd)
      IF (zero<=root5a) THEN
        root5 = root5a
      ELSE
        root5 = root5b
      END IF
      root5 = min(max(root5,zero), chi2)
      psi2 = chi2 - root5
    END IF
    !
    psconv5 = abs(psi2-psi20) <= eps*psi20
    psi20 = psi2
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.0D0*root7, zero)
    nai = max(waer(1)-2.0D0*root1-root3-root5, zero)
    so4i = max(waer(2)-waer(6)-root7-root1, zero)
    nh4i = max(waer(3)-root4, zero)
    no3i = max(waer(4)-root5, zero)
    cli = max(waer(5)-root4-root3, zero)
    cai = zero
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !      IF (HI.LE.TINY) HI = SQRT(AKW)
    !      OHI   = AKW/HI
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv7 .AND. psconv1 .AND. psconv4 .AND. psconv3 .AND. psconv5) &
           GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCU2A')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 IF (cli<=tiny .AND. waer(5)>tiny) THEN !No disslv Cl-;solid only
    DO i = 1, nions
      molal(i) = zero
    END DO
    DO i = 1, ngasaq
      gasaq(i) = zero
    END DO
    CALL calcu1a
  ELSE ! OK, aqueous phase present
    a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
    a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
    a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
    !
    gnh3 = nh4i/hi/a2
    ghno3 = hi*no3i/a3
    ghcl = hi*cli/a4
    !
    gasaq(1) = nh3aq
    gasaq(2) = claq
    gasaq(3) = no3aq
    !
    cnh42s4 = zero
    cnh4no3 = zero
    cnh4cl = chi4 - psi4
    cnacl = chi3 - psi3
    cnano3 = chi2 - psi2
    cna2so4 = chi1 - psi1
    cmgso4 = zero
    ck2so4 = chi7 - psi7
    ccaso4 = min(waer(6), waer(2))
  END IF
  !
  RETURN
  !
END SUBROUTINE calcu2a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCU1
! *** CASE U1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SO4RAT > 2.0), (DUST + SODIUM) RICH: R(Cr+Na)>2; DUST POOR: Rcr<2.
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : K2SO4, CASO4, NA2SO4, MGSO4, NH4CL, NANO3, NACL, NH4NO3
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  LOGICAL :: exac
  LOGICAL :: exan
  LOGICAL :: exsc
  LOGICAL :: exsn
  !
  ! *** SOLVE FOR DRY CASE AND SEE WHICH SOLIDS ARE POSSIBLE **************
  !
  scase = 'U1 ; SUBCASE 1'
  CALL calcu1a ! SOLID
  scase = 'U1 ; SUBCASE 1'
  !
  exan = cnh4no3 > tiny ! NH4NO3
  exac = cnh4cl > tiny ! NH4CL
  exsn = cnano3 > tiny ! NANO3
  exsc = cnacl > tiny ! NACL
  !
  ! *** REGIME DEPENDS ON RELATIVE HUMIDITY AND POSSIBLE SPECIES **********
  !
  IF (exan .OR. exac .OR. exsc .OR. exsn) THEN ! *** WATER POSSIBLE
    IF (rh>=drmm1) THEN
      scase = 'U1 ; SUBCASE 2' ! MDRH
      CALL calcmdrpii(rh, drmm1, drnh4no3, calcu1a, calcu2a)
      scase = 'U1 ; SUBCASE 2'
    END IF
    !
  ELSE IF (.NOT. exan) THEN ! *** NH4NO3 = 0
    IF (exac .AND. exsn .AND. exsc) THEN
      IF (rh>=drmm2) THEN
        scase = 'U1 ; SUBCASE 3'
        CALL calcmdrpii(rh, drmm2, drnano3, calcu1a, calcu3a)
        scase = 'U1 ; SUBCASE 3'
      END IF

    ELSE IF (.NOT. exac .AND. exsn .AND. exsc) THEN
      IF (rh>=drmr1) THEN
        scase = 'U1 ; SUBCASE 4'
        CALL calcmdrpii(rh, drmr1, drnano3, calcu1a, calcu3a)
        scase = 'U1 ; SUBCASE 4'
      END IF

    ELSE IF (.NOT. exac .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr2) THEN
        scase = 'U1 ; SUBCASE 5'
        CALL calcmdrpii(rh, drmr2, drnacl, calcu1a, calcu3a) !, CALCR4A)
        scase = 'U1 ; SUBCASE 5'
      END IF

    ELSE IF (.NOT. exac .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr3) THEN
        scase = 'U1 ; SUBCASE 6'
        CALL calcmdrpii(rh, drmr3, drnano3, calcu1a, calcu3a)
        scase = 'U1 ; SUBCASE 6'
      END IF

    ELSE IF (exac .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr4) THEN
        scase = 'U1 ; SUBCASE 7'
        CALL calcmdrpii(rh, drmr4, drnacl, calcu1a, calcu3a) !, CALCR4A)
        scase = 'U1 ; SUBCASE 7'
      END IF

    ELSE IF (exac .AND. .NOT. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr5) THEN
        scase = 'U1 ; SUBCASE 8'
        CALL calcmdrpii(rh, drmr5, drnh4cl, calcu1a, calcu3a) !, CALCR5)
        scase = 'U1 ; SUBCASE 8'
      END IF

    ELSE IF (exac .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr6) THEN
        scase = 'U1 ; SUBCASE 9'
        CALL calcmdrpii(rh, drmr6, drnano3, calcu1a, calcu3a)
        scase = 'U1 ; SUBCASE 9'
      END IF
    END IF
    !
  ELSE IF (.NOT. exac) THEN ! *** NH4CL  = 0
    IF (exan .AND. exsn .AND. exsc) THEN
      IF (rh>=drmr7) THEN
        scase = 'U1 ; SUBCASE 10'
        CALL calcmdrpii(rh, drmr7, drnh4no3, calcu1a, calcu2a)
        scase = 'U1 ; SUBCASE 10'
      END IF

    ELSE IF (exan .AND. .NOT. exsn .AND. exsc) THEN
      IF (rh>=drmr8) THEN
        scase = 'U1 ; SUBCASE 11'
        CALL calcmdrpii(rh, drmr8, drnh4no3, calcu1a, calcu2a)
        scase = 'U1 ; SUBCASE 11'
      END IF

    ELSE IF (exan .AND. .NOT. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr9) THEN
        scase = 'U1 ; SUBCASE 12'
        CALL calcmdrpii(rh, drmr9, drnh4no3, calcu1a, calcu2a)
        scase = 'U1 ; SUBCASE 12'
      END IF

    ELSE IF (exan .AND. exsn .AND. .NOT. exsc) THEN
      IF (rh>=drmr10) THEN
        scase = 'U1 ; SUBCASE 13'
        CALL calcmdrpii(rh, drmr10, drnh4no3, calcu1a, calcu2a)
        scase = 'U1 ; SUBCASE 13'
      END IF
    END IF
    !
  ELSE IF (.NOT. exsn) THEN ! *** NANO3  = 0
    IF (exan .AND. exac .AND. exsc) THEN
      IF (rh>=drmr11) THEN
        scase = 'U1 ; SUBCASE 14'
        CALL calcmdrpii(rh, drmr11, drnh4no3, calcu1a, calcu2a)
        scase = 'U1 ; SUBCASE 14'
      END IF

    ELSE IF (exan .AND. exac .AND. .NOT. exsc) THEN
      IF (rh>=drmr12) THEN
        scase = 'U1 ; SUBCASE 15'
        CALL calcmdrpii(rh, drmr12, drnh4no3, calcu1a, calcu2a)
        scase = 'U1 ; SUBCASE 15'
      END IF
    END IF
    !
  ELSE IF (.NOT. exsc) THEN ! *** NACL   = 0
    IF (exan .AND. exac .AND. exsn) THEN
      IF (rh>=drmr13) THEN
        scase = 'U1 ; SUBCASE 16'
        CALL calcmdrpii(rh, drmr13, drnh4no3, calcu1a, calcu2a)
        scase = 'U1 ; SUBCASE 16'
      END IF
    END IF
  END IF
  !
  RETURN


  !      IF (RH.LT.DRMM1) THEN
  !         SCASE = 'U1 ; SUBCASE 1'
  !         CALL CALCU1A              ! SOLID PHASE ONLY POSSIBLE
  !         SCASE = 'U1 ; SUBCASE 1'
  !      ELSE
  !         SCASE = 'U1 ; SUBCASE 2'  ! LIQUID & SOLID PHASE POSSIBLE
  !         CALL CALCMDRPII (RH, DRMM1, DRNH4NO3, CALCU1A, CALCU2A)
  !         SCASE = 'U1 ; SUBCASE 2'
  !         ENDIF
  !C
  !      RETURN
  !C
END SUBROUTINE calcu1
!
!=======================================================================
!
! *** ISORROPIA CODE
! *** SUBROUTINE CALCU1A
! *** CASE U1A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0); CRUSTAL+SODIUM RICH (CRNARAT >= 2.0); CRUSTAL POOR (CRRAT<2)
!     2. THERE IS ONLY A SOLID PHASE
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcu1a (  )
  USE mo_isrpia  
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: frcl
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frnh3
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: so4fr
  !
  ! *** CALCULATE SOLIDS *************************************************
  !
  ccaso4 = min(waer(6), waer(2)) ! CCASO4
  so4fr = max(waer(2)-ccaso4, zero)
  cafr = max(waer(6)-ccaso4, zero)
  ck2so4 = min(0.5D0*waer(7), so4fr) ! CK2SO4
  frk = max(waer(7)-2.D0*ck2so4, zero)
  so4fr = max(so4fr-ck2so4, zero)
  cmgso4 = min(waer(8), so4fr) ! CMGSO4
  frmg = max(waer(8)-cmgso4, zero)
  so4fr = max(so4fr-cmgso4, zero)
  cna2so4 = max(so4fr, zero) ! CNA2SO4
  frna = max(waer(1)-2.D0*cna2so4, zero)
  !
  cnh42s4 = zero
  !
  cnano3 = min(frna, waer(4))
  frno3 = max(waer(4)-cnano3, zero)
  frna = max(frna-cnano3, zero)
  !
  cnacl = min(frna, waer(5))
  frcl = max(waer(5)-cnacl, zero)
  frna = max(frna-cnacl, zero)
  !
  cnh4no3 = min(frno3, waer(3))
  frno3 = max(frno3-cnh4no3, zero)
  frnh3 = max(waer(3)-cnh4no3, zero)
  !
  cnh4cl = min(frcl, frnh3)
  frcl = max(frcl-cnh4cl, zero)
  frnh3 = max(frnh3-cnh4cl, zero)
  !
  ! *** OTHER PHASES ******************************************************
  !
  water = zero
  !
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  RETURN
  !
END SUBROUTINE calcu1a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW13
! *** CASE W13
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4
!     4. Completely dissolved: CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw13 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep

    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! ION CONCENTRATIONS
    !
    nai = waer(1)
    so4i = max(waer(2)-waer(6), zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    ki = waer(7)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW13')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cmgso4 = zero
  ck2so4 = zero
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = zero
  kcl = zero
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw13
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW12
! *** CASE W12
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4
!     4. Completely dissolved: CA(NO3)2, CACL2, KNO3, KCL, MGSO4,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw12 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psi9o = -great ! GREAT = 1.D10
  root9 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7))
      cc = waer(7)*(waer(2)-waer(6)) + 0.25D0*waer(7)*waer(7)
      dd = -0.25*(waer(7)*waer(7)*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0, (waer(2)-waer(6)), chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = waer(3)
    no3i = waer(4)
    cli = waer(5)
    cai = zero
    nai = waer(1)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW12')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = zero
  kcl = zero
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw12
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW11
! *** CASE W11
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3
!     4. Completely dissolved: CA(NO3)2, CACL2, KCL, MGSO4,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw11 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  !
  psi9o = -great
  psi13o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13)
      cc = (waer(7)-root13)*(waer(2)-waer(6)) + 0.25D0*(waer(7)-root13)**  &
           2.0
      dd = -0.25*((waer(7)-root13)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13, (waer(2)-waer(6)), chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)+waer(7)-2.D0*root9)
      cc = waer(4)*(waer(7)-2.D0*root9) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,zero))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = waer(3)
    no3i = max(waer(4)-root13, zero)
    cli = waer(5)
    cai = zero
    nai = waer(1)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW11')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = zero
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw11
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW10
! *** CASE W10
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4
!     4. Completely dissolved: CA(NO3)2, CACL2, KCL,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw10 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  !
  psi9o = -great
  psi13o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a

  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi10 = cmgso4
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13)
      cc = (waer(7)-root13)*(waer(2)-waer(6)) + 0.25D0*(waer(7)-root13)**  &
           2.0
      dd = -0.25*((waer(7)-root13)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13, (waer(2)-waer(6)), chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)+waer(7)-2.D0*root9)
      cc = waer(4)*(waer(7)-2.D0*root9) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,zero))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = waer(3)
    no3i = max(waer(4)-root13, zero)
    cli = waer(5)
    cai = zero
    nai = waer(1)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW10')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = zero
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw10
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW9
! *** CASE W9
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw9 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv14
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi14o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root14
  DOUBLE PRECISION :: root14a
  DOUBLE PRECISION :: root14b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  psconv14 = .TRUE.
  !
  psi9o = -great
  psi13o = -great
  psi14o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  root14 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi10 = cmgso4
  chi14 = ckcl
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    a14 = xk20*(water/gama(20))**2.0 ! KCL       <==> K+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13-root14)
      cc = (waer(7)-root13-root14)*(waer(2)-waer(6)) +                     &
           0.25D0*(waer(7)-root13-root14)**2.0
      dd = -0.25*((waer(7)-root13-root14)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13-root14, (waer(2)-waer(6)),     &
           chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)+waer(7)-2.D0*root9-root14)
      cc = waer(4)*(waer(7)-2.D0*root9-root14) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! POTASSIUM CLORIDE
    !
    IF (ki*cli>a14) THEN
      bb = -(waer(5)+waer(7)-2.D0*root9-root13)
      cc = waer(5)*(waer(7)-2.D0*root9-root13) - a14
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root14a = 0.5D0*(-bb-dd)
      root14b = 0.5D0*(-bb+dd)
      IF (zero<=root14a) THEN
        root14 = root14a
      ELSE
        root14 = root14b
      END IF
      root14 = min(max(root14,zero), chi14)
      psi14 = chi14 - root14
    END IF
    psconv14 = abs(psi14-psi14o) <= eps*psi14o
    psi14o = psi14
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13-root14, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = waer(3)
    no3i = max(waer(4)-root13, zero)
    cli = max(waer(5)-root14, zero)
    cai = zero
    nai = waer(1)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13 .AND. psconv14) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW9')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = zero
  cnacl = zero
  cnano3 = zero
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = chi14 - psi14
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw9
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW8
! *** CASE W8
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw8 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv14
  LOGICAL :: psconv5
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi14o
  DOUBLE PRECISION :: psi5o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root14
  DOUBLE PRECISION :: root14a
  DOUBLE PRECISION :: root14b
  DOUBLE PRECISION :: root5
  DOUBLE PRECISION :: root5a
  DOUBLE PRECISION :: root5b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  psconv14 = .TRUE.
  psconv5 = .TRUE.
  !
  psi9o = -great
  psi13o = -great
  psi14o = -great
  psi5o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  root14 = zero
  root5 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi10 = cmgso4
  chi14 = ckcl
  chi5 = cnh4cl
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    a14 = xk20*(water/gama(20))**2.0 ! KCL       <==> K+
    a5 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13-root14)
      cc = (waer(7)-root13-root14)*(waer(2)-waer(6)) +                     &
           0.25D0*(waer(7)-root13-root14)**2.0
      dd = -0.25*((waer(7)-root13-root14)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13-root14, (waer(2)-waer(6)),     &
           chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)+waer(7)-2.D0*root9-root14)
      cc = waer(4)*(waer(7)-2.D0*root9-root14) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,zero))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! POTASSIUM CLORIDE
    !
    IF (ki*cli>a14) THEN
      bb = -(waer(5)-root5+waer(7)-2.D0*root9-root13)
      cc = (waer(5)-root5)*(waer(7)-2.D0*root9-root13) - a14
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root14a = 0.5D0*(-bb-dd)
      root14b = 0.5D0*(-bb+dd)
      IF (zero<=root14a) THEN
        root14 = root14a
      ELSE
        root14 = root14b
      END IF
      root14 = min(max(root14,zero), chi14)
      psi14 = chi14 - root14
    END IF
    psconv14 = abs(psi14-psi14o) <= eps*psi14o
    psi14o = psi14
    !
    ! AMMONIUM CLORIDE
    !
    IF (nh4i*cli>a5) THEN
      bb = -(waer(5)+waer(3)-root14)
      cc = (waer(5)-root14)*waer(3) - a5
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root5a = 0.5D0*(-bb-dd)
      root5b = 0.5D0*(-bb+dd)
      IF (zero<=root5a) THEN
        root5 = root5a
      ELSE
        root5 = root5b
      END IF
      root5 = min(max(root5,zero), chi5)
      psi5 = chi5 - root5
    END IF
    psconv5 = abs(psi5-psi5o) <= eps*psi5o
    psi5o = psi5
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13-root14, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = max(waer(3)-root5, zero)
    no3i = max(waer(4)-root13, zero)
    cli = max(waer(5)-root14-root5, zero)
    cai = zero
    nai = waer(1)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13 .AND. psconv14 .AND. psconv5) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW8')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = chi5 - psi5
  cnacl = zero
  cnano3 = zero
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = chi14 - psi14
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw8
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW7
! *** CASE W7
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NANO3, NH4NO3
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw7 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv14
  LOGICAL :: psconv5
  LOGICAL :: psconv7
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi14o
  DOUBLE PRECISION :: psi5o
  DOUBLE PRECISION :: psi7o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root14
  DOUBLE PRECISION :: root14a
  DOUBLE PRECISION :: root14b
  DOUBLE PRECISION :: root5
  DOUBLE PRECISION :: root5a
  DOUBLE PRECISION :: root5b
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: root7a
  DOUBLE PRECISION :: root7b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  psconv14 = .TRUE.
  psconv5 = .TRUE.
  psconv7 = .TRUE.
  !
  psi9o = -great
  psi13o = -great
  psi14o = -great
  psi5o = -great
  psi7o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  root14 = zero
  root5 = zero
  root7 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi10 = cmgso4
  chi14 = ckcl
  chi5 = cnh4cl
  chi7 = cnacl
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    a14 = xk20*(water/gama(20))**2.0 ! KCL       <==> K+
    a5 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    a7 = xk8*(water/gama(1))**2.0 ! NaCl      <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13-root14)
      cc = (waer(7)-root13-root14)*(waer(2)-waer(6)) +                     &
           0.25D0*(waer(7)-root13-root14)**2.0
      dd = -0.25*((waer(7)-root13-root14)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13-root14, (waer(2)-waer(6)),     &
           chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)+waer(7)-2.D0*root9-root14)
      cc = waer(4)*(waer(7)-2.D0*root9-root14) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,zero))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! POTASSIUM CLORIDE
    !
    IF (ki*cli>a14) THEN
      bb = -(waer(5)-root5-root7+waer(7)-2.D0*root9-root13)
      cc = (waer(5)-root5-root7)*(waer(7)-2.D0*root9-root13) - a14
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root14a = 0.5D0*(-bb-dd)
      root14b = 0.5D0*(-bb+dd)
      IF (zero<=root14a) THEN
        root14 = root14a
      ELSE
        root14 = root14b
      END IF
      root14 = min(max(root14,zero), chi14)
      psi14 = chi14 - root14
    END IF
    psconv14 = abs(psi14-psi14o) <= eps*psi14o
    psi14o = psi14
    !
    ! AMMONIUM CLORIDE
    !
    IF (nh4i*cli>a5) THEN
      bb = -(waer(5)+waer(3)-root14-root7)
      cc = (waer(5)-root14-root7)*waer(3) - a5
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root5a = 0.5D0*(-bb-dd)
      root5b = 0.5D0*(-bb+dd)
      IF (zero<=root5a) THEN
        root5 = root5a
      ELSE
        root5 = root5b
      END IF
      root5 = min(max(root5,zero), chi5)
      psi5 = chi5 - root5
    END IF
    psconv5 = abs(psi5-psi5o) <= eps*psi5o
    psi5o = psi5
    !
    ! SODIUM CLORIDE
    !
    IF (nai*cli>a7) THEN
      bb = -(waer(5)+waer(1)-root14-root5)
      cc = (waer(5)-root14-root5)*waer(1) - a7
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root7a = 0.5D0*(-bb-dd)
      root7b = 0.5D0*(-bb+dd)
      IF (zero<=root7a) THEN
        root7 = root7a
      ELSE
        root7 = root7b
      END IF
      root7 = min(max(root7,zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi7o) <= eps*psi7o
    psi7o = psi7
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13-root14, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = max(waer(3)-root5, zero)
    no3i = max(waer(4)-root13, zero)
    cli = max(waer(5)-root14-root5-root7, zero)
    cai = zero
    nai = max(waer(1)-root7, zero)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13 .AND. psconv14 .AND. psconv5 .AND.        &
           psconv7) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW7')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = chi5 - psi5
  cnacl = chi7 - psi7
  cnano3 = zero
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = chi14 - psi14
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw7

!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW6
! *** CASE W6
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL, NANO3
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2, NH4NO3
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw6 (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv14
  LOGICAL :: psconv5
  LOGICAL :: psconv7
  LOGICAL :: psconv8
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi14o
  DOUBLE PRECISION :: psi5o
  DOUBLE PRECISION :: psi7o
  DOUBLE PRECISION :: psi8o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root14
  DOUBLE PRECISION :: root14a
  DOUBLE PRECISION :: root14b
  DOUBLE PRECISION :: root5
  DOUBLE PRECISION :: root5a
  DOUBLE PRECISION :: root5b
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: root7a
  DOUBLE PRECISION :: root7b
  DOUBLE PRECISION :: root8
  DOUBLE PRECISION :: root8a
  DOUBLE PRECISION :: root8b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  psconv14 = .TRUE.
  psconv5 = .TRUE.
  psconv7 = .TRUE.
  psconv8 = .TRUE.
  !
  psi9o = -great
  psi13o = -great
  psi14o = -great
  psi5o = -great
  psi7o = -great
  psi8o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  root14 = zero
  root5 = zero
  root7 = zero
  root8 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi10 = cmgso4
  chi14 = ckcl
  chi5 = cnh4cl
  chi7 = cnacl
  chi8 = cnano3
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    a14 = xk20*(water/gama(20))**2.0 ! KCL       <==> K+
    a5 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    a7 = xk8*(water/gama(1))**2.0 ! NaCl      <==> Na+
    a8 = xk9*(water/gama(3))**2. ! NaNO3     <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13-root14)
      cc = (waer(7)-root13-root14)*(waer(2)-waer(6)) +                     &
           0.25D0*(waer(7)-root13-root14)**2.0
      dd = -0.25*((waer(7)-root13-root14)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13-root14, (waer(2)-waer(6)),     &
           chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)-root8+waer(7)-2.D0*root9-root14)
      cc = (waer(4)-root8)*(waer(7)-2.D0*root9-root14) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,zero))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! POTASSIUM CLORIDE
    !
    IF (ki*cli>a14) THEN
      bb = -(waer(5)-root5-root7+waer(7)-2.D0*root9-root13)
      cc = (waer(5)-root5-root7)*(waer(7)-2.D0*root9-root13) - a14
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root14a = 0.5D0*(-bb-dd)
      root14b = 0.5D0*(-bb+dd)
      IF (zero<=root14a) THEN
        root14 = root14a
      ELSE
        root14 = root14b
      END IF
      root14 = min(max(root14,zero), chi14)
      psi14 = chi14 - root14
    END IF
    psconv14 = abs(psi14-psi14o) <= eps*psi14o
    psi14o = psi14
    !
    ! AMMONIUM CLORIDE
    !
    IF (nh4i*cli>a5) THEN
      bb = -(waer(5)+waer(3)-root14-root7)
      cc = (waer(5)-root14-root7)*waer(3) - a5
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root5a = 0.5D0*(-bb-dd)
      root5b = 0.5D0*(-bb+dd)
      IF (zero<=root5a) THEN
        root5 = root5a
      ELSE
        root5 = root5b
      END IF
      root5 = min(max(root5,zero), chi5)
      psi5 = chi5 - root5
    END IF
    psconv5 = abs(psi5-psi5o) <= eps*psi5o
    psi5o = psi5
    !
    ! SODIUM CLORIDE
    !
    IF (nai*cli>a7) THEN
      bb = -(waer(5)+waer(1)-root8-root14-root5)
      cc = (waer(5)-root14-root5)*(waer(1)-root8) - a7
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root7a = 0.5D0*(-bb-dd)
      root7b = 0.5D0*(-bb+dd)
      IF (zero<=root7a) THEN
        root7 = root7a
      ELSE
        root7 = root7b
      END IF
      root7 = min(max(root7,zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi7o) <= eps*psi7o
    psi7o = psi7
    !
    ! SODIUM NITRATE
    !
    IF (nai*no3i>a8) THEN
      bb = -(waer(4)-root13+waer(1)-root7)
      cc = (waer(4)-root13)*(waer(1)-root7) - a8
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root8a = 0.5D0*(-bb-dd)
      root8b = 0.5D0*(-bb+dd)
      IF (zero<=root8a) THEN
        root8 = root8a
      ELSE
        root8 = root8b
      END IF
      root8 = min(max(root8,zero), chi8)
      psi8 = chi8 - root8
    END IF
    psconv8 = abs(psi8-psi8o) <= eps*psi8o
    psi8o = psi8
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13-root14, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = max(waer(3)-root5, zero)
    no3i = max(waer(4)-root13-root8, zero)
    cli = max(waer(5)-root14-root5-root7, zero)
    cai = zero
    nai = max(waer(1)-root7-root8, zero)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13 .AND. psconv14 .AND. psconv5 .AND.        &
           psconv7 .AND. psconv8) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW6')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = chi5 - psi5
  cnacl = chi7 - psi7
  cnano3 = chi8 - psi8
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = chi14 - psi14
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw6
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW5
! *** CASE W5
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL, NANO3, NH4NO3
!     4. Completely dissolved: CA(NO3)2, CACL2,
!                              MG(NO3)2, MGCL2
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw5 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  IF (waer(4)>tiny) THEN ! NO3 EXIST, WATER POSSIBLE
    scase = 'W5 ; SUBCASE 1'
    CALL calcw5a
    scase = 'W5 ; SUBCASE 1'
  ELSE ! NO3, CL NON EXISTANT
    scase = 'W1 ; SUBCASE 1'
    CALL calcw1a
    scase = 'W1 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmp5) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcw1a
      scase = 'W5 ; SUBCASE 2'
      RETURN
    ELSE
      scase = 'W5 ; SUBCASE 3' ! MDRH REGION (CaSO4, K2SO4, KNO3, KCL, MGSO4,
      !                                                    NANO3, NACL, NH4NO3, NH4CL)
      CALL calcmdrpii(rh, drmp5, drnh4no3, calcw1a, calcw6)
      scase = 'W5 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcw5
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW5A
! *** CASE W5A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL,
!                          NANO3, NH4NO3
!     4. Completely dissolved: CA(NO3)2, CACL2, MG(NO3)2, MGCL2
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw5a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv14
  LOGICAL :: psconv5
  LOGICAL :: psconv7
  LOGICAL :: psconv8
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi14o
  DOUBLE PRECISION :: psi5o
  DOUBLE PRECISION :: psi7o
  DOUBLE PRECISION :: psi8o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root14
  DOUBLE PRECISION :: root14a
  DOUBLE PRECISION :: root14b
  DOUBLE PRECISION :: root5
  DOUBLE PRECISION :: root5a
  DOUBLE PRECISION :: root5b
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: root7a
  DOUBLE PRECISION :: root7b
  DOUBLE PRECISION :: root8
  DOUBLE PRECISION :: root8a
  DOUBLE PRECISION :: root8b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  psconv14 = .TRUE.
  psconv5 = .TRUE.
  psconv7 = .TRUE.
  psconv8 = .TRUE.
  !
  psi9o = -great
  psi13o = -great
  psi14o = -great
  psi5o = -great
  psi7o = -great
  psi8o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  root14 = zero
  root5 = zero
  root7 = zero
  root8 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi10 = cmgso4
  chi14 = ckcl
  chi5 = cnh4cl
  chi7 = cnacl
  chi8 = cnano3
  chi6 = cnh4no3
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    a14 = xk20*(water/gama(20))**2.0 ! KCL       <==> K+
    a5 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    a7 = xk8*(water/gama(1))**2.0 ! NaCl      <==> Na+
    a8 = xk9*(water/gama(3))**2. ! NaNO3     <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13-root14)
      cc = (waer(7)-root13-root14)*(waer(2)-waer(6)) +                     &
           0.25D0*(waer(7)-root13-root14)**2.0
      dd = -0.25*((waer(7)-root13-root14)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13-root14, (waer(2)-waer(6)),     &
           chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)-root8+waer(7)-2.D0*root9-root14)
      cc = (waer(4)-root8)*(waer(7)-2.D0*root9-root14) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,zero))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! POTASSIUM CLORIDE
    !
    IF (ki*cli>a14) THEN
      bb = -(waer(5)-root5-root7+waer(7)-2.D0*root9-root13)
      cc = (waer(5)-root5-root7)*(waer(7)-2.D0*root9-root13) - a14
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root14a = 0.5D0*(-bb-dd)
      root14b = 0.5D0*(-bb+dd)
      IF (zero<=root14a) THEN
        root14 = root14a
      ELSE
        root14 = root14b
      END IF
      root14 = min(max(root14,zero), chi14)
      psi14 = chi14 - root14
    END IF
    psconv14 = abs(psi14-psi14o) <= eps*psi14o
    psi14o = psi14
    !
    ! AMMONIUM CLORIDE
    !
    IF (nh4i*cli>a5) THEN
      bb = -(waer(5)+waer(3)-root14-root7)
      cc = (waer(5)-root14-root7)*waer(3) - a5
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root5a = 0.5D0*(-bb-dd)
      root5b = 0.5D0*(-bb+dd)
      IF (zero<=root5a) THEN
        root5 = root5a
      ELSE
        root5 = root5b
      END IF
      root5 = min(max(root5,zero), chi5)
      psi5 = chi5 - root5
    END IF
    psconv5 = abs(psi5-psi5o) <= eps*psi5o
    psi5o = psi5
    !
    ! SODIUM CLORIDE
    !
    IF (nai*cli>a7) THEN
      bb = -(waer(5)+waer(1)-root8-root14-root5)
      cc = (waer(5)-root14-root5)*(waer(1)-root8) - a7
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root7a = 0.5D0*(-bb-dd)
      root7b = 0.5D0*(-bb+dd)
      IF (zero<=root7a) THEN
        root7 = root7a
      ELSE
        root7 = root7b
      END IF
      root7 = min(max(root7,zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi7o) <= eps*psi7o
    psi7o = psi7
    !
    ! SODIUM NITRATE
    !
    IF (nai*no3i>a8) THEN
      bb = -(waer(4)-root13+waer(1)-root7)
      cc = (waer(4)-root13)*(waer(1)-root7) - a8
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root8a = 0.5D0*(-bb-dd)
      root8b = 0.5D0*(-bb+dd)
      IF (zero<=root8a) THEN
        root8 = root8a
      ELSE
        root8 = root8b
      END IF
      root8 = min(max(root8,zero), chi8)
      psi8 = chi8 - root8
    END IF
    psconv8 = abs(psi8-psi8o) <= eps*psi8o
    psi8o = psi8
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13-root14, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = max(waer(3)-root5, zero)
    no3i = max(waer(4)-root13-root8, zero)
    cli = max(waer(5)-root14-root5-root7, zero)
    cai = zero
    nai = max(waer(1)-root7-root8, zero)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13 .AND. psconv14 .AND. psconv5 .AND.        &
           psconv7 .AND. psconv8) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW5')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = chi5 - psi5
  cnacl = chi7 - psi7
  cnano3 = chi8 - psi8
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = chi14 - psi14
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw5a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW4
! *** CASE W4
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw4 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  IF (waer(4)>tiny) THEN ! NO3 EXIST, WATER POSSIBLE
    scase = 'W4 ; SUBCASE 1'
    CALL calcw4a
    scase = 'W4 ; SUBCASE 1'
  ELSE ! NO3, CL NON EXISTANT
    scase = 'W1 ; SUBCASE 1'
    CALL calcw1a
    scase = 'W1 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmp4) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcw1a
      scase = 'W4 ; SUBCASE 2'
      RETURN
    ELSE
      scase = 'W4 ; SUBCASE 3' ! MDRH REGION (CaSO4, K2SO4, KNO3, KCL, MGSO4,
      !                                                    MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL)
      CALL calcmdrpii(rh, drmp4, drmgno32, calcw1a, calcw5a)
      scase = 'W4 ; SUBCASE 3'
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcw4
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW4A
! *** CASE W4A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL,
!                          NANO3, NH4NO3, MG(NO3)2
!     4. Completely dissolved: CA(NO3)2, CACL2, MGCL2
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw4a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv14
  LOGICAL :: psconv5
  LOGICAL :: psconv7
  LOGICAL :: psconv8
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi14o
  DOUBLE PRECISION :: psi5o
  DOUBLE PRECISION :: psi7o
  DOUBLE PRECISION :: psi8o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root14
  DOUBLE PRECISION :: root14a
  DOUBLE PRECISION :: root14b
  DOUBLE PRECISION :: root5
  DOUBLE PRECISION :: root5a
  DOUBLE PRECISION :: root5b
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: root7a
  DOUBLE PRECISION :: root7b
  DOUBLE PRECISION :: root8
  DOUBLE PRECISION :: root8a
  DOUBLE PRECISION :: root8b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  psconv14 = .TRUE.
  psconv5 = .TRUE.
  psconv7 = .TRUE.
  psconv8 = .TRUE.
  !
  psi9o = -great
  psi13o = -great
  psi14o = -great
  psi5o = -great
  psi7o = -great
  psi8o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  root14 = zero
  root5 = zero
  root7 = zero
  root8 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi10 = cmgso4
  chi14 = ckcl
  chi5 = cnh4cl
  chi7 = cnacl
  chi8 = cnano3
  chi6 = cnh4no3
  chi15 = cmgno32
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    a14 = xk20*(water/gama(20))**2.0 ! KCL       <==> K+
    a5 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    a7 = xk8*(water/gama(1))**2.0 ! NaCl      <==> Na+
    a8 = xk9*(water/gama(3))**2. ! NaNO3     <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13-root14)
      cc = (waer(7)-root13-root14)*(waer(2)-waer(6)) +                     &
           0.25D0*(waer(7)-root13-root14)**2.0
      dd = -0.25*((waer(7)-root13-root14)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13-root14, (waer(2)-waer(6)),     &
           chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)-root8+waer(7)-2.D0*root9-root14)
      cc = (waer(4)-root8)*(waer(7)-2.D0*root9-root14) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,zero))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! POTASSIUM CLORIDE
    !
    IF (ki*cli>a14) THEN
      bb = -(waer(5)-root5-root7+waer(7)-2.D0*root9-root13)
      cc = (waer(5)-root5-root7)*(waer(7)-2.D0*root9-root13) - a14
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root14a = 0.5D0*(-bb-dd)
      root14b = 0.5D0*(-bb+dd)
      IF (zero<=root14a) THEN
        root14 = root14a
      ELSE
        root14 = root14b
      END IF
      root14 = min(max(root14,zero), chi14)
      psi14 = chi14 - root14
    END IF
    psconv14 = abs(psi14-psi14o) <= eps*psi14o
    psi14o = psi14
    !
    ! AMMONIUM CLORIDE
    !
    IF (nh4i*cli>a5) THEN
      bb = -(waer(5)+waer(3)-root14-root7)
      cc = (waer(5)-root14-root7)*waer(3) - a5
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root5a = 0.5D0*(-bb-dd)
      root5b = 0.5D0*(-bb+dd)
      IF (zero<=root5a) THEN
        root5 = root5a
      ELSE
        root5 = root5b
      END IF
      root5 = min(max(root5,zero), chi5)
      psi5 = chi5 - root5
    END IF
    psconv5 = abs(psi5-psi5o) <= eps*psi5o
    psi5o = psi5
    !
    ! SODIUM CLORIDE
    !
    IF (nai*cli>a7) THEN
      bb = -(waer(5)+waer(1)-root8-root14-root5)
      cc = (waer(5)-root14-root5)*(waer(1)-root8) - a7
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root7a = 0.5D0*(-bb-dd)
      root7b = 0.5D0*(-bb+dd)
      IF (zero<=root7a) THEN
        root7 = root7a
      ELSE
        root7 = root7b
      END IF
      root7 = min(max(root7,zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi7o) <= eps*psi7o
    psi7o = psi7
    !
    ! SODIUM NITRATE
    !
    IF (nai*no3i>a8) THEN
      bb = -(waer(4)-root13+waer(1)-root7)
      cc = (waer(4)-root13)*(waer(1)-root7) - a8
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root8a = 0.5D0*(-bb-dd)
      root8b = 0.5D0*(-bb+dd)
      IF (zero<=root8a) THEN
        root8 = root8a
      ELSE
        root8 = root8b
      END IF
      root8 = min(max(root8,zero), chi8)
      psi8 = chi8 - root8
    END IF
    psconv8 = abs(psi8-psi8o) <= eps*psi8o
    psi8o = psi8
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13-root14, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = max(waer(3)-root5, zero)
    no3i = max(waer(4)-root13-root8, zero)
    cli = max(waer(5)-root14-root5-root7, zero)
    cai = zero
    nai = max(waer(1)-root7-root8, zero)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13 .AND. psconv14 .AND. psconv5 .AND.        &
           psconv7 .AND. psconv8) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW4')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = chi5 - psi5
  cnacl = chi7 - psi7
  cnano3 = chi8 - psi8
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = chi14 - psi14
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw4a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW3
! *** CASE W3
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw3 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** REGIME DEPENDS ON THE EXISTANCE OF WATER AND OF THE RH ************
  !
  !      IF (WAER(4).GT.TINY .AND. WAER(5).GT.TINY) THEN ! NO3,CL EXIST, WATER POSSIBLE
  !         SCASE = 'W3 ; SUBCASE 1'
  !         CALL CALCW3A
  !         SCASE = 'W3 ; SUBCASE 1'
  !      ELSE                                      ! NO3, CL NON EXISTANT
  !         SCASE = 'W1 ; SUBCASE 1'
  !         CALL CALCW1A
  !         SCASE = 'W1 ; SUBCASE 1'
  !      ENDIF
  !
  CALL calcw1a

  IF (water<=tiny) THEN
    IF (rh<drmp3) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcw1a
      scase = 'W3 ; SUBCASE 2'
      RETURN
    ELSE
      scase = 'W3 ; SUBCASE 3' ! MDRH REGION (CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
      !                                                    MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL)
      CALL calcmdrpii(rh, drmp3, drcano32, calcw1a, calcw4a)
      scase = 'W3 ; SUBCASE 3'
    END IF
  ELSE ! NO3, CL NON EXISTANT
    scase = 'W3 ; SUBCASE 1'
    CALL calcw3a
    scase = 'W3 ; SUBCASE 1'
  END IF
  !
  RETURN
  !
END SUBROUTINE calcw3
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW3A
! *** CASE W3A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL,
!                          NANO3, NH4NO3, CA(NO3)2, MG(NO3)2
!     4. Completely dissolved: CACL2, MGCL2
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw3a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv14
  LOGICAL :: psconv5
  LOGICAL :: psconv7
  LOGICAL :: psconv8
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi14o
  DOUBLE PRECISION :: psi5o
  DOUBLE PRECISION :: psi7o
  DOUBLE PRECISION :: psi8o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root14
  DOUBLE PRECISION :: root14a
  DOUBLE PRECISION :: root14b
  DOUBLE PRECISION :: root5
  DOUBLE PRECISION :: root5a
  DOUBLE PRECISION :: root5b
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: root7a
  DOUBLE PRECISION :: root7b
  DOUBLE PRECISION :: root8
  DOUBLE PRECISION :: root8a
  DOUBLE PRECISION :: root8b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  psconv14 = .TRUE.
  psconv5 = .TRUE.
  psconv7 = .TRUE.
  psconv8 = .TRUE.
  !
  psi9o = -great
  psi13o = -great
  psi14o = -great
  psi5o = -great
  psi7o = -great
  psi8o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  root14 = zero
  root5 = zero
  root7 = zero
  root8 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi10 = cmgso4
  chi14 = ckcl
  chi5 = cnh4cl
  chi7 = cnacl
  chi8 = cnano3
  chi6 = cnh4no3
  chi15 = cmgno32
  chi12 = ccano32
  chi11 = ccaso4
  !C
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    a14 = xk20*(water/gama(20))**2.0 ! KCL       <==> K+
    a5 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    a7 = xk8*(water/gama(1))**2.0 ! NaCl      <==> Na+
    a8 = xk9*(water/gama(3))**2. ! NaNO3     <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13-root14)
      cc = (waer(7)-root13-root14)*(waer(2)-waer(6)) +                     &
           0.25D0*(waer(7)-root13-root14)**2.0
      dd = -0.25*((waer(7)-root13-root14)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13-root14, (waer(2)-waer(6)),     &
           chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)-root8+waer(7)-2.D0*root9-root14)
      cc = (waer(4)-root8)*(waer(7)-2.D0*root9-root14) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,zero))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! POTASSIUM CLORIDE
    !
    IF (ki*cli>a14) THEN
      bb = -(waer(5)-root5-root7+waer(7)-2.D0*root9-root13)
      cc = (waer(5)-root5-root7)*(waer(7)-2.D0*root9-root13) - a14
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root14a = 0.5D0*(-bb-dd)
      root14b = 0.5D0*(-bb+dd)
      IF (zero<=root14a) THEN
        root14 = root14a
      ELSE
        root14 = root14b
      END IF
      root14 = min(max(root14,zero), chi14)
      psi14 = chi14 - root14
    END IF
    psconv14 = abs(psi14-psi14o) <= eps*psi14o
    psi14o = psi14
    !
    ! AMMONIUM CLORIDE
    !
    IF (nh4i*cli>a5) THEN
      bb = -(waer(5)+waer(3)-root14-root7)
      cc = (waer(5)-root14-root7)*waer(3) - a5
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root5a = 0.5D0*(-bb-dd)
      root5b = 0.5D0*(-bb+dd)
      IF (zero<=root5a) THEN
        root5 = root5a
      ELSE
        root5 = root5b
      END IF
      root5 = min(max(root5,zero), chi5)
      psi5 = chi5 - root5
    END IF
    psconv5 = abs(psi5-psi5o) <= eps*psi5o
    psi5o = psi5
    !
    ! SODIUM CLORIDE
    !
    IF (nai*cli>a7) THEN
      bb = -(waer(5)+waer(1)-root8-root14-root5)
      cc = (waer(5)-root14-root5)*(waer(1)-root8) - a7
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root7a = 0.5D0*(-bb-dd)
      root7b = 0.5D0*(-bb+dd)
      IF (zero<=root7a) THEN
        root7 = root7a
      ELSE
        root7 = root7b
      END IF
      root7 = min(max(root7,zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi7o) <= eps*psi7o
    psi7o = psi7
    !
    ! SODIUM NITRATE
    !
    IF (nai*no3i>a8) THEN
      bb = -(waer(4)-root13+waer(1)-root7)
      cc = (waer(4)-root13)*(waer(1)-root7) - a8
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root8a = 0.5D0*(-bb-dd)
      root8b = 0.5D0*(-bb+dd)
      IF (zero<=root8a) THEN
        root8 = root8a
      ELSE
        root8 = root8b
      END IF
      root8 = min(max(root8,zero), chi8)
      psi8 = chi8 - root8
    END IF
    psconv8 = abs(psi8-psi8o) <= eps*psi8o
    psi8o = psi8
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13-root14, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = max(waer(3)-root5, zero)
    no3i = max(waer(4)-root13-root8, zero)
    cli = max(waer(5)-root14-root5-root7, zero)
    cai = zero
    nai = max(waer(1)-root7-root8, zero)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13 .AND. psconv14 .AND. psconv5 .AND.        &
           psconv7 .AND. psconv8) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW3')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = chi5 - psi5
  cnacl = chi7 - psi7
  cnano3 = chi8 - psi8
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = chi14 - psi14
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw3a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW2
! *** CASE W2
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
!     THERE ARE THREE REGIMES IN THIS CASE:
!     1. CACL2(s) POSSIBLE. LIQUID & SOLID AEROSOL (SUBROUTINE CALCL2A)
!     2. CACL2(s) NOT POSSIBLE, AND RH < MDRH. SOLID AEROSOL ONLY
!     3. CACL2(s) NOT POSSIBLE, AND RH >= MDRH. SOLID & LIQUID AEROSOL
!
!     REGIMES 2. AND 3. ARE CONSIDERED TO BE THE SAME AS CASES W1A, W2B
!     RESPECTIVELY
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
!
SUBROUTINE calcw2 (  )
  USE mo_isrpia
  IMPLICIT NONE
  INTEGER :: i
  !
  ! *** FIND DRY COMPOSITION **********************************************
  !
  CALL calcw1a
  !
  ! *** REGIME DEPENDS UPON THE POSSIBLE SOLIDS & RH **********************
  !
  IF (ccacl2>tiny) THEN
    scase = 'W2 ; SUBCASE 1'
    CALL calcw2a
    scase = 'W2 ; SUBCASE 1'
  END IF
  !
  IF (water<=tiny) THEN
    IF (rh<drmp2) THEN ! ONLY SOLIDS
      water = tiny
      DO i = 1, nions
        molal(i) = zero
      END DO
      CALL calcw1a
      scase = 'W2 ; SUBCASE 2'
    ELSE
      IF (cmgcl2>tiny) THEN
        scase = 'W2 ; SUBCASE 3' ! MDRH (CaSO4, CA(NO3)2, K2SO4, KNO3, KCL, MGSO4, MGCL2,
        !                                                  MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL)
        CALL calcmdrpii(rh, drmp2, drmgcl2, calcw1a, calcw3a)
        scase = 'W2 ; SUBCASE 3'
      END IF
      IF (water<=tiny .AND. rh>=drmp3 .AND. rh<drmp4) THEN
        scase = 'W2 ; SUBCASE 4' ! MDRH (CaSO4, K2SO4, KNO3, KCL, MGSO4, CANO32,
        !                                                  MG(NO3)2, NANO3, NACL, NH4NO3, NH4CL)
        CALL calcmdrpii(rh, drmp3, drcano32, calcw1a, calcw4a)
        scase = 'W2 ; SUBCASE 4'
      END IF
      IF (water<=tiny .AND. rh>=drmp4 .AND. rh<drmp5) THEN
        scase = 'W2 ; SUBCASE 5' ! MDRH (CaSO4, K2SO4, KNO3, KCL, MGSO4,
        !                                                  MGNO32, NANO3, NACL, NH4NO3, NH4CL)
        CALL calcmdrpii(rh, drmp4, drmgno32, calcw1a, calcw5a)
        scase = 'W2 ; SUBCASE 5'
      END IF
      IF (water<=tiny .AND. rh>=drmp5) THEN
        scase = 'W2 ; SUBCASE 6' ! MDRH (CaSO4, K2SO4, KNO3, KCL, MGSO4,
        !                                                  NANO3, NACL, NH4NO3, NH4CL)
        CALL calcmdrpii(rh, drmp5, drnh4no3, calcw1a, calcw6)
        scase = 'W2 ; SUBCASE 6'
      ELSE
        water = tiny
        DO i = 1, nions
          molal(i) = zero
        END DO
        CALL calcw1a
        scase = 'W2 ; SUBCASE 2'
      END IF
    END IF
  END IF
  !
  RETURN
  !
END SUBROUTINE calcw2
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW2A
! *** CASE W2A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. THERE IS BOTH A LIQUID & SOLID PHASE
!     3. SOLIDS POSSIBLE : CaSO4, K2SO4, KNO3, MGSO4, KCL, NH4CL, NACL,
!                          NANO3, NH4NO3, CA(NO3)2, MG(NO3)2, MGCL2
!     4. Completely dissolved: CACL2
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw2a (  )
  USE mo_isrpia
  USE mo_solut
  IMPLICIT NONE
  DOUBLE PRECISION :: akw
  DOUBLE PRECISION :: bb
  DOUBLE PRECISION :: cai
  DOUBLE PRECISION :: cc
  DOUBLE PRECISION :: claq
  DOUBLE PRECISION :: cli
  DOUBLE PRECISION :: dd
  DOUBLE PRECISION :: del
  DOUBLE PRECISION :: gg
  DOUBLE PRECISION :: hi
  DOUBLE PRECISION :: hso4i
  INTEGER :: i
  INTEGER :: islv
  INTEGER :: kcl
  DOUBLE PRECISION :: ki
  DOUBLE PRECISION :: mgi
  DOUBLE PRECISION :: nai
  DOUBLE PRECISION :: nh3aq
  DOUBLE PRECISION :: nh4i
  DOUBLE PRECISION :: no3aq
  DOUBLE PRECISION :: no3i
  DOUBLE PRECISION :: ohi
  LOGICAL :: psconv13
  LOGICAL :: psconv14
  LOGICAL :: psconv5
  LOGICAL :: psconv7
  LOGICAL :: psconv8
  LOGICAL :: psconv9
  DOUBLE PRECISION :: psi13o
  DOUBLE PRECISION :: psi14o
  DOUBLE PRECISION :: psi5o
  DOUBLE PRECISION :: psi7o
  DOUBLE PRECISION :: psi8o
  DOUBLE PRECISION :: psi9o
  DOUBLE PRECISION :: root13
  DOUBLE PRECISION :: root13a
  DOUBLE PRECISION :: root13b
  DOUBLE PRECISION :: root14
  DOUBLE PRECISION :: root14a
  DOUBLE PRECISION :: root14b
  DOUBLE PRECISION :: root5
  DOUBLE PRECISION :: root5a
  DOUBLE PRECISION :: root5b
  DOUBLE PRECISION :: root7
  DOUBLE PRECISION :: root7a
  DOUBLE PRECISION :: root7b
  DOUBLE PRECISION :: root8
  DOUBLE PRECISION :: root8a
  DOUBLE PRECISION :: root8b
  DOUBLE PRECISION :: root9
  DOUBLE PRECISION :: so4i
  !
  ! *** SETUP PARAMETERS ************************************************
  !
  frst = .TRUE.
  calain = .TRUE.
  calaou = .TRUE.
  !
  psconv9 = .TRUE.
  psconv13 = .TRUE.
  psconv14 = .TRUE.
  psconv5 = .TRUE.
  psconv7 = .TRUE.
  psconv8 = .TRUE.
  !
  psi9o = -great
  psi13o = -great
  psi14o = -great
  psi5o = -great
  psi7o = -great
  psi8o = -great ! GREAT = 1.D10
  !
  root9 = zero
  root13 = zero
  root14 = zero
  root5 = zero
  root7 = zero
  root8 = zero
  !
  ! *** CALCULATE INITIAL SOLUTION ***************************************
  !
  CALL calcw1a
  !
  chi9 = ck2so4 ! SALTS
  chi13 = ckno3
  chi10 = cmgso4
  chi14 = ckcl
  chi5 = cnh4cl
  chi7 = cnacl
  chi8 = cnano3
  chi6 = cnh4no3
  chi15 = cmgno32
  chi12 = ccano32
  chi16 = cmgcl2
  chi11 = ccaso4
  !
  psi1 = cna2so4 ! SALTS DISSOLVED
  psi5 = cnh4cl
  psi6 = cnh4no3
  psi7 = cnacl
  psi8 = cnano3
  psi9 = ck2so4
  psi10 = cmgso4
  psi11 = ccaso4
  psi12 = ccano32
  psi13 = ckno3
  psi14 = ckcl
  psi15 = cmgno32
  psi16 = cmgcl2
  psi17 = ccacl2
  !
  CALL calcmr ! WATER
  !
  nai = waer(1) ! LIQUID CONCENTRATIONS
  so4i = max(waer(2)-waer(6), zero)
  nh4i = waer(3)
  no3i = waer(4)
  cli = waer(5)
  cai = waer(6)
  ki = waer(7)
  mgi = waer(8)
  !
  hso4i = zero
  nh3aq = zero
  no3aq = zero
  claq = zero
  !
  ! *** SOLVE EQUATIONS ; WITH ITERATIONS FOR ACTIVITY COEF. ************
  !
  DO i = 1, nsweep
    !
    a9 = xk17*(water/gama(17))**3.0 ! K2SO4     <==> K+
    a13 = xk19*(water/gama(19))**2.0 ! KNO3      <==> K+
    a14 = xk20*(water/gama(20))**2.0 ! KCL       <==> K+
    a5 = xk14*(water/gama(6))**2.0 ! NH4Cl     <==> NH4+
    a7 = xk8*(water/gama(1))**2.0 ! NaCl      <==> Na+
    a8 = xk9*(water/gama(3))**2. ! NaNO3     <==> Na+
    akw = xkw*rh*water*water ! H2O       <==> H+
    !
    ! POTASSIUM SULFATE
    !
    IF (ki*ki*so4i>a9) THEN
      bb = -((waer(2)-waer(6))+waer(7)-root13-root14)
      cc = (waer(7)-root13-root14)*(waer(2)-waer(6)) +                     &
           0.25D0*(waer(7)-root13-root14)**2.0
      dd = -0.25*((waer(7)-root13-root14)**2.0*(waer(2)-waer(6))-a9)
      CALL poly3(bb, cc, dd, root9, islv)
      IF (islv/=0) root9 = tiny
      root9 = min(root9, waer(7)/2.0-root13-root14, (waer(2)-waer(6)),     &
           chi9)
      root9 = max(root9, zero)
      psi9 = chi9 - root9
    END IF
    psconv9 = abs(psi9-psi9o) <= eps*psi9o
    psi9o = psi9
    !
    ! POTASSIUM NITRATE
    !
    IF (ki*no3i>a13) THEN
      bb = -(waer(4)-root8+waer(7)-2.D0*root9-root14)
      cc = (waer(4)-root8)*(waer(7)-2.D0*root9-root14) - a13
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root13a = 0.5D0*(-bb-dd)
      root13b = 0.5D0*(-bb+dd)
      IF (zero<=root13a) THEN
        root13 = root13a
      ELSE
        root13 = root13b
      END IF
      root13 = min(max(root13,zero), chi13)
      psi13 = chi13 - root13
    END IF
    psconv13 = abs(psi13-psi13o) <= eps*psi13o
    psi13o = psi13
    !
    ! POTASSIUM CLORIDE
    !
    IF (ki*cli>a14) THEN
      bb = -(waer(5)-root5-root7+waer(7)-2.D0*root9-root13)
      cc = (waer(5)-root5-root7)*(waer(7)-2.D0*root9-root13) - a14
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root14a = 0.5D0*(-bb-dd)
      root14b = 0.5D0*(-bb+dd)
      IF (zero<=root14a) THEN
        root14 = root14a
      ELSE
        root14 = root14b
      END IF
      root14 = min(max(root14,zero), chi14)
      psi14 = chi14 - root14
    END IF
    psconv14 = abs(psi14-psi14o) <= eps*psi14o
    psi14o = psi14
    !
    ! AMMONIUM CLORIDE
    !
    IF (nh4i*cli>a5) THEN
      bb = -(waer(5)+waer(3)-root14-root7)
      cc = (waer(5)-root14-root7)*waer(3) - a5
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root5a = 0.5D0*(-bb-dd)
      root5b = 0.5D0*(-bb+dd)
      IF (zero<=root5a) THEN
        root5 = root5a
      ELSE
        root5 = root5b
      END IF
      root5 = min(max(root5,zero), chi5)
      psi5 = chi5 - root5
    END IF
    psconv5 = abs(psi5-psi5o) <= eps*psi5o
    psi5o = psi5
    !
    ! SODIUM CLORIDE
    !
    IF (nai*cli>a7) THEN
      bb = -(waer(5)+waer(1)-root8-root14-root5)
      cc = (waer(5)-root14-root5)*(waer(1)-root8) - a7
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root7a = 0.5D0*(-bb-dd)
      root7b = 0.5D0*(-bb+dd)
      IF (zero<=root7a) THEN
        root7 = root7a
      ELSE
        root7 = root7b
      END IF
      root7 = min(max(root7,zero), chi7)
      psi7 = chi7 - root7
    END IF
    psconv7 = abs(psi7-psi7o) <= eps*psi7o
    psi7o = psi7
    !
    ! SODIUM NITRATE
    !
    IF (nai*no3i>a8) THEN
      bb = -(waer(4)-root13+waer(1)-root7)
      cc = (waer(4)-root13)*(waer(1)-root7) - a8
      dd = sqrt(max(bb*bb-4.D0*cc,tiny))
      root8a = 0.5D0*(-bb-dd)
      root8b = 0.5D0*(-bb+dd)
      IF (zero<=root8a) THEN
        root8 = root8a
      ELSE
        root8 = root8b
      END IF
      root8 = min(max(root8,zero), chi8)
      psi8 = chi8 - root8
    END IF
    psconv8 = abs(psi8-psi8o) <= eps*psi8o
    psi8o = psi8
    !
    ! ION CONCENTRATIONS ; CORRECTIONS
    !
    ki = max(waer(7)-2.D0*root9-root13-root14, zero)
    so4i = max(waer(2)-waer(6)-root9, zero)
    nh4i = max(waer(3)-root5, zero)
    no3i = max(waer(4)-root13-root8, zero)
    cli = max(waer(5)-root14-root5-root7, zero)
    cai = zero
    nai = max(waer(1)-root7-root8, zero)
    mgi = waer(8)
    !
    ! SOLUTION ACIDIC OR BASIC?
    !
    gg = 2.D0*so4i + no3i + cli - nai - nh4i - 2.D0*cai - ki - 2.D0*mgi
    IF (gg>tiny) THEN ! H+ in excess
      bb = -gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      hi = 0.5D0*(-bb+sqrt(dd))
      ohi = akw/hi
    ELSE ! OH- in excess
      bb = gg
      cc = -akw
      dd = bb*bb - 4.D0*cc
      ohi = 0.5D0*(-bb+sqrt(dd))
      hi = akw/ohi
    END IF
    !
    ! UNDISSOCIATED SPECIES EQUILIBRIA
    !
    IF (hi>ohi) THEN
      !         CALL CALCAMAQ2 (-GG, NH4I, OHI, NH3AQ)
      !         HI    = AKW/OHI
      !         HSO4I = ZERO
      !      ELSE
      !         GGNO3 = MAX(2.D0*SO4I + NO3I - NAI - NH4I - 2.D0*CAI
      !     &           - KI - 2.D0*MGI, ZERO)
      !         GGCL  = MAX(GG-GGNO3, ZERO)
      !         IF (GGCL .GT.TINY) CALL CALCCLAQ2 (GGCL, CLI, HI, CLAQ) ! HCl
      !         IF (GGNO3.GT.TINY) THEN
      !            IF (GGCL.LE.TINY) HI = ZERO
      !            CALL CALCNIAQ2 (GGNO3, NO3I, HI, NO3AQ)              ! HNO3
      !         ENDIF
      !
      ! CONCENTRATION ADJUSTMENTS ; HSO4 minor species.
      !
      CALL calchs4(hi, so4i, zero, del)
    ELSE
      del = zero
    END IF
    so4i = so4i - del
    hi = hi - del
    hso4i = del
    !         IF (HI.LE.TINY) HI = SQRT(AKW)
    ohi = akw/hi
    !
    IF (hi<=tiny) THEN
      hi = sqrt(akw)
      ohi = akw/hi
    END IF
    !
    ! *** SAVE CONCENTRATIONS IN MOLAL ARRAY ******************************
    !
    molal(1) = hi
    molal(2) = nai
    molal(3) = nh4i
    molal(4) = cli
    molal(5) = so4i
    molal(6) = hso4i
    molal(7) = no3i
    molal(8) = cai
    molal(9) = ki
    molal(10) = mgi
    !
    ! *** CALCULATE WATER **************************************************
    !
    CALL calcmr
    !
    ! *** CALCULATE ACTIVITIES OR TERMINATE INTERNAL LOOP *****************
    !
    IF (frst .AND. calaou .OR. .NOT. frst .AND. calain) THEN
      CALL calcact
    ELSE
      IF (psconv9 .AND. psconv13 .AND. psconv14 .AND. psconv5 .AND.        &
           psconv7 .AND. psconv8) GO TO 100
    END IF
  END DO
  !cc      CALL PUSHERR (0002, 'CALCW2')    ! WARNING ERROR: NO CONVERGENCE
  !
  ! *** CALCULATE GAS / SOLID SPECIES (LIQUID IN MOLAL ALREADY) *********
  !
100 a2 = (xk2/xkw)*r*temp*(gama(10)/gama(5))**2. ! NH3  <==> NH4+
  a3 = xk4*r*temp*(water/gama(10))**2. ! HNO3 <==> NO3-
  a4 = xk3*r*temp*(water/gama(11))**2. ! HCL  <==> CL-
  !
  gnh3 = nh4i/hi/a2
  ghno3 = hi*no3i/a3
  ghcl = hi*cli/a4
  !
  gasaq(1) = nh3aq
  gasaq(2) = claq
  gasaq(3) = no3aq
  !
  cnh42s4 = zero
  cnh4no3 = zero
  cnh4cl = chi5 - psi5
  cnacl = chi7 - psi7
  cnano3 = chi8 - psi8
  cmgso4 = zero
  ck2so4 = chi9 - psi9
  ccaso4 = min(waer(6), waer(2))
  ccano32 = zero
  ckno3 = chi13 - psi13
  kcl = chi14 - psi14
  cmgno32 = zero
  cmgcl2 = zero
  ccacl2 = zero
  !
  RETURN
  !
END SUBROUTINE calcw2a
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW1
! *** CASE W1
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
!     THERE ARE TWO POSSIBLE REGIMES HERE, DEPENDING ON RELATIVE HUMIDITY:
!     1. WHEN RH >= MDRH ; LIQUID PHASE POSSIBLE (MDRH REGION)
!     2. WHEN RH < MDRH  ; ONLY SOLID PHASE POSSIBLE (SUBROUTINE CALCP1A)
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY,
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS & ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw1 (  )
  USE mo_isrpia
  IMPLICIT NONE
  !
  ! *** REGIME DEPENDS UPON THE AMBIENT RELATIVE HUMIDITY *****************
  !
  IF (rh<drmp1) THEN
    scase = 'W1 ; SUBCASE 1'
    CALL calcw1a ! SOLID PHASE ONLY POSSIBLE
    scase = 'W1 ; SUBCASE 1'
  ELSE
    scase = 'W1 ; SUBCASE 2' ! LIQUID & SOLID PHASE POSSIBLE
    CALL calcmdrpii(rh, drmp1, drcacl2, calcw1a, calcw2a)
    scase = 'W1 ; SUBCASE 2'
  END IF
  !
  RETURN
  !
END SUBROUTINE calcw1
!
!=======================================================================
!
! *** ISORROPIA CODE II
! *** SUBROUTINE CALCW1A
! *** CASE W1A
!
!     THE MAIN CHARACTERISTICS OF THIS REGIME ARE:
!     1. SULFATE POOR (SULRAT > 2.0) ; Rcr+Na >= 2.0 ; Rcr > 2)
!     2. SOLID AEROSOL ONLY
!     3. SOLIDS POSSIBLE : CaSO4, CA(NO3)2, CACL2, K2SO4, KNO3, KCL, MGSO4,
!                          MG(NO3)2, MGCL2, NANO3, NACL, NH4NO3, NH4CL
!
! *** COPYRIGHT 1996-2008, UNIVERSITY OF MIAMI, CARNEGIE MELLON UNIVERSITY
! *** GEORGIA INSTITUTE OF TECHNOLOGY
! *** WRITTEN BY CHRISTOS FOUNTOUKIS AND ATHANASIOS NENES
!
!=======================================================================
!
SUBROUTINE calcw1a (  )
  USE mo_isrpia
  IMPLICIT NONE
  DOUBLE PRECISION :: cafr
  DOUBLE PRECISION :: clfr
  DOUBLE PRECISION :: frk
  DOUBLE PRECISION :: frmg
  DOUBLE PRECISION :: frna
  DOUBLE PRECISION :: frno3
  DOUBLE PRECISION :: so4fr
  !
  ! *** CALCULATE SOLIDS **************************************************
  !
  ccaso4 = min(waer(2), waer(6)) !SOLID CASO4
  cafr = max(waer(6)-ccaso4, zero)
  so4fr = max(waer(2)-ccaso4, zero)
  ck2so4 = min(so4fr, 0.5D0*waer(7)) !SOLID K2SO4
  frk = max(waer(7)-2.D0*ck2so4, zero)
  so4fr = max(so4fr-ck2so4, zero)
  cmgso4 = so4fr !SOLID MGSO4
  frmg = max(waer(8)-cmgso4, zero)
  cnacl = min(waer(1), waer(5)) !SOLID NACL
  frna = max(waer(1)-cnacl, zero)
  clfr = max(waer(5)-cnacl, zero)
  ccacl2 = min(cafr, 0.5D0*clfr) !SOLID CACL2
  cafr = max(cafr-ccacl2, zero)
  clfr = max(waer(5)-2.D0*ccacl2, zero)
  ccano32 = min(cafr, 0.5D0*waer(4)) !SOLID CA(NO3)2
  cafr = max(cafr-ccano32, zero)
  frno3 = max(waer(4)-2.D0*ccano32, zero)
  cmgcl2 = min(frmg, 0.5D0*clfr) !SOLID MGCL2
  frmg = max(frmg-cmgcl2, zero)
  clfr = max(clfr-2.D0*cmgcl2, zero)
  cmgno32 = min(frmg, 0.5D0*frno3) !SOLID MG(NO3)2
  frmg = max(frmg-cmgno32, zero)
  frno3 = max(frno3-2.D0*cmgno32, zero)
  cnano3 = min(frna, frno3) !SOLID NANO3
  frna = max(frna-cnano3, zero)
  frno3 = max(frno3-cnano3, zero)
  ckcl = min(frk, clfr) !SOLID KCL
  frk = max(frk-ckcl, zero)
  clfr = max(clfr-ckcl, zero)
  ckno3 = min(frk, frno3) !SOLID KNO3
  frk = max(frk-ckno3, zero)
  frno3 = max(frno3-ckno3, zero)
  !
  ! *** OTHER PHASES ******************************************************
  !
  water = zero
  !
  gnh3 = zero
  ghno3 = zero
  ghcl = zero
  !
  RETURN
  !
END SUBROUTINE calcw1a

END MODULE art_isorropia
